
%%
%% Self-Guided Ergo 5 interaction tour, Part 2. (proofs.thy)
%%

% This is the second part of a self-guided introduction to the
% Ergo proof tool.
%
% At this stage, it is assumed that you are familiar with the emacs
% interface to Ergo, and that you have Ergo loaded.
%
% If Ergo is not currently loaded (with the cpred theory), place
% the cursor at the beginning of the following line, and press F5
% (or C-c C-t).

[cpred].


%%
%% Contents of this file.
%%

% Sequent calculus "recap".
% An introduction to proofs using Ergo.
% Gumtree.
% Writing Tactics.
% The rule browser.
% Quantifiers, substitutions and Unification.
% Review.            

%%
%% Sequent calculus "recap".
%%

% In a sequent calculus proof, the main data structure is a sequent.
%
% A sequent consists of a 'goal' (i.e. a formula to be proved) and
% hypotheses (assumed to be true for the proof of the goal).
%
% A proof in sequent calculus is typically done in a 'goal directed' or
% 'backward' style starting with what is to be proved and applying
% inference rules to reduce the problem to simpler problems until all
% (sub) problems have been proved.
% A sequent claculus proof in Ergo constructs a 'proof tree'.
% A proof tree is a tree with each node being a sequent, 
% where the root node is the sequent representing
% the theorem being proved. The children of each node is related to
% the node by the application of an inference rule.
%
% A node of a proof tree is 'closed' if an inference rule has been applied
% to the node, and 'open' otherwise.
%
% In Ergo the hypotheses of each sequent are partitioned into contexts. 
%
% Ergo takes a very flexible view of contexts - users may declare 
% their own context. However, the built-in classical sequent calculus theories
% come with the predeclared context 'hyp' which reflects the required 
% semantics of classical sequent calculus.
%
% Ergo has two kinds of inference rules -- primitive rules
% and derived rules. Primitive rules are the basic rules
% which define a theory, and are ideally as few as possible.
% Derived rules can be proven from primitive rules (or by other
% derived rules, provided we don't create circular proofs).
%
% Primitive inference rules are identified by the keyword 
% "axiom" at the start of its declaration.
%
% In Ergo, rules are typically "schematic" - i.e. they contain vaiables
% that stand for terms or formulae. When a rule is applied, variables in
% the rule are unified with terms and formulae of the sequent being proved.
%
% Because rules are schematic, proofs of rules involve variables. 
% Typically, the variables occurring in the statement of the rule being proved
% should not be instantiated - as this would overspecialise the proof.
% As a consequence, Ergo "freezes" variables occurring in the statement
% of the rule to be proved. A frozen variable cannot be instantiated and this
% is usually what is required.
% 
% In some proofs, the values for some variables are to be determined during
% the proof (e.g. when prove/1 is used to simplify or transform an expression
% to a yet-to-be-determined value). 
% In such cases the variables in question need to be thawed before they can 
% be instantiated. 
%
% WARNING: variables introduced into a proof when a rule is applied - e.g.
% by using imp_elim or ex_intro -  are not frozen and so can be instantiated.
% This means you need to be careful when applying certain kinds of tactics
% and rules as they may lead to infinite computation. For example, if the
% goal is a thawed variable then the application of the and_intro rule
% will cause this variable to be instantiated to the conjunction of
% 2 variables - leading to 2 proof nodes, both of which have thawed
% variables as goals. If a tactics was called that repeatedly called
% and_intro then an infinite computation would arise.
%
%%
%% An introduction to proofs using Ergo.
%%

% We'll begin by performing a proof of a very simple
% theorem.
%
% First we need to enter the theorem into Ergo's 
% database:

theorem first === A => (B and C) => B.

% At our disposal, we have hundreds of inference rules
% already in the system as part of the cpred theory (and 
% its ancestors). The proof of 'first' makes use of the 
% imp_intro, and_elim_c and assump inference rules.

show_rule(assump).
show_rule(and_elim_c).
show_rule(imp_intro).

% We will look at the details of rules in a moment, but for now we
% just give the meaning of these three rules.
%
% The assump rule says that B can be proved if B is in the hyp
% context - i.e. it is an hypothesis.
%
% The and_elim_c is a form of and-elimination using the hyp context.
% It states that, if (A and B) is an hypothesis, then it can be replaced by the
% hypothesis A and the hypothesis B.
%
% The imp_intro rule says that, in order to prove A => B  (i.e. A implies B),
% it is enough to prove B under the hypothesis A.


% Now, we begin its proof:

prove_rule(first).

% In doing a proof, Ergo constructs a proof tree, with the inital
% goal as the root of the tree and all the unproven nodes as the leaves
% of the proof tree.
%
% Initially, the (only) open node is the goal at the root of the tree.
% The interface lists all the open nodes with a node identifier and
% the prompt consists of the list of open node identifiers in left-to-right
% order in the proof tree.
%
%
% The following rule will take an implication, and create
% a new subgoal with the LHS as a hypothesis. We do this
% twice because there are two implications.

rule(imp_intro).
rule(imp_intro).

% With each rule application, note that we have added
% a new "context term" to 'hyp' (the hypotheses), and
% it has been indicated by adding the context term 
% identifier to the list of hypotheses.
%
% Now we have a conjunction (a term with an 'and' in it)  
% as a hypothesis, this can be split up using the and_elim_c
% rule.

rule(and_elim_c).

% Finally, we can complete the proof by the law of assumption
% because we have a hypotheses, B, that matches the current 
% subgoal.

rule(assump).

% These inference rules usually take additional arguments.
% The and_elim_c rule, for example, can have an extra argument
% to specify the hypothesis that contains the conjunction that
% is to be "split up".
% Ergo will search for a valid application of the rule if 
% these arguments are left off.
%
% We now have a different prompt to indicate that there are
% no remaining subgoals -- that we've sucessfully completed
% the proof.
%
% At this point, we can save the proof. We need to specify
% the root node of the proof, which was 1 ('1' was the number
% to the left of the first subgoal in the proof, and will 
% almost always be 1).

save_proof(first_proof,1).

% Now we'll exit the proof, and return to Ergo's database
% interface.

quit_proof.

% We'll now look at inference rules in more detail.
% We start wilh a simple rule.

show_rule(and_intro).

% Below the line in the rule (which can be formed from anything between 
% 4 and 60 dash characters) we have the conclusion, and above it, 
% are the premises of the rule.
%
% If the rule is applied to a node with a goal of the form
% "A and B",  Ergo will "branch" at that node, into two new 
% open nodes, one with a goal being "A" and the other, "B".
%
% On the left hand side of the ':::' we have variables representing 
% node identifiers. These are typically not needed when
% defining new theorems, and so are optional.
%
% Looking once more at and_elim_c we see notation for 
% adding and deleting from 'hyp'.

show_rule(and_elim_c).

% Much of this is optional, in "short-hand" (as it appears
% in the theory source file) it can be represented as;
%
%  theorem and_elim_c(A:thy(term),B:thy(term)) ===
%        hyp---[CId ::: A and B]+++[A,B] ---> C
%        ---------------------------------------
%        Node :::                             C
%      provided
%        context_search(hyp, Node, CId, A and B).  
%
% First note that if the rule names have arguments (as above) then
% they must have types associated with them. This is so that the interpreter
% can decide what needs to be deabbreviated before being used.
%
% The provided section is used to restrict the use 
% of a rule. The condition in the 'provided' clause
% above states that 'A and B' must be a term of the
% 'hyp' context of the current node (Node).
%
% Above the line we see notation for removing hypotheses from ('---') and 
% adding hypothes to ('+++') the hyp context.
% 
% The ---> arrow separates the contexts from the goal of the sequent.
%
% We'll now look at another simple proof:

theorem second === A => B => (A and B).

prove_rule(second).

% Again, we'll 'remove' the implications:
%
% (In a goal directed proof, we use introduction rules
% when we want to "eliminate" and elimination rules 
% when we want to "introduce").

rule(imp_intro).
rule(imp_intro).

% We now have one open node which contains a conjunction,
% which happens to exactly match the conclusion of the 
% and_intro rule we saw previously.

rule(and_intro).

% Ergo now indicates that we have two open nodes. We can 
% discharge each node by the assump rule.

[rule(assump),rule(assump)].

% Above, we applied two rules in parallel, this will be
% explained in a moment, but first we'll try a few variants.

undo.

% Ergo has undone the last command - the state of the proof
% has been restored to what it was before we applied
% the last command.

[one,rule(assump)].
rule(assump).

% This time one node was left untouched, and the other was
% discharged, and then the remaining node was discharged
% in the subsequent command.

undo.
undo.

[4].

% Note that we now have a different prompt (([5]) [4]:::), this
% indicates that rules (or tactics) will only be applied
% to node 4 (and its context).
%
% The second list of node IDs is the list of current nodes. The first list is
% the list of open nodes that are not current. In this example node 5 is
% an open node but is not a current node.
%
% Now discharge node 4:

rule(assump).

% And then node 5:

[5].
rule(assump).


% We'll quit the proof (without saving), and start again:

quit_proof.

prove_rule(second).

rule(imp_intro);rule(imp_intro);rule(and_intro);[rule(assump),rule(assump)].

% The semicolon is called a tactic combinator, it is used
% to combine several steps into one command.
%
% We'll save this proof too, this time under the name of
% the theorem. Once this has been done with save_postulate_proof,
% the theorem is marked as proven and cannot be reproved.

save_postulate_proof(1).

quit_proof.

%%
%% Gumtree.
%%

% Gumtree is both an interface to Ergo and a tactic language for Ergo.
% The proofs above use the Gumtree interface, and this is the default
% interface for all proofs in Ergo.
%
% So far, we have seen a mixture of Gumtree commands, and the
% Gumtree tactic language. The Gumtree tactic language is used
% to carry out the steps of a proof, whereas Gumtree commands
% (such as 'undo' and 'save_postulate_proof') relate to the Ergo
% environment itself.
%
% Any Gumtree code composed from the Gumtree tactic language can
% be called a "tactic", but in general the term "tactic" refers 
% to Gumtree code that has been specifically declared as a tactic.
%
% We have already seen some of the Gumtree tactic language:
%
%     rule(RuleName)
%         applies the inference rule RuleName.
%
%     [Tactic1,Tactic2,...]
%         applies several tactics in "parallel"
%         
%     one
%         leaves exactly one proof node unchanged
%
%     Tactic1;Tactic2;...
%         sequences the execution of tactics,
%         do Tactic1, then do Tactic2, then ....
%
% For a list of gumtree code combinators, view the following
% help topic.

help(tactic_combine_cmds).

% Other handy Gumtree primitive tactics can be found here:

help(proof_tree_cmds).

% In practice, only a small subset of Gumtree commands are 
% regularly needed, some more of these are:
%
%     skip
%         leaves any number of proof nodes unchanged
%
%     Tactic1 '|' Tactic2
%         executes Tactic1. if Tactic1 fails, execute Tactic2


%%
%% Writing Tactics.
%%

% If we recognised that we were going to prove a lot of theorems
% that contained implications and conjunctions, we could write
% a tactic to simplify things.
%
% A tactic may be declared while in the middle of a proof, but it
% is usually preferable (if possible) to declare them at the
% database prompt, or even better, within a file that we load.
%
% We declare a tactic as follows.
%
% The doc part of the declaration is optional and if supplied adds
% a documentation node to the system.
% The mode describes the types of the arguments of the tactic.
% The arity describes the number of nodes input and number output.
% In this case there are no constraints on the arity and so a variable is used.

tactic
  imp_or_conj
doc
  'Apply one of the rules imp_intro, and_intro or and_elim_c.'
mode
  imp_or_conj
arity
  _.

% The following shows all the doc nodes for user defined Gumtree tactics.

help(gumtree_user_tactics, global).

% The definition of the tactic is made with one or more clauses as follows.

tactic imp_or_conj ===
        rule(imp_intro)
        '|'
        rule(and_intro)
        '|'
        rule(and_elim_c).

% NOTE:
% We could also define the same tactic using three clauses as follows.
% tactic imp_or_conj === rule(imp_intro).
% tactic imp_or_conj === rule(and_intro).
% tactic imp_or_conj === rule(and_elim_c).

% This tactic will apply one of the three rules we've been using
% in our proofs. If the first rule fails (ie. it cannot be applied),
% it will try the second, and if that fails it will try the third.
%
% We'll use it to prove 'first' again.

prove_rule(first).

imp_or_conj.
imp_or_conj.
imp_or_conj.

% If we had many related problems, this kind of tactic would make
% proofs far simpler, and would automate a great deal of the process.
%
% If we try once more, none of the rules will be appropriate, so
% the entire tactic will fail.

imp_or_conj.

% We can finish the proof:

rule(assump).

quit_proof.

% Gumtree, like Prolog, backtracks upon failure. This makes it 
% easy to code tactics that search for proofs. This will be covered 
% in more detail later.

% Now we'll try a more ambitious tactic.
%
% We will attempt to code a kind of meta-tactic, it will repeatedly
% apply another tactic until it fails.
%
% Gumtree has no looping constructs, so we must use recursion.
%
% An initial attempt might appear something like this:
%
% tactic repeat_tactic(Tactic) ===
%         Tactic;repeat_tactic(Tactic).
%
% The problem with this is that it will either fail, or enter 
% an infinite loop. This is because if Tactic fails, the entire 
% command will fail, but if Tactic succeeds, it will just be 
% executed again (forever, or until it fails).
%
% A better solution follows:

tactic
  repeat_tactic(Tactic)
doc
  'Repeatedly apply Tactic until it no longer applies.'
mode
  repeat_tactic(tactic)
arity
  _.

tactic repeat_tactic(Tactic) ===
        Tactic;repeat_tactic(Tactic)
        '|'
        skip.

% This code will keep executing Tactic until it fails, and then
% execute a skip.
%
% Variable names follow Prolog naming conventions; all variables 
% must begin with a capital letter.
%
% Lets test the tactic;

prove_rule(first).
repeat_tactic(imp_or_conj).
rule(assump).

% Before quitting the proof, we'll examine the underlying inference
% rule applications.

show_subproof(1).

% If we examine our first proof, we notice that it has an identical
% sequence of rule applications (excluding an initial premises([])
% command).

show_proof_file('cpred/first_proof.proof').

% This is despite the fact that we entered a different sequence of 
% commands.

show_proof_script_file('cpred/first_proof.proof').

quit_proof.

% We now look at two more tactics - many and deep - both of which are
% defined in tacticals.gum.
% Here is the definition of many.
%
% tactic
%   many(Tactic)
% doc
%   'Try applying Tactic to each open node.',nl,
%   'Alternatives are removed from each successful application of Tactic.'
% mode
%   many(tactic)
% arity
%   _.
% 
% tactic many(Tactic) ===
%     !(
%         zero
%         '|'
%         [!(one;Tactic '|' one), many(Tactic)]
%     ).
% 
% Let's look at this tactic in detail. First consider
% one;Tactic 
% The 'one' forces the composition to have input arity 1 and hence
% Tactic is called with input arity 1. As we will see in a moment, this
% forces Tactic to be called on the first open node.
% Now consider
% !(one;Tactic '|' one)
% This attempts to call Tactic (with input arity 1) and if that succeeds
% then the '!' removes any alternatives within Tactic as well as the 
% alternative 'one'.
% If Tactic fails then 'one' is called.
% The input arity of the tactic !(one;Tactic '|' one) is 1.
% Now consider
% [!(one;Tactic '|' one), many(Tactic)]
% The gumtree 'parallel' tactic takes a list of tactics and attempts
% to partition the list of open nodes (with one partition for each tactic
% in the tactic list) so that each tactic's input arity matches the
% number of open nodes in the corresponding partition. Each tactic is
% applied to the nodes of its partition. Note that, for tactics that
% can take different input arities, the gumtree parallel tactic will
% search for different ways of partitioning the open nodes so that
% all the tactics will succeed.
% In our example, becuase Gumtree knows !(one;Tactic '|' one) has
% input arity 1 then the only partition that could work is to choose
% the first open node for !(one;Tactic '|' one) and the remainder for
% many(Tactic).
% Consequently, 
% [!(one;Tactic '|' one), many(Tactic)]
% will apply the tactic
% !(one;Tactic '|' one)
% to the (list containing) the first open node and apply the tactic
% many(Tactic)
% to the remaining open nodes.
% Now if we consider the execution of the tactic
% [!(one;Tactic '|' one), many(Tactic)]
% we see that it attempts to apply Tactic to each open node until
% it runs out of nodes. What happens then? Well, this tactic fails
% since its input arity is at least one.
% To stop the entire tactic failing, we therefore provide 'zero' as
% an alternative. zero has input arity 0, and 
% [!(one;Tactic '|' one), many(Tactic)]
% has input arity at least 1. Together they cover the possible input
% open nodes.
%
% Here is the definition of deep.
% tactic
%   deep(Tac)
% doc
%   'Try applying Tactic to each open node, and then apply deep(Tac)',nl,
%   'to each successful result. Alternatives are removed from',nl,
%   'each successful application of Tactic.'
% mode
%   deep(tactic)
% arity
%   _.
% 
% tactic deep(Tac) ===
%     !(
%         zero
%         '|'
%         [!(one;Tac;deep(Tac) '|' one), deep(Tac)]
%     ).
%
% We note that many and deep are very similar. The only difference is that
% deep recursively applies itself to any open nodes produced by applying
% Tac (one;Tac;deep(Tac)).

%%
%% The rule browser.
%%

% Ergo has a rule browser, as a means of keeping track of the many
% inference rules in a theory.
%
% When performing a proof, it is helpful to have a basic familiarity
% with the inference rules of a theory. 
% You may find that you keep the theory files (.thy) open in another
% window, and continually scan the file for 'clues' to performing
% the current proof, or for inference rules that might be useful
% at some point.
%
% If you have an inference rule in mind, but can't remember its name
% the rule browser can save a great deal of time:

theorem rules_demo === (A => true) and (B <=> B).

prove_rule(rules_demo).

% At this point we may have forgotten the and_intro rule,
% so we will attempt to search for it;
%
% We know the rule has a conclusion 'A and B', and two
% premises; 'A' and 'B'.
%
% We can find all rules with a conclusion with a conjunction
% as follows;

rules(concl(A and B)).


% We can restrict this search to rules with only two 
% premises;

rules([concl(A and B),premises(2)]).

% And we've found it!

rule(and_intro).

% Now we need to discharge each remaining node.

rules(concl(A => true)).

% There are two rules which apply in this situation.
%
% Ergo also has a use command, which will apply any rule
% that satisfy a query to rules.

[2].

use([concl(A => true),frozen(true)]).

% Notice that it has applied the first rule that matched
% the search. If we wanted the second rule, we can
% use the retry command, which causes the current
% result to fail, and forces Ergo to backtrack and search
% for a different solution.
% NOTE: The Heuristic unification warning occurs because some rule
% is tried that generates a delayed unification problem and when
% a solution is attempted it reaches the heuristic depth limit.
% When the depth limit is reached this warning is produced and the
% attempted unification fails.
% A delayed unification  problem is produced when a unification is attempted
% that may have more than one solution. Rather than trying to guess which
% is the required solution, the system delays the unification problem in
% the hope that some future computation will cause the problem to be
% sufficiently simplified to have either a unique solution, or the unification
% fails. If this does not happen then the user is eventually required to
% find a solution. The system is supplied with a heuristic that can be
% used to find unifiers for such delayed problems. We will see examples of this
% later.

retry.

% At this point, the rule we wanted has been applied, so
% we proceed;
%
% This time we will search for a rule that proves the current
% goal.

[3].
use([goal, premises(0)]).

% And we've completed the proof.

show_subproof(1).

save_postulate_proof(1).

quit_proof.

% More information about the query language for rules can
% be found in help;

help(query_language).


%%
%% Quantifiers, substitutions and Unification.
%%

% Theories in Ergo can make use of quantifiers and substitutions.
% cpred is a theory of classical predicate calculus, and as such
% makes use of these capabilities of Ergo.
%
% Qu-Prolog references (particularly the Qu-Prolog user manual)
% provide detailed discussion of the properties of quantifiers, 
% substitutions and the unification algorithms. Because Ergo is 
% implemented in Qu-Prolog, these details are valuable for serious
% Ergo users.
% While reading these references it may be useful to note that 
% variables present in the theorem statement are 'frozen' when
% a proof is started, but variables introduced during a proof 
% are free to be unified (ie. they're 'thawed').
%
% Ergo (and hence Qu-Prolog) variables are really meta-variables and
% are divided into 2 groups. The first group are like standard Prolog
% variables and the second group are specific to Qu-Prolog. They are
% called object variables and are really meta-variables that range over
% variables at the object level. Consequently object variables can only
% be instantiated to other object variables.

% Object variables normally need to be escaped with a single
% exclamation mark (eg. !x), but some may have been predeclared 
% in the current theory, for convenience.

show(object_vars).


% We'll now attempt a proof of the following theorem.

theorem ex_not_not__not_all_not ===
         ex x (not not A)
         ------------------
         not (all x (not A)).

% This theorem has a premise. To prove the rule, we must
% reach a point where the only open node matches the premise.
%
% The simplification rule not_all_iff_ex_not appears to
% be suitable for this proof.

show_rule(not_all_iff_ex_not).

% And so we begin;

prove_rule(ex_not_not__not_all_not).

% not_all_iff_ex_not has no premises, hence we need to apply some other
% inference rules so that we can use not_all_iff_ex_not as
% a way of transforming the goal (instead of just discharging 
% a node).

rule(r__iff__l).

% Note that this rule has introduced a new variable, B.
% If B was in the form 'ex [x0] not not A', then we could
% use not_all_iff_ex_not to dischage the second open node.
%
% Fortunately, a process called unification allows us to
% directly apply this rule. The two terms will be unified
% so that they are equal before the rule is applied.
%
% Before applying the rule, we will examine it again,

show_rule(not_all_iff_ex_not).

% Notice that this time, the rule is expressed in terms 
% of C (instead of A). This is because a variable in an
% inference rule is free to be unified (ie. made equal)
% with any variable or term (subject to the conditions
% in the provided clause). The output routines do not
% use the letter A because this variable is already in 
% use in the proof.
%
% Variables present in the original problem to be proved
% are 'frozen', that is, they cannot be unified with anything
% (except themselves). Any other variable introduced in a
% proof is 'fair game' for unification.
%
% So, we can apply the rule:

[one, rule(not_all_iff_ex_not)].

% 'B' has been instantiated to 'ex [x1] not (not [x1/x0]A)'.
% This has made it possible to discharge the second open node
% and now the node that was 'B' is now 'ex [x1] not (not [x1/x0]A).
%
% When unifying quantified terms (as above) the unification
% algorithm usually does a `change of bound variable' as
% part of the process. This change of bound variable is reflected
% in the proof by the appearance of the substituted term [x1/x0]A.
%
% In most cases, rules involving quantified terms are applied
% in cases where change of bound variables are not required to
% produce a unifier.
%
% We can redo the above step using a variant of rule (srule)
% that first attempts to unify proof terms as structures.
% In this case the bound variables of the goal and the inference
% rules are unified and so no change of bound variable is required.
%
% If the terms do not unify structurally then srule reverts to 
% normal unification.

undo.
[one, srule(not_all_iff_ex_not)].

% In this case the substitution does not appear because the
% terms unified structurally.
% When using rules involving quantifiers it is usually better to
% use srule than rule because fewer substitutions
% and fewer not_free_in constraints are generated.
% This can have a significant effect on the efficiency of proofs
% (as well readability for users).
% NOTE: In some proofs, however, not doing a change of bound variables 
% (by using srule) can cause a proof to become overconstrained by adding
% not_free_in constraints between variables in the statement of the
% theorem. When this happens the proof cannot be saved, and the theorem
% is not proved. When this happens, it is necessary to revert to using rule
% instead of srule at the required points of the proof.
%
% In either case the new goal unifies with the premise, without
% adding additional constraints, so we can save this proof:

save_postulate_proof(1).

quit_proof.

% Next we will look at some caveats with working with quantifiers
% that appear in the following proof;

theorem ex__not_all_not ===
         ex x A
         ------------------
         not (all x not A).

prove_rule(ex__not_all_not).

% Since we've already proven ex_not_not__not_all_not, we can use
% it now;

rule(ex_not_not__not_all_not).

rule(ex_elim).

% ex_elim introduces several constraints to the proof that may
% prevent us from saving the proof (Ergo will notice that the
% proof has been over constrained). This is prevented by applying
% ex_elim as soon as possible (as we have done here).
%
% For now, we'll work on the second open node.

[4].

rule(ex_intro).

rule(id__not_not).

% At this point, we've simplified the 'not not', and we can
% unify with the hypothesis:

rule(assump).

[3].

% If we try to save at this point, Ergo will only return a
% failure message:

save_postulate_proof(1).

% We can get around this problem by unifying the current
% term with the premise, and then retrying any delayed
% unification problems.

term(ex [x0] A).

{incomplete_retry_delays}.

% Now we should be able to save:

save_postulate_proof(1).

quit_proof.

% Note that, although using srule above makes the proof look simpler,
% we still end up with a delayed unification problem that needs to be
% solved.

% As yet another example, we'll try the entire proof
% again, but without making use of not_all_iff_ex_not.
%
% We have to give it a different name, because there is
% already a saved and proven theorem called ex__not_all_not
% (from the previous proof).

theorem ex__not_all_not2 ===
         ex x A
         ------------------
         not (all x not A).


% But beforehand, we'll need another theorem:

theorem deny_not ===
         hyp +++ [A] ---> not A
         ----------------------
         not A.

prove_rule(deny_not).

rule(case_or_not).
[one, rule(assump)].

% 'not A' unifies with 'not B' to instantiate B to A.
% This means that the hypothesis of Node 2 is now A.
%
% And we have then completed the proof.

save_postulate_proof(1).

quit_proof.


% So we begin our proof,

prove_rule(ex__not_all_not2).

% Again, we must apply ex_elim as soon as possible.

rule(ex_elim).


[3].

% Now we'll use our new theorem.

rule(deny_not).


rule(all_elim_c).

% If we add a hypothesis '[C/x0]A', we can discharge
% node 5 (because there would be a contradiction in the
% hypothesis. But, to add it to the hypothesis requires 
% us to prove '[C/x0]A' (which is exactly what we want to
% do).

rule(hyp_add([C/x0]A)).
[rule(contradiction),one].

% Now we can use the rule of assumption to unify B with
% [C/x0]A.

rule(assump).

% If, instead of there being just one thawed variable for 
% unification, there were many; we can specify a particular
% hypothesis number, as follows:

undo.

rule(assump(1)).


[2].

% Ergo won't let us save at this point;

save_postulate_proof(1).

% So, we'll unify it with the premise, and solve any
% delayed unification problems.

term(ex x A).

{incomplete_retry_delays}.

% Constraints associated with proofs can be displayed
% using show constraints.
help show_constraints.

show_constraints.

% At this point, we have a solution which hasn't over-constrained
% the proof, so we can save.

save_postulate_proof(1).

quit_proof.




%%
%% Review.
%%

% If you wish to reinforce these concepts, it may be helpful
% to practice with some very simple proofs:

theorem imp_not_iff_false === 
      A => (not A <=> false).

theorem not_not_not_and_imp_not_or_not === 
      not not (not (A and B) => (not A or not B)).


% A simple, but interesting tactic would split all conjuncts. You might like
% to consider at this point how you would code such a tactic.
% The tactic deep could be used directly to define this tactic, or you
% could use deep as inspiration for writing this tactic.
%
% If you are having trouble, there is already such a tactic defined -
% see split_all_ands in jprop/window_tacs.gum

% Declare and define a tactic called split_all_conjuncts.

% Debugging a tactic usually involves inventing proofs to 
% test the tactic.

theorem split_all_conjuncts_test ===
     (true and true) and ((A => A) and ((A <=> A) and (true => true))).

prove_rule(split_all_conjuncts_test).

split_all_conjuncts.

[rule(true),rule(true),rule(imp_reflex),rule(iff_reflex),rule(imp_reflex)].

quit_proof.

% This concludes the introduction to the emacs interface for Ergo.
%
% If you wish to proceed, please open the file tactics.thy now.








