
%%
%% Self-Guided Ergo 5 interaction tour, Part 3. (tactic.thy)
%%

% This is the third part of a self-guided introduction to the
% Ergo proof tool.
%
% At this stage, it is assumed that you are familiar with the emacs
% interface to Ergo, that you can perform some simple proofs and
% write simple tactics, and that you have Ergo loaded with the
% cpred theory.
%
% If Ergo is not currently loaded (with the cpred theory), place
% the cursor at the beginning of the following line, and press F5
% (or C-c C-t).

[cpred].

% This tutorial makes use of a rule proved as an exercise in the
% previous tutorial.
%
% One possible proof follows. Step through this proof (with F5)
% to load the theorem into Ergo, if you have not already added
% (and proven) this inference rule.

theorem imp_not_iff_false_1 === A => (not A <=> false).

prove_rule(imp_not_iff_false_1).

rule(imp_intro).         % "Remove" the outer implication.

rule(r__iff__l).         % A kind of tranformation,
[one,rule(iff_decomp)].  % by unification.

rule(r__or).             % The RHS of the disjunction is obviously
                         % true so we only have to prove that.

rule(and_intro).         % Now we'll prove each side of this
                         % conjunction individually.

[6].
rule(r__iff__l);         % We "transform" 'not not A' into 'A',
[one,rule(not_not)].
rule(assump).            % And match with a hypothesis.

[7].
rule(not_false1).

save_postulate_proof(1).

quit_proof.

%%
%% Window Inference and other System Tactics
%%

%%
%% Contents of this file.
%%

% Window inference "recap".
% Other system tactics.
% System tactics "in action".
% Review.

%%
%% Window inference "recap".
%%

% Instead of forming a proof tree, a window inference proof is
% based on the idea of a stack. The transformational nature of 
% a window inference proof is a powerful concept that can
% be applied to Ergo 5 style proofs, and some familiarity with the
% window inference tactics that come with the system can help
% simplify difficult proofs.
%
% In a window inference proof, we have the idea of a stack of
% windows. Each window consists of a term and a window relation 
% (which can be any reflexive, transitive relation).
%
% The term that appears in the window at the top of the stack is 
% called the focus.
%
% In window inference,
%   -  we can apply a transformation rule, which transforms the 
%      focus using a rule in the current theory.
%   -  we can open a window on a subterm: this pushes a new window
%      onto the stack so that the focus is now the subterm
%   -  we can close a window: this 'pops' the top frame and
%      replaces the old subterm with the newly transformed subterm.
%
% Ergo is distributed with a tactic file, window_tacs.gum, in jprop
% that declares many tactics for window inference. Most of the system theories
% already include the theorems required to make the window inference
% tactics possible.
%
% Users familiar with Ergo 4 should find the transition to Ergo 5
% relatively straightforward.
%
% To use window inference, we must load the tactic file.

load_tactic_file(window_tacs, jprop).

% In Ergo, a window is represented as a single node of the
% following form;
%
%        F WR G
%
% Where WR is the window relation, F is the focus and G is
% typicall an unbound variable used as a placeholder for a 
% transformed term.
%
% For example, we might want to prove that we can transform
% ((A and false) and (B and not A)) into false under the
% window relation <=>.
% (ie. we want to show that they are equivalent).

theorem first_window === 
    ((A and true) and (B and not A)) <=> false.

% Notice that this can already be considered a window. We have
% the LHS of the equivalence as our focus, and the RHS is our
% eventual transformed term.

prove_rule(first_window).

% The outermost operator in the focus is the 'and'. Since it
% has two subterms (ie. it is a binary operator) we have two 
% subterms that we can "open into".
%
% The 'op' tactic is used to open a window on a subterm, its
% argument is the path to the subterm.

op([1]).

% This has opened onto the first subterm of the focus (ie. the
% left hand side of the conjunction). The first open node (node 4)
% contains the new window that is at the top of the window stack.
% If we look at the LHS of the window relation in this node, we
% see that the focus is now the subterm which we opened onto.
% (ie. the LHS of the original focus (A and false), has become
% the current focus.
%
% The RHS of node 4 is a new variable. This variable now also 
% appears in the original window in place of the subterm which
% we had opened into.
%
% At this point, we can apply a transformation rule.

trans(rule(and_true)).

% The trans rule (or its variants) can only use inference 
% rules that have no premises (in the Ergo 5 sense of the word).

show_rule(and_true).

% This rule could be read as saying that under the window
% relation, logical equivalence, 'D and true' can be transformed
% to 'D'.
% 
% The window inference tactics make use of the transitive property
% of the window relation to perform this step. If the inference
% rule uses a relation that is stronger than the current window
% relation, the tactic will automatically make the current 
% relation sufficiently stronger to apply the rule.
%
% At this point we can close the window;

close_win.

% Behind the scenes, close_win simply applies a reflexive rule.
% This causes the 'place holder' variable to be instantiated
% to the transformed focus, and closes the node. Notice now
% the LHS of the conjunction in the focus is the
% transformed subterm.
%
% Proceeding;

op([2]).
op([2]).

% In fact, these two commands can be merged into one;

undo.
undo.

op([2,2]).

% Notice that as a window is opened, extra hypotheses are
% "picked up".
%
% Now we'll use a theorem we've proven previously.

trans_ante(rule(imp_not_iff_false_1), rule(assump)).

% The trans_ante tactic applies a rule that has antecedents,
% and takes a tactic (typically a list of inference rules)
%  that it should use to discharge the antecedents.
%
% Inference rules used by Ergo to carry out window transformations
% are typically of one of the
% following forms.
%
%          F WR G    (trans)
%    A => (F WR G)   (trans_ante)
%               G    (trans_t)
%
% The window inference tactics will automatically attempt to
% reverse or strengthen the window relation, so that inference
% rules need not exactly match the above pattern. (ie. a rule of
% the form G WR2 F would still be perfectly suitable to use
% with trans, provided WR2 is the 'reverse' of the window relation).

close_win.

op([2]).

trans(rule(and_false)).

close_win.

trans(rule(and_false)).

% At this point we have completed the window inference, we can
% apply the reflexive rule of logical equivalence, or use the
% close_win tactic to do this for us (despite the fact that we
% didn't technically open a corresponding window in the first 
% place).

close_win.

save_postulate_proof(1).

quit_proof.

% The window inference tactics are reasonably powerful tools
% for use in complex proofs. A hybrid style of proof that uses both
% window inference and sequent calculus styles can be very effective. 
% For example, using close_win as above is
% not technically part of the window inference proof, but in
% this situation proved to be a useful strategy.
%
% Another example:

theorem second_window ===
   all x (x or not x).

prove_rule(second_window).

% In this case, the window inference tactics might be helpful
% but they can't be applied directly because we're in a 
% sequent calculus style proof.
%
% 'second_window' can be proven in many ways, but we'll use
% window inference. The start_window_proof tactic is helpful here:

start_window_proof.

% This tactic transforms a goal G into the equivalent goal
% G <= true   (<=  (follows) is the reverse of => (implication)).
% This is now in the form
% Focus WR Result
% which makes it a window.
%
% During a "pure" windows style proof we might want to display only
% information about the "window at the top of the stack". 
% We can change the display to do this as follows.
% At any time we can revert to the standard display with stop_wi.
% WARNING : before the proof is finished you will need to use
% stop_wi otherwise this display mode will carry over to the next
% proof which may not display correctly.

start_wi.

% Now we can open on the body of the quantified term,

op([b([x0])]).

% This time, the path to the subterm included a directive
% to open on the body of a quantified term, using the bound
% variables given as an argument.
%
% Or more simply:

undo.

op([x0]).

trans_t(rule(or_not)).

% trans_t is used for transforming a window to true.
%
% Notice that the inference rule or_not has no antecedents
% nor any kind of relation.

show_rule(or_not).

% Now we can close the window, and complete the proof.

close_win.

trans(rule(all_true)).

close_win.

% Change back to standard interface (for next proof).

stop_wi.

save_postulate_proof(1).

quit_proof.

% For a detailed understanding of the window inference tactics
% take some time to read through the source and comments of 
% the window_tacs.gum file in the jprop directory.
% 
% For more information about window inference proofs, you may
% wish to read the seminal paper about window inference: 
% `Formalizing a Hierarchical Structure of Practical Mathematical
% Reasoning', by Peter J. Robinson and John Staples, Journal of 
% Logic and Computation, Volume 3, Number 1, 1993, pages 47-61.           
%
% One last note on (non-hybrid) window inference:
% There are only two ways to add additional hypotheses
%   - use the add_hyp tactic which performs forward inference
%     on the hypotheses directly.
%   - open on a subterm
%
% This means that to add the LHS of an implication to the hypotheses,
% in a window inference proof, a window is opened on the RHS of the
% implication. The LHS of the implication will appear as a hypothesis
% in the subwindow (the opening rule is used in window inference, 
% despite the fact that in an Ergo 5 proof, imp_intro would be used).


%%
%% Other system tactics.
%%

% Ergo is distributed with with a large collection of system
% tactics. These tactics can simplify common proof strategies, 
% and may be useful examples for the creation of your own tactics.
%
% We will examine a few of the more interesting tactics, but
% the best references are the tactic files themselves, which
% include some documentation on the use of each tactic.
%
% A tactic designed for a particular theory can also be used in any
% theory that inherits from the original theory for which the tactic
% was defined.
%
% Some of the most useful tactic files that are available are listed below.
% Most of these files provide sequent calculus (SC) and window 
% inference (WI) versions of each tactic.
%
% global:
%     tacticals.gum   A collection of tactics for combining tactics.
% jprop:
%     window_tacs.gum Tactics to support window inference in Ergo
%     init_proof.gum  Initialises a proof that has premises.
%     simplify.gum    Simplifies the focus or current term.
%     fwd_hyps.gum    Performs forward inference on hypotheses.
%     horn.gum        Adds hypotheses, using horn clauses. (SC only)
%     ac.gum          Rewrites associative-commutative expressions. (WI only)
%     later.gum       Helps a proof appear to be a DAG instead of a tree.
%     unfold.gum      Unfolds/folds definitions in subterms.
%     elf.gum         Tactic to be run  before user interactions.
% iprop:
%     window_tacs.gum Tactics to do window transformations w.r.t. false
% cprop:
%     tableaux.gum    Performs an automatic tableau proof. (SC only)
%
% For a complete overview of the range of tactics available,
% browse the directory structure located under $ERGOHOME/theories
% for files with the .gum extension.
%
% Some tactics are "available" in both window inference versions
% (usually suffixed with '_wi') and sequent calculus versions.
% Other tactics are clearly of a transformational or direct proof
% nature and so have only been written in one style.
% This is typically not a problem, particularly if a hybrid
% approch to the use of the window inference tactics is being used.


%%
%% System tactics "in action".
%%

% init_proof prepares a proof that has premises, so that at the
% conclusion of the proof, all the premises are in the correct
% order.
%
% For example,

load_tactic_file(init_proof, jprop).

theorem init_proof_demo ===
       A,
       B
       ---------------
       (A and B) and A.

% Now we'll prove this, without the init_proof tactic.
% Our aim is to have two open nodes, 'A' and 'B', in that order.

prove_rule(init_proof_demo).

rule(and__and).
rule(and_and__and_and).

rule(and_intro).

[4].
rule(r__iff__l).
[one,rule(and_idempotent)].

% Even for this simple proof, it requires some careful fiddling
% and forethought to get the proof right.
%
% We'll try again, with the init_proof tactic.

quit_proof.

prove_rule(init_proof_demo).
init_proof.
[4].
rule(and_intro).
[rule(and_intro), one].
[rule(assump),rule(assump),rule(assump)].

% This time, while the proof may have been longer, it was
% far more straightforward. It consisted of an obvious application
% of and_intro, followed by assump -- no tricky manipulation
% using the associative and commutative laws was necessary.
%
% The tacticals.gum tactic file could make this proof even easier:

quit_proof.

load_global_tactic_file(tacticals).

prove_rule(init_proof_demo).
init_proof.
[4].
deep(rule(and_intro)).
many(rule(assump)).

% The deep tactic is a meta-tactic that applies the argument 
% to every node, and then applies it again to any newly created nodes.
% The many tactic simply attempts to apply a tactic to as _many_
% open nodes as it can.

save_postulate_proof(1).

quit_proof.

% Other tactics similarly embody common proof strategies,
% in fact, for certain classes of problems, tactics can
% be written for automatic proofs.
%
% The tableaux.gum tactic can automatically prove most
% theorems in classical propositional logic (provided
% they _can_ indeed be proven).

load_tactic_file(cprop_tableaux,cprop).

theorem tableaux_theorem ===
      (A and B <=> false and B) or (A <=> true) and ((C => B) or (B => C)).

prove_rule(tableaux_theorem).

tableau.

save_postulate_proof(1).

quit_proof.

% The proof constructed by the tactic can be viewed as follows.

show_proof(tableaux_theorem).

%
% In some earlier proofs we used a rule to allow us to transform a focus
% to false. This process can be simplified by loading the tactic file
% iprop/window_tacs.gum. This tactic file defines tactics for transforming
% the focus to false.
%
% This means that if you are in a theory that inherits from iprop then it
% would be better to load the iprop version of window_tacs.gum if you
% intend to do window inference. This gives you access to both the jprop
% and iprop tactics for window inference.
%

%
% The elves tactic.
%
% In some of the window inference style proofs above you will have noticed
% that we needed to apply several simplification rules that rewrote formulae
% containing true (or false) as subformulae.
%
% To simplify this we have reused an idea from Ergo4: elves.
%
% In Ergo, the elves tactic (defined in jprop/elf.gum) is a tactic that,
% is executed before each user interaction in the
% Gumtree interpreter. This is a table driven tactic based on two tables - one
% for window inference and one for sequent calculus. The window inference
% table is already populated with several simplifications.
%
load_tactic_file(elf, jprop).
help(elves).
help(elves_type).

% System declared table entries for window inference can be found in 
% most theories.
load_tactic_file(elf_entries).
show_tactic_table(elves_table_wi).

% To use the window proof elves:
set elves_type === window.


%%
%% Review.
%%

% The most frequently used meta-tactics are found in tacticals.gum. 
%
% The sources and comments in these files are essential reading
% if you plan to use Ergo for anything but the simplest proofs.


% This concludes the introduction to the Ergo system tactics.
%
% If you wish to proceed, please open the file advanced.thy now.

abandon.












