;; Major modes for interacting with an inferior Ergo process
;; and for editing theory files.

;; $Id: ergo-mode-svrc.el,v 2.3 2003/05/08 22:03:08 itp Exp $

(provide 'ergo-mode)
(require 'comint)
(require 'assoc)

(defconst ergo-mode-version "$Revision: 2.3 $")


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; User options
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar ergo-program "ergo"
  "*Command to invoke to run Ergo.")

(put 'ergo-program 'variable-interactive "FErgo program: ")

(defvar ergo-theory-mode-hooks nil
  "*Hooks run after starting Ergo theory mode.")

(defvar inferior-ergo-mode-hooks nil
  "*Hooks run after starting an inferior Ergo.")

(defvar ergo-debug nil
  "*Flag controlling whether the Ergo filter is being debugged.
If non-nil, all filtered output is written into the *Ergo Debug* buffer.")
(put 'ergo-debug 'variable-interactive
     '(list (y-or-n-p "Debug ergo? ")))

(defvar ergo-active-face 'bold
  "*What face should we use for mouse-active regions?")
(put 'ergo-active-face 'variable-interactive
     '(list (read-face-name "Face for mouse-active Ergo text: ")))

(defvar ergo-selected-face 'highlight
  "*What face should we use for selected mouse-active regions?")
(put 'ergo-active-face 'variable-interactive
     '(list (read-face-name "Face for mouse-selected Ergo text: ")))

(defvar ergo-emph-face
  (if (face-differs-from-default-p 'italic)
      'italic
    'region)
  "*What face should we use for emphasized text?")
(put 'ergo-emph-face 'variable-interactive
     '(list (read-face-name "Face for emphasized Ergo text: ")))

  
(defvar ergo-easy-menu-max-pane 24
  "*Maximum number of entries to put on a single popup-menu pane.")

(defvar ergo-do-mouse-bindings t
  "*Should Ergo set up its own mouse bindings?")

(defvar ergo-logging t
  "*Should we log commands in proof scripts?")

(defvar ergo-help-ring-max 20
  "*Maximum number of help buffers to retain in the ring.")

(defvar ergo-proof-frame-alist nil
  "*Alist of values for proof frame creation.
If not set, this will be filled in by looking in the X resources database
for keys `Emacs.ergo-proof.geometry' and `Emacs.ergo-proof.font'.")

(defvar ergo-proof-script-frame-alist nil
  "*Alist of values for proof script frame creation.
If not set, this will be filled in by looking in the X resources database
for keys `Emacs.ergo-proof-script.geometry' and 
`Emacs.ergo-proof-script.font'.")

(defvar ergo-help-frame-alist nil
  "*Alist of values for help frame creation.
If not set, this will be filled in by looking in the X resources database
for keys `Emacs.ergo-help.geometry' and `Emacs.ergo-help.font'.")

(defvar ergo-rules-frame-alist nil
  "*Alist of values for rules frame creation.
If not set, this will be filled in by looking in the X resources database
for keys `Emacs.ergo-rules.geometry' and `Emacs.ergo-rules.font'.")

(defvar ergo-browser-frame-alist nil
  "*Alist of values for browser frame creation.
If not set, this will be filled in by looking in the X resources database
for keys `Emacs.ergo-browser.geometry' and `Emacs.ergo-browser.font'.")

(defvar ergo-proof-frame-panes
  nil
  "*List specifying the panes of the proof frame.
The format is a list of row-entries describing rows.
Each row-entry is a dotted pair whose car gives the height of the pane and
whose cdr is either a string, naming a pane, or a list of column-entries.
Each column-entry is a dotted pair whose car gives the width of the pane and
whose cdr is the name of the pane.
Heights and widths are specified as a proportion of the total height/width of
the frame.
The pane names are:
  \"subproof_stack\"
  \"hypotheses\"
  \"proof_control\"
  \"subproof\").
If this variable is nil, its value will be set from the X resource
`Emacs.ergo-proof-frame.panes', if it exists, or to a suitable default value
(ergo-default-proof-frame-panes).")
(put 'ergo-proof-frame-panes 'variable-interactive
     '(list (read-from-minibuffer
               "Proof frame pane list: "
               (format "%S" (or ergo-proof-frame-panes
                                ergo-default-proof-frame-panes))
               nil
               'read
               nil)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Major global variables
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defconst ergo-prompt-regexp
  (format "^\\(%s\\|%s\\|%s\\|%s\\|%s\\) *"
          "No [Tt]heory:"                 ; theory management: 'No theory: '
          "\\[+['a-z][A-Za-z0-9_' ]+\\]+:"; theory management: '[[name]]: '
          "[0-9][0-9a-zA-Z]*%"            ; Ergo 4 proofs:     'num% '
          "\\[[0-9A-Z, ]*\\]:::"          ; Ergo 5 proofs:     '[Nodes]::: '
          "\\(\\[[^]]+\\] *\\)?|? ?\\?-"  ; prolog:            '[b1i1] | ?- '
          )
  "Regexp matching any of Ergo's prompts.")

(defvar ergo-inferior-font-lock-keywords
  (list
   (list ergo-prompt-regexp
            '(0 font-lock-type-face))
  )
  "Font lock keywords for the *ergo* buffer.")

(defconst ergo-error-regexp
  "\\([Ee]rror:\\|(failed)\\|failed:\\)"
  "Regexp matching any of Ergo's error messages.")

(defconst ergo-magic-table
  '(("NORMAL"         ergo-normal-output-handler partial)
    ("FOCUS"          ergo-focus-handler)
    ("INDEX"          ergo-index-handler)
    ("THEORIES"       ergo-theories-handler)
    ("HELP"           ergo-help-handler)
    ("XREF"           ergo-xref-handler local)
    ("EMPH"           ergo-emph-handler local)
    ("QUERY"          ergo-query-handler)
    ("RULES"          ergo-rules-handler)
    ("SYNC"           ergo-sync-handler)
    ("HYPS"           ergo-hypotheses-handler)
    ("BROWSEANCS"     ergo-browse-ancs-handler)
    ("BROWSEHYPS"     ergo-browse-hyps-handler)
    ("BROWSESUBP"     ergo-browse-subp-handler)
    ("BROWSEQUIT"     ergo-browse-quit-handler)
    ("SUBPROOF"       ergo-subproof-handler)
    ("CONTEXT"        ergo-subproof-stack-handler)
    ("PARAM"          ergo-param-handler)
    ("STARTPROOF"     ergo-start-proof-handler)
    ("ENDPROOF"       ergo-end-proof-handler)
    ("WINDOW"         ergo-window-handler)
    ("BUTTON"         ergo-button-handler)
    ("HINTS"          ergo-hints-handler)
    ("LOG"            ergo-log-handler)
    ("EDITELVES"      ergo-edit-elves-handler)
    ("SAVEELVES"      ergo-save-elves-handler)
    ("DIALOG"         ergo-dialog-handler)
    ("DAVINCI"        ergo-davinci-handler)
    ("SENDSTRING"     ergo-sendstring-handler)
    ("PROOFSCRIPT"    ergo-proofscript-handler)
    ;;("SYMBOLS"        ergo-symbols-handler)
    ("END"            ergo-end-handler endlocal)
    )               
  "Table of (STRING FUNCTION . OPTIONS) records for handling MAGIC commands.
When a command is received, the current function is called with arg `nil', then
the new function is called with arg `t'.  Non-magic output is passed to the
current function as a string.
Most functions will receive whole lines only; an option 'partial means the
function will accept partial lines.
An option 'local means that the handler should be enabled locally; after
MAGIC END, the previous handler should resume.")

(defconst ergo-noise-table
  '(("\\[c\\]ontinue/.*: " . abort)
    )
  "Table of (REGEXP . ACTION) pairs for handling `noise' in MAGIC output.
If a match for any REGEXP occurs while we are handling magic, the corresponding
ACTION will be taken.
Possible ACTION values are:
  discard -- discard the offending string entirely.
  display -- display the string (and don't deliver it to the handler), but
             continue processing magic as before.
  abort   -- revert to the normal handler.
  (call FUNCTION) -- call FUNCTION with a single argument: the string.  After
     FUNCTION returns, dispatching of the output following the string continues
     (though FUNCTION may change the magic state)."
)

(defconst ergo-theorem-regexp
  (concat
   "^\\(theorem\\) +"
   "'?\\([^ \t\n']+\\)'?[ \t\n]+==="
   "\\([^.]\\|\\.[^\n]\\)*\\.$")
  "Regular expression matching an Ergo theorem statement.  The second
  parenthesized component should match the theorem name.")

(defconst ergo-tactic-file-commands
  '(("add_tactic_file")
    ("add_global_tactic_file")
    ("recompile_tactic_file")
    ("recompile_global_tactic_file")
    ("reconsult_tactic_file")
    ("reconsult_global_tactic_file")))

(defvar ergo-buffer nil
  "The buffer to use for Ergo output.")

(defvar ergo-process nil
  "The process to send Ergo input to.")

(defvar ergo-current-theory nil
  "Emacs's best guess at Ergo's current theory, or nil.")

(defvar ergo-output-handlers '((ergo-normal-output-handler partial))
  "How should we deal with Ergo output?
A list whose car is a (FUNCTION . OPTIONS) pair; FUNCTION will receive the
  output. 
An entry is pushed according to the ergo-magic-table upon seeing a MAGIC
command, and popped on seeing an ENDMAGIC command.")

(defvar ergo-magic-disabled nil
  "Non-nil if magic processing is completely disabled.
This can be set by calling `ergo-toggle-magic-processing' with a negative
argument.")

(defvar ergo-selection-text nil
  "The text that \\[ergo-mouse-insert-selection] will insert.
If nil, \\[ergo-mouse-insert-selection] will insert the X selection.")

(defvar ergo-browser-ancs-buffer nil
  "The buffer used to browse subproof ancestors.")

(defvar ergo-browser-hyps-buffer nil
  "The buffer used to browse subproof hypotheses.")

(defvar ergo-browser-subp-buffer nil
  "The buffer used to browse subproof steps.")

(defvar ergo-browser-frame nil
  "The frame used for the browser.")

(defvar ergo-help-buffers nil
  "Ring of buffers for displaying Ergo help.")

(defvar ergo-help-window nil
  "The window used to display Ergo help.")

(defvar ergo-subproof-stack-buffer nil
  "The buffer used to display current subproof stack.")

(defvar ergo-proof-control-buffer nil
  "The buffer used to display proof control information.")

(defvar ergo-subproof-buffer nil
  "The buffer used to display the current subproof.")

(defvar ergo-hypotheses-buffer nil
  "The buffer used to display current hypotheses.")

(defvar ergo-proof-script-buffer nil
  "The buffer currently being used to build a proof script.")

(defvar ergo-proof-script-window nil
  "The window used to display the current proof script.")

(defvar ergo-current-window-level 0
  "Emacs's idea of the current window level (an integer).
This is used to determine indentation in the proof script file.")

(defvar ergo-current-window-id "?"
  "Emacs's idea of the current window ID (a string).")

(defvar ergo-last-focus-select nil
  "The active Ergo focus selection.
This variable is nil if there is no selection, otherwise it is a cons
whose car is the place clicked, and whose cdr is the selected path.")


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Keymaps
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defvar ergo-menu (make-sparse-keymap "Ergo")
  "Keymap used for Ergo menu.")

(defvar ergo-common-map (make-sparse-keymap)
  "Keymap for bindings common to all Ergo buffers.")
(define-key ergo-common-map [down-mouse-1] 'ergo-mouse-select)
(define-key ergo-common-map [mouse-1]  'ignore)
(define-key ergo-common-map "\C-hm"    'ergo-describe-mode)
(define-key ergo-common-map "\C-c\C-c" 'ergo-interrupt)
(define-key ergo-common-map "\C-c\C-e" 'ergo-end-of-buffer)
(define-key ergo-common-map "\C-c\C-l" 'ergo-refresh-interface)
(define-key ergo-common-map "\C-c\C-w" 'ergo-toggle-wrapping)
(define-key ergo-common-map "\C-c\C-d" 'run-ergo)
(define-key ergo-common-map "\C-hm"    'ergo-describe-mode)
(define-key ergo-common-map [menu-bar ergo] (cons "Ergo" ergo-menu))

(defvar ergo-readonly-common-map (cons 'keymap ergo-common-map)
  "Keymap for bindings common to all read-only Ergo buffers.")
(define-key ergo-readonly-common-map "<" 'ergo-scroll-left)
(define-key ergo-readonly-common-map ">" 'ergo-scroll-right)
(define-key ergo-readonly-common-map "h" 'ergo-describe-mode)

(defvar ergo-theory-mode-map (cons 'keymap ergo-common-map)
  "Keymap for editing theory source files and proof scripts.")
(define-key ergo-theory-mode-map "\C-c\C-b" 'ergo-send-buffer)
(define-key ergo-theory-mode-map "\C-c\C-r" 'ergo-send-region)
(define-key ergo-theory-mode-map "\C-c\C-y" 'ergo-send-rest-of-buffer)
(define-key ergo-theory-mode-map "\C-c\C-t" 'ergo-send-term)
(define-key ergo-theory-mode-map "\C-c\C-p" 'ergo-send-previous-term)
(define-key ergo-theory-mode-map "\t"       'indent-according-to-mode)
(define-key ergo-theory-mode-map "\M-\C-a"  'ergo-backward-term)
(define-key ergo-theory-mode-map "\M-\C-e"  'ergo-forward-term)
(define-key ergo-theory-mode-map "\C-c\C-s" 'ergo-find-this-proof-script)

;; inferior-ergo-mode-map is a copy of comint-mode-map that inherits
;; ergo-common-map.
(defvar inferior-ergo-mode-map (copy-keymap comint-mode-map)
  "Keymap for the Ergo interaction buffer.")
(let ((m inferior-ergo-mode-map))
  (while (consp (cdr m)) (setq m (cdr m)))
  (rplacd m 'ergo-common-map))
(define-key inferior-ergo-mode-map "\C-m" 'ergo-send-input)
(define-key inferior-ergo-mode-map [menu-bar ergo] (cons "Ergo" ergo-menu))

(if (not ergo-do-mouse-bindings)
    nil
  (define-key inferior-ergo-mode-map [mouse-1] 'ignore)
  (define-key inferior-ergo-mode-map [down-mouse-1] 'ergo-mouse-select)
  (define-key inferior-ergo-mode-map [mouse-2] 'ergo-insert-x-sel)
  (define-key inferior-ergo-mode-map [S-mouse-2] 'ergo-insert-sel)
  (define-key inferior-ergo-mode-map [C-S-mouse-2] 'ergo-insert-and-send-sel)
  (define-key ergo-theory-mode-map [S-mouse-2] 'ergo-insert-sel)
  (define-key ergo-theory-mode-map [C-S-mouse-2] 'ergo-insert-and-send-sel)
  ;; global bindings!
  (let ((fb (or (lookup-key global-map [menu-bar file])
                (lookup-key global-map [menu-bar files])))
        (hb (or (lookup-key global-map [menu-bar help])
                (lookup-key global-map [menu-bar help-menu]))))
    (if (keymapp fb)
        (define-key-after fb
          [tear-off]
          '("Tear off window" . ergo-tear-off-window)
          'make-frame
          ))
    (if (keymapp hb)
        (define-key-after hb
          [ergo]
          '("Ergo" . ergo-help)
          'emacs-tutorial
          ))))

(defvar ergo-proof-control-mode-map
  (cons 'keymap (cons "Ergo Proof-Control"  ergo-readonly-common-map))
  "Keymap used for Ergo Proof-Control buffers.")
(define-key ergo-proof-control-mode-map [down-mouse-2] 'ergo-mouse-select)
(define-key ergo-proof-control-mode-map [mouse-2] 'ignore)

(defvar ergo-subproof-mode-map
  (cons 'keymap (cons "Ergo Subproof"  ergo-readonly-common-map))
  "Keymap used for Ergo Subproof buffers.")
(define-key ergo-subproof-mode-map [mouse-2] 'ergo-send-sel)
(define-key ergo-subproof-mode-map "x" 'ergo-expand-subproof-line)

(defvar ergo-hypotheses-mode-map
  (cons 'keymap (cons "Ergo Hypotheses"  ergo-readonly-common-map))
  "Keymap used for Ergo Hypotheses buffers.")
(define-key ergo-hypotheses-mode-map [mouse-2] 'ergo-send-sel)
(define-key ergo-hypotheses-mode-map "x" 'ergo-expand-hypotheses-line)
(define-key ergo-hypotheses-mode-map "d" 'ergo-delete-hypotheses-line)
(define-key ergo-hypotheses-mode-map "D" 'ergo-discharge-hypotheses-line)
(define-key ergo-hypotheses-mode-map "f" 'ergo-hypotheses-focus-hyp)
(define-key ergo-hypotheses-mode-map "u" 'ergo-hypotheses-use-this-hyp)

(defvar ergo-subproof-stack-mode-map
  (cons 'keymap (cons "Ergo Subproof-Stack"  ergo-readonly-common-map))
  "Keymap used for Ergo Subproof-Stack buffers.")
(define-key ergo-subproof-stack-mode-map [down-mouse-2] 'ergo-mouse-select)
(define-key ergo-subproof-stack-mode-map [mouse-2] 'ignore)
(define-key ergo-subproof-stack-mode-map "x" 'ergo-expand-subproof-stack-line)


(defvar ergo-rules-mode-map
  (cons 'keymap (cons "Ergo Rules"  ergo-readonly-common-map))
  "Keymap used for Ergo Rules buffers.")
(define-key ergo-rules-mode-map "u" 'ergo-rules-use-rule)
(define-key ergo-rules-mode-map "f" 'ergo-rules-fwd-rule)
(define-key ergo-rules-mode-map "a" 'ergo-rules-apply-rule)
(define-key ergo-rules-mode-map "r" 'ergo-rules-rule-rule)
(define-key ergo-rules-mode-map "i" 'ergo-rules-irule-rule)
(define-key ergo-rules-mode-map "s" 'ergo-rules-select-rule)
(define-key ergo-rules-mode-map [mouse-2] 'ergo-send-sel)
(define-key ergo-rules-mode-map "p" 'ergo-rules-previous)
(define-key ergo-rules-mode-map "\C-?" 'ergo-rules-previous)
(define-key ergo-rules-mode-map "n" 'ergo-rules-next)
(define-key ergo-rules-mode-map " " 'ergo-rules-next)
(define-key ergo-rules-mode-map "q" 'ergo-kill-current-buffer-and-window)
(define-key ergo-rules-mode-map "\C-c\C-c" 'ergo-rules-stop-search)

(defvar ergo-selected-rule-mode-map
  (cons 'keymap (cons "Ergo Selected-Rule"  ergo-readonly-common-map))
  "Keymap used for Ergo Selected-Rule buffers.")
(define-key ergo-selected-rule-mode-map "\C-c\C-c" 'ergo-rules-stop-search)

(defvar ergo-help-mode-map
  (cons 'keymap (cons "Ergo Help"  ergo-readonly-common-map))
  "Keymap used for Ergo Help buffers.")
(define-key ergo-help-mode-map [down-mouse-2] 'ergo-mouse-select)
(define-key ergo-help-mode-map [mouse-2] 'ignore)
(define-key ergo-help-mode-map "p" 'ergo-help-previous)
(define-key ergo-help-mode-map "n" 'ergo-help-next)
(define-key ergo-help-mode-map "k" 'ergo-help-kill)
(define-key ergo-help-mode-map " " 'scroll-up)
(define-key ergo-help-mode-map "\C-?" 'scroll-down)

(defvar ergo-browser-mode-map
          (cons 'keymap (cons "Ergo Browser"  ergo-readonly-common-map))
  "Keymap used for Ergo Browser buffers.")
(define-key ergo-browser-mode-map [down-mouse-2] 'ergo-mouse-select)
(define-key ergo-browser-mode-map [mouse-2] 'ignore)
(define-key ergo-browser-mode-map " " 'scroll-up)
(define-key ergo-browser-mode-map "\C-?" 'scroll-down)

(defvar ergo-query-template-mode-map
          (cons 'keymap (cons "Ergo Query Template"  ergo-common-map))
  "Keymap used for Ergo Query-Template buffers.")
(define-key ergo-query-template-mode-map "\C-k" 'ergo-query-delete-field)
(define-key ergo-query-template-mode-map "\C-n" 'ergo-query-next-field)
(define-key ergo-query-template-mode-map [return] 'ergo-query-next-field)
(define-key ergo-query-template-mode-map [down] 'ergo-query-next-field)
(define-key ergo-query-template-mode-map "\C-p" 'ergo-query-previous-field)
(define-key ergo-query-template-mode-map [up] 'ergo-query-previous-field)
(define-key ergo-query-template-mode-map "\C-c\C-c" 'ergo-rules-stop-search)
(substitute-key-definition 'ergo-send-buffer 'ergo-query-send
                           ergo-query-template-mode-map ergo-theory-mode-map)
(substitute-key-definition 'ergo-send-region 'ergo-query-send
                           ergo-query-template-mode-map ergo-theory-mode-map)
(substitute-key-definition 'ergo-send-term 'ergo-query-send
                           ergo-query-template-mode-map ergo-theory-mode-map)
(substitute-key-definition 'ergo-send-previous-term 'ergo-query-send
                           ergo-query-template-mode-map ergo-theory-mode-map)
(substitute-key-definition 'undo 'ergo-query-undo
                           ergo-query-template-mode-map global-map)
(substitute-key-definition 'advertised-undo 'ergo-query-undo
                           ergo-query-template-mode-map global-map)

(define-key ergo-menu [magic]
  '("Toggle Magic Processing" . ergo-toggle-magic-processing))
(define-key ergo-menu [logging]
  '("Toggle Logging" . ergo-toggle-logging))
;;ERGO4;; (define-key ergo-menu [elves]
;;ERGO4;;  '("Edit Elves" . ergo-edit-elves))
(define-key ergo-menu [send-buffer]
  '("Send Entire Buffer" . ergo-send-buffer))
(define-key ergo-menu [send-rest]
  '("Send Rest of Buffer" . ergo-send-rest-of-buffer))
(define-key ergo-menu [send-region]
  '("Send Region" . ergo-send-region))
(define-key ergo-menu [refresh]
  '("Refresh Interface" . ergo-refresh-interface))
;;ERGO4;; (define-key ergo-menu [default-proof-frame]
;;ERGO4;;   '("Show Default Proof Frame" . ergo-show-default-proof-frame))
(define-key ergo-menu [cd] '("Change Theory" . ergo-cd))

;(define-key ergo-menu [tactics] '("Tactics" . ergo-op-tactic))

(define-key ergo-menu [show]
  (cons "Show" (make-sparse-keymap "Show")))

;;;;(define-key ergo-menu [show other]
;;;;  (cons "Other Theory" (make-sparse-keymap "Other Theory")))
;;(define-key ergo-menu [show current]
;;  (cons "Current Theory" (make-sparse-keymap "Current Theory")))
;;(define-key ergo-menu [show global]
;;  (cons "Global" (make-sparse-keymap "Global")))

;;(define-key ergo-menu [show global tactics]
;;  (cons "Tactics" (make-sparse-keymap "Tactics")))
;;(define-key ergo-menu [show global syntax]
;;  (cons "Syntax and I/O" (make-sparse-keymap "Syntax and I/O")))
;;(define-key ergo-menu [show global thy]
;;  (cons "Theory management" (make-sparse-keymap "Theory management")))

;;(define-key ergo-menu [show current tactics]
;;  (cons "Tactics" (make-sparse-keymap "Tactics")))
;;(define-key ergo-menu [show current syntax]
;;  (cons "Syntax and I/O" (make-sparse-keymap "Syntax and I/O")))
;;(define-key ergo-menu [show current thy]
;;  (cons "Theory management" (make-sparse-keymap "Theory management")))
;;(define-key ergo-menu [show current windows]
;;  (cons "Window inference" (make-sparse-keymap "Window inference")))

;;ERGO4;; (define-key ergo-menu [show windows]
;;ERGO4;;   (cons "Window inference" (make-sparse-keymap "Window inference")))
(define-key ergo-menu [show tactics]
  (cons "Tactics" (make-sparse-keymap "Tactics")))
(define-key ergo-menu [show syntax]
  (cons "Syntax" (make-sparse-keymap "Syntax")))
(define-key ergo-menu [show thy]
  (cons "Theory management" (make-sparse-keymap "Theory management")))

(defmacro define-ergo-show (command submenu name)
  "Define show COMMAND in SUBMENU with NAME."
  (let (fn deffn)
    (setq fn (intern (format "ergo-send-show-%s" command))
          deffn (` (defun (, fn) ()
                     (, (format "Send `show(%s)' to the Ergo process."
                                command))
                     (interactive)
                     (ergo-send-string (, (format "show(%s).\n"
                                                  command)) 'echo))))
    (list 'progn
          deffn
          (` (define-key ergo-menu [show (, submenu) (, (intern command))]
               '((, name) . (, fn)))))))

;;ERGO4;; (define-ergo-show "global_tactic_file" tactics "Global Tactic Files")
(define-ergo-show "tactic_file" tactics "Tactic Files")
;;ERGO4;; (define-ergo-show "tactic" tactics "Tactics")

;;(define-ergo-show "write_depth" syntax "write_depth")
;;(define-ergo-show "verbosity" syntax "verbosity")
(define-ergo-show "abbrev" syntax "Abbreviations")
(define-ergo-show "global_operator" syntax "Global Operators")
(define-ergo-show "operator" syntax "Operators")
;;ERGO4;; (define-ergo-show "allnfi" syntax "Nfis")
(define-ergo-show "object_vars" syntax "Object Variables")
(define-ergo-show "symbol" syntax "Symbols")

;;(define-ergo-show "path" thy "path")
;;(define-ergo-show "dbms_interface" thy "dbms_interface")
;;(define-ergo-show "subthy_owner" thy "subthy_owner")
;;(define-ergo-show "theorem_action" thy "theorem_action")
;;(define-ergo-show "save_proof" thy "save_proof")
;;(define-ergo-show "default_prover" thy "default_prover")
;;(define-ergo-show "interpretation" thy "interpretation")
;;(define-ergo-show "interpreted_theory" thy "interpreted_theory")
;;(define-ergo-show "parent" thy "parent")
;;ERGO4;; (define-ergo-show "rulegroup" thy "Rule Groups")
(define-ergo-show "rule" thy "Rules")
(define-ergo-show "visible_theory" thy "Visible Theories")
(define-ergo-show "ancestor" thy "Ancestors (of this theory)")
(define-ergo-show "status" thy "Status (of this theory)")
(define-ergo-show "theory" thy "All Theories")

;;ERGO4;; (define-ergo-show "openrule" windows "Window Opening Rules")
;;ERGO4;; (define-ergo-show "reverse_of" windows "Reverse Relations")
;;ERGO4;; (define-ergo-show "stronger_than" windows "Stronger Relations")
;;ERGO4;; (define-ergo-show "equivalence" windows "Equivalences")
;;ERGO4;; (define-ergo-show "preorder" windows "Preorders")

;;ERGO4;; (define-key ergo-menu [rules] '("Rules" . ergo-list-focus-rules))
(define-key ergo-menu [help] '("Help" . ergo-help))
;;(define-key ergo-menu [symbols] '("Symbols" . ergo-symbol-menu))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Macros
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmacro ergo-with-buffer (buffer &rest forms)
  "With BUFFER selected and writeable, execute FORMS."
  (`
   (let ((obuf (current-buffer)))
     (, (if buffer (list 'set-buffer buffer)))
     (let ((inhibit-read-only t)
           (vwin (if (, buffer) (get-buffer-window (, buffer) t))))
       (prog1
           (progn
             (if vwin (goto-char (window-point vwin)))
             (,@ forms))
         (if vwin (set-window-point vwin (point)))))
     (set-buffer obuf))))

(defmacro ergo-define-magic (string function &rest options)
  "Define an entry in ergo-magic-table (q.v.).
Args are STRING, FUNCTION and OPTIONS (rest of args).
STRING is an index, so we override any existing entry for STRING."
  `(aput 'ergo-magic-table ,string ',(cons function options)))



;;(def-edebug-spec ergo-with-buffer (sexp body))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Major modes
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar ergo-theory-mode-syntax-table
  (copy-syntax-table text-mode-syntax-table))
(modify-syntax-entry ?% "<" ergo-theory-mode-syntax-table)
(modify-syntax-entry ?\n ">" ergo-theory-mode-syntax-table)
(modify-syntax-entry ?/ ". 14b" ergo-theory-mode-syntax-table)
(modify-syntax-entry ?* ". 23b" ergo-theory-mode-syntax-table)
(modify-syntax-entry ?\" "\"" ergo-theory-mode-syntax-table)

(defun ergo-theory-mode ()
  "Major mode for editing Ergo theories.
\\[run-ergo] starts an Ergo process.
\\[ergo-forward-term] and \\[ergo-backward-term] skip backwards and forwards over terms.
\\[ergo-end-of-buffer] redisplays the Inferior Ergo window so the prompt is near the bottom.
\\[ergo-send-term] sends the following term to the Ergo process.
\\[ergo-send-previous-term] sends the preceding term to the Ergo process.
\\[ergo-send-region] sends the current region to the Ergo process.
\\[ergo-send-buffer] sends the whole current buffer to the Ergo process.
\\[ergo-send-rest-of-buffer] sends from point to the end of the current buffer.
\\[ergo-send-nfi], \\[ergo-send-quit], \\[ergo-send-undo] send nfi, quit and undo commands to the Ergo process.
\\[ergo-send-print-region] sends a print command with the current region as an argument.

\\{ergo-theory-mode-map}"
  (interactive)
  (make-local-variable 'indent-line-function)
  (setq mode-name "Ergo"
        major-mode 'ergo-theory-mode
        indent-line-function 'ergo-theory-indent-line
        comment-start "%")
  (use-local-map ergo-theory-mode-map)
  (set-syntax-table ergo-theory-mode-syntax-table)
  (run-hooks 'ergo-theory-mode-hooks))

(defun inferior-ergo-mode ()
  "Major mode for running Ergo.
\\{inferior-ergo-mode-map}"
  (interactive)
  (comint-mode)
  (setq major-mode 'inferior-ergo-mode
        mode-name "Inferior-Ergo")
  (use-local-map inferior-ergo-mode-map)
  (make-variable-buffer-local 'comint-prompt-regexp)
  (setq comint-prompt-regexp ergo-prompt-regexp)
  (make-variable-buffer-local 'font-lock-defaults)
  (setq font-lock-defaults 
	'(ergo-inferior-font-lock-keywords t nil nil beginning-of-line))
  (run-hooks 'inferior-ergo-mode-hooks))

(defun run-ergo (&optional local-ergo-program ensure-visible)
  "Run Ergo."
  (interactive (if current-prefix-arg
                   (list (read-string "Ergo program: " ergo-program)
                         t)
                 (list nil t)))
  (or (memq ergo-buffer (buffer-list))
      (setq ergo-buffer (get-buffer-create "*ergo*")))
  (or ergo-process (get-buffer-process ergo-buffer))
  (save-excursion
    (if (and ergo-process (eq (process-status ergo-process) 'run))
        nil
      (set-buffer ergo-buffer)
      (setq ergo-process
            (get-buffer-process 
             (make-comint "ergo" (or local-ergo-program ergo-program))))
      (set-process-filter ergo-process 'ergo-filter)
      (inferior-ergo-mode)
      (setq comint-process-echoes nil)
      ;; can't use echo/exec for this, as the handler isn't yet loaded!!!
;;ERGO4;;       (ergo-send-string "require_global_tactic_file(emacs).\n" nil)
;;ERGO;;       (ergo-send-string "require_global_tactic_file(emacs5).\n" nil)
      (ergo-send-string (format "set emacs_magic === %s"
                                (if (and window-system
                                         (not ergo-magic-disabled))
                                    "default"
                                  "[]"))
                        'echo)))
  (if (and ensure-visible
           (not (ergo-buffer-visible-p ergo-buffer)))
      (display-buffer ergo-buffer))
  ergo-process)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Term navigation
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; A term ends with a full stop followed by white space that isn't inside a
; string or comment.  It starts at the beginning of the buffer, or after a
; term-end marker, except that leading comments or whitespace are ignored.
(defun ergo-forward-term (&optional n)
  "Move point to the end of term.  With arg, do it that many times."
  (interactive "p")
  (or n (setq n 1))
  (if (<= n 0)
      (ergo-backward-term (- 1 n))
    (while (> n 0)
      (let ((start (point)))
        (while (and (re-search-forward "\\." nil t)
                    (or (let ((parse (parse-partial-sexp start (point))))
                          (or (nth 3 parse) (nth 4 parse))) ; string or comment
 			(or (and (looking-at "[ \t\n]") ; graphic chars
				 (let ((gra (skip-chars-backward "\-/+*<=>#@$\\&~`:.?|^")))
				   (let ((ign (skip-chars-forward "\-/+*<=>#@$\\&~`:.?|^")))
				     (< gra -1)))) 
			    (not (looking-at "[ \t\n]"))))))
        (skip-chars-forward " \t")
        (if (or (eolp) (looking-at "%.*$"))
            (forward-line 1))
        (setq n (1- n))))))

(defun ergo-backward-term (n)
  "Move backward to start of term.  With arg, do it that many times."
  (interactive "p")
  (if (< n 0)
      (ergo-forward-term (- n))
    (let ((terms nil) (count 0) (orig (point)))
      (goto-char (point-min))
      (while (< (point) orig)
        (setq terms (cons (cons count (point)) terms)
              count (1+ count))
        (ergo-forward-term 1)
        (skip-chars-forward " \t\n"))
      (goto-char (or (cdr-safe (assoc (- count n) terms)) (point-min))))))

(defun ergo-calculate-indent ()
  "Calculate indentation for a line of Ergo source."
  (let (prev pos)
    (save-excursion
      (setq prev (buffer-substring (progn (beginning-of-line 0) (point))
                                   (progn (end-of-line 1) (point)))
            pos (progn (back-to-indentation) (current-column))))
    (save-match-data 
      (cond ((string-match "===$" prev) (+ pos 2))
            ((and (eq (current-buffer) ergo-proof-script-buffer)
                  ergo-current-window-level)
             (* 2 (1- ergo-current-window-level)))
            (t pos)))))
            
(defun ergo-theory-indent-line ()
  "Indent current line as Ergo source."
  (interactive)
  (let ((col (ergo-calculate-indent)))
    (back-to-indentation)
    (if (= (current-column) col)
        nil
      (delete-region (save-excursion (beginning-of-line nil) (point))
                     (point))
      (insert-char ?\  col))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Talking to the inferior process.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defvar ergo-send-region-pos (make-marker)
  "Current position in script buffer.")
(defvar ergo-send-region-end nil
  "Finish position in script buffer.  Nil after a failure.")

(defun ergo-send-region (start end &optional macro)
  "Send the current region to the Ergo process.
From a program, arguments START and END delimit the region.
With optional (prefix) argument MACRO, save the region in a temporary file and
send it as a macro.  Otherwise, send it one command at a time, and watch for
errors."
  (interactive "r\nP")
  (if macro
      (let ((tmpf (make-temp-name "/tmp/esr")))
        (write-region start end tmpf nil nil)
        (ergo-macro-file tmpf 'delete))
    (set-marker ergo-send-region-pos start)
    (setq ergo-send-region-end end)
    (ergo-send-region-loop)))

(defun ergo-send-region-loop ()
  (if (and ergo-send-region-end
           (< ergo-send-region-pos ergo-send-region-end))
      (ergo-with-buffer (marker-buffer ergo-send-region-pos)
        (goto-char ergo-send-region-pos)
        (ergo-send-term 'nosync)
        (set-marker ergo-send-region-pos (point))
        (ergo-synchronize 'ergo-send-region-loop))
    (ergo-do-deferred-output-maybe)))

(defun ergo-send-buffer (&optional macro)
  "Send the entire current buffer to the Ergo process.
With an optional (prefix) argument MACRO non-nil, send the buffer's file as a
macro.  Otherwise, send it one command at a time, and watch for errors."
  (interactive "P")
  (if (null macro)
      (ergo-send-region (point-min) (point-max))
    (if (and (buffer-modified-p)
             (y-or-n-p "Save buffer first "))
        (save-buffer))
    (ergo-macro-file (buffer-file-name))))

(defun ergo-send-rest-of-buffer (&optional macro)
  "Send the rest of the current buffer to the Ergo process.
The region sent is from the current point to the end of buffer.
With optional (prefix) argument MACRO, save the region in a temporary file and
send it as a macro.  Otherwise, send it one command at a time, and watch for
errors."
  (interactive "P")
  (ergo-send-region (point) (point-max) macro))

(defun ergo-macro-file (file &optional delete)
  "Send FILE as a macro to Ergo."
  (interactive "fSend macro file: ")
  (setq file (expand-file-name file))
  (if delete
      (ergo-send-string (format "macro_and_delete('%s').\n" file)
                        'echo nil 'update)
    (ergo-send-string (format "macro('%s').\n" file)
                      'echo nil 'update)))

(defconst ergo-input-chunk-size
  (if (boundp 'comint-input-chunk-size)
      (symbol-value 'comint-input-chunk-size)  ;; pre-19.29
    500 ;; 19.30 value
    ))

(defun ergo-send-string (str &optional special focus update)
  "Send the string STR to the Ergo process.
Optional arg SPECIAL controls the way the command is sent to Ergo.
SPECIAL=echo causes Ergo to echo the command before parsing it.
SPECIAL=log does echo, and also logs the command in the proof script if
  there is one and if logging is enabled.
SPECIAL=exec causes Ergo to execute the command immediately, without
  reprompting.
SPECIAL=select means that the command is not sent to Ergo at all; instead, it
  becomes the ergo selection (and can be pasted with \\[ergo-insert-sel]).
With optional third arg FOCUS, if there is a selected subfocus, make STR apply
  to that selection.
With optional fourth arg UPDATE, force a display update after the operation."
  (interactive "sString: ")
  (run-ergo nil t)
  (if (and focus (cdr-safe ergo-last-focus-select))
      (setq str (concat (ergo-focus-cmd (cdr ergo-last-focus-select))
                        " ---> " str)))
  (cond ((and (memq special '(echo log))
              (not (and (eq (string-match "[ \t\n]*" str) 0)
                        (= (match-end 0) (length str)))))
         (or (string-match "\\.\\s-*$" str)
             (setq str (concat str ".")))
         (or (= (aref str (1- (length str))) ?\n)
             (setq str (concat str "\n")))
         (let ((pos 0))
           (while (setq pos (string-match "[\"\\]" str pos))
             (setq str (concat (substring str 0 pos) "\\" (substring str pos))
                   pos (+ pos 2))))
         (if (eq special 'log) (ergo-log str))
         (ergo-with-buffer ergo-buffer
           (ring-insert comint-input-ring (substring str 0 -1)))
         (process-send-string ergo-process
                              (format "'READER SPECIAL''ECHO'(\"%s\").\n"
                                      str)))
        ((eq special 'exec)
         (process-send-string ergo-process
                              (format "'READER SPECIAL''EXEC'(emacs(%s)).\n"
                                      str)))
        ((eq special 'select)
         (ergo-set-selection str))
        (t
         ;; ordinary case: send what we have.
         (process-send-string ergo-process str)))
  (if update
      (ergo-synchronize 'ergo-do-deferred-output-maybe))
  ergo-process)

(defun ergo-send-input ()
  "Send typed input to Ergo process.
Does same as comint-send-input, then requests a proof frame update."
  (interactive)
  (run-ergo)
  (comint-send-input)
  (ergo-synchronize 'ergo-do-deferred-output-maybe))

(defun ergo-send-term (&optional noupdate)
  "Send the term after point to the Ergo process."
  (interactive)
  (ergo-send-string
      (buffer-substring (point)
                        (progn (ergo-forward-term) (point)))
      (and (not ergo-magic-disabled) 'echo)
      nil
      (not noupdate)))

(defun ergo-do-deferred-output-maybe ()
  (ergo-send-string "emacs_do_deferred_output" 'exec))


(defun ergo-send-previous-term ()
  "Send the term before point to the Ergo process."
  (interactive)
  (insert "\n")
  (ergo-send-string
      (buffer-substring (save-excursion (ergo-backward-term 1) (point))
                        (point))
      'echo nil 'update))

(defun ergo-end-of-buffer ()
  "Move point forward to the last prompt in the Ergo buffer.
If point is already at or beyond the last prompt, do nothing.
Never switches windows."
  (interactive)
  (let ((obuf (current-buffer))
        (owin (selected-window)))
    (set-buffer ergo-buffer)
    (select-window (get-buffer-window ergo-buffer t))
    (if (>= (window-point) (process-mark ergo-process))
        nil
      (goto-char (point-max))
      (recenter -2))
    (set-buffer obuf)
    (select-window owin)))

(defun ergo-interrupt ()
  "Interrupt the Ergo process."
  (interactive)
  (interrupt-process ergo-process))

(defun ergo-refresh-interface ()
  "Refresh the Emacs-Ergo interface."
  (interactive)
  (ergo-send-string "emacs_refresh" 'echo))

(defun ergo-toggle-wrapping ()
  "Toggle line wrapping, by toggling the truncate-lines variable."
  (interactive)
  (setq truncate-lines (not truncate-lines))
  ;; The following line works around a bug in Emacs 19.28
  ;; (the cursor moves, but the buffer is not redrawn).
  ;; It is not necessary in Emacs 19.30.
  (switch-to-buffer (current-buffer)))

(defun ergo-sendstring-handler (str)
  "Inject STR on Ergo's input, after echoing."
  (if (stringp str)
      (ergo-send-string str 'echo)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Output filtering
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmacro ergo-handler-error (&rest args)
  "Signal an error in a handler."
  (` (progn
       (ergo-debug "handler-error: \"%s\"\n" (format (,@ args)))
       (ergo-set-output-handler 'ergo-normal-output-handler '(partial) 'now)
       (error (,@ args)))))

(defconst ergo-magic-regexp "MAGIC \\([A-Z]+\\)\\.?\n")

(defconst ergo-filter-regexp
  (concat "\\("
          ergo-magic-regexp
          "\\)\\|"
          ergo-error-regexp
          (if ergo-noise-table
              (concat "\\|\\("
                      (mapconcat (function (lambda (ent) (car ent)))
                                 ergo-noise-table
                                 "\\)\\|\\(")
                      "\\)"))))

(defun ergo-filter (proc str)
  "Filter function controlling disposition of Ergo output.
This function will extract any MAGIC commands and set ergo-output-handlers
appropriately."
  (ergo-debug "Filter got \"%s\"\n" str)
  (if ergo-magic-disabled
      (ergo-handle-output '((ergo-magic-disabled partial)) str)
    (save-match-data
      (progn ;;;;save-excursion
        (let ((pos 0)
              (case-fold-search nil)
              magic noise error mstart mend)
          (while (string-match ergo-filter-regexp str pos)
            (setq mstart (match-beginning 0)
                  mend (match-end 0))
            ;; magic, error or noise??
            (cond ((string= (substring str mstart (+ mstart 6)) "MAGIC ")
                   ;; magic = the string to index the magic table
                   (setq magic (ergo-matched-text str 2)
                         error nil
                         noise nil))
                  ((eq (string-match ergo-error-regexp str mstart) mstart)
                   ;; error = the kind of error
                   (setq error (ergo-matched-text str 0)
                         magic nil
                         noise nil))
                  (t
                   ;; noise = the noisy string
                   (setq noise (substring str mstart mend)
                         error nil
                         magic nil)))
            ;; send any preceding text to the old handler
            (if (> mstart pos)
                (ergo-handle-output ergo-output-handlers
                                    (substring str pos mstart)))
            (cond (magic
                   ;; process magic:
                   ;; look up the magic table ...
                   (let ((table ergo-magic-table)
                         (handler nil)
                         (handler-options nil))
                     (while (and table (not handler))
                       (if (string= (car (car table)) magic)
                           (setq handler (nth 1 (car table))
                                 handler-options (nthcdr 2 (car table))))
                       (setq table (cdr table)))
                     ;; ... and call the handler.
                     (if handler
                         (ergo-set-output-handler handler handler-options)
                       (ergo-handler-error "No handler for MAGIC %s" magic))))
                  (noise
                   ;; process noise:
                   ;; look up the noise table ...
                   (let ((table ergo-noise-table)
                         (action nil))
                     (while (and table (not action))
                       (if (string-match (car (car table)) noise)
                           (setq action (cdr (car table))))
                       (setq table (cdr table)))
                     ;; ... and do the action
                     (ergo-debug "noise handler: \"%s\", action %S\n"
                                 noise action)
                     (cond ((eq action 'discard)
                            ;; silently discard the noise.
                            (ergo-debug "discarded noise: %s\n" noise))
                           ((eq action 'display)
                            ;; let the user hear, but carry on.
                            (ergo-debug "displayed noise: %s\n" noise)
                            (ergo-handle-output '((ergo-normal-output-handler partial))
                                                noise))
                           ((eq action 'abort)
                            ;; stop, stop, stop all the magic ...
                            (ergo-set-output-handler
                             'ergo-normal-output-handler
                             '(partial) 'now)
                            (ergo-handle-output ergo-output-handlers noise))
                           ((eq (car-safe action) 'call))
                           (t (ergo-handler-error "noise confusion: %s"
                                                  noise)))))
                   (error
                    ;; process an error
                    (let (start end)
                      (setq ergo-send-region-end nil)
                      (ergo-with-buffer ergo-buffer
                        (setq start (point)))
                      (ergo-debug "error handler: \"%s\"\n" error)
                      (comint-output-filter ergo-process error)
                      (ergo-with-buffer ergo-buffer (setq end (point)))
                      (add-text-properties start end
                                           '(face highlight) ergo-buffer)
                      (message "Ergo says ``%s'' here" error)
                      (ding))))
            (setq pos mend))
          ;; send any remaining text to the new handler
          (if (< pos (length str))
              (ergo-handle-output ergo-output-handlers (substring str pos))))
        ))))

(defun ergo-debug (fmt &rest args)
  "Maybe produce some debugging output.
Depends on the value of debug-ergo-filter."
  (if ergo-debug
      (save-excursion
        (set-buffer (get-buffer-create "*Ergo Debug*"))
        (goto-char (point-max))
        (insert (apply 'format fmt args)))))

(defvar ergo-output-pending "")

(defun ergo-handle-output (handlers str)
  "Send some Ergo output to the current handler.
We can buffer output here if a partial line is received and the handler will
not accept partial lines."
  (setq str (concat ergo-output-pending str))
  (let (hfn hopts)
    (if (null handlers)
        (setq hfn 'ergo-normal-output-handler
              hopts '(partial))
      (setq hfn (car (car handlers))
            hopts (cdr (car handlers))))
    (if (memq 'partial hopts)
        ;; send it in one chunk
        (progn
          (ergo-handle-output-chunk hfn str)
          (setq ergo-output-pending ""))
      ;; send one whole line at a time
      (let ((pos 0) (epos))
        (while (string-match ".*\n" str pos)
          (setq epos (match-end 0))
          (ergo-handle-output-chunk hfn (substring str pos epos))
          (setq pos epos))
        (setq ergo-output-pending (substring str pos))))))

(defun ergo-handle-output-chunk (hfn str)
  (ergo-debug "%s \"%s\"\n" hfn str)
  (let ((obuf (current-buffer)))
    (set-buffer ergo-buffer)
    (funcall hfn str)
    (set-buffer obuf)))

(defun ergo-set-output-handler (handler options &optional now)
  "Set ergo-output-handlers to HANDLER with OPTIONS.
Also disable the old handler (after delivering any pending output) and enable
the new one.
If 'local is one of the options, the previous handler is not disabled, and a
subsequent 'endlocal reverts to the previous handler.
With optional arg NOW, discard all pending output."
  (let ((ohandler (or (car-safe ergo-output-handlers)
                      '(ergo-normal-output-handler partial)))
        (obuf (current-buffer)))
    (set-buffer ergo-buffer)
    (if (or now (string= ergo-output-pending ""))
        nil
      (ergo-handle-output-chunk (car ohandler) ergo-output-pending)
      (setq ergo-output-pending ""))
    (cond ((memq 'local options)
           (setq ergo-output-handlers
                 (cons (cons handler options) ergo-output-handlers))
           (ergo-debug "%s SUSPEND\n" (car ohandler))
           (ergo-debug "%s ON\n" handler)
           (funcall handler t))
          ((memq 'endlocal options)
           (or (memq 'local (cdr ohandler))
               (ergo-handler-error "Magic: endlocal when not in local."))
           (ergo-debug "%s OFF\n" (car ohandler))
           (funcall (car ohandler) nil)
           (setq ergo-output-handlers (cdr ergo-output-handlers))
           (ergo-debug "%s RESUME\n" (car (car ergo-output-handlers))))
          (t 
           (setq ergo-output-handlers (list (cons handler options)))
           (ergo-debug "%s OFF\n" (car ohandler))
           (funcall (car ohandler) nil)
           (ergo-debug "%s ON\n" handler)
           (funcall handler t)))
    (set-buffer obuf)))

(defun ergo-normal-output-handler (str)
  "Handle some normal Ergo output."
  (if (stringp str)  ; ignore on-off switches
      (save-match-data
        (comint-output-filter ergo-process str)
        (if (string-match "^\\[+'?\\([A-Za-z0-9_]+\\)'?\\]+: $" str) ;a prompt
            (setq ergo-current-theory (substring str
                                                 (match-beginning 1)
                                                 (match-end 1)))))))
(defun ergo-magic-disabled (str)
  "Handle some Ergo output with magic processing disabled."
  (comint-output-filter ergo-process str))

(defun ergo-toggle-magic-processing (&optional arg)
  "Toggle magic processing.
A positive prefix arg turns magic on, a negative one turns it off.
A C-u arg toggles emacs magic without affecting Ergo."
  (interactive "P")
  (cond ((listp arg)
         (setq ergo-magic-disabled (not ergo-magic-disabled)))
        ((< (prefix-numeric-value arg) 0)
         (setq ergo-magic-disabled t))
        ((> (prefix-numeric-value arg) 0)
         (setq ergo-magic-disabled nil)))
  (or (consp arg)
      (if ergo-magic-disabled
          (progn
            (setq ergo-output-handlers '((ergo-normal-output-handler partial)))
            (ergo-send-string "set emacs_magic === []" 'echo)
            (message "Magic processing is off"))
        (ergo-send-string "set emacs_magic === default"
                          'echo nil 'update)
        (message "Magic processing is on"))))

(defun ergo-end-handler (arg)
  "Pop the handler stack, reverting to the previous handler.
The current handler must be 'local."
  ;; Do nothing: the dispatcher does it all!
)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Theories
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun ergo-cd ()
  "Put up a menu of theories and CD."
  (interactive)
  (run-ergo nil t)
  (ergo-send-string "send_theories" 'exec))

(defvar ergo-theories nil)

(defun ergo-theories-handler (str)
  "Handler used to create a menu of theories."
  (cond ((eq str t)
         (setq ergo-theories nil))
        ((null str)
         (if ergo-theories
             (ergo-easy-menu "Select Theory" ergo-theories 'send-echo)
           (error "No theories")))
        ((stringp str)
         (if (string-match "^.*\n" str)
             (let ((thy (substring str (match-beginning 0)
                                   (1- (match-end 0)))))
               (setq ergo-theories (cons (cons thy (format "cd(%s)." thy))
                                         ergo-theories)))))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Help and xrefs
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun ergo-help-mode ()
  "Major mode for Ergo help buffers.
The buffers are arranged in a ring.
Use `\\[ergo-help-previous]' and `\\[ergo-help-next]' \
to cycle through the ring.
Use mouse-1 to follow a cross-reference.
Use `\\[scroll-up]' and `\\[scroll-down]' to scroll the display.

\\{ergo-help-mode-map}"
  (setq buffer-read-only t
        major-mode 'ergo-help-mode
        mode-name "Ergo-Help")
  (use-local-map ergo-help-mode-map))

(defun ergo-help-handler (str)
  "Ergo is sending some help.
Add a new help buffer to the ring and display it in the help window."
  (cond ((eq str t)
         (setq ergo-help-buffers
               (cons (generate-new-buffer " Ergo Help") ergo-help-buffers))
         (let ((waste (nthcdr ergo-help-ring-max ergo-help-buffers)))
           (if (null waste)
               nil
             (mapcar 'kill-buffer waste)
             (setcdr (nthcdr (1- ergo-help-ring-max) ergo-help-buffers) nil)))
         (save-excursion
           (set-buffer (car ergo-help-buffers))
           (ergo-help-mode))
         (if (window-live-p ergo-help-window)
             (let ((owin (selected-window)))
               (select-window ergo-help-window)
               (switch-to-buffer (car ergo-help-buffers))
               (ergo-make-frame-visible (window-frame ergo-help-window))
               (select-window owin))
           (let ((pop-up-frames t)
                 (default-frame-alist (ergo-frame-parameters
                                         'ergo-help-frame-alist
                                         "ergo-help"
                                         default-frame-alist)))
             (setq ergo-help-window
                   (display-buffer (car ergo-help-buffers))))))
        ((stringp str)
         (ergo-with-buffer (car ergo-help-buffers)
           (if (string-match "^%%%%node: \\(.*\\)\n" str)
               (rename-buffer (format " Ergo Help (%s)"
                                      (ergo-matched-text str 1)) 'unique)
             (insert str))))))

(defvar ergo-xref nil)
(defvar ergo-xref-cmd nil)

(defun ergo-xref-handler (str)
  "Ergo is sending a help cross-reference.
Gather it up and send it to the old handler."
  (cond ((eq str t)
         (setq ergo-xref "" ergo-xref-cmd nil))
        ((null str)
         (ergo-handle-output
            (cdr ergo-output-handlers)
            (ergo-button ergo-xref ergo-xref-cmd 'echo)))
        ((string-match "^COMMAND$" str)
         (setq ergo-xref (substring ergo-xref 0 -1)
               ergo-xref-cmd ""))
        (ergo-xref-cmd
         (setq ergo-xref-cmd (concat ergo-xref-cmd str)))
        (t         
         (setq ergo-xref (concat ergo-xref str)))))

(defvar ergo-emph nil)

(defun ergo-emph-handler (str)
  "Ergo is sending text to emphasize
Gather it up and send it to the old handler."
  (cond ((eq str t)
         (setq ergo-emph ""))
        ((null str)
         (add-text-properties 0 (length ergo-emph)
                              (list 'face ergo-emph-face)
                              ergo-emph)
         (ergo-handle-output
            (cdr ergo-output-handlers)
            ergo-emph))
        (t         
         (setq ergo-emph (concat ergo-emph str)))))


(defun ergo-help-previous ()
  "Rotate the help buffer ring backwards.
That is, display the node that is one older than the currently-displayed node,
if there is one.  If the currently-displayed node is the oldest node available,
cycle back to the newest."
  (interactive)
  (setq ergo-help-buffers
        (append (cdr ergo-help-buffers) (list (car ergo-help-buffers))))
  ;; bypass deleted buffers
  (while (not (buffer-name (car ergo-help-buffers)))
    (setq ergo-help-buffers (cdr ergo-help-buffers)))
  (switch-to-buffer (car ergo-help-buffers)))

(defun ergo-help-next ()
  "Rotate the help buffer ring forwards.
That is, display the node that is one newer than the currently-displayed node,
if there is one.  If the currently-displayed node is the newest node available,
cycle back to the oldest."
  (interactive)
  (let* ((last-two (nthcdr (- (length ergo-help-buffers) 2) ergo-help-buffers))
         (last (cdr last-two)))
    (setcdr last-two nil)
    (setq ergo-help-buffers
          (append last ergo-help-buffers)))
  ;; bypass deleted buffers
  (while (not (buffer-name (car ergo-help-buffers)))
    (setq ergo-help-buffers (cdr ergo-help-buffers)))
  (switch-to-buffer (car ergo-help-buffers)))

(defun ergo-help-kill ()
  "Kill this help buffer, unless it is the last."
  (interactive)
  (if (null (cdr-safe ergo-help-buffers))
      (error "Cannot kill the last remaining help buffer."))
  (kill-buffer (car ergo-help-buffers))
  ;; bypass deleted buffers
  (while (not (buffer-name (car ergo-help-buffers)))
    (setq ergo-help-buffers (cdr ergo-help-buffers)))
  (switch-to-buffer (car ergo-help-buffers)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Indexed terms
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar ergo-focus-start nil
  "The buffer position where an Ergo focus started.
This is used by MAGIC INDEX to make the region mouse sensitive.")

(defvar ergo-focus-end nil
  "The buffer position where an Ergo focus ended.
This is used by MAGIC INDEX to make the region mouse sensitive.")

(defvar ergo-current-index nil
  "Accumulator for an index.")

(defun ergo-focus-handler (str)
  "Ergo is sending us a focus.
Remember where it starts so we can make it mouse-sensitive."
  ;; ASSUMPTION: it is to go in the subproof pane.
  (ergo-with-buffer ergo-subproof-buffer
    (goto-char (point-max))
    (cond ((eq str t)
           (let ((pos (point-min)) npos)
             (while (setq pos (text-property-not-all pos (point-max)
                                                     'ergo-focus-cmd nil))
               (setq npos (text-property-any pos (point-max) 'ergo-focus-cmd nil))
               (remove-text-properties pos npos '(ergo-focus-cmd nil face nil))
               (setq pos npos)))
           (setq ergo-focus-start (point)))
          ((eq str nil)
           (setq ergo-focus-end (point)))
          ((stringp str)
           (insert str)))))

(defun ergo-index-handler (str)
  "Ergo is sending us an index.
Use it to mouse-sensitize a region."
  ;; ASSUMPTION: it is to go in the subproof pane.
  ;; ASSUMPTION: the regions are perfectly-nested, and come innermost-first.
  (cond ((eq str t)
         (ergo-set-focus-selection nil)
         (setq ergo-current-index "")
         (or ergo-focus-start
             (ergo-handler-error "Got INDEX without FOCUS!")))
        ((null str)
         (ergo-with-buffer ergo-subproof-buffer
           (add-text-properties
              ergo-focus-start (1- ergo-focus-end)
              (list 'face ergo-active-face
                    'ergo-focus-cmd ergo-current-index))))
        ((stringp str)
         (setq ergo-current-index (concat str ergo-current-index)))))

(defun ergo-add-focus-properties (start end cmd)
  "Augment ergo-focus-cmd properties from START to END with CMD."
  ;; assumption: the existing focus properties of the whole region are the
  ;; same.
  (ergo-debug "add-focus-properties %d-%d\n" start end cmd)
  (put-text-property start end 'ergo-focus-cmd
                     (cons (cons start (cons end cmd))
                           (get-text-property start 'ergo-focus-cmd))))

;; Encoding for numbers used in indexes.
;; Single-digit encoding: D or @DDDD
;; Two-digit encoding:   DD or @DDDD
;; Each D is a base-64 digit.
;; Single-digit encoding is used in path vectors; two-digit encoding is
;; used for character positions.
;; Each subterm is indexed <start-pos> <end-pos> <path-vec> <newline>.
;; Path-vec is a sequence of single-digit-encoded numbers.

(defun ergo-decode-number (str pos n)
  "Decode part of STR starting at POS using N-digit encoding.
Returns a pair (VAL . NPOS), where VAL is the value and NPOS is the position
where we finished.  Returns nil if POS is past the end or pointing to a newline
character."
  (cond ((>= pos (length str))
         nil)
        ((= (aref str pos) ?#)
         (ergo-decode-number str (1+ pos) 4))
        ((= (aref str pos) ?\n)
         nil)
        (t
         (let ((val 0))
           (while (> n 0)
             (setq val (+ (* val 64) (aref str pos) -48)
                   pos (1+ pos)
                   n   (1- n)))
           (cons val pos)))))

(defun ergo-decode-index-line (str pos)
  "Decode index line in STR starting at POS.
Returns nil if past end of str, otherwise a 4-tuple
\(NPOS START END PATH)."
  (if (>= pos (length str))
      nil
    (let (start end path val)
      (setq val   (ergo-decode-number str pos 2)
            start (car val)
            pos   (cdr val)
            val   (ergo-decode-number str pos 2)
            end   (car val)
            pos   (cdr val))
      (while (setq val (ergo-decode-number str pos 1))
        (setq path (cons (car val) path)
              pos  (cdr val)))
      (if (= (aref str pos) ?\n)
          (setq pos (1+ pos)))
      (list pos
            start
            end
            path))))
            

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Starting and ending proofs
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defvar ergo-current-proof-theory nil)
(defvar ergo-current-proof-name nil)
(defvar ergo-current-proof-extra nil)
(defvar ergo-displayed-proof-theory nil)
(defvar ergo-displayed-proof-name nil)

(defun ergo-start-proof-handler (str)
  (cond ((stringp str)
         (cond ((string-match "^theory: \\(.*\\)\n" str)
                (setq ergo-current-proof-theory
                      (ergo-matched-text str 1)))
               ((string-match "^name: \\(.*\\)\n" str)
                (setq ergo-current-proof-name
                      (ergo-matched-text str 1)))
               ((string-match "^extra: \\(.*\\)\n" str)
                (setq ergo-current-proof-extra
                      (ergo-matched-text str 1)))
               (t (ergo-handler-error "start proof: \"%s\"" str))))
        ((null str)
         nil)))

(defun ergo-end-proof-handler (arg)
  (if arg
      nil
    (setq ergo-current-proof-theory nil
          ergo-current-proof-name nil
          ergo-current-proof-extra nil)
    (ergo-hide-proof-frame)))

(defun ergo-window-handler (arg)
  (cond ((and (null arg)
              (or (not (string= ergo-current-proof-theory
                                ergo-displayed-proof-theory))
                  (not (string= ergo-current-proof-name
                                ergo-displayed-proof-name))))
         (ergo-realize-proof-frame ergo-current-proof-theory
                                   ergo-current-proof-name
                                   ergo-current-proof-extra))
        ((not (stringp arg))
         nil)
        ((string-match "thy: " arg)
         (setq ergo-current-proof-theory (substring arg 5 -1))
         (ergo-debug "window-handler: thy is %s, old is %s\n"
                     ergo-current-proof-theory
                     ergo-displayed-proof-theory))
        ((string-match "name: " arg)
         (setq ergo-current-proof-name (substring arg 6 -1))
         (ergo-debug "window-handler: name is %s, old is %s\n"
                     ergo-current-proof-name
                     ergo-displayed-proof-name))
        ((string-match "id: " arg)
         (setq ergo-current-window-id (substring arg 4 -1)))
        ((string-match "level: " arg)
         (setq ergo-current-window-level (string-to-int (substring arg 7))))
))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Proof frame
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; TO DO: User window management in the proof frame section!

(defvar ergo-proof-frame nil)

(defconst ergo-default-proof-frame-panes
  '(
    (2 . ((3 . "hypotheses") (1 . "proof_control")))
    (3 . "subproof"))
  "Default value for the variable ergo-proof-frame-panes (q.v.).")

(defun ergo-show-default-proof-frame ()
  "Create a new frame showing Ergo proof state.
If called interactively, destroys existing proof frame first (so you can use
this if your proof frame is messed up)."
  (interactive)
  (if (and (frame-live-p ergo-proof-frame)
           (if (interactive-p)
               (progn (delete-frame ergo-proof-frame) nil)
             (ergo-make-frame-visible ergo-proof-frame)
             t))
      nil
    (setq ergo-proof-frame-panes
          (or ergo-proof-frame-panes
              (ergo-parse-panes-resource "ergo-proof-frame")
              ergo-default-proof-frame-panes))
    (setq ergo-proof-frame
          (make-frame (ergo-frame-parameters 'ergo-proof-frame-alist
                                             "ergo-proof"
                                             '((name . "Ergo Proof Frame")
                                               (height . 34)
                                               (width . 80)))))
    (select-frame ergo-proof-frame)
    (let ((panenames (ergo-make-panes ergo-proof-frame-panes
                                      (frame-height)
                                      (frame-width))))
      (ergo-send-string (concat "set emacs_proof_panes === ["
                                (mapconcat 'identity panenames ",")
                                "].")
                        'echo nil 'update))))

;; We are going to do some nasty stuff in here.
(defun ergo-parse-panes-resource (component))
(defun ergo-make-panes (panes frmht frmwd))
(eval-when-compile (setq byte-compile-warnings nil))

(defun ergo-parse-panes-str ()
  ;; Reads RES and modifies IND from caller's scope.
  ;; It's recursive, too!!!
  (let ((done nil) (val nil))
    (while (and (< ind (length res)) (not done))
      (cond ((memq (aref res ind) '(?\  ?\t ?\n ?, ?=))
             (setq ind (1+ ind)))
            ((= (aref res ind) ?\[)
             (setq ind (1+ ind)
                   val (cons (cons nil (ergo-parse-panes-str))
                             val)))
            ((= (aref res ind) ?\])
             (setq done t
                   ind (1+ ind)))
            ((eq (string-match "[-_a-zA-Z]+" res ind) ind)
             (let ((resname (substring res ind (match-end 0)))
                   (resind 0))
               ;; replace - by _ (backwards compatability)
               (while (< resind (length resname))
                 (if (= (aref resname resind) ?-)
                     (aset resname resind ?_))
                 (setq resind (1+ resind)))
               (setq val (cons (cons nil resname) val)
                     ind (match-end 0))))
            ((eq (string-match "[0-9]+" res ind) ind)
             (rplaca (car val)
                     (string-to-int (substring res ind (match-end 0))))
             (setq ind (match-end 0)))
            (t (setq done t))))
    (nreverse val)))

(defun ergo-parse-panes-resource (component)
  ;; Look up the X resource `Emacs.COMPONENT.panes' and return a panes list.
  (let ((res (x-get-resource "panes" "Panes"
                             component (capitalize component)))
        (ind 0))
    (if (null res)
        nil
      (prog1
          (ergo-parse-panes-str)
        (if (< ind (length res))
            (error "Could not parse panes resource for %s" component))))))

(defun ergo-make-panes (panes frmht frmwd)
  ;; creates panes in current buffer; returns list of all pane names
  (let ((htprop 0) (panenames nil))
    (let ((p panes))
      (while p (setq htprop (+ (car (car p)) htprop) p (cdr p))))
    (while (cdr panes)
      (split-window-vertically (/ (* (car (car panes)) frmht) htprop))
      (ergo-make-pane-row (cdr (car panes)) frmwd)
      (other-window 1)
      (setq panes (cdr panes)))
    (ergo-make-pane-row (cdr (car panes)) frmwd)
    panenames))

(defun ergo-make-pane-row (panes frmwd)
  ;; modifies panenames from caller's scope!
  (if (consp panes)
      (let ((wdprop 0))
        (let ((p panes))
          (while p (setq wdprop (+ (car (car p)) wdprop) p (cdr p))))
        (while (cdr panes)
          (split-window-horizontally (/ (* (car (car panes)) frmwd) wdprop))
          (setq panenames (cons (cdr (car panes)) panenames))
          (switch-to-buffer (ergo-make-buffer (cdr (car panes))))
          (other-window 1)
          (setq panes (cdr panes)))
        (setq panenames (cons (cdr (car panes)) panenames))
        (switch-to-buffer (ergo-make-buffer (cdr (car panes)))))
    (setq panenames (cons panes panenames))
    (switch-to-buffer (ergo-make-buffer panes))))

;; OK, that's enough nasty stuff for now.
(eval-when-compile (setq byte-compile-warnings
                         (if (boundp 'byte-compile-warning-types)
                             byte-compile-warning-types
                           t)))

(defconst ergo-buffers
  '(
    ("subproof_stack" "*Ergo Subproof-Stack*"
        ergo-subproof-stack-mode ergo-subproof-stack-buffer nil)
    ("subproof" "*Ergo Subproof*"
        ergo-subproof-mode ergo-subproof-buffer nil)
    ("hypotheses" "*Ergo Hypotheses*"
        ergo-hypotheses-mode ergo-hypotheses-buffer nil)
    ("proof_control" "*Ergo Proof-Control*"
        ergo-proof-control-mode ergo-proof-control-buffer nil)
    ("query" "*Ergo Query*"
        ergo-query-template-mode ergo-query-buffer nil)
    ("rules" "*Ergo Rules*"
        ergo-rules-mode ergo-matched-rules-buffer nil)
    ("selected-rule" "*Ergo Selected-Rule*"
        ergo-selected-rule-mode
        ergo-selected-rule-buffer nil)
    ("interaction" "*ergo*"
        inferior-ergo-mode ergo-buffer nil)
    ("help" nil
        ergo-help-mode nil ergo-help-window)
    ("proof-script" name
        ergo-proof-script-mode nil ergo-proof-script-window)
    )
  "List specifying mapping between ergo buffer IDs, buffer names, modes and
symbols giving buffer and window objects.")

(defun ergo-make-buffer (id)
  "Set up an Ergo buffer.
Returns the buffer object."
  (let ((details (assoc id ergo-buffers)))
    (or details (error "%s is not an Ergo buffer id." id))
    (let ((name   (or (nth 1 details) "*scratch*"))
          (mode   (nth 2 details))
          (bufsym (nth 3 details))
          (winsym (nth 4 details))
          buf)
      (set-buffer (setq buf (get-buffer-create name)))
      (if bufsym (set bufsym buf))
      (if winsym (set winsym (selected-window)))
      (if mode (funcall mode))
      buf)))

(defun ergo-realize-proof-frame (thy name extra)
  "Display a proof frame.
Set its title to THY:NAME EXTRA and initialize all the buffers."
  (ergo-show-default-proof-frame)
  (setq ergo-displayed-proof-theory thy
        ergo-displayed-proof-name name)
  (modify-frame-parameters
     ergo-proof-frame
     (list (cons 'name (format "Ergo Proof Frame              %s:%s %s"
                               (or thy "")
                               (or name "")
                               (or extra "")))))
  (if (and ergo-logging thy name)
      (ergo-find-proof-script thy name nil nil)))

(defun ergo-hide-proof-frame ()
  "Make the proof frame invisible."
  (setq ergo-displayed-proof-theory nil
        ergo-displayed-proof-name nil)
  (if (frame-live-p ergo-proof-frame)
      (make-frame-invisible ergo-proof-frame))
  (if (and ergo-proof-script-buffer
           (buffer-modified-p ergo-proof-script-buffer)
           (buffer-file-name ergo-proof-script-buffer))
      (save-excursion
        (set-buffer ergo-proof-script-buffer)
        (save-buffer)
        (if (window-live-p ergo-proof-script-window)
        (save-window-excursion
          (select-window ergo-proof-script-window)
          (if (one-window-p)
              (delete-frame)))))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Subproof Stack pane
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun ergo-subproof-stack-mode ()
  "Major mode for the subproof-stack buffer.
\\{ergo-subproof-stack-mode-map}"
  (setq truncate-lines t
        major-mode 'ergo-subproof-stack-mode
        mode-name "Ergo-Subproof-Stack"
        buffer-read-only t)
  (use-local-map ergo-subproof-stack-mode-map))

(defun ergo-subproof-stack-handler (str)
  (if (bufferp ergo-subproof-stack-buffer)
      (ergo-with-buffer ergo-subproof-stack-buffer
        (cond ((eq str t)
               (goto-char (point-max))
               (delete-region (point-min) (point)))
              ((stringp str)
               (ergo-insert-subproof str))
              ((null str)
               (sit-for 0)))
        (goto-char (point-max))
        (set-window-point (get-buffer-window ergo-subproof-stack-buffer 'visible)
                          (point)))))

(defun ergo-expand-subproof-stack-line ()
  "Expand the currently-selected line of the Ergo subproof stack.
That is, fully display the term, ignoring write_depth.
Currently, this redisplays the whole subproof stack pane."
  (interactive)
  (ergo-send-string "expand_subproof_stack_line(*)" 'exec))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Hypotheses pane
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun ergo-hypotheses-mode ()
  "Major mode for hypothesis buffer.
\\{ergo-hypotheses-mode-map}"
  (setq truncate-lines t
        buffer-read-only t
        major-mode 'ergo-hypotheses-mode
        mode-name "Ergo-Hypotheses")
  (use-local-map ergo-hypotheses-mode-map))
        
;; hyp format:
;;  ADD: <discharge_cmd><number><kind_marker> <term>
;;  DEL: delete <number>
;;  CLR: clear
(defconst ergo-hyps-add-hyp-regexp " *\\([a-z_]*\\)\\([0-9]+\\)\\([:%?*]\\) \\(.*\\)\n")
(defconst ergo-hyps-del-hyps-regexp "delete \\([0-9]+\\)\n")
(defconst ergo-hyps-clear-hyps-regexp "clear\n")

(defun ergo-hypotheses-handler (str)
  (if (not (stringp str))
      nil
    (cond ((string-match ergo-hyps-add-hyp-regexp str)
           (ergo-hyps-add-hyp
              (ergo-matched-text str 2)
              (ergo-matched-text str 3)
              (ergo-matched-text str 4)
              (ergo-matched-text str 1)))
          ((string-match ergo-hyps-del-hyps-regexp str)
           (ergo-hyps-del-hyps
              (string-to-int (ergo-matched-text str 1))))
          ((string-match ergo-hyps-clear-hyps-regexp str)
           (ergo-hyps-clear-hyps))
          ((string-match " *...\n" str)
           nil)
          (t (ergo-handler-error "hypotheses: bad update: \"%s\"" str)))))

(defun ergo-expand-hypotheses-line ()
  "Expand the currently-selected Ergo hypothesis.
That is, fully display the term, ignoring write_depth.
Currently, this redisplays the whole hypotheses pane."
  (interactive)
  (ergo-send-string "expand_hypotheses_line(*)" 'exec))

(defun ergo-hyps-add-hyp (num kind hyp &optional discharge)
  (if (string= discharge "") (setq discharge nil))
  (ergo-with-buffer ergo-hypotheses-buffer
    (goto-char (point-max))
    (let ((pt (point)))
      (insert (format "%s%s %s\n" num kind hyp))
      (add-text-properties pt (1- (point))
                           (list 'ergo-hyps-command (format "hyp:(%s)." num)
                                 ;;;                              WAS:  hyp
                                 'face ergo-active-face
                                 'ergo-discharge-cmd discharge)))))

(defun ergo-hyps-del-hyps (num)
  (ergo-with-buffer ergo-hypotheses-buffer
    (goto-char (point-min))
    (while (search-forward-regexp "^[0-9]+" nil t)
      (if (>= (car (read-from-string (ergo-matched-text nil 0)))
              num)
          (delete-region (progn (beginning-of-line nil) (point))
                         (progn (beginning-of-line 2) (point)))))      
    (goto-char (point-max))))

(defun ergo-hyps-clear-hyps ()
  (ergo-with-buffer ergo-hypotheses-buffer
    (erase-buffer)))

(defun ergo-hyps-select-hyp ()
  "Select the hypothesis under the mouse."
  (interactive)
  (let ((epos (event-start last-command-event)) start end)
    (ergo-with-buffer (window-buffer (posn-window epos))
      (ergo-replace-text-properties (point-min) (point-max)
                                    'face ergo-selected-face ergo-active-face)
      (goto-char (posn-point epos))
      (beginning-of-line nil)
      (setq start (point))
      (end-of-line nil)
      (setq end (point))
      (put-text-property start end 'face ergo-selected-face)
      (ergo-send-string (get-text-property start 'ergo-hyps-command)
                        'select 'focus))))

(defun ergo-hypotheses-focus-hyp ()
  "Focus on the selected hypothesis."
  (interactive)
  (goto-char (or (text-property-any (point-min) (point-max)
                                    'face ergo-selected-face)
                 (error "No hypothesis is selected.")))
  (or (looking-at "\\([0-9]+\\). \\(.*\\)$")
      (error "Help, I'm lost in the hyps buffer!"))
  (ergo-send-string (format "hyp:(%s).\n" (ergo-matched-text nil 1))
                    'log nil 'update))


(defun ergo-delete-hypotheses-line ()
  "Delete the selected hypothesis."
  (interactive)
  (goto-char (or (text-property-any (point-min) (point-max)
                                    'face ergo-selected-face)
                 (error "No hypothesis is selected.")))
  (or (looking-at "\\([0-9]+\\). \\(.*\\)$")
      (error "Help, I'm lost in the hyps buffer!"))
  (ergo-send-string (format "delete_hyp(hyp:(%s)).\n"
                            (ergo-matched-text nil 1)) 
                    'log nil 'update))

(defun ergo-discharge-hypotheses-line ()
  "Discharge the selected conjecture."
  (interactive)
  (goto-char (or (text-property-any (point-min) (point-max)
                                    'face ergo-selected-face)
                 (error "No hypothesis is selected.")))
  (or (looking-at "\\([0-9]+\\). \\(.*\\)$")
      (error "Help, I'm lost in the hyps buffer!"))
  (ergo-send-string (format "%s(hyp:(%s)).\n"
                            (or (get-text-property (point) 'ergo-discharge-cmd)
                                "discharge_conjecture")
			    (ergo-matched-text nil 1))
                    'log nil 'update))

(defun ergo-hypotheses-use-this-hyp ()
  "Use the currently selected hypothesis to transform the focus."
  (interactive)
  (let ((pos (text-property-any (point-min) (point-max)
                                'face ergo-selected-face))
        (hyp "_"))
    (if (null pos)
        nil
      (goto-char pos)
      (or (looking-at "\\([0-9]+\\). \\(.*\\)$")
          (error "Help, I'm lost in the hyps buffer!"))
      (setq hyp (ergo-matched-text nil 1)))
    (ergo-send-string (format "use(hyp:(%s)).\n" hyp)
                      'log 'focus 'update)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Proof contol pane
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun ergo-proof-control-mode ()
  "Major mode for Ergo proof-control buffers.
\\{ergo-proof-control-mode-map}"
  (setq major-mode 'ergo-proof-control-mode
        mode-name "Ergo-Proof-Control"
        truncate-lines t
        buffer-read-only t)
  (ergo-with-buffer nil
    (erase-buffer)
    (ergo-update-buttons)
    (ergo-update-parameter-displays)
    (use-local-map ergo-proof-control-mode-map)))

(defvar ergo-control-params nil
  "The parameters to display in the Proof Control pane.")

(defvar ergo-control-displayed-params nil
  "The parameters currently displayed in the Proof Control pane.")

(defconst ergo-param-regexp "\\([^:\n]+\\): +\\([^\n]+\\)\n")
;; regexp is Label: OtherData
;; OtherData is Value
;;           or Name <TAB> Kind <TAB> Value

(defun ergo-param-handler (str)
  (if (not (stringp str))
      nil
    (or (string-match ergo-param-regexp str)
        (ergo-handler-error "param handler: bad data: \"%s\"" str))
    (let ((object (ergo-matched-text str 1))
          (value  (ergo-matched-text str 2)))
      (if (string-match "\\([^\t\n]+\\)\t\\([^\t\n]+\\)\t\\([^\t\n]+\\)"
                        value)
          (ergo-update-parameter object
                                 (ergo-matched-text value 1)
                                 (ergo-matched-text value 2)
                                 (ergo-matched-text value 3))
        (ergo-update-parameter object nil "normal" value)))))

(defun ergo-update-parameter (label name kind value)
  (let ((orec (assoc name ergo-control-params)))
    (if orec
        (rplacd orec (list label kind value))
      (setq ergo-control-params (cons (list name label kind value)
                                      ergo-control-params)))
    (if (and ergo-proof-control-buffer (buffer-name ergo-proof-control-buffer))
        (ergo-update-parameter-displays))))

(defun ergo-update-parameter-displays (&optional all)
  (interactive '(t))
  "Update all displayed parameters."
  (if all
      (setq ergo-control-displayed-params nil))
  (mapcar
     (function (lambda (p)
                 (or (equal p (assoc (car p) ergo-control-displayed-params))
                     (apply 'ergo-update-parameter-display p))))
     (reverse ergo-control-params)))
      
(defun ergo-update-parameter-display (name label kind value)
  "Add/update/delete a parameter display."
  (if (string= kind "numeric")
      (let ((val (car (read-from-string value))))
        (add-text-properties
           0 (length value)
           (list 'face ergo-active-face
                 'mouse-face ergo-selected-face
                 'ergo-parameter (list name (list 'numeric 0 (+ val 5))))
           value)))
  (ergo-with-buffer ergo-proof-control-buffer
    (goto-char 0)
    (search-forward-regexp (concat "^" (regexp-quote label) ":") nil 'move)
    (beginning-of-line nil)
    (delete-region (point) (progn (beginning-of-line 2) (point)))
    (if (string= kind "delete")
        nil
      (let ((tab (make-string (max (- 12 (length label)) 1) ?\ )))
        (insert label ":" tab value "\n")))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Subproof pane
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun ergo-subproof-mode ()
  "Major mode for viewing an Ergo subproof.
\\{ergo-subproof-mode-map}"
  (setq major-mode 'ergo-subproof-mode
        mode-name "Ergo-Subproof"
        truncate-lines t
        buffer-read-only t)
  (use-local-map ergo-subproof-mode-map))

(defun ergo-subproof-handler (str)
  (ergo-with-buffer ergo-subproof-buffer
    (cond ((eq str t)
	   nil)
          ((null str)
           (sit-for 0))
	  ((string-match "^clear.*" str)
           (goto-char (point-max))
           (delete-region (point-min) (point))
           (ergo-insert-subproof
                (format "This is subproof {{{%s}}}.\n\n"
                        ergo-current-window-id)))
          (t
           (ergo-insert-subproof str)))
    (goto-char (point-max))
    (set-window-point (get-buffer-window ergo-subproof-buffer 'visible)
                      (point))))

;; also used by stack viewer and browser
(defun ergo-insert-subproof (str)
  (goto-char (point-max))
  (let (opos (pos 0) subn)
    (while (string-match "{{{[^}]*}}}" str pos)
      (setq opos pos
            pos (match-end 0)
            subn (substring str (+ (match-beginning 0) 3) (- pos 3)))
      (insert (substring str opos (match-beginning 0))
              (ergo-button subn
                           (format "emacs_browse_subproof(%s)" subn)
                           'exec)))
    (insert (substring str pos))))

(defun ergo-expand-subproof-line ()
  "Expand the currently-selected line of the Ergo subproof.
That is, fully display the term, ignoring write_depth.
Currently, this redisplays the whole subproof pane."
  (interactive)
  (ergo-send-string "expand_subproof_line(*)" 'exec))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Browser
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun ergo-browser-mode ()
  "Major mode for browsing Ergo proofs.
\\{ergo-browser-mode-map}"
  (setq major-mode 'ergo-browser-mode
        mode-name "Ergo-Browser"
        truncate-lines t
        buffer-read-only t)
  (use-local-map ergo-browser-mode-map))

(defun ergo-browse-ancs-handler (arg)
  (cond ((eq arg t)
         (setq ergo-browser-ancs-buffer
               (get-buffer-create "*Browse Ancestors*"))
         (ergo-with-buffer ergo-browser-ancs-buffer
           (erase-buffer)
           (ergo-browser-mode)))
        ((null arg)
         nil)
        (t
         (ergo-with-buffer ergo-browser-ancs-buffer
           (ergo-insert-subproof arg)))))

(defun ergo-browse-hyps-handler (arg)
  (cond ((eq arg t)
         (setq ergo-browser-hyps-buffer
               (get-buffer-create "*Browse Hypotheses*"))
         (ergo-with-buffer ergo-browser-hyps-buffer
           (erase-buffer)
           (ergo-browser-mode)))
        ((null arg)
         nil)
        (t
         (ergo-with-buffer ergo-browser-hyps-buffer
           (ergo-insert-subproof arg)))))

(defun ergo-browse-subp-handler (arg)
  (cond ((eq arg t)
         (setq ergo-browser-subp-buffer
               (get-buffer-create "*Browse Subproof*"))
         (ergo-with-buffer ergo-browser-subp-buffer
           (erase-buffer)
           (ergo-browser-mode))
         (if (frame-live-p ergo-browser-frame)
             (ergo-make-frame-visible ergo-browser-frame)
           (set-buffer ergo-browser-ancs-buffer)
           (select-frame
            (setq ergo-browser-frame
                  (make-frame (ergo-frame-parameters
                                 'ergo-browser-frame-alist
                                 "ergo-browser"
                                 '((name . "Ergo Browser")
                                   (height . 40)
                                   (width . 80))))))
           (split-window-vertically 12)
           (other-window 1)
           (switch-to-buffer ergo-browser-hyps-buffer)
           (split-window-vertically 12)
           (other-window 1)
           (switch-to-buffer ergo-browser-subp-buffer)))
        ((null arg)
         nil)
        (t
         (ergo-with-buffer ergo-browser-subp-buffer
           (if (string-match "^{{{.*}}}\n" arg)
               (progn
                 (insert "Browsing subproof ")
                 (ergo-insert-subproof (substring arg 0 -1))
                 (insert "\t\t"
                         (ergo-button "Quit Browser"
                                      "emacs_quit_browser.\n"
                                      nil)
                         "\t\t"
                         (ergo-button "Write Transcript"
                                      "summarize_subproof"
                                      'exec)
                         "\n"))
             (ergo-insert-subproof arg))))))

(defun ergo-browse-quit-handler (arg)
  (if arg
      nil
    (make-frame-invisible ergo-browser-frame)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Dialog boxes
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar ergo-dialog-prompt nil)
(defvar ergo-dialog-data nil)
(defvar ergo-dialog-default nil)

(defun ergo-dialog-handler (arg)
  (cond ((null arg)
         (ergo-send-string
            (format "%s\n"
                    (or (x-popup-dialog t
                                        (append (list ergo-dialog-prompt)
                                                (reverse ergo-dialog-data)
                                                (list nil)))
                        ergo-dialog-default))
            'log nil 'update))
        ((eq arg t)
         (sit-for 0) ;; make sure any other input is displayed
         (setq ergo-dialog-prompt nil
               ergo-dialog-data nil))
        ((string-match "^{{{answer}}} \\(.*\\)$" arg)
         (setq ergo-dialog-data
               (cons (cons (ergo-matched-text arg 1) (ergo-matched-text arg 1))
                     ergo-dialog-data)))
        ((string-match "^{{{default}}} \\(.*\\)$" arg)
         (setq ergo-dialog-default (ergo-matched-text arg 1)
               ergo-dialog-data
               (cons (cons (format "<<%s>>" ergo-dialog-default)
                           ergo-dialog-default)
                     ergo-dialog-data)))
        (t
         (setq ergo-dialog-prompt (concat ergo-dialog-prompt arg)))))
                

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Communication with da Vinci
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defvar davinci-process nil)
(defvar davinci-buffer nil)

(defun ergo-davinci-handler (arg)
  (interactive '("start\n"))
  (cond ((string= arg "start\n")
         (if (and davinci-process (eq (process-status davinci-process) 'run))
             (kill-process davinci-process))
         (let ((pipe (make-temp-name "/tmp/dv5.")))
           (call-process "mkfifo" nil nil nil pipe)
           (setq davinci-buffer (get-buffer-create "*da Vinci*"))
           ;; Deleted; I'm not sure where this should happen, but it surely
           ;; isn't here.  For now, I'll make users do it for themselves.
           ;;(setenv "DAVINCIHOME" "/homes/reftool/daVinci_V2.0/")
           (setq davinci-process
                 (start-process 
		    "daVinci" davinci-buffer "sh" "-c"
		    (format "$DAVINCIHOME/daVinci -pipe <%s" pipe)))
           (ergo-send-string (format "daVinci(connect('%s'))" pipe) 'exec)
           (set-process-filter davinci-process 'davinci-filter)
           (set-process-sentinel davinci-process 'davinci-sentinel)))))

(defvar davinci-output-pending "")

(defun davinci-filter (proc str)
  (ergo-with-buffer davinci-buffer
    (insert str))
  (setq str (concat davinci-output-pending str))
  (let ((pos 0) npos)
    (while (setq npos (string-match "\n" str pos))
      (davinci-output-handler (substring str pos npos))
      (setq pos (1+ npos)))
    (setq davinci-output-pending (substring str pos))))

(defun davinci-sentinel (proc str)
  (message "davinci connection closed")
  (ergo-send-string "daVinci(close)" 'exec))

(defvar davinci-selected-nodes nil)

(defun davinci-output-handler (str)
  (cond ((string-match "menu_selection(\"\\([^\"]*\\)\")" str)
         (ergo-send-string (format "daVinci(menu(%s, [%s]))"
                            (ergo-matched-text str 1)
                            (mapconcat 'identity davinci-selected-nodes ","))
                    'exec))
        ((string-match "node_selections_labels(\\[\\([^]]*\\)\\])" str)
         (let ((sel (davinci-node-list-to-nodes (ergo-matched-text str 1) 0)))
           (if (equal sel davinci-selected-nodes)
               nil
             (setq davinci-selected-nodes sel)
             (ergo-send-string
              (format "daVinci(selected([%s]))"
                      (mapconcat 'identity sel ","))
              'exec))))))

(defun davinci-node-list-to-nodes (str pos)
  (if (< pos (length str))
      (let ((npos (string-match "\"" str (1+ pos))))
        (cons (substring str (1+ pos) npos)
              (davinci-node-list-to-nodes str (+ npos 2))))))
  


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Synchronization
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar ergo-sync-awaiting nil
  "Synchronization tag we are waiting for.")

(defvar ergo-sync-nowait nil)

(defun ergo-synchronize (&optional nowait)
  "Wait until Ergo is accepting commands.
With optional arg NOWAIT, don't wait.  If NOWAIT is a string, emacs will write
the tag value into the file named by NOWAIT when synchronization happens.
The caller should call `accept-process-output' and should examine
ergo-sync-awaiting to be sure that synchronization has occurred.
If NOWAIT is an atom with a function definition, it is called (with no
arguments) when synchronization happens.
Does nothing if magic processing is currently disabled."
  (interactive)
  (if ergo-magic-disabled
      nil
    (let ((tag (int-to-string (nth 1 (current-time)))))
      (setq ergo-sync-awaiting (regexp-quote tag))
      (ergo-send-string (format "synchronize_emacs(%s)" tag) 'exec))
    (setq ergo-sync-nowait nowait)
    (or nowait
        (while ergo-sync-awaiting
          (accept-process-output)
          (if (null ergo-sync-awaiting)
              nil
            (message "Waiting for Ergo ...")
            (sit-for 1)))
        (message ""))))

(defun ergo-sync-handler (str)
  (cond ((not (stringp str))
         nil)
        ((or (null ergo-sync-awaiting)
             (not (string-match ergo-sync-awaiting str)))
;;         (setq ergo-sync-awaiting nil)
;;         (ergo-handler-error "Synchronization error!")
         nil)
        ((stringp ergo-sync-nowait)
         (write-region (concat ergo-sync-awaiting "\n") nil ergo-sync-nowait)
         (setq ergo-sync-awaiting nil))
        ((fboundp ergo-sync-nowait)
         (setq ergo-sync-awaiting nil)
         (let ((fun ergo-sync-nowait))
           (setq ergo-sync-nowait nil)
           (funcall fun)))
        (t
         (setq ergo-sync-awaiting nil))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Selections
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun ergo-set-focus-selection (point &optional path ind offset)
  ;; select path around point
  ;; can select a shorter path if there is no exact match
  (if (and ergo-subproof-buffer (buffer-name ergo-subproof-buffer))
      (ergo-with-buffer ergo-subproof-buffer
        (ergo-replace-text-properties (point-min) (point-max) 'face
                                      ergo-selected-face ergo-active-face)
        (if (null point)
            (setq ergo-last-focus-select nil)
          (let ((match nil) (start offset) (end offset))
            (while (not match)
              ;; while there is no exact one subterm whose index is exactly
              ;; path, keep trying a shorter path ...
              (setq ergo-last-focus-select (cons point path))
              (ergo-set-selection (concat (ergo-focus-cmd path) "."))
              (let ((ind ind))
                (while ind
                  (if (equal (nth 2 (car ind)) path)
                      (setq match t))
                  ;; path is a prefix of index path, enlarge the selected
                  ;; region to incorporate the entire index entry range
                  (if
                      (let ((ip (nth 2 (car ind)))
                            (pp path))
                        (while (and pp ip)
                          (if (= (car pp) (car ip))
                              (setq pp (cdr pp) ip (cdr ip))
                            (setq ip nil)))
                        (null pp))
                      (setq start (min start (nth 0 (car ind)))
                            end (max end (nth 1 (car ind)))))
                  (setq ind (cdr ind))))
              (setq path (reverse (cdr (reverse path)))))
            (setq start (+ start point (- offset))
                  end (+ end point (- offset)))
            (ergo-add-nonnl-text-properties start end
                                            (list 'face ergo-selected-face))
            (x-set-selection nil (buffer-substring start end)))))))

(defun ergo-focus-cmd (path)
  ;; PATH is a position list; return Ergo vector as a string.
  (concat "[" (mapconcat 'identity path ",") "]"))

(defun ergo-add-nonnl-text-properties (start end props)
  "Add properties to the non-nl text from START to END."
  (save-excursion
    (goto-char start)
    (skip-chars-forward "[ \t\n]")
    (let ((scol (current-column)))
      (while (< (point) end)
        (setq start (point))
        (end-of-line nil)
        (add-text-properties start (min end (point)) props)
        (forward-to-indentation 1)
        (if (> (current-column) scol) (move-to-column scol))))))

(eval-when-compile (setq byte-compile-warnings nil))
(defun ergo-mouse-select (event)
  "Ergo handler for down-mouse-1 click.
If on a sensitive region, highlight the selected region, make it the X
selection, and make the associated command the Ergo selection.
Otherwise, cancel the Ergo selection and do mouse-1's normal action."
  (interactive "e")
  (let* ((win (posn-window (event-start event)))
         (buf (window-buffer win))
         (obuf (current-buffer))
         (owin (selected-window))
         (pos (posn-point (event-start event)))
         (props (text-properties-at pos buf)))
    (cond ((member 'ergo-focus-cmd props)
           (ergo-with-buffer buf
             (let* ((start (previous-single-property-change (1+ pos)
                                                            'ergo-focus-cmd))
                    (end (next-single-property-change pos 'ergo-focus-cmd))
                    (ind (get-text-property pos 'ergo-focus-cmd))
                    (offset (- pos start))
                    path)
               ;; decode the index if necessary
               (if (stringp ind)
                   (let ((ipos 0) (newind nil) dec)
                     (while (setq dec (ergo-decode-index-line ind ipos))
                       (setq newind (cons (cdr dec) newind)
                             ipos (car dec)))
                     (setq ind newind)
                     (add-text-properties start end (list 'ergo-focus-cmd ind))))
               ;; make the selection
               (if (eq (car-safe ergo-last-focus-select) pos)
                   (if (cdr-safe ergo-last-focus-select)
                       ;; repeat click; enlarge the selection
                       (setq path (reverse
                                   (cdr (reverse (cdr ergo-last-focus-select)))))
                     ;; repeat click; deselect all
                     (setq pos nil path nil))
                 ;; first click; make indexed selection
                 (let ((i ind))
                   (while i
                     (if (and (<= (nth 0 (car i)) offset)
                              (>  (nth 1 (car i)) offset))
                         (setq path (nth 2 (car i))
                               i nil)
                       (setq i (cdr i))))))
               (ergo-set-focus-selection pos path ind offset)
               ;; discard the up event
               (read-event)
               ;; get the cursor out of the way
               (set-window-point win (point-max))
               (set-buffer obuf))))
          ((member 'query-frobbable props)
           (set-buffer buf)
           (goto-char pos)
           (ergo-query-frob pos))
          ((member 'ergo-parameter props)
           (ergo-set-focus-selection nil)
           (set-window-point win pos)
           (let* ((prop (get-text-property pos 'ergo-parameter buf))
                  (parm (nth 0 prop))
                  (type (nth 1 prop)))
             (if (eq (car-safe type) 'numeric)
                 (ergo-numeric-parameter-menu parm
                                              (nth 1 type)
                                              (nth 2 type)))))
          ((member 'ergo-apprules-command props)
           (select-window win)
           (goto-char pos)
           (ergo-rules-select-rule))
          ((member 'ergo-hint props)
           (select-window win)
           (goto-char pos)
           (ergo-rules-select-rule))
          ((member 'ergo-hyps-command props)
           (select-window win)
           (ergo-hyps-select-hyp))
          ((member 'ergo-button props)
           (let ((args (get-text-property pos 'ergo-button buf)))
             (if (stringp (car args))
                 ;; command is a string: send to Ergo
                 (progn
                   (apply 'ergo-send-string args)
                   (ergo-set-focus-selection nil))
               ;; command is a form: do it in Lisp
               (select-window win)
               (goto-char pos)
               (eval (car args)))))
          (t
	   (if ergo-subproof-buffer (ergo-set-focus-selection nil))
           (setq ergo-selection-text nil)
           (funcall (lookup-key global-map [down-mouse-1]) event)))))
(eval-when-compile (setq byte-compile-warnings
                         (if (boundp 'byte-compile-warning-types)
                             byte-compile-warning-types
                           t)))

(defun ergo-set-selection (cmd)
  "Set the Ergo selection to CMD."
  (message "Use %s to execute the command `%s'"
           (substitute-command-keys "\\[ergo-send-sel]")
           cmd)
  (setq ergo-selection-text cmd))

(defun ergo-insert-sel ()
  "Insert the Ergo selection in a buffer.
If in the Ergo buffer, move to the end first."
  (interactive "@")
  (ergo-insert (or ergo-selection-text (x-selection) "") nil))

(defun ergo-insert-x-sel ()
  "Insert the X selection in a buffer.
If in the Ergo buffer, move to the end first."
  (interactive "@")
  (ergo-insert (or (x-selection) "") nil))

(defun ergo-insert-and-send-sel ()
  "Insert the Ergo selection in a buffer, and send it to Ergo."
  (interactive "@")
  (ergo-insert (concat (or ergo-selection-text (x-selection) "") "\n") 'send))

(defun ergo-send-sel ()
  "Send the Ergo selection to Ergo.
If logging is active, log it in the proof script as well."
  (interactive)
  (ergo-send-string (or ergo-selection-text (x-selection) "")
                    'log nil 'update))

(defun ergo-insert (txt sendp)
  (ergo-end-of-buffer)
  (if (and sendp (eq (current-buffer) ergo-buffer))
      nil
    (insert txt))
  (if sendp
      (ergo-send-string txt 'echo nil 'update)))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Proof scripts and logging
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun ergo-toggle-logging (&optional arg)
  "Change whether commands are automatically logged in the proof script.
With arg, set logging iff arg is positive."
  (interactive "P")
  (setq ergo-logging
        (if (null arg)
            (not ergo-logging)
          (> (prefix-numeric-value arg) 0))))

(defun ergo-log-handler (str)
  "Log STR in the proof script if there is one and logging is enabled.
Do not log it if an exact match already appears."
  (if (stringp str)
      (ergo-log str)))

(defun ergo-log (str)
  "Log STR in the proof script if appropriate.
Do nothing if there is no proof script, if logging is disabled or if STR
already appears exactly."
  (if (and ergo-logging
           ergo-proof-script-buffer
           (buffer-name ergo-proof-script-buffer))
      (let ((win (and (window-live-p ergo-proof-script-window)
                      (eq (window-buffer ergo-proof-script-window)
                          ergo-proof-script-buffer)
                      ergo-proof-script-window)))
        (ergo-with-buffer ergo-proof-script-buffer
          (goto-char (if win (window-point win) (point-max)))
          (ergo-theory-indent-line)
          (if (looking-at (regexp-quote str))
              (goto-char (match-end 0))
            (insert str))
          (if win (set-window-point win (point)))))))

(defun ergo-find-this-proof-script (&optional insert)
  "Find the proof script for the theorem around point.
If the proof script file is empty or does not exist, or if optional prefix
argument INSERT is non-nil, insert the theorem in the script buffer."
  (interactive "P")
  (if (or (looking-at ergo-theorem-regexp)
          (save-excursion
            (ergo-backward-term 1)
            (looking-at ergo-theorem-regexp)))
      nil
    (error "Not on a theorem!"))
  (ergo-find-proof-script (substring buffer-file-name 0 -4)
                          (ergo-matched-text nil 2)
                          (ergo-matched-text nil 0)
                          'insert))

(defun ergo-find-proof-script (thy thmname thm &optional insert)
  "Find the proof script for THY:THMNAME.
If the proof script file is empty or does not exist, or if optional prefix
argument INSERT is non-nil, insert THM in the script buffer."
  (if (eq (nth 0 (file-attributes thy)) t)
      ;; it is a directory
      nil
    (make-directory thy nil))
  (setq ergo-proof-script-buffer
        (find-file-noselect (format "%s/%s.psc" thy thmname)))
  (if (window-live-p ergo-proof-script-window)
      (ergo-make-frame-visible (window-frame ergo-proof-script-window))
    (setq ergo-proof-script-window
          (let ((pop-up-frames t)
                (default-frame-alist (ergo-frame-parameters
                                      'ergo-proof-script-frame-alist
                                      "ergo-proof-script"
                                      default-frame-alist)))
            (display-buffer ergo-proof-script-buffer))))
  (select-window ergo-proof-script-window)
  (switch-to-buffer ergo-proof-script-buffer)
  (ergo-theory-mode)
  (if (and thm (or insert (= (buffer-size) 0)))
      (let (end)
        (goto-char (point-min))
        ;;(insert "set_theorem_action(prove).\n")
        (insert thm)
        (setq end (point))
        (save-excursion
          (goto-char (point-min))
          (if (re-search-forward "[ \t\n]*:-[ \t\n]*rerun" end t)
              (replace-match ""))))))

;; handler for editing Ergo 5 proof scripts, which are not normally in files.
(defvar ergo-proofscript-buffer nil)
(defun ergo-proofscript-handler (arg)
  (cond ((null arg)
         (display-buffer ergo-proofscript-buffer)
         (ergo-with-buffer ergo-proofscript-buffer
           (ergo-theory-mode)))
        ((eq arg t)
         (setq ergo-proofscript-buffer (generate-new-buffer "*Proof Script*")))
        ((stringp arg)
         (ergo-with-buffer ergo-proofscript-buffer
           (insert arg)))))
         


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Buttons and Popup Menus
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar ergo-control-buttons nil
  "The buttons to display in the Proof Control pane.
Note, this list is in *reverse* order.")

(defun ergo-button-handler (arg)
;; TO DO: support options other than log, focus, update in ergo-button-handler.
  (if (stringp arg)
      (if (string-match "^\\(.*\\):\\(.*\\)" arg)
          (ergo-add-button (ergo-matched-text arg 1)
                           (ergo-matched-text arg 2)
                           'log 'focus 'update)
        (ergo-handler-error "bad button: \"%s\"" arg))))

(defun ergo-add-button (label command &optional special focus update)
  "Add a button to the Proof Control pane.
Args are LABEL and COMMAND (both strings).
Optional args SPECIAL, FOCUS, UPDATE are as in ergo-send-string.
If a button with the same label is already there, it is deleted first, so this
   command may be used to move a button to the end.
If COMMAND is nil or the empty string, no button is added (but one may be
deleted)."
  (interactive (list (read-string "Label: ")
                     (read-string "Command: ")
                     'log
                     'focus))
  (if (assoc label ergo-control-buttons)
      (adelete 'ergo-control-buttons label))
  (if (and (stringp command) (not (string= command "")))
      (aput 'ergo-control-buttons label (list command special focus update)))
  (if (and ergo-proof-control-buffer (buffer-name ergo-proof-control-buffer))
      (ergo-update-buttons)))

(defun ergo-update-buttons ()
  (interactive)
  (ergo-with-buffer ergo-proof-control-buffer
    (let ((wd (window-width (get-buffer-window (current-buffer)
                                               ergo-proof-frame))))
      (goto-char (point-min))
      (if (search-forward "\n\n" nil t)
          (delete-region (point-min) (point)))
      (mapcar (function (lambda (b)
                          (if (> (+ (current-column) 2 (length (car b))) wd)
                              (insert "\n"))
                          (insert (apply 'ergo-button b)
                                  "  ")))
              (reverse ergo-control-buttons)))
    (insert "\n\n")))

(defun ergo-button (label command &optional special focus update)
  "Return a string representing an Ergo button.
LABEL is the label, COMMAND is the command to invoke.
If COMMAND is not a string, it is instead a form to exceute in Lisp when the
button is pressed.
Optional args SPECIAL, FOCUS, UPDATE are as in ergo-send-string."
  (let ((str (copy-sequence label)))
    (add-text-properties 0 (length str)
                         (list 'face ergo-active-face
                               'mouse-face ergo-selected-face
                               'ergo-button (list command special focus update))
                         str)
    str))

(defun ergo-easy-menu (prompt entries &optional doit)
  "Pop up a menu with PROMPT and ENTRIES.
The ENTRIES may be a list of strings or a list of lists whose cars are
strings (like a completion table).
If optional arg DOIT is absent or nil, the entire entry (string or list) is
returned.
Otherwise, a value is extracted:
  - if the entry is \"string\" or (\"string\"), the value is that string;
  - if the entry is (\"string\" . value), (value non-nil), the value is value.
The value is then used according to DOIT:
  - cut: the (string) value is put into the cut buffer;
  - send: the (string) value is sent to Ergo as a command;
  - send-echo: the (string) value is sent to Ergo as a command to be echoed;
  - return: the value is returned;
  - lisp: the value is evaluated as a Lisp expression, and the result returned;
  - submenu: the value is passed as the ENTRIES argument to ergo-easy-menu,
    using the same prompt and with DOIT=internal;
  - internal: the value is a cons, whose car is a new DOIT value and whose cdr
    is the real value."
  (let (ents npanes panes entry num)
    (setq ents
          (mapcar (function (lambda (ent)
                              (if (stringp ent)
                                  (cons ent ent)
                                (cons (car ent) ent))))
                  entries)
          npanes
          (1+ (/ (1- (length ents)) ergo-easy-menu-max-pane)))
    ;(message "%d panes for %d ents" npanes (length ents))
    (setq panes (list (cons prompt ents)))
    (while (> npanes 1)
      (setq num (/ (length ents) npanes))
      (setq ents (nthcdr num ents))
      (rplacd (nthcdr (1- num) (cdr (car panes))) nil)
      (setq npanes (1- npanes)
            panes (cons (cons prompt ents) panes)))
    (setq entry (x-popup-menu t (cons prompt panes)))
    (ergo-debug "menu: selected %S\n" entry)
    (if (or (null doit) (null entry))
        entry
      (let ((val (cond ((stringp entry)
                        entry)
                       ((null (cdr-safe entry))
                        (car-safe entry))
                       (t
                        (cdr-safe entry)))))
        (ergo-easy-menu-doit val doit prompt)))))

(defun ergo-easy-menu-doit (val doit prompt)
  (cond ((eq doit 'internal)
         (ergo-easy-menu-doit (cdr val) (car val) prompt))
        ((eq doit 'return)
         val)
        ((eq doit 'cut)
         (ergo-debug "menu: cutting \"%s\"\n" val)
         (ergo-set-selection val))
        ((eq doit 'send)
         (ergo-debug "menu: sending \"%s\"\n" val)
         (ergo-send-string val 'exec))
        ((eq doit 'send-echo)
         (ergo-debug "menu: sending \"%s\"\n" val)
         (ergo-send-string (concat val "\n") 'log nil 'update))
        ((eq doit 'lisp)
         (eval val))
        ((eq doit 'submenu)
         (ergo-easy-menu prompt val 'internal))
        (t (error "ergo-easy-menu: bad DOIT"))))

(defun ergo-numeric-parameter-menu (parm min max)
  (let ((menu nil) (more (` (">>" lisp .
                             (ergo-numeric-parameter-menu (, parm)
                                                          (, (1+ max))
                                                          (, (+ max 10)))))))
    (while (>= max min)
      (setq menu (cons (cons (int-to-string max)
                             (cons 'send
                                   (format "set_parameter(%s, %d)" parm max)))
                       menu)
            max (1- max)))
    ;;;(setq menu (cons more menu))
    (ergo-easy-menu parm menu 'internal)))

;;(defun ergo-symbol-menu ()
;;  "Pop up a menu of symbols of the current theory."
;;  (interactive)
;;  (ergo-send-string "emacs_symbols" 'exec))

;;(defvar ergo-symbols nil)
;;(defun ergo-symbols-handler (arg)
;;  (cond ((null arg)
;;         (insert (ergo-easy-menu "Symbols" ergo-symbols)))
;;        ((eq arg t)
;;         (setq ergo-symbols nil))
;;        ((string-match "^\\([^.\t\n]+\\)\\.\\([^\t\n]+\\)\t\\([^\n]+\\)$" arg)
;;         (setq ergo-symbols
;;               (cons
;;                  (ergo-matched-text arg 2)
;;                     ergo-symbols)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Miscellaneous
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun ergo-frame-parameters (sym component &optional defaults)
  ; Return in SYM a frame parameter alist for COMPONENT using DEFAULTS.
  (or (symbol-value sym)
      (let ((geom (x-get-resource "geometry" "Geometry"
                                  component (capitalize component)))
            (font (x-get-resource "font" "Font"
                                  component (capitalize component)))
            alist)
        (if geom (setq alist (x-parse-geometry geom)))
        (if font (setq alist (cons (cons 'font font) alist)))
        (set sym defaults)
        (aput sym 'user-position t)
        (aput sym 'user-size t)
        (while alist
          (aput sym (car (car alist)) (cdr (car alist)))
          (setq alist (cdr alist)))
        (symbol-value sym))))
          
(defun ergo-op-tactic (&optional cmd file)
  "Do CMD on tactic FILE."
  (interactive
   (save-match-data
     (let ((dir default-directory) (cmd) (file))
       (if (eq last-input-event 'tactics)
           (progn
             (setq cmd
                   (car (ergo-easy-menu "Command" ergo-tactic-file-commands)))
             (or (string-match "global" cmd)
                 (setq dir (concat dir ergo-current-theory "/")))
             (setq file
                   (let ((files (directory-files dir nil "\\.ql$")))
                     (and files (ergo-easy-menu "File" files))))
             (list cmd file))))))
  (let ((dir default-directory))
    (or cmd (setq cmd (completing-read "Command: " ergo-tactic-file-commands)))
    (or (string-match "global" cmd)
        (setq dir (concat dir ergo-current-theory "/")))
    (or file (setq file (read-file-name "File: " dir)))
    (setq file (file-name-nondirectory file))
    (if (string-match "\\..*$" file)
        (setq file (substring file 0 (match-beginning 0))))
    (ergo-send-string (format "%s(%s).\n" cmd file) 'echo)))
             
(defun ergo-help ()
  "Get context-sensitive help on Ergo."
  (interactive)
  (ergo-send-string "help.\n" 'echo))

(defun ergo-describe-mode ()
  "Get help on the current window."
  (interactive)
  (let ((pop-up-frames t))
    (describe-mode)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Useful non-Ergo-specific user functionality
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun ergo-tear-off-window (&optional window)
  (interactive)
  (or window
      (setq window
            (if (memq (event-basic-type last-command-event)
                      '(mouse-1 mouse-2 mouse-3))
                (posn-window (event-start last-command-event))
              (selected-window))))
  (let* ((buf (window-buffer window))
	 (frame (make-frame)))
    (if (one-window-p 'nomini)
        (switch-to-buffer (other-buffer buf))
      (delete-window window))
    (select-frame frame)
    (switch-to-buffer buf)
    frame))

(defun ergo-kill-current-buffer-and-window ()
  "Kill the current buffer and its window.
If the window is the only one on the frame, delete the frame."
  (interactive)
  (let ((buf (current-buffer)))
    (if (eq (window-buffer (selected-window)) buf)
        (if (one-window-p)
            (delete-frame)
          (delete-window)))
  (kill-buffer buf)))

(defun ergo-scroll-left (&optional arg)
  "Scroll selected window display ARG columns left.
Default for ARG is half of window width."
  (interactive "P")
  (scroll-left (or arg (/ (window-width) 2))))

(defun ergo-scroll-right (&optional arg)
  "Scroll selected window display ARG columns right.
Default for ARG is half of window width."
  (interactive "P")
  (scroll-right (or arg (/ (window-width) 2))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Useful non-Ergo-specific Lisp functionality
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun ergo-buffer-visible-p (buffer)
  "Return non-nil if BUFFER is shown in some window of some visible,
non-iconic frame.  Actually returns the window in which it's shown."
  (and buffer
       (setq buffer (get-buffer buffer))
       (let ((win (get-buffer-window buffer t)))
         (and win
              (let ((vis (frame-visible-p (window-frame win))))
                (and vis
                     (not (eq vis 'icon))
                     win))))))


(defun ergo-matched-text (str idx)
  "Return the IDX'th matched substring of STR.
If STR is nil, use the current buffer."
  (if str
      (substring str (match-beginning idx) (match-end idx))
    (buffer-substring (match-beginning idx) (match-end idx))))

(defun ergo-replace-text-properties (start end prop oldval newval
                                     &optional newprop)
  "Replace some text properties in the current buffer.
For every character from START to END whose PROP property has the value OLDVAL,
replace the value by NEWVAL.
With optional arg NEWPROP, replace the property name as well."
  (let ((pstart))
  (while (and start (< start end))
    (setq pstart (text-property-any start end prop oldval)
          start  (and pstart (text-property-not-all pstart end prop oldval)))
    (if (null pstart)
        nil
      (if newprop
          (remove-text-properties pstart (or start end) (list prop nil)))
      (add-text-properties pstart (or start end)
                           (list (or newprop prop) newval))))))

(defun ergo-tab-to-column (col)
  "Move right to column COL, adding spaces if necessary.
If the current column is already beyond COL, do nothing."
  (if (< (current-column) col)
      (move-to-column col 'force)))

(defun ergo-split-string (string char)
  "Split STRING at every CHAR, giving a list of strings."
  (let ((res nil)
        (start 0)
        (end 0))
    (while (< end (length string))
      (if (/= (aref string end) char)
          nil
        (setq res (cons (substring string start end) res)
              start (1+ end)))
      (setq end (1+ end)))
    (if (<= start end)
        (setq res (cons (substring string start) res)))
    (nreverse res)))

(defun ergo-remove-duplicates (list)
  "Return LIST with duplicates removed."
  (let ((res nil) (l list))
    (while l
      (if (member (car l) res)
          nil
        (setq res (cons (car l) res)))
      (setq l (cdr l)))
    (nreverse res)))
  
(defun ergo-insert-with-properties (props &rest strings)
  "Adding PROPERTIES, insert STRINGS (rest of args)."
  (let ((start (point))
        (inhibit-read-only t))
    (apply 'insert strings)
    (add-text-properties start (point) props)))

(defun ergo-make-frame-visible (frame)
  "Make frame visible, even to the extent of raising it."
  (make-frame-visible frame)
  (raise-frame frame))

(defmacro ergo-with-buffer-tangible (&rest forms)
  (if (string< emacs-version "19.30")
      (` (let ((inhibit-point-motion-hooks t))
           (ergo-replace-text-properties (point-min) (point-max)
                                    'intangible t t 'dummy-intangible)
           (prog1 (progn (,@ forms))
             (ergo-replace-text-properties (point-min) (point-max)
                                      'dummy-intangible t t 'intangible))))
    (` (let ((inhibit-point-motion-hooks t))
         (,@ forms)))))

(require 'ergo-rules)
(require 'ergo-elves)
(provide 'ergo)

;; Local Variables:
;; eval: (put 'ergo-with-buffer 'lisp-indent-hook 'defun)
;; End:
