%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: postulates.ql,v 1.27 2001/06/04 01:24:08 itp Exp $

#include "../lib/ergo.h"

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Embryonic support for POSTULATES
%%
%% A postulate is a rule that is asserted without proof (like an axiom), but
%% for which a proof can later be furnished (like a theorem).  This file
%% defines commands to add and to prove postulates.  Later, it should be
%% possible to enable and disable the use of postulates, both individually and
%% _en masse_.  When a postulate is disabled, neither that postulate nor any
%% theorem that depends on it should be usable.  Disabling postulates is
%% essential to the soundness of prove_postulate, otherwise unsound circular
%% inference is possible.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% add_postulate(Name, AssumpList, Stmt, Body).
%%
%%  Add a postulate to the current theory or a subtheory thereof.
%%
%%  Name must be an atom or a Sub:Name pair, where Sub is a subtheory of the
%%  current theory.  The Name must not already in use for any rule in the
%%  current theory.
%%
%%  Ass must be a closed list of terms that are legal in the current 
%%  theory.  Currently, only the empty list is supported.
%%
%%  Term must be a term that is legal in the current theory.
%%
%%  mode add_postulate(@nonvar, @list(term), @term, @nonvar).
%%
add_postulate(QName, Ass, Stmt, Body) :-
    decode_subthy_rule(QName, Dest, Name),
    check_no_name_clash_rule(Dest, Name),
    check_or_error(
        ( nonvar(Stmt)
        ; Ass=[_|_]
        ),
        ['postulate ',q(Name),' is too general.']),
    check_term(Stmt),
    check_terms(Ass),
    collect_vars((Ass,Stmt), Vars),
    check_axiom_body(Body, Vars),
    assert_axthm(Name, postulate, Ass, Stmt, Body, Dest).


%% prove_postulate(Name, Tact).
%% prove_postulate(Name, Tact, Ask).
%%
%%  Prove a postulate or a list of postulates in the current theory or a
%%  subtheory thereof.
%%
%%  Name must be an atom or a Sub:Name pair, where Sub is a subtheory of the
%%  current theory or a list of names.
%%  Tact must be a tactic term used to discharge the postulate.
%%  Ask is the Ask argument of add_theorem.  By default it is `ask'.
%%
%%  mode prove_postulate(@nonvar, @term).
%%  mode prove_postulate(@nonvar, @term, @atom).
%%
prove_postulate(QName, Tact) :-
    prove_postulate(QName, Tact, ask, _).

prove_postulate(QName, Tact, Stmt) :-
    prove_postulate(QName, Tact, ask, Stmt).


prove_postulate(QNameList, Tact, Ask, Stmt) :-
    list(QNameList),
    !,
    (
        member(QName, QNameList),
        catch(prove_postulate(QName, Tact, Ask,Stmt),
              abort_proof(_),
              true),             
        fail
    ;
        true
    ).

prove_postulate(QName, Tact, Ask, Stmt) :-
    decode_subthy_rule(QName, Dest, Name),
    %% We take *two* copies of the rule, one to start the proof off
    %% and the other one to check against the resulting theorem to
    %% ensure that it is identical to the original postulate (that is,
    %% no instantiations and no extra constraints).
    check_or_error(
        query_theory(rule(Name, postulate, _, Stmt0, Body0), Dest),
        [q(Name),' is not a postulate in the theory ',q(Dest),'.']),
     freeze_vars([Stmt0,Body0]),
     check_or_error(
      (retract_theory(rule(Name, postulate, _, Stmt1, Body), Dest),
       structural_unify(Stmt, Stmt1)),
     [q(Name),' is not a postulated as ',q(Stmt),' in the theory ',q(Dest),'.']),
    writelist(['Warning: avoid circular reasoning!', nl]),
    unwind_protect(
      (with_disabled_postulates(
          [Dest:Name],
          add_theorem(Dest:Name, [], Stmt1, Ask,
                      (Body, accept_constraints, Tact))),
        query_theory(rule(Name, theorem(Deps), _, Stmt2, Body2), Dest),
        %% now we do a one-sided unification (the *0 vars are frozen).
        collect_vars(Stmt0, Vars0),
        collect_vars(Stmt2, Vars2),
        Vars0 = Vars2,        
        ( Stmt2 = Stmt0 ->
            true
        ;
            writelist(['Error: postulate was instantiated during the proof!',
                       nl]),
            retract_theory(rule(Name, _,_,_,_), Dest),
            fail
        ),
        %% now check that all constraints in Body2 are also in Body0.
        prolog_code_diff(Body2, Body0, Diff),
        ( Diff == true ->
            true
        ;
            %% Unfortunately, it is a bit confusing to print out Diff
            %% here, because its variables are named differently to
            %% the variables in the copy that went to add_theorem.
            writelist(['Error: constraints (eg. not_free_in''s)',
                         ' were added to the postulate.',nl]),
            retract_theory(rule(Name, _,_,_,_), Dest),
            fail
        ),
        ( is_circular_proof(Dest:Name, Deps) ->
            writelist(['Error: you used circular reasoning!',nl]),
            retract_theory(rule(Name, _,_,_,_), Dest),
            fail
        ;
            true
        ) ->
            writelist(['Proved postulate ', q(Dest), ':', q(Name), nl])
      ),
      (\+ query_theory(rule(Name, _, _, _, _), Dest) ->
          (add_postulate(Dest:Name, [], Stmt, Body) ->
              writelist(['Failed to prove postulate ', q(Dest), ':',
                         q(Name), nl])
          ;
              writelist(['Oops, your postulate ', q(Dest), ':', q(Name),
                         ' got lost.  Sorry!', nl])
          )
      ;
          true
     )),
    (
        query_theory(rule(Name, theorem(_), _, _, _), Dest)
    ;
        fail
    ).


%% is_circular_proof(Thm, ThmDeps).
%%
%%  Succeeds if adding Thm (which must be of the form Thy:Name)
%%  to the current theory, with dependencies ThmDeps, would
%%  cause the dependency graph between postulates to become circular.
%%
%%  mode is_circular_proof(@fullrulename, @list(fullrulename)).
%%
is_circular_proof(Thm, Deps) :-
    member(Thm, Deps),
    !.
is_circular_proof(Thm, Deps) :-
    member(Thy:T, Deps),
    query_theory(rule(T, theorem(TDeps), _, _, _), Thy),
    is_circular_proof(Thm, TDeps),
    !.


%%%%%%
%% Enabling and disabling postulates.  In all cases, Posts may be a list of
%% Thy:Name, or the atom 'all'.
%% disabled_postulates(Posts) -- return currently disabled postulates.
%% set_disabled_postulates(Posts) -- set currently disabled postulates.
%% disable_postulates(Posts) -- add to currently disabled postulates.
%% enable_postulates(Posts) -- remove from currently disabled postulates.
%% with_disabled_postulates(Posts, Goal) -- call Goal with Posts disabled.
%% THIS IS CURRENTLY A STUB, until the prover is changed to take notice (which
%% should not happen until the dependency management syste m is sorted out
%% properly.)
%%%%%%

%%disabled_postulates(Posts) :-
%%    query_global(disabled_postulates(Posts)).

%%set_disabled_postulates(Posts) :-
%%    retractall_global(disabled_postulates(_)),
%%    assert_global(disabled_postulates(Posts)).

%%disable_postulates(Posts) :-
%%    disabled_postulates(Old),
%%    ( Old = all ->
%%        true
%%    ; Posts = all ->
%%        set_disabled_postulates(all)
%%    ; union_list(Old, Posts, New),
%%      set_disabled_postulates(New)
%%    ).

%%enable_postulates(Posts) :-
%%    disabled_postulates(Old),
%%    ( Old = all ->
%%        check_or_error(
%%            Posts = all,
%%            ['cannot selectively enable when all postulates are disabled.']),
%%        set_disabled_postulates([])
%%    ; Posts = all ->
%%        set_disabled_postulates([])
%%    ; diff_list(Old, Posts, New),
%%        set_disabled_postulates(New)
%%    ).

%%with_disabled_postulates(Posts, Goal) :-
%%    disabled_postulates(DPs),
%%    unwind_protect((disable_postulates(Posts), call_predicate(Goal)),
%%                   set_disabled_postulates(DPs)).
with_disabled_postulates(_Posts, Goal) :-
    call_predicate(Goal).

