%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: syntax1.ql,v 1.43 2003/12/19 01:32:49 itp Exp $

#include "../lib/ergo.h"

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% This file contains procedures for checking various syntactic
%% well-formedness conditions for terms of the current theory.
%% Most procedures with names like check_* generate error messages
%% when they fail.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% check_atom(Name).
%%
%%  Succeeds iff Name is an atom.
%%
%%  mode check_atom(@atom).
%%
check_atom(Name) :-
    atom(Name),
    !.
check_atom(Name) :-
    handle_error([q(Name),' is not an atom.']).


%% check_std_nonvar(Term).
%%
%%  Succeeds iff std_nonvar(Term).
%%
%%  mode check_atom(@nonvar).
%%
check_std_nonvar(Term) :-
    std_nonvar(Term),
    !.
check_std_nonvar(Term) :-
    handle_error([q(Term),' must be an atom or a compound term',nl,
                  'whose functor is an atom.']).


%% check_legal_filename(Name).
%%
%%  Succeeds iff Name is an acceptable atom for part of a filename.
%%
%%  mode check_legal_filename(@atom).
%%
check_legal_filename(Name) :-
    check_legal_rule_name(Name).


%% check_truefalse(Name).
%%
%%  Succeeds iff Name is either 'true' or 'false'.
%%
%%  mode check_truefalse(@atom).
%%
check_truefalse(Name) :-
    check_or_error(
        ( Name==true ; Name==false ),
        [q(Name),' must be true or false.']).


%% check_integer(Int).
%%
%%  Succeeds iff Int is an integer.
%%
%%  mode check_integer(@integer).
%%
check_integer(Int) :-
    check_or_error(
        integer(Int),
        [q(Int),' is not an integer.']).


%% check_integer(Int).
%%
%%  Succeeds iff Int is an integer.
%%
%%  mode check_integer(@integer).
%%
check_posint(PosInt) :-
    check_or_error(
        (integer(PosInt),
         0 < PosInt
        ),
        [q(PosInt),' is not a positive integer.']).


%% check_non_internal_atom(Name).
%%
%%  Succeeds iff Name is an atom and is not an internal name.
%%
%%  mode check_non_internal_atom(@atom).
%%
check_non_internal_atom(Name) :-
    check_atom(Name),
    check_or_error(
        \+ is_internal_name(Name),
        [q(Name),' must not contain a ''.'' character.']).


%% is_legal_cmd(Term).
%% check_legal_cmd(Term).
%%
%%  Succeeds iff Term is a a legal command: an atom or a compound
%%  term whose functor is an atom.
%%
%%  mode is_legal_cmd(@nonvar).
%%  mode check_legal_cmd(@nonvar).
%%
is_legal_cmd(Term) :- atom(Term),!.
is_legal_cmd(Term) :- compound(Term),functor(Term,F,_),atom(F).

check_is_legal_cmd(Term) :-
    check_or_error(
        is_legal_cmd(Term),
        [q(Term),' is not a legal command.']).


%% check_legal_rule_name(Name).
%%
%%  True if Name is an atom and does not contain non-printing
%%  characters or '/'s etc.
%%
%%  mode check_legal_rule_name(@atom).
%%
check_legal_rule_name(Name) :-
    \+ atom(Name),
    !,
    handle_error(['illegal rule/file name: ',q(Name),'.  (Not an atom).']).
check_legal_rule_name('') :-
    !,
    handle_error(['illegal rule/file name: '''' (empty atom)']).
check_legal_rule_name(Name) :-
    %% atom(Name).
    atom_codes(Name, Codes),
    \+ (member(C, Codes),
        \+ legal_rule_name_code(C,Name)).

%% We use theorem names as file names, so need to restrict the
%% allowable characters.  The current restrictions are somewhat
%% UNIX-specific.  We ban '/' because UNIX doesn't allow it in
%% file names, '.' because other systems don't like it and its
%% a disaster to allow it in theory names, and all non-printing
%% or whitespace characters.
%% This relies on the ASCII character set and doesn't handle
%% 8-bit chars yet (maybe it shouldn't anyway).
%%
legal_rule_name_code(47,Name) :-
    !,
    handle_error(['''/'' is not allowed in rule/file name: ',q(Name)]).
legal_rule_name_code(46,Name) :-
    !,
    handle_error(['''.'' is not allowed in rule/file name: ',q(Name)]).
legal_rule_name_code(C,_) :-
    %% \+ member(C, "/."),
    32 < C, C < 127,
    !.
legal_rule_name_code(C,Name) :-
    !,
    handle_error(['illegal char (',q(C),') in rule/file name: ',q(Name)]).


%% check_no_name_clash(Thy, Name, Arity).
%%
%%  Succeeds if Name (the internal form of a new function/quantifier)
%%  does not clash with any of the symbols already defined in theory Thy.
%%
%%  mode check_no_name_clash(@atom, @atom, @list(ground)).
%%
check_no_name_clash(Thy, Name, []) :-
    \+ (query_theory(symbol(Name,_,_), Thy)),
    !.
check_no_name_clash(Thy, Name, [quant(L)|_]) :-
    \+ (query_theory(symbol(Name,[quant(L)|_],_), Thy)),
%% TO DO: extend check_no_name_clash for open bndvar lists.
    !.
check_no_name_clash(Thy, Name, [N|_]) :-
    integer(N),
    N > 0,
    \+ (query_theory(symbol(Name,[N|_],_), Thy)),
    \+ (query_theory(symbol(Name,[],_), Thy)),
    !.
check_no_name_clash(_, Name, _) :-
    handle_error([q(Name),' is already declared/defined.']).


%% check_no_name_clash_const(Thy, Const).
%%
%%  Succeeds if the parameterized constant Const is not defined in
%%  theory Thy and it does not clash with other symbols in that theory.
%%
%%  mode check_no_name_clash_const(@atom, @compound).
%%
check_no_name_clash_const(Thy, Const) :-
    query_theory(symbol(Const,_,Decl), Thy),
    !,
    handle_error([q(Const),' is already ',q(Decl),'.']).
check_no_name_clash_const(_, _).


%% check_no_name_clash_obvar(Thy, Prefix).
%%
%%  Since object variables are not prefixed with the theory name
%%  like functions and quantifiers, object variable prefixes cannot
%%  clash with function or quantifier names.
%%  Therefore, all we need to check is that Prefix has not already 
%%  been declared as an object variable prefix.
%%
%%  mode check_no_name_clash_obvar(@atom, @atom).
%%
check_no_name_clash_obvar(Thy, Prefix) :-
    check_or_error(
        \+ (query_theory(object_vars(Prefix), Thy)),
        ['object variable prefix ',q(Prefix),' is already used.']).


%% check_no_name_clash_rule(Thy, Name).
%%
%%  Rules have their own name space.
%%
%%  mode check_no_name_clash_rule(@atom, @atom).
%%
check_no_name_clash_rule(Thy, Name) :-
    check_or_error(
        \+ (query_theory(rule(Name,_,_,_,_), Thy)),
        ['rule name ',q(Name),' is already used.']).


%% check_arity(Arity).
%%
%%  Succeeds if Arity is a legal arity for a name.
%%
%%  mode check_arity(@list(ground)).
%%
check_arity(Arity) :-
    check_or_error(
        closed_list(Arity),
        [q(A),' is not a legal arity list.']),
    \+ (member(A,Arity),
        \+ check_arity1(A)
       ).

check_arity1(I) :-
    integer(I),
    0 < I,
    !.
%% PJR
check_arity1(quant(L)) :- 
    var(L),
    !.
check_arity1(quant(L)) :-
    list(L),
    check_arity_quant(L),
    !.
check_arity1(A) :-
    handle_error([q(A),' is not a legal arity.']).

check_arity_quant([]).
check_arity_quant([T|TS]) :-
    check_or_error(
        ( T == typed
        ; T==untyped
        ),
        [q(T),' must be ''typed'' or ''untyped''.']),
    check_arity_quant(TS).



%% check_axiom_body(Body, Vars).
%%
%%  Succeeds iff Body contains only distinctness conditions and
%%  not_free_in conditions
%%  whose LHS is an object variable in Vars and whose RHS is a term
%%  all of whose variables are in Vars.
%%
%%  mode check_axiom_body(@nonvar, @list(anyvar)).
%%
check_axiom_body((A,B), Vars) :-
    !,
    check_axiom_body(A, Vars),
    check_axiom_body(B, Vars).
check_axiom_body(B, Vars) :-
    check_axiom_body_cond(B, Vars).

check_axiom_body_cond(true, _) :-
    !.
check_axiom_body_cond(distinct(L), Vars) :-
    !,
    check_or_error(
        map(obvar, [L]),
        [q(distinct(L)),' contains non-object-variables.']),
    check_or_error(
        subset_var(L, Vars),
        [q(distinct(L)),' contains variables that do not appear ',
         'in the axiom statement.']),
    !.
check_axiom_body_cond(not_free_in(X,T), Vars) :-
    obvar(X),
    !,
    check_or_error(
        member_var(X, Vars),
        [q(X),' does not appear in the axiom statement.']),
    collect_vars(T, TVars),
    check_or_error(
        subset_var(TVars, Vars),
        [q(T),' contains variables that do not appear ',
         'in the axiom statement.']),
    !.
check_axiom_body_cond(C, _) :-
    handle_error(['illegal axiom condition: ', q(C)]).


%% check_distinct_metavars(Vars).
%%
%%  Succeeds iff Vars is a list of distinct variables.
%%
%%  mode check_distinct_metavars(@list(var)).
%%
check_distinct_metavars([]).
check_distinct_metavars([V|VS]) :-
    member_eq(V,VS),
    !,
    handle_error([q(V), ' occurs more than once.']).
check_distinct_metavars([V|VS]) :-
    std_var(V),
    %% \+ member_eq(V,VS),
    !,
    check_distinct_metavars(VS).
check_distinct_metavars([V|_]) :-
    %% \+ std_var(V),
    handle_error([q(V), ' is not a meta-variable.']).


%% check_term(Term).
%% check_terms(TermList).
%%
%%  Succeeds iff Term (or each member of TermList, in the second case)
%%  is a well-formed term of the current theory.
%%
%%  mode check_term(@term).
%%  mode check_terms(@list(term)).
%%
check_terms(T) :-
    \+ (member(Term, T),
        \+ check_term(Term)
       ).

check_term(Term) :- 
    %% simplify_term(Term, SimpTerm),
%% TO DO: is a simplify_term call necessary/desirable in check_term?
    check_subterm(Term).

check_subterm(Term) :- 
    atom(Term),
    !,
    check_or_error(
        known_symbol(Term, []), 
        ['unknown constant ',q(Term),'.']).
check_subterm(Term) :-
    sub(Term),
    !,
    substitute(Term, Sub, SubTerm),
    check_subterm(SubTerm),
    check_sub(Sub).
check_subterm(Term) :- 
    var(Term),
    %% and \+ sub(Term).
    !.
check_subterm(Term) :- 
    obvar(Term),
    %% and \+ sub(Term).
    !.
check_subterm(Term) :- 
    quant(Term),
    %% and \+ sub(Term).
    !,
    quantify(Term, Q, Vars, Body),
    check_atom(Q),
    check_bndvar_list(Vars, QArity, _, Types),
    check_or_error(
        known_symbol(Q, [quant(QArity)]),
        ['unknown quantifier ',q(Q),' of arity ',q(QArity)]),
    check_terms(Types),
    check_term(Body).
check_subterm(Term) :-
    is_param_const(Term),
    !,
    param_const(Term,Name,_),
    check_atom(Name),
    known_symbol(Term, when(_)).
check_subterm(Term) :- 
    functor(Term, Func, N),
    %%% PJR - testing higher-order stuff
    ( var(Func)
      ->
        true
      ;
        check_atom(Func),
        check_or_error(
            known_symbol(Func, [N]), 
            ['unknown function ',q(Func),'/',q(N),'.'])
    ),
    check_term_args(Term, N).


check_sub(Sub) :-
    \+ (member(H, Sub),
        member((R/_), H),
        \+ check_term(R)).




check_term_args(Term, N) :-
    (N > 0 ->
        arg(N, Term, Arg),
        check_term(Arg),
        N1 is N - 1,
        check_term_args(Term, N1)
    ;
        true
    ).



%% check_bndvar_list(List, Arity, ObVars, Types).
%%
%%  Succeeds if List is a legal closed bound variable list.
%%  Arity is a list of 'typed' or 'untyped' atoms, describing
%%  which members of List have types.
%%
%%  mode check_bndvar_list(@list(term),?list(atom),?list(obvar),?list(term)).
%%  mode check_bndvar_list(?list(term),@list(atom),?list(obvar),?list(term)).
%%
check_bndvar_list(V, Arity, _, Types) :-
    var(V),
    var(Arity),
    !,
    Types = [].
check_bndvar_list([], [], [], []) :-
    !.
check_bndvar_list([!x|BL], [untyped|Arity2], [!x|ObVars2], Types) :-
    !,
    check_bndvar_list(BL, Arity2, ObVars2, Types).
check_bndvar_list([!x:T|BL], [typed|Arity2], [!x|ObVars2], [T|Types2]) :-
    !,
    check_bndvar_list(BL, Arity2, ObVars2, Types2).
check_bndvar_list(V, _, _, _) :-
    nonvar(V),
    !,
    handle_error(['illegal bound variable list: ', q(V)]).
check_bndvar_list(_, A, _, _) :-
    nonvar(A),
    !,
    handle_error(['illegal quantifier arity: ', q(quant(A))]).


%% bound_vars_to_arity(BoundVars, VarsArity).
%%
%%  If BoundVars is the bound variable list from a quantifier,
%%  then quant(VarsArity) is the arity of that quantifier.
%%
%%  mode bound_vars_to_arity(@list(nonvar), ?list(ground)).
%%
bound_vars_to_arity(Vars, _) :-
    var(Vars),
    !.
bound_vars_to_arity([], []).
bound_vars_to_arity([X:_|Vars], [typed|BArity]) :-
    obvar(X),
    bound_vars_to_arity(Vars, BArity).
bound_vars_to_arity([X|Vars], [untyped|BArity]) :-
    obvar(X),
    bound_vars_to_arity(Vars, BArity).

/* NOT USED.
 * The utility of a skeletal quantification is dubious.
%% private ergo_term(Term, Thy, Name, Arity).
%%
%%  Term has functor Thy.Name and Arity.
%%
%%  mode ergo_term(+thy(nonvar), ?theory, ?atom, ?arity).
%%  mode ergo_term(-thy(nonvar), +theory, +atom, +arity).
%%
ergo_term(Term, Thy, Name, Arity) :-
    atom(Term),
    !,
    internal_name(Term, Thy, Name),
    Arity = [].
ergo_term(Term, Thy, Name, Arity) :-
    compound(Term),
    !,
    functor(Term, Fun, Ar),
    internal_name(Fun, Thy, Name),
    Arity = [Ar].
ergo_term(Term, Thy, Name, Arity) :-
    quant(Term),
    !,
    quantify(Term, Quant, Vars, _),
    internal_name(Quant, Thy, Name),
    bound_vars_to_arity(Vars, BArity),
    Arity = [quant(BArity)].
ergo_term(Term, Thy, Name, []) :-
    !,
    internal_name(Term, Thy, Name).
ergo_term(Term, Thy, Name, [Arity]) :-
    integer(Arity),
    !,
    internal_name(Fun, Thy, Name),
    functor(Term, Fun, Arity).
ergo_term(Term, Thy, Name, [quant(BArity)]) :-
    !,
    internal_name(Quant, Thy, Name),
    arity_to_bound_vars(BArity, Vars),
    quantify(Term, Quant, Vars, _).

%% arity_to_bound_vars(VarsArity, BoundVars).
%%
%%  If quant(VarsArity) is the arity of a quantifier, then BoundVars is a
%%  skeletal bound variable list for that quantification.
%%
%%  mode arity_to_bound_vars(@list(ground), ?list(nonvar)).
%%
arity_to_bound_vars([], []).
arity_to_bound_vars([typed|BArity], [_X:_T|Vars]) :-
    arity_to_bound_vars(BArity, Vars).
arity_to_bound_vars([untyped|BArity], [_X|Vars]) :-
    arity_to_bound_vars(BArity, Vars).
*/


%% arity_description(Arity, Kind).
%%
%%  Kind is an atom that describes terms whose arity is Arity.
%%  The possible values of Kind are: constant, quantifier, function.
%%
%%  mode arity_description(+arity, ?atom).
%%
arity_description([],           constant  ) :- !.
arity_description(when(_),      param_const) :- !.
arity_description([quant(_)|_], quantifier) :- !.
arity_description([N|_],        function  ) :- integer(N).



%% Procedures for checking opening rule syntax.
%%

check_ruleterm([]) :-
    !.
check_ruleterm([Term|List]) :-
    Term = '--->'(Term1,Term2),
    check_term(Term1),
    check_term(Term2),
    check_ruleterm(List).

check_ruleequ([]) :-
    !.
check_ruleequ(['--->'(Eq1,Eq2)|List]) :-
    check_preorder(Eq1),
    check_preorder(Eq2),
    check_ruleequ(List).


%% check_preorder(Rel).
%%
%%  Succeeds iff Rel is a known preorder of the current theory,
%%  or the reverse (rev/1) of such a preorder.
%%
%%  mode check_preorder(@atom).
%%  mode check_preorder(@gcomp).
%%
check_preorder(Rel) :-
    is_preorder(Rel),
    !.
check_preorder(Rel) :-
    handle_error(['unknown preorder relation ',q(Rel),'/2.']).



%% is_param_const(Term).
%%
%%  Succeeds iff Term is a parameterized constant.
%%
%%  mode is_param_const(@compound).
%%
is_param_const(Term) :-
    compound(Term),
    functor(Term, F, 2),
    F == param_const.


%% param_const(Term, Name, Value).
%%
%%  Term is a parameterized constant, containing Name and Value.
%%
%%  mode param_const(?compound, ?atom, ?term).
%%
param_const(param_const(Name,Value), Name, Value).
