%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: theory_dag.ql,v 1.91 2003/12/19 01:32:49 itp Exp $

#include "../lib/ergo.h"

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% The procedures in this file manage the Ergo theory graph.
%%
%% There are two kinds of theories: 'real' theories and 'subtheories'.
%% A _subtheory_ (the name may change) is a specially named group of theorems
%% (and, later, perhaps some other things, including definitions) within some
%% owner theory.  Everything in the subtheory is conceptually also in the owner
%% theory, though we may later provide mechanisms to hide subtheories.  One
%% never switches (`cd') to a subtheory; rather, one switches to its owner
%% theory, then chooses to add theorems to the subtheory and use rules from it.
%%
%% There are two kinds of (directed) arcs in the theory graph: inheritance 
%% and interpretation.  The inheritance arcs are constrained to be acyclic
%% and must always go from a real theory to another real theory.
%% The interpretation arcs can go from any real theory (and the subtheories
%% of that theory are interpreted as well) to a subtheory of any other
%% real theory.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% add_theory(Name).
%%
%%  Add the theory `Name' to the dbms.
%%
%%  mode add_theory(@atom).
%%
add_theory(Name) :- 
    check_atom(Name),
    check_or_error(
        legal_theory_name(Name),
        [q(Name),' is not a legal theory name.']),
    check_or_error(
        \+ is_theory_name(Name),
        ['there is already a theory called ',q(Name),'.']),
    create_theory_database(Name),
    create_theory_searcher(Name),
    assert_theory(status(open), Name),
    assert_global(ancestor(Name, [])),
    atom_concat2(ancestor_of_, Name, ANC),
    assert_global(ANC(Name)),
    announce_event(theory(Name,created)),
    cd(Name).

%% legal_theory_name(Name).
%%
%%  Succeeds if Name is an atom composed of [A-Za-z0-9_] chars
%%  and it is not reserved by Ergo.
%%
%%  mode legal_theory_name(+atom).
%%
legal_theory_name(Name) :-
    atom_codes(Name, Codes),
    \+ (member(C, Codes),
        \+ legal_theory_name_char(C)),
    \+ invalid_theory_name(Name).

legal_theory_name_char(C) :- 0'A =< C, C =< 0'Z.
legal_theory_name_char(C) :- 0'a =< C, C =< 0'z.
legal_theory_name_char(C) :- 0'0 =< C, C =< 0'9.
legal_theory_name_char(95).  % underscore.
legal_theory_name_char(35).  % hash.  (To be removed when Ergo supports
                             %         nested subtheories).

% table of unwanted theory/subtheory names.
invalid_theory_name(conj).
invalid_theory_name(curr).
invalid_theory_name(global).
invalid_theory_name(hyp).
invalid_theory_name(proof).
invalid_theory_name(user).


%% delete_theory(Name).
%%
%%  Delete the theory `Name' (and its subtheories) from the dbms.
%%
%%  mode delete_theory(@atom).
%%
delete_theory(Name) :-
    %% check that Name is a known theory.
    is_real_theory_name(Name),
    !,
    %% check that Name has no children.
    check_or_error(
        \+ query_global(parent(_, Name)),
        ['you must delete dependent theories first.']),
    %% check that Name is not the current theory.
    check_or_error(
        \+ global_state_lookup(current_dir,Name),
        ['you cannot delete the current theory!']),
    %% Check that there are no suspended proofs for this theory.
%% TO DO: Check in delete_theory that there are no suspended proofs
%% in the subtheories too!
    check_or_error(
        ( get_current_proofs(Proofs),
          \+ member(Name:_, Proofs)),
        ['there are suspended proofs in theory ',q(Name)]),
    delete_subtheories_of(Name),
    atom_concat2(ancestor_of_, Name, ANC),
    retractall_global(ANC(_)),
    delete_theory_database(Name),
    retractall_global(ancestor(Name,_)),
    retractall_global(parent(Name,_)),
    retractall_global(subthy_owner(_,Name)),
    retractall_global(needs_saving(Name)),
    announce_event(theory(Name,deleted)).
delete_theory(Name) :-
    writelist(['Theory ',q(Name),
               ' does not exist, so delete was ignored.',nl]).

%% delete_theory_and_dependents(Thy).
%%
%%  Delete the theory `Thy' and all its descendants from the dbms.
%%  Note: this predicate succeeds (with a warning message) if Thy
%%  does not exist.
%%
%%  mode delete_theory_and_dependents(@atom).
%%

delete_theory_and_dependents(Thy) :- 
    debug(delete_theory, ['delete theory ',w(Thy)]),
    check_or_error(
        \+ ancestor(Thy),
        ['get out of the current theory before deleting it!']),
    (   query_global(parent(Child, Thy)),
        debug(delete_theory, ['delete theory recurs ',w(Child)]),
        delete_theory_and_dependents(Child),
        fail
    ;
        true
    ),
    !,
    delete_theory(Thy).


%% inherit_theory(Parent).
%%
%%  Inherit Parent into the current theory.
%%  Fails if this would introduce circularities in the inheritance graph.
%%
%%  Functions etc. in inherited theories can never clash, because their
%%  names always include the theory name.
%%
%%  mode inherit_theory(@atom).
%%
inherit_theory(Parent) :-
    check_open(CD),
    check_is_real_theory_name(Parent),
    check_or_error( 
        \+  ancestor(Parent), 
        [q(CD), ' already inherits ', q(Parent)]),
    query_global(ancestor(Parent, PList)),
    query_global(ancestor(CD, CList)),
    union_list([Parent|PList], CList, FinalList),
    check_or_error(
        \+ member(CD, FinalList),
        ['inheriting ',q(Parent),' would give a circular theory structure.']),
    global_state_set(current_dir,none),
    unwind_protect(
        load_ancestors(Parent),
        global_state_set(current_dir,CD)),
    atom_concat2(ancestor_of_, CD, ANC),
    retractall_global(ancestor(CD, _)),
    assert_global(ancestor(CD, FinalList)),
    retractall_global(ANC(_)),
    assert_global(ANC(CD)),
    (member(AThy, FinalList), assert_global(ANC(AThy)),fail;true),
    assert_global(parent(CD, Parent)),
    announce_event(theory(CD, inherited(Parent))).


%% add_subtheory(SubThy).
%%
%%  Create a new, empty subtheory of the current theory, named SubThy.
%%  Unlike add_theory, this does not select the new theory.
%%
%%  mode add_subtheory(@atom).
%%
add_subtheory(Name) :-
    check_atom(Name),
    check_or_error(
        legal_theory_name(Name),
        [q(Name),' is not a legal theory name.']),
    check_or_error(
        \+ is_theory_name(Name),
        ['there is already a theory called ',q(Name),'.']),
    check_current_theory(CD),
    check_is_real_theory_name(CD),
    % subtheories are treated as `ancestors' of their owner theories, since
    % the subthy should be searched whenever the owner is.
    % subthy_owner(Sub, Thy) -- the owner of Sub is Thy.
    assert_global(subthy_owner(Name, CD)),
    create_theory_database(Name),
    create_theory_searcher(Name),
    assert_theory(status(open), Name),
    % now we have created the subtheory, but its ancestor list, and the
    % ancestor list of the owning theory and all its other subtheories
    % must be updated.
    retract_global(ancestor(CD, CList)),
    union_list([Name],    CList,   FinalList),
    assert_global(ancestor(CD, FinalList)),
    atom_concat2(ancestor_of_, CD, ANC),
    retractall_global(ANC(_)),
    assert_global(ANC(CD)),
    (member(AThy, FinalList), assert_global(ANC(AThy)),fail;true),
    update_subtheory_ancestors(CD, FinalList),
    announce_event(theory(Name,created)).

%% private update_subtheory_ancestors(Thy, ThyAncestors).
%%
%%  Updates any ancestor lists in all subtheories of Thy.
%%
%%  mode update_subtheory_ancestors(@theory, @list(theory)).
%%
update_subtheory_ancestors(Thy, An) :-
    findall(SubThy, query_global(subthy_owner(SubThy, Thy)), SubTheories),
    update_all_subtheory_ancestors(SubTheories, Thy, An).

update_all_subtheory_ancestors([], _Thy, _An).
update_all_subtheory_ancestors([SubThy|SubTheories], Thy, An) :-
    delete(SubThy, An, SubAn),
    retractall_global(ancestor(SubThy, _)),
    FinalList = [Thy|SubAn],
    assert_global(ancestor(SubThy, FinalList)),
    atom_concat2(ancestor_of_, SubThy, ASubThy),
    retractall_global(ASubThy(_)),
    assert_global(ASubThy(SubThy)),
    (member(AThy, FinalList), assert_global(ASubThy(AThy)), fail;true),
    update_all_subtheory_ancestors(SubTheories, Thy, An).


%% delete_subtheory(Name).
%%
%%  Delete the subtheory `Name' from the dbms.
%%  This should not be callable by users, since it may delete things
%%  (eg. symbols declared in Name) that the owning theory uses.
%%
%%  mode delete_subtheory(@atom).
%%
delete_subtheory(Name) :-
    check_or_error(
        query_global(subthy_owner(Name, Owner)),
        [q(Name),' is not a subtheory.']),
    delete_subtheory(Name, Owner).


%% private delete_subtheory(Name, Owner).
%%
%%  mode delete_subtheory(@atom, @atom).
%%
delete_subtheory(Name, Owner) :-
    retractall_global(subthy_owner(Name,Owner)),
    atom_concat2(ancestor_of_, Owner, AOwner),
    retract_global(AOwner(Name)),
    atom_concat2(ancestor_of_, Name, AName),
    retractall_global(AName(_)),
    retractall_global(ancestor(Name,_)),
    %% now update the owner's ancestor list.
    retract_global(ancestor(Owner, OwnerAncestors0)),
    delete(Name, OwnerAncestors0, OwnerAncestors),
    assert_global(ancestor(Owner, OwnerAncestors)),
    update_subtheory_ancestors(Owner, OwnerAncestors),
    delete_theory_database(Name),
    announce_event(theory(Name,deleted)).


%% private delete_subtheories_of(Name).
%%
%%  Delete all subtheories of `Name' from the dbms.
%%  This should only be called when the owner theory is also
%%  about to be deleted, since it does not update the owner's
%%  ancestor information correctly.
%%
%%  mode delete_subtheories_of(@atom).
%%
delete_subtheories_of(Owner) :-
    check_is_real_theory_name(Owner),
    query_global(subthy_owner(Sub, Owner)),  %% NB retract does not backtrack!
    retract_global(subthy_owner(Sub, Owner)),
    atom_concat2(ancestor_of_, Sub, ASub),
    retractall_global(ASub(_)),
    retractall_global(ancestor(Sub,_)),
    delete_theory_database(Sub),
    announce_event(theory(Sub,deleted)),
    fail.
delete_subtheories_of(_).


%% ancestor_of(Thy, Ancestor).
%% strict_ancestor(Thy, Ancestor).
%% real_strict_ancestor(Thy, Ancestor).
%%
%%  Succeeds iff Ancestor is inherited (directly or indirectly) by Thy.
%%  Note that ancestor_of is reflexive, but strict_ancestor is not.
%%  Also note that subtheories are returned as ancestors, except by
%%    real_strict_ancestor.
%%
%% ancestor(Ancestor).
%%  Ancestor is an ancestor of the current theory.  Reflexive.
%%
%%  mode ancestor_of(@theory, ?theory).
%%  mode strict_ancestor(@theory, ?theory).
%%  mode real_strict_ancestor(@theory, ?theory).
%%  mode ancestor(?theory).
%%
ancestor_of(CD, PT) :-
    query_global(ancestor_of(CD, PT)).

strict_ancestor(CD, PT) :-
    query_global(ancestor_of(CD, PT)),
    CD \= PT.

% This works because subtheories never have ancestors of their own.
real_strict_ancestor(CD, PT) :-
    strict_ancestor(CD, PT),
    \+ query_global(subthy_owner(PT, _)).

ancestor(PT) :-
    global_state_lookup(current_dir,CD),
    CD \= none,
    query_global(ancestor_of(CD, PT)).


%% is_theory_name(Name).
%% is_real_theory_name(Name).
%% is_subtheory_name(Name).
%% check_is_theory_name(Name).
%% check_is_real_theory_name(Name).
%% check_is_subtheory_name(Name).
%%
%%  Succeeds iff Name is the name of a known (though not necessarily
%%  loaded) theory or subtheory.  The 'check_*' version generates an error
%%  message when it fails.  The `*real*' versions restrict the test to genuine
%%  theories (and exclude subtheories).  The `*subtheory*' versions restrict
%%  the test to subtheories.
%%
%%  mode is_theory_name(@atom).
%%  mode is_real_theory_name(@atom).
%%  mode is_subtheory_name(@atom).
%%  mode check_is_theory_name(@atom).
%%  mode check_is_real_theory_name(@atom).
%%  mode check_is_subtheory_name(@atom).
%%
is_theory_name(Name) :- 
    atom(Name),
    query_global(ancestor(Name, _)).

is_real_theory_name(Name) :- 
    is_theory_name(Name),
    \+ is_subtheory_name(Name).

is_subtheory_name(Name) :-
    atom(Name),
    query_global(subthy_owner(Name, _)).

check_is_theory_name(Name) :-
    check_or_error(
        is_theory_name(Name),
        ['unknown theory: ',q(Name)]).

check_is_real_theory_name(Name) :-
    check_is_theory_name(Name),
    check_or_error(
        \+ is_subtheory_name(Name),
        [q(Name),' is a subtheory.']).

check_is_subtheory_name(Name) :-
    check_or_error(
        is_subtheory_name(Name),
        ['unknown subtheory: ',q(Name)]).



%% check_ancestor(Thy).
%%
%%  Succeeds iff there is a current theory and Thy is one of its
%%  ancestors or a subtheory of one of its ancestors (including itself).
%%
%%  mode check_ancestor(@theory).
%%
check_ancestor(Thy) :-
    atom(Thy),
    ancestor(Thy),
    !.
check_ancestor(Thy) :-
    handle_error([q(Thy),' is not inherited by the current theory.']).


%% check_strict_ancestor(Thy).
%%
%%  Succeeds iff there is a current theory and Thy is one of its
%%  ancestors (not a subtheory and not the current theory itself).
%%
%%  mode check_strict_ancestor(@theory).
%%
check_strict_ancestor(Thy) :-
    atom(Thy),
    ancestor(Thy),
    \+ global_state_lookup(current_dir,Thy),
    \+ is_subtheory_name(Thy),
    !.
check_strict_ancestor(Thy) :-
    global_state_lookup(current_dir,CD),
    !,
    handle_error([q(Thy),' is not a strict ancestor of ',q(CD)]).
check_strict_ancestor(_) :-
    !,
    handle_error(['no current theory']).


%% cd(Thy).
%%
%%  Change to the given theory.
%%  If Thy=global, this will deselect any current theory.
%%
%%  mode cd(+atom).
%%
cd(_) :-
    in_proof,
    !,
    handle_error(['Cannot change theories within a proof.',nl,
                  'Use the `suspend'' command to suspend the current proof.']).
cd(global) :-
    !,
    global_state_set(current_dir,none).
cd(Name) :- 
    global_state_lookup(current_dir,Name),
    !.
cd(Name) :-
    check_is_theory_name(Name),
    debug(cd, ['checked name']),
    global_state_set(current_dir,none),
    load_ancestors(Name),
    debug(cd, ['loaded ancestors']),
    global_state_set(current_dir,Name),
    announce_event(theory(Name,selected)).


%% load_ancestors(Name).
%%
%%  Load theory Name, along with its ancestors and subtheories.
%%  There must not be any current theory selected when this is called.
%%
%%  mode load_ancestors(+atom).
%%
load_ancestors(Name) :-
    loaded_theory(Name,_),
    !.
load_ancestors(Name) :-
    % first load the real ancestors ...
    \+  (query_global(parent(Name, T)),
         \+  load_ancestors(T)),
    % next, if this is a subtheory, load its owner theory ...
    \+  (query_global(subthy_owner(Name, T)),
         \+ load_ancestors(T)),
    % then load the theory itself, ...
    load_theory(Name),
    % and finally its subtheories.
    \+  (query_global(subthy_owner(T, Name)),
         \+ load_ancestors(T)).


%% private load_theory(Name).
%%
%%  Loads a single theory.
%%  There must not be any current theory selected when this is called, and
%%  there will be none when it returns.  The newly-loaded theory will be
%%  selected temporarily while the theory-specific tactics are loaded.
%%
%%  mode load_theory(+atom).
%%
load_theory(Name) :-
    loaded_theory(Name,_),
    !.
load_theory(Name) :-
    create_theory_database(Name),
    consult_thy(Name, V, P),
    check_theory_file_version(Name,V,P),
    load_tactics(Name),
    announce_event(theory(Name,loaded)).


%% private check_theory_file_version(Name,Version,Path).
%%
%%  Decides if old theory files should be updated.
%%
check_theory_file_version(_Name,Version,_Path) :-
    theory_file_version(Version),
    !.
check_theory_file_version(Name,_,Path) :-
    (Path == '.' ->
        touch(Name),
        writelist(['Theory ',q(Name),' is stored in an old format.',
                   '  It will be updated when it is next saved.', nl])
    ;
        theory_file_version(Ver),
        writelist(['Theory ',q(Name),' is stored in an old format.',nl,
                   '  Please ask its owner to update it to version ',w(Ver),
                   ' format.', nl])
    ).


%% private load_tactics(Name).
%%
%%  Loads/consults all tactic files associated with theory Name.
%%  There must not be any current theory selected when this is called.
%%  However, so that goals and theory terms in tactic files work correctly,
%%  this procedure sets the current theory to Name while loading the files, and
%%  announces the selection of a theory (so that theory abbreviations and
%%  operators are set up for the tactic file).
%%  
%%  mode load_tactics(@theory).
%%
load_tactics(Name) :-
    PRE(load_tactics, global_state_lookup(current_dir,none)),
    debug(cd, ['loading tactics for ', w(Name)]),
    %% The comment below relates to an earlier version of Ergo.
    %% The current version does not seem to suffer from this problem.
    %% To avoid unnecessary loading we first check if there are any tactic
    %% files to load.
    %% We'd like to do that for the sake of operators, but it is ridiculously
    %%   expensive:
    %% From first Ergo prompt, to do `cd(zsyntax).' takes 37 sec without the
    %% announcement, and 22 minutes with it!  Enable_operators is the
    %% bottleneck.
    ( query_theory(tactic_file(_), Name) ->
        global_state_set(current_dir,Name),
        announce_event(theory(Name,selected)),
        (
            query_theory(tactic_file(File), Name),
            concat_atom([Name, '/', File], FName), 
            \+(loaded(FName,_,_)),
            operate_tactic_file(Name, load, File),
            fail
        ;
            true
        ),
        !,
        global_state_set(current_dir,none)
    ;
        true
    ).

%% load_global_tactics.
%%
%%  Loads/consults all global tactic files.
%%
%%  mode load_global_tactics.
%%
load_global_tactics :-
    query_global(tactic_file(File)),
    \+ loaded(File,_,_),
    operate_tactic_file(global, load, File),
    fail.
load_global_tactics.

%% close_theory.
%% close_theory(Thy).
%% 
%%  Change the status of theory/subtheory Thy to 'closed'.
%%  It is allowable for Thy to be closed already.
%%  The single argument form defaults Thy to the current theory.
%%
%%  mode close_theory.
%%  mode close_theory(@theory).
%%
close_theory :-
    check_current_theory(Thy),
    close_theory(Thy).

close_theory(Thy) :-
    retractall_theory(status(_), Thy),
    assert_theory(status(closed), Thy).


%% check_current_theory(Thy).
%%
%%  Succeeds iff there is a current theory and Thy is its name.
%%
%%  mode check_current_theory(?atom).
%%
check_current_theory(Thy) :-
    check_or_error(
        (global_state_lookup(current_dir,Thy), Thy \= none),
        ['no current theory.']).


%% check_open.
%% check_open(Thy).
%% 
%%  Checks that the current theory is open.
%%  Thy is the name of the current theory.
%%  Error messages are generated if there is no current
%%  theory, or the current theory is closed.
%%
%%  mode check_open.
%%  mode check_open(?atom).
%%
check_open :-
    check_open(_).

check_open(Theory) :-
    check_or_error(
        (global_state_lookup(current_dir,Theory), Theory \= none),
        ['no current theory.']),
    check_or_error(
        query_theory(status(open), Theory),
        ['theory ',q(Theory),' is closed.']).



op_arity(xf, 1).
op_arity(yf, 1).
op_arity(fy, 1).
op_arity(fx, 1).
op_arity(xfy, 2).
op_arity(yfx, 2).
op_arity(xfx, 2).
op_arity(A, A).


%% is_internal_name(Name).
%% internal_name(InternalName, TheoryName, Name).
%%
%%  InternalName is an atom corresponding to 'TheoryName.Name'.
%%  is_internal_name succeeds when Name has an embedded period
%%  (and it is not the first character in Name).
%%
%%  mode is_internal_name(@atom).
%%  mode internal_name(@atom, ?atom, ?atom).
%%  mode internal_name(?atom, @atom, @atom).
%%
is_internal_name(IName) :-
    atom(IName),
    atom_search(IName,1,'.',Pos),
    Pos > 1.

internal_name(IName, Thy, Name) :-
    atom(IName),
    !,
    split_atom(IName, Thy, '.', Name).
internal_name(IName, Thy, Name) :-
    %% construct IName from Thy and Name.
    PRE(internal_name, (atom(Name),atom(Thy))),
    concat_atom([Thy,'.',Name], IName).


%% private split_atom(Atom, Part1, Sep, Part2).
%%
%%  Succeeds if concat_atom([Part1,Sep,Part2], Atom) is true.
%%  Sep must be a single character atom and both Part1 and
%%  Part2 must be at least one character long.
%%
%%  mode split_atom(@atom, ?atom, @atom, ?atom).
%%
split_atom(IName, Thy, Sep, Name) :-
    atom(IName),
    !,
    %% split IName into Thy and Name, if possible.
    atom_search(IName,1,Sep,Pos),
    Pos > 1,
    atom_length(IName, Len),
    ThyLen  is Pos - 1,
    PosPlus1 is Pos + 1,
    NameLen is Len - Pos,
    NameLen > 0,
    sub_atom(IName, 1, ThyLen, Thy),
    sub_atom(IName, PosPlus1, NameLen, Name).


current_dir(T) :-
    global_state_lookup(current_dir,T), 
    T \= none.
