%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: expansions.ql,v 1.54 2001/06/04 01:24:08 itp Exp $

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% WARNING!
%% Some expansions should be able to run OUTSIDE Ergo.
%% Beware of using Ergo-defined procedures in this file!
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% This file contains procedures for managing the different kinds of term
%% expansion.
%%
%% There are five kinds of expansion, each defined in its own file:
%% 
%%  1. DECLARATION EXPANSIONS: for Ergo command declarations.
%%    Defined in decl_expansions.ql.
%% 
%%  2. MACRO EXPANSIONS.
%%    Defined in macro_expansions.ql.
%% 
%%  3. THEORY EXPANSIONS: for @thy(...) terms.
%%    Defined in thy_expansions.ql.
%%
%%  4. PREFIX EXPANSIONS: for adding module prefix to procedures.
%%    Defined in prefix_expansions.ql.
%%
%%  5. SINGLETON VARIABLE CHECKING.
%%    Defined in uniqvars.ql.
%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% consult_with_expansions(Expansions, File).
%% reconsult_with_expansions(Expansions, File).
%% compile_with_expansions(Expansions, File).
%% expand_with_expansions(Expansions, File).
%%
%%  Consult/reconsult/compile/expand File with Expansions turned on.
%%  Expansions can include:
%%    decl(ThyOrGlobal), macro, thy(Thy), prefix(Thy), uniqvars.
%%
%%  mode consult_with_expansions(@list(nonvar), @atom).
%%  mode reconsult_with_expansions(@list(nonvar), @atom).
%%  mode compile_with_expansions(@list(nonvar), @atom).
%%  mode expand_with_expansions(@list(nonvar), @atom).
%%
consult_with_expansions(Exps, File) :-
    unwind_protect((setup_expansions(Exps, File),
                    consult(File)),
                   remove_expansions(Exps)).
reconsult_with_expansions(Exps, File) :-
    unwind_protect((setup_expansions(Exps, File),
                    reconsult(File)),
                   remove_expansions(Exps)).

compile_with_expansions(Exps, File) :-
    unwind_protect((setup_expansions(Exps, File),
                    fcompile(File, [compiler_heap(1000),
                                    string_table(1000),
                                    compiler_scratchpad(1000),
                                    assemble_only(true)])),
                   remove_expansions(Exps)).

expand_with_expansions(Exps, File) :-
    unwind_protect((setup_expansions(Exps, File),
                    fcompile(File, [compiler_heap(1000),
                                    compiler_scratchpad(1000),
                                    string_table(1000),
                                   expand_only(true)])),
                   remove_expansions(Exps)).


%% private setup_expansions(Expansions, File).
%%
%%  Set up Expansions for consulting or compiling File.
%%
%%  mode setup_expansions(@list(nonvar), @atom).
%%
setup_expansions([], _).
setup_expansions([Exp|Exps], File) :-
    ( Exp = decl(Thy) ->
        enable_declaration_expansions(Thy)
    ; Exp = macro ->
        enable_macro_expansions
    ; Exp = thy(Thy) ->
        enable_theory_expansions(Thy)
    ; Exp = prefix(Thy) ->
        enable_prefix_expansions(Thy, File)
    ; Exp = uniqvars ->
       enable_uniqvars_expansions
    ;
        exp_program_error(unrecoverable,
                          setup_expansions([Exp|Exps], File),
                          ['bad expansion ', w(Exp)])
    ),
    setup_expansions(Exps, File).

%% private remove_expansions(Exps).
%%
%%  Remove Ergo expansions Exps.
%%
%%  mode remove_expansions(@list(atom)).
%%
remove_expansions([]).
remove_expansions([Exp|Exps]) :-
    ( Exp = decl(_) ->
        disable_declaration_expansions
    ; Exp = macro ->
        disable_macro_expansions
    ; Exp = thy(_) ->
        disable_theory_expansions
    ; Exp = prefix(_) ->
        disable_prefix_expansions
    ; Exp = uniqvars ->
        disable_uniqvars_expansions
    ;
        exp_program_error(unrecoverable,
                          remove_expansions([Exp|Exps]),
                          ['remove_expansions: bad expansion ', w(Exp)])
    ),
    remove_expansions(Exps).

?- dynamic(ergo_expansions/1).

%% add_ergo_expansion(HigherGoal).
%% remove_ergo_expansion(HigherGoal).
%%
%%  Add/remove HigherGoal(Input, Output, InputVars) to/from the list of Ergo
%%  expansions.
%%
add_ergo_expansion(Goal) :-
    ergo_expansions(_),
    !,
    assertz(ergo_expansions(Goal)).
add_ergo_expansion(Goal) :-
    %% there were no prior goals,
    assertz(ergo_expansions(Goal)),
    add_linking_clause(term_expansion, do_ergo_expansion, 3).

remove_ergo_expansion(Goal) :-
    retract(ergo_expansions(Goal)),
    ergo_expansions(_),
    !.
remove_ergo_expansion(_) :-
    %% there are no remaining goals,
    del_linking_clause(term_expansion, do_ergo_expansion, 3).

%% do_ergo_expansion(Input, Output, InputVars).
%% do_ergo_expansion(Exps, Input, Output, InputVars).
%%
%%  Do expansion according to current ergo_expansions.
%%
do_ergo_expansion(T0, T, V) :-
    findall(Exp, ergo_expansions(Exp), Exps),
    do_ergo_expansion(Exps, T0, T, V).

do_ergo_expansion([], T, T, _).
do_ergo_expansion([Goal|Goals], T0, T, V) :-
    call_predicate(Goal(T0, T1, V)),
    !,
    do_ergo_expansion(Goals, T1, T, V).
do_ergo_expansion([_Goal|Goals], T0, T, V) :-
    %% the goal failed,
    do_ergo_expansion(Goals, T0, T, V).

%% nonvar_functor(Term).
%%
%%  Term is NOT a compound term with a variable functor.
%%  That is, Term is one of:
%%   -- a simple term,
%%   -- a compound term with a non-variable functor,
%%   -- a quantification,
%%   -- a substituted version of any of the above.
%%  
%%  mode nonvar_functor(@term).
%%
nonvar_functor(Term) :-
    compound(Term),
    !,
    functor(Term, Fun, _),
    nonvar(Fun).
nonvar_functor(_).

%% TO DO: Get rid of the temporary error handling mechanism for program and
%% data erros.
%% TEMPORARY DEFINITIONS:
%% Ergo should catch these exceptions and use its handle_error mechanism.
exp_program_error(Sev, Goal, Msg) :-
    exception(ergo_program(Sev, Goal, Msg)).

exp_data_error(Sev, Goal, Msg) :-
    append(Msg, [nl], Msg1),
    exception(ergo_data(Sev, Goal, Msg1)).

%% We need this here in case we run outside Ergo.
exp_debug(WriteList) :-
    ( clause(dbg(expansions), _) ->
        debug(expansions, WriteList)
    ;
        true
    ).
