%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: rule_dms.ql,v 1.10 2006/02/06 00:52:12 itp Exp $
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%  This file contains procedures for managing dependencies
%%  between theorems.  It assumes that for each proved theorem
%%  Thy:Thm a fact in theory Thy of the form:
%%
%%    infrule_status(Thm, [T1:N1,T2:N2,...,Tn:Nn]).
%%
%%  The theorems in the list are those that were used directly
%%  in the proof of Thy:Thm.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


library command
  dependent_rules(Rules, Dependents)
extends
  general_cmds
doc
  'Dependents is the list of all theorems whose proofs depend',nl,
  '(directly or indirectly) on Rules.',
  para,
  'Rules and Dependents are both lists of rule names of the form',nl,
  'Thy:Name, where Thy is a theory that is accessible from the',nl,
  'current theory and Name is the short name of a rule in Thy.'
mode
  dependent_rules(@list(rulename), ?list(rulename)).

dependent_rules(StartRules, Deps) :-
    %% We assume that the 'Deps@Table' array implicit parameter
    %% is unused and is thus free for use by this procedure.
    %% To help ensure this and clean up space, we wrap this
    %% whole procedure with calls to clear the array.
    %% We also cut to remove any choice points.
    ip_array_clear('Deps@Table'),
    dependent_rule2(StartRules, Deps),
    ip_array_clear('Deps@Table'),
    !.

dependent_rule2(StartRules, Deps) :-
    check_current_theory(_),
    list(StartRules),
    debug(dependencies, ['start rules are: ',q(StartRules)]),
    ( StartRules = [TopThy:_] ->
        check_ancestor(TopThy)
    ;
        StartRules = [_|_]
    ),
    dep_insert_start_rules(StartRules,StartVar),
    %%
    %% Now find all accessible theorems that are not in StartRules.
    %% If TopThy is nonvar, we reduce the search space by only looking
    %% in theories between TopThy and the current theory.  This is
    %% sound because the theorems in theories above TopThy cannot
    %% depend upon the rule in StartRules.  This optimization is
    %% harder when there are multiple start rules, so we don't bother.
    %% (The single rule case is used by proof_start_postulate, so
    %% is worth optimizing).
    findall(Thm,
            dep_non_startrule_theorem(TopThy,Thm),
            Theorems),
    %%
    %% Now build the dependency graph in the 'Deps@Table' array.
    %%   Each rule in StartRules is bound to start(FreshVar).
    %%   Each rule in Theorems, (relying directly on [T1,..,Tn] say),
    %%     (where Ti are in Theorems)
    %%     is bound to [dep(C1),..,dep(Cn)], where Ci is what Ti is bound to.
    debug(dependencies, ['Building dependency graph.  StartVar=',q(StartVar)]),
    %% Add theorems of Theorems to 'Deps@Table'
    add_theorems_to_deps_table(Theorems),
    map(dep_process_theorem, [Theorems]),
    %%
    %% Finally, return all rules in Theorems that are reachable
    %% from one of the StartRules (ie. from StartVar).  The reachability
    %% check is done by Qu-Prolog's occurs check.
    debug(dependencies, ['Filtering dependent theorems.']),
    filter(dep_upon(StartVar), Theorems, Deps).


dep_non_startrule_theorem(TopThy, Thy:Name) :-
    ancestor(Thy),
    ( atom(TopThy) ->
        ancestor_of(Thy, TopThy)
    ;
        true
    ),
    query_theory(infrule_status(Name, L), Thy),
    list(L),  % only return theorems
    dep_lookup_rule(Thy:Name,Value),
    ( std_nonvar(Value),Value = start(_) ->
        fail
    ;
        true
    ).

dep_process_theorem(Thy:Name) :-
    query_theory(infrule_status(Name, L), Thy),
    dep_lookup_rule(Thy:Name, dep(V)),
    get_all_dep_rule_vars(L, L2),
    V = L2.

get_all_dep_rule_vars([], []).
get_all_dep_rule_vars([H|T], [V|Vs]) :-
    dep_lookup_rule(H, Value),
    std_nonvar(Value),
    Value = _(V),
    !,
    get_all_dep_rule_vars(T, Vs).
get_all_dep_rule_vars([_|T], Vs) :-
    get_all_dep_rule_vars(T, Vs).

dep_upon(StartVar, Rule) :-
    dep_lookup_rule(Rule, Value),
    %%
    %% Value is now a tree of all the 'identity variables' of all
    %% the rules that are used in its proof (transitive closure).
    %% If the tree includes StartVar, then StartVar=Value will fail,
    %% because of the Qu-Prolog occurs check.  In this case we
    %% succeed, to indicate that Rule is used (directly or indirectly)
    %% to prove one or more of the rules in StartRules.
    debug(dependencies, [left(30,[q(Rule)]),' ',q(StartVar),' ?= ',q(Value)]),
    \+ StartVar = Value,
    debug(dependencies, ['    is dependent.']).

dep_insert_start_rules([], _).
dep_insert_start_rules([Thy:Name|Rules], StartVar) :-
    atom(Thy),
    atom(Name),
    !,
    dep_set_rule(Thy:Name,start(StartVar)),
    dep_insert_start_rules(Rules, StartVar).
dep_insert_start_rules([Rule|_], _) :-
    handle_error(['dependent_rules/2: ',q(Rule),' is not in Thy:Name form.']).

add_theorems_to_deps_table([]).
add_theorems_to_deps_table([Rule|Rules]) :-
    dep_set_rule(Rule, dep(_)),
    add_theorems_to_deps_table(Rules).   


    
%% private.  These procedures implement a table: ThyxName --> Value.
%%
%%  Since 'assignment' is done via unification, the lookup and set
%%  procedures are the same.  However, we distinguish them to make
%%  tracing easier.
%%
dep_lookup_rule(Thy:Name,Value) :-
    concat_atom([Thy,':',Name], Atom),
    ip_lookup('Deps@Table', Atom, Value),
    debug(dependencies, ['    ',left(30,[q(Atom)]),' = ',q(Value)]).

dep_set_rule(Thy:Name,Value) :-
    concat_atom([Thy,':',Name], Atom),
    ip_lookup('Deps@Table', Atom, Value0),
    debug(dependencies, [left(30,[q(Atom)]),' ',q(Value0),' := ',q(Value)]),
    Value0 = Value.
