%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% interpret(AbsTheory, SubTheory, Instantiations, Reparentings).
%%
%%  AbsTheory is the name of the abstract theory to be interpreted.
%%  The interpretation goes into the current theory, in SubTheory which is used
%%   to identify the interpretation in future.
%%  Instantiations is a list of AbsTerm-->ConTerm maplets for instantiating the
%%   declared constants and functions of the theory; the constraints on each
%%   AbsTerm and ConTerm are:
%%     1. AbsTerm must be either an atom or a compound term with all arguments
%%        distinct metavariables.
%%     2. ConTerm must be a legal term of the target (current) theory.
%%     3. ConTerm must contain no free object variables.
%%     4. ConTerm must contain no metavariables other than those in AbsTerm.
%%  Reparentings is a list of Theory-->SubTheory maplets that explains how each
%%   ancestor of AbsTheory is to be mapped to an existing interpreted SubTheory
%%   of the current theory.  If no symbols in some ancestor are to be
%%   instantiated, that ancestor need not be named here, and it becomes an
%%   ancestor of the current theory (if it is not one already).
%%
%%  mode interpret(+atom, +atom, @list(term), +list(term)).
%%
interpret(Abs, Sub, Inst0, Repar) :-
    % get Abs loaded: I'm scared to use load_ancestors directly, because it
    % seems to mess up the database.  But doing it this way wastes an awful lot
    % of time.
    current_dir(CD),
    process_reparentings(Repar, CD, Inst0, Inst),
    check_interpretation(Inst),
    cd(Abs),
    cd(CD),
    add_subtheory(Sub),
    !,
    record_interpretation(Abs, Sub, Inst),
    % interpret theory facts (fail-driven loop)
    ( theory_fact(Name, Arity),
      functor(Fact, Name, Arity),
      query_theory(Fact, Abs),
      interpret_fact(Name, Fact, Sub, Inst, Repar),
      fail
    ; true
    ).

%% interpret_fact(FactName, Fact, SubThy, Instns, Repars)
%% mode interpret_fact(+atom, @term, +atom, @list(term), +list(term))
%% Try to interpret a theory fact Fact (with principal functor FactName) into
%% SubThy using the given instantiations and reparentings.
% abbrev/4, compiled_rule/8, default_prover/2, equivalence/1, interpretation/2,
% interpreted_theory/1, object_vars/1, openrule/8, operator/4, preorder/1,
% reverse_of/2, rule/5, rulegroup/5, save_clauses/3, status/1, stronger_than/2,
% symbol/3, tactic_file/1, visible_theory/1
% Currently only interprets rules.
interpret_fact(rule, rule(Name, Type, Ass, Stmt, Body),
               Sub, Inst, Repar) :-
    !,
    interpret_rule(Type, Name, Ass, Stmt, Body, Sub, Inst).
interpret_fact(Name, _, _, _, _) :-
    ( member(Name, [status, operator, visible_theory]) -> true
    ; writelist(['Sorry, don''t know how to interpret ''', Name, '''.', nl]),
      fail
    ).

%% interpret_rule(Type, Name, AssumpList, Statement, Body, SubThy, Instms)
%% mode interpret_rule(@term, +atom, @term, @term, @term, +atom, @list(term))
%% Interpret a rule (axiom, postulate or theorem) into SubThy using the given
%% instantiations.
%%  Axiom     --> postulate
%%  Theorem   --> theorem (with no proof obligation)
%%  Postulate --> theorem?
interpret_rule(Type, Name, Ass, Stmt, Body, Sub, Inst) :-
    % do the instantiation
    interpret_terms([Stmt,Body|Ass], Sub, [StmtI,BodyI|AssI]),
    % check it's a legal term of the current theory (may be redundant as it is
    % done in check_interpretation??)
    check_term(StmtI),
    check_terms(AssI),
    collect_vars((AssI,StmtI), Vars),
    check_axiom_body(BodyI, Vars),
    ( Type = axiom ->
        assert_theory(rule(Name, postulate, AssI, StmtI, BodyI), Sub)
    ; Type = theorem(Deps) ->
        translate_dependencies(Deps, Sub, TDeps),
        assert_theory(rule(Name, theorem(TDeps), AssI, StmtI, BodyI), Sub)
    ; Type = postulate ->
        assert_theory(rule(Name, theorem([Thy:Name]), AssI, StmtI, BodyI), Sub)
    ; writelist(['Sorry, don''t know how to interpret ''', Type, '''.', nl])
    ).

%% interpret_terms(Terms0, SubThy, Terms).
%% mode interpret_terms(@list(term), +atom, -list(term)).
%% Apply SubThy's instantiation function recursively to Terms0 giving Terms.
interpret_terms([], _, []).
interpret_terms([T0|Ts0], Sub, [T|Ts]) :-
    interpret_term(T0, Sub, T),
    interpret_terms(Ts0, Sub, Ts).

%% interpret_term(Term0, SubThy, Term).
%% mode interpret_term(@term, +atom, -term).
%% Apply SubThy's instantiation function recursively to Term0 giving Term.
interpret_term(T0, Sub, T) :-
    transform_subterms(interpret_subterm(Sub), T0, T).

%% interpret_subterm(SubThy, Term0, Term).
%% mode interpret_term(+atom, @term, -term).
%% Apply SubThy's instantiation function once to Term0 giving Term.
% Yes, using query_global this way *is* faster than getting it once and using
% copy_term!  (But only by a small margin.)
interpret_subterm(Sub, T0, T) :-
    ( var(T0) -> T = T0
    ; query_theory(interpretation(T0, T), Sub) -> true
    ; T = T0
    ).

%% record_interpretation(AbsThy, SubThy, Instn)
%% mode record_interpretation(+atom, +atom, +list(term))
record_interpretation(Abs, Sub, []) :-
    assert_theory(interpreted_theory(Abs), Sub).
record_interpretation(Abs, Sub, [(A-->C)|Maps]) :-
    assert_theory(interpretation(A, C), Sub),
    record_interpretation(Abs, Sub, Maps).

%% check_interpretation(Instn)
%% mode check_interpretation(+list(term))
%% Check that Instn is a valid interpretation into the current theory.
check_interpretation([]).
check_interpretation([(Abs-->Con) | Map]) :-
    Abs =.. [F|AbsVars],
    check_distinct_metavars(AbsVars),
    check_term(Con),
    collect_free_obvars(Con, [], _),
    collect_metavars(Con, ConVars),
    subset_var(ConVars, AbsVars),
    check_interpretation(Map).

%% process_reparentings(Reparentings, CD, Inst0, Inst)
%% mode process_reparentings(+list(term), +atom, @list(term), -list(term))
%% Process Reparentings into CD, adding the inherited instantiations to Inst0
%% giving Inst.
process_reparentings([], _, Inst, Inst).
process_reparentings([(Abs-->Sub) | Repar], CD, Inst0, Inst) :-
    check_or_writelist(
        query_global(subthy_owner(Sub, CD)),
        ['Error: ', Sub, ' is not a subtheory of ', CD, '.', nl]),
    check_or_writelist(
        query_theory(interpreted_theory(Abs), Sub),
        ['Error: ', Sub, ' was not created by interpreting ', Abs, '.', nl]),
    bagof((A-->C), query_theory(interpretation(A,C), Sub), Inst1),
    append(Inst1, Inst0, Inst2),
    process_reparentings(Repar, CD, Inst2, Inst).

%% translate_dependencies(Deps, Sub, TDeps)
%% Take a list of theorem dependencies and translate it according to the
%% interpretation for Sub.
% not yet implemented.
translate_dependencies(Deps, _, Deps).
