%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: newvars.ql,v 1.35 2000/11/15 06:29:47 itp Exp $

#include "../lib/ergo.h"

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% The procedures in this file handle thawing and freezing of
%% proof variables, well-formedness checking of instantiations
%% of those variables, and promoting/demoting of proof variables.
%%
%% Proof variables (which includes meta-variables and object variables)
%% are split into two disjoint sets.  One set contains the "fixed" 
%% variables, which are typically the variables that appear in the
%% original statement of the theorem being proved.  The other set
%% contains the "new" variables, which are typically variables
%% that have been introduced during the proof.  Proof commands can
%% instantiate new variables or restrict them by adding not_free_in
%% constraints, but are not generally allowed to instantiate/restrict
%% fixed variables.  Most of the time, both kinds of proof variables
%% are frozen (to prevent instantiation), but the thawing_new_vars/1
%% procedure allows the new variables to be thawed during the 
%% execution of some code.
%%
%% "Promoting" a variable means moving it from the "new" set
%% into the "fixed" set, and "demoting" means the opposite.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% thawing_new_vars(Code).
%%
%%  Executes Code with all new variables unfrozen.
%%  If Code is non-deterministic, then thawing_new_vars will be too.
%%
%%  Any new variables added by Code must be put into the
%%  'newvars' implicit parameter, so that they can be syntax
%%  checked and added to the list of known new variables.
%%
%%  Calls to thawing_new_vars can be nested, though this may not be
%%  good style.  That is, Code may contain calls to thawing_new_vars.
%%
%%  Code is called using call_predicate, so it should not directly
%%  contain cuts or inline macros.
%%
%%  mode thawing_new_vars(+nonvar).
%%
thawing_new_vars(Code) :-
    ip_lookup(newvar_flag, true),
    !,
    ip_lookup(newvars, Old), 
    ip_set(newvar_flag, fail),
    ip_set(newvars, []),
    thaw_term(Old),
    call_predicate(Code),
    ip_lookup(newvars, L),
    check_terms(L),
    check_terms(Old),
    freeze_term((L,Old), NewVars),
    ip_set(newvars, NewVars),
    ip_set(newvar_flag, true).
thawing_new_vars(Code) :-
    %% newvar_flag is not true, so this call to thawing_new_vars
    %% has occurred when the new variables are already thawed!
    %% Consequently, we do not thaw/freeze them again.
    ip_lookup(newvars, Old),
    ip_set(newvars, []),
    Code,
    ip_lookup(newvars, L),
    check_terms(L),
    check_terms(Old),
    collect_thawvars((L,Old), NewVars),
    ip_set(newvars, NewVars).



%% private thawed_sublist(Vars, Tail, Thawed).
%% private thawed_sublist(Vars, Tail, Thawed_Union_Tail, Frozen).
%%
%%  This is only for internal use by the proof subsystem.
%%
%%  What is its mode?
%%
thawed_sublist(A, B, C) :- thawed_sublist(A, B, C, _).

thawed_sublist([], I, I, []).
thawed_sublist([A|B], I, C, [A|D]) :-
    frozen_var(A),
    !,
    thawed_sublist(B, I, C, D).
thawed_sublist([A|B], I, [A|C], D) :-
    thawed_sublist(B, I, C, D).



proof command
  inst_var(Var, Term)
extends
  var_proofcmds,
  var_treecmds
doc
  'Demotes Var (changing it from a fixed proof variable into a new',nl,
  'proof variable), then unifies Var with Term.'
mode
  inst_var(?thy(term), ?term_spec).

inst_var(Term1, Term2) :- 
    demote_var(Term1),
    thawing_new_vars(once((Term1 = Term2, incomplete_retry_delays))),
    accept_constraints.



proof command
  unify_focus_goal
extends
  var_proofcmds
doc
  'Demotes any fixed variables in the focus and goal of the current',nl,
  'window, then unifies the focus and the goal.'
mode
  unify_focus_goal.

unify_focus_goal :-
    is_focus(F),
    is_goal(G),
    demote_var(F),
    demote_var(G),
    thawing_new_vars(once((F = G, incomplete_retry_delays))),
    accept_constraints.


%% command match_terms(Term1, Term2).
%%
%%  Unifies Term1 and Term2 with new variables unfrozen.
%%
%%  mode match_terms(?thy(term), ?thy(term)).
%%
match_terms(Term1, Term2) :-
    thawing_new_vars(once((Term1 = Term2, incomplete_retry_delays))).


%% assimilate_new_vars(Term).
%%
%%  Freeze all (object and meta) variables in Term that are not already
%%  frozen and add them to the 'newvars' list (an implicit parameter).
%%
%%  mode assimilate_new_vars(@term).
%%
assimilate_new_vars([]) :-
    !.
assimilate_new_vars(Vars) :-
    thawing_new_vars(assimilate_new_vars2(Vars)).

assimilate_new_vars2(Vars) :-
    ip_lookup(newvars, L),
    collect_thawvars((L,Vars), NewVars),
    ip_set(newvars, NewVars).


proof command
  promote_var(Term)
extends
  var_proofcmds
doc
  'Promotes all the variables in Term so that they become ''fixed''',nl,
  'variables of the proof, rather than ''new'' variables.',nl,
  'This means that ordinary proof transformations will not be',nl,
  'able to instantiate these variables (until they are demoted).'
mode
  promote_var(@term).

promote_var(Var) :- 
    ensure_constraints_accepted(
        ['All recent constraints must be accepted before',nl,
         'any variables can be promoted.',nl]),
    internal_promote_var(Var),
    accept_constraints.


internal_promote_var(Var) :- 
    collect_allvars(Var, VarList),
    ip_lookup(fixedvars, PL),
    thawing_new_vars(
        (thawed_sublist(VarList, PL, NewPL),
          freeze_list(NewPL))),
    ip_set(fixedvars, NewPL).


proof command
  demote_var(Term)
extends
  var_proofcmds
doc
  'Changes all ''fixed'' proof variables that appear in Term into',nl,
  '''new'' variables of the proof.',nl,
  'This means that ordinary proof transformations will be',nl,
  'able to instantiate these variables.'
mode
  demote_var(@term).

demote_var(Var) :-
    ensure_constraints_accepted(
        ['All recent constraints must be accepted before',nl,
         'fixed variables can be demoted or instantiated.',nl]),
    ip_lookup(newvars, NL),
    ip_lookup(fixedvars, PL),
    collect_vars(Var, VarList),
%% TO DO: \+ '$check_delayed_problem'('='), ??? in demote_var.
    thaw_list(VarList),
    thawed_sublist(PL, NL, NewNL, NewPL),
    (ip_lookup(newvar_flag, true) -> freeze_vars(NewNL) ; true),
    ip_set(newvars, NewNL),
    ip_set(fixedvars, NewPL),
    accept_constraints.


%% is_newvarlist(VarList).
%%
%%  VarList is the current list of new proof variables.
%%
%%  mode is_newvarlist(?list(var)).
%%
is_newvarlist(NV) :-
    ip_lookup(newvars, NV).


%% is_fixedvarlist(VarList).
%%
%%  VarList is the current list of fixed proof variables.
%%
%%  mode is_fixedvarlist(?list(var)).
%%
is_fixedvarlist(PV) :-
    ip_lookup(fixedvars, PV).


