%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%%
%% CVS $Id: tactics.ql,v 1.47 2001/06/04 01:24:09 itp Exp $

#include "../lib/ergo.h"

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% This file defines various ways of invoking proof tactics and
%% determining the current 'prover' (ie. user interface tactic
%% for proofs).  The current prover is called whenever the proof
%% tactic being executed is a variable.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



%% set_prover(Name, Arity).
%%
%%  Set the current prover to be Name/Arity.
%%  That is, Name/Arity will be called whenever call_tac attempts to
%%  execute a variable.
%%  This procedure should only be called when a proof is in progress.
%%
%%  mode set_prover(@atom, @integer).
%%
set_prover(Name,Arity) :-
    check_prover_name(Name, Arity),
    ip_lookup(current_prover, prover(_,_,Num0)),
    (integer(Num0) ->
        Num is Num0 + 1
    ;
        Num = 0
    ),
    ip_set(current_prover, prover(Name,Arity,Num)).


%% get_prover(Name, Arity).
%%
%%  Name/Arity is the current prover that will be called whenever
%%  an uninstantiated tactic is found during a proof.
%%
%%  mode get_prover(?atom, ?integer).
%%
get_prover(Name, Arity) :-
    ip_lookup(current_prover, Prover),
    nonvar(Prover),
    Prover = prover(Name, Arity, _).


%% use_prover(Name, Arity, Code).
%%
%%  Execute Code with the prover set to Name/Arity.
%%  Restore the original prover if Code succeeds and the user
%%  has not done a set_prover call within Code.
%%
%%  mode use_prover(@atom, @integer, +compound).
%%
use_prover(Name, Arity, Code) :-
    check_prover_name(Name, Arity),
    ip_lookup(current_prover, prover(Name0,Arity0,Count0)),
    Count is Count0 + 1,
    ip_set(current_prover, prover(Name,Arity,Count)),
    call_tac(Code),
    !,
    (ip_lookup(current_prover, prover(Name,Arity,Count)) ->
        ip_set(current_prover, prover(Name0,Arity0,Count0))
    ;
        true
    ).


%% check_prover_name(Name, Arity).
%%
%%  Checks that Name is an atom and Arity is 0 or 1.
%%  Generates an error message otherwise.
%%
%%  mode check_prover_name(@atom, @integer).
%%
check_prover_name(Name, Arity) :-
    check_atom(Name),
    check_or_error(
        (
            Arity == 0
        ;
            Arity == 1
        ),
        ['prover must be ',q(Name),'/0 or ',q(Name),'/1.']).



%% command add_tactic_wrap_info(Thy, Call, Msg, SubMsgs).
%%
%%  Defines what information should be recorded in a proof record
%%  when the Call tactic is executed via wrap_tac.  Thy must be
%%  'global' or the name of a theory.
%%
%%  At the point in the proof where Call is passed to wrap_tac,
%%  Call and Msg are recorded.   Call contains the (unprefixed) name
%%  and arguments of the tactic, and Msg is a writelist that briefly
%%  describes the total effect of the tactic.
%%
%%  SubMsgs is a list of writelists.  One of these writelists is
%%  recorded at the point where this tactic starts a subproof that
%%  it does not control.  If the number of such subproofs is not
%%  fixed, then SubMsgs must be an open list.
%%
%%  Eg. add_tactic_wrap_info(jprop, 'jprop:cases'(E),
%%                           ['By case analysis on: ',thy(E)],
%%                           [['True case.'],['False case.']]).
%%
%%  mode add_tactic_wrap_info(@atom, @cmd, @writelist, @list(writelist)).
%%
add_tactic_wrap_info(Thy, Call, Msg, SubMsgs) :-
    check_atom(Thy),
    check_is_legal_cmd(Call),
    check_is_legal_writelist(Msg),
    check_or_error(
        list(SubMsgs),
        ['illegal subproof message list ',q(SubMsgs),' for tactic ',q(F/N)]),
    functor(Call,F,N),
    ( Thy = global ->
        InternalF = F,
        ICall = Call
    ;
        concat_atom([Thy,':',F], InternalF),
        functor(ICall, InternalF, N),
        same_args(Call, ICall, 1, N)
    ),
    make_wrap_list(SubMsgs, 1, CallId, SubProofs, WrappedSubProofs),
    ( SubMsgs == [] ->
        FinalCall = ICall
    ;
        N1 is N + 1,
        functor(FinalCall,InternalF,N1),
        same_args(Call, FinalCall, 1, N),
        arg(N1, FinalCall, WrappedSubProofs)
    ),
    update_theory(tactic_wrapper(ICall, SubProofs, 
                                 wrap_func(FinalCall,CallId,Msg,ICall)),
                  Thy).


%% private make_wrap_list(MsgList, ArgNum, OwnerNum, TacList, WrappedTacList).
%%
%%  Takes a (possibly open) list of sub-proof messages and generates
%%  two corresponding lists:
%%    - TacList is a list of variables that will eventually be
%%       instantiated into the tactics for each of the subproofs,
%%    - WrappedTacList contains the same variables, but with an
%%       appropriate wrap_arg/4 around each one.
%%
%%  ArgNum is the number of this subproof in the owning tactics
%%  list of subproofs (should be 1 initially).  OwnerNum is a
%%  number that will be allocated when the owning tactic is executed.
%%  It should be the same as the second argument to wrap_func/4.
%%  Its purpose is to uniquely identify that particular invocation of
%%  that tactic within the proof.  For the convenience of proof reading
%%  programs, make_wrap_list copies OwnerNum into all the subproof
%%  wrappers so that subproofs can be easily related to the tactic
%%  that caused them.
%%
%%  mode make_wrap_list(?list(writelist),@integer,@var,-list(var),-list(compound)).
%%
make_wrap_list(Label, _, _, _, _) :-
    var(Label),
    !.
make_wrap_list([], _, _, [], []).
make_wrap_list([L|Labels], ArgNum, OwnerId, [A|Args], [WrappedA|Rest]) :-
    WrappedA = wrap_arg(A, OwnerId, L, ArgNum),
    ArgNum1 is ArgNum+1,
    make_wrap_list(Labels, ArgNum1, OwnerId, Args, Rest).


%% wrap_tac(TacticCall).
%%
%%  If TacticCall is a variable it is processed by call_tac/1.
%%  If TacticCall is nonvar, and a call to a declared tactic, then
%%  that tactic is called in such a way that the proof record reflects
%%  the effects of that call.
%%
%%  A special form of TacticCall allows the calling tactic to instantiate
%%  some or all of the code valued arguments of the called tactic:
%%
%%      TacticCall = (Tactic ---> ArgList)
%%
%%  Examples:
%%
%%      wrap_tac(conj)       - calls conj with no code arguments instantiated
%%
%%      wrap_tac(conj ---> [discharge, trans_true(...), _])
%%                           - calls conj, specifying the code for the
%%                             first two conjuncts, but also limiting the
%%                             number of conjuncts to exactly three.
%%
%%      wrap_tac(conj ---> [discharge, trans_true  | _]
%%                           - doesn't limit the nr of conjuncts.
%%
%%  mode wrap_tac(+nonvar).
%%
wrap_tac(Call) :-
    var(Call),
    !,
    call_tac(Call).
wrap_tac('--->'(Call,SubProofs)) :-
    !,
    wrap_tac2(Call, SubProofs).
wrap_tac(Call) :-
    wrap_tac2(Call, []).

wrap_tac2(Call,SubProofs) :-
    functor(Call,F,_),
    %% If we cannot split F into Thy:TacticName
    %% then it must be a global tactic.
    ( atom_search(F,1,':',Pos) ->
        ThyLen is Pos - 1,
        sub_atom(F, 1, ThyLen, Thy)
    ;
        Thy = global
    ),
    ( query_theory(tactic_wrapper(Call, SubProofs, RealCall), Thy) ->
        call_predicate(RealCall)
    ; SubProofs == [] ->
        call_predicate(Call)
    ;
        '--->'(Call, SubProofs)
    ).



%% call_tac(Tactic).
%%
%%  This (or wrap_tac/1) should be used to execute all proof
%%  that may contain uninstantiated pieces of code.  It is similar
%%  to call/1, but invokes the current prover whenever a variable
%%  is about to be called.  Also, it only handles the ',' operator,
%%  not ';', '->' etc.  In addition, if Tactic is of the form
%%  Call ---> SubProofs, it appends SubProofs to the arguments of
%%  Call, then executes the resulting tactic.
%%
%%  mode call_tac(?term).
%%
call_tac(T) :-
    var(T),
    !,
    ip_lookup(current_prover, Prover),
    ( var(Prover) ->
        handle_error(['no proof interface.  Call set_prover/2.'])
    ; Prover = prover(Name,0,_) ->
        call_predicate(Name)
    ; Prover = prover(Name,1,_) ->
        functor(Tac, Name, 1),
        arg(1, Tac, T),
        call_predicate(Tac)
    ;
        handle_error(program, ['bad proof interface.'])
    ).
call_tac((T1,T2)) :-
    !,
    call_tac(T1),
    call_tac(T2).
call_tac(T) :-
    call_predicate(T).


%% '--->'(Tac, SubProofs).
%%
%%  Tacks SubProofs on as the last argument of Tac, then
%%  calls the resulting predicate.
%%
%%  mode '--->'(+cmd, ?list(cmd)).
%%
'--->'(Tac,SubProofs) :-
    !,
    functor(Tac,F,A),
    A1 is A + 1,
    functor(Tac2,F,A1),
    same_args(Tac,Tac2,1,A),
    arg(A1, Tac2, SubProofs),
    debug(call_tac,['Constructed the call: ',q(Tac2)]),
    call_predicate(Tac2).



%% wrap_func(Tactic, WrapNum, TacMsg, Summary).
%% wrap_arg( Tactic, WrapNum, ArgMsg, ArgNum).
%%
%%   Executes Tactic and annotates the proof tree with the remaining
%%   arguments.
%%
%%   Wrappers are used to associate a portion of the proof tree
%%   with some external concept of organization.  Wrappers are typically
%%   used to mark the effects of tactics, so as to distinguish proof
%%   steps generated by tactics from those which the user or heuristics
%%   have initiated.
%%
%%   In describing the effects of a tactic, wrappers must model the handover of
%%   control which takes place when a tactic calls one of its tactic valued
%%   arguments.  If this argument effectively returned control to the user,
%%   this should be reflected in the proof record.
%%
%%   The effect of both of the above procedures on the current proof
%%   is simply to execute their first argument, Tactic.  However, in
%%   the resulting proof record, a wr/4 wrapper will be placed around the
%%   effects of that tactic.  Eg. wr(Annotation, Hyps0, Hyps, SubProof),
%%   where Hyps0 and Hyps give the number of hypotheses before/after
%%   SubProof and Annotation is:
%%       wf(WrapNum, TacMsg, Summary) or
%%       wa(WrapNum, ArgMsg, ArgNum)
%%
%%   By convention, Summary should be the tactic call that generated
%%   the wf wrapper, but excluding any subproof arguments.
%%
%%   The Msg arguments should be writelist/1 messages that can be
%%   displayed by a proof browser, to describe the wrapped portion of the
%%   proof in a suitably abstract way.  Eg. in terms of a case analysis,
%%   rather than applications of various low level rules.
%%
%%   Example.
%%   ========
%%
%%   Consider a tactic fred/2 with two subproofs:
%%
%%      ?- global_tactic(fred(Num), fred(@integer),
%%                       ['Using fred ',q(Num)],
%%                       [['true case'], ['false case']],
%%                       ['fred(Num).',nl,'More doc...']).
%%
%%      fred(Num,[Tac1,Tac2]) :-
%%           call_tac(Tac1),     %% (these calls may be
%%           wrap_tac(mary),     %%  interspersed with other
%%           wrap_tac(Tac2),     %%  processing or non-wrapped
%%           call_tac(mary).     %%  tactic calls).
%%
%%   where mary/0 is some other declared tactic with no subproofs.
%%
%%   Calling wrap_tac(fred(2) ---> [_,_]) would lead to the call
%%
%%      wrap_func(fred(2,[wrap_arg(A,Id,['true case'],1),
%%                        wrap_arg(B,Id,['false case'],2),
%%                Id,
%%                ['Using fred ',q(2)],
%%                fred(2)).
%%
%%   which would result in the following proof tree:
%%
%%       wr( wf(11, ['Using fred: ',q(2)], fred(2)),
%%           15, 18,  %% fred/2 created three new hypotheses.
%%           [ wf( wa(11,['true case'],1), ...),   %% Tac1 wrapped. (*)
%%             wf( wf(12,['Mary...'],mary), ...),  %% mary wrapped.
%%             wf( wa(11,['false case'],2), ...),  %% Tac2 wrapped.
%%             ...                                 %% mary steps, not wrapped.
%%           ])
%%
%%  (*) Note that when a tactic is called via wrap_tac, then its
%%      subproofs are always wrapped, regardless of whether they are
%%      called via call_tac or wrap_tac.  Conversely, if a tactic
%%      is not called via wrap_tac, then neither it nor its arguments
%%      are wrapped.
%%
%%  mode wrap_func(+cmd, +cmd, @writelist, ?integer).
%%  mode wrap_arg( +cmd, +cmd, @writelist, @integer).
%%
wrap_func(Tactic, WrapNum, Msg, Summary) :-
    nonvar(Summary),
    nonvar(Tactic),
    !,
    \+ in_itrans,
    %% Allocate the (probably unique) identity number for this call.
    ip_lookup(wrapnr, OldId),
    alloc_wrap_id(OldId, WrapNum, NewId),
    ip_set(wrapnr, NewId),
    wrap_proof_steps(Tactic,WrapNum,Msg,tactic(Summary),
                     wf(WrapNum,Msg,Summary)).
wrap_func(_, _, _, Summary) :-
    var(Summary),
    !,
    handle_illegal_arg(wrap_func, 4, nonvar, Summary).
wrap_func(Tac, _, _, _) :-
    handle_illegal_arg(wrap_func, 1, code, Tac).


wrap_arg(Tactic, WrapNum, Msg, ArgNum) :-
    integer(WrapNum), %% Should have been allocated by wrap_func.
    !,
    \+ in_itrans,
    wrap_proof_steps(Tactic,WrapNum,Msg,subproof(ArgNum),
                     wa(WrapNum,Msg,ArgNum)).
wrap_arg(_, WrapNum, _, _) :-
    handle_illegal_arg(wrap_arg, 2, integer, WrapNum).


alloc_wrap_id(OldId, NewId, NewId) :-
    NewId is OldId + 1,
    !.
alloc_wrap_id(OldId, WrapNum, NewId) :-
    integer(WrapNum),
    %% This means that the caller (probably rerun) is requesting us
    %% to use a given id number, which is different to the
    %% id number that we are about to generate.  
    ( WrapNum > OldId ->
        %% Start allocating ids from WrapNum upwards (thus skipping
        %% some numbers).  We can still guarantee uniqueness.
        NewId = WrapNum
    ;
        %% Generate a warning, because WrapNum id number has probably
        %% already been used and we are about to use it again.
        %% However, allocate future ids from OldId+1 upwards as usual.
        writelist(['WARNING: reusing tactic ID. number ',q(WrapNum),nl]),
        NewId is OldId + 1
    ),
    !.
alloc_wrap_id(_, WrapNum, _) :-
    handle_illegal_arg(wrap_func, 2, integer, WrapNum).


wrap_proof_steps(Tactic,WrapNum,Msg,Summary,Annotation) :-
    %% Experiment: use push/pop_subproof for wrappers, but copy the hyp list to
    %% the parent when the wrapped tactic has finished.  The wrapped tactic
    %% call thus contributes a single line to the current subproof, but any
    %% hypotheses added/deleted by the tactic affect the caller window.
    is_focus(Focus0),
    push_simple_subproof(Focus0, []),
    ip_lookup(hypnr, Nr0),
    wrap_call_announce(Tactic,WrapNum,Msg,Summary),
    ip_lookup(hypnr, Nr),
    ip_lookup(hyplist, Hyps),
    pop_subproof(Focus, CDep, UDep, SubPrf, false),
    %% Update the caller's hyps from the tactic's subproof
    ip_set(hypnr, Nr),
    ip_set(hyplist, Hyps),
    debug(wrap_proof_steps, ['After tactic ',q(WrapNum),':',q(Summary),nl,
                             '    new proof: ',q(SubPrf),nl,
                             '    udep list: ',q(UDep),nl,
                             '    hypnr is : ',q(Nr)]),
    %% Now add one step that encapsulates all of Tactic's work.
    add_proof_step(Focus,wr(Annotation,Nr0,Nr,SubPrf),CDep,UDep).

%% private wrap_call_announce(Tactic,WrapNum,Msg,Kind).
%%
%%  Executes Tactic and generates announcements about its progress.
%%  WrapNum is an integer that uniquely identifies a particular
%%  invocation of a tactic.  Msg is a writelist and Kind is
%%  either tactic(Call) or subproof(Num).
%%
%%  mode wrap_call_announce(+cmd,+integer,@writelist,@compound).
%%
wrap_call_announce(Tactic,WrapNum,Msg,Kind) :-
    announce_event(wrap_tac(WrapNum,Msg,Kind,call)),
    ( call_tac(Tactic) ->
        announce_event(wrap_tac(WrapNum,Msg,Kind,exit))
    ;
        announce_event(wrap_tac(WrapNum,Msg,Kind,fail)),
        fail
    ).


