%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: annotations.ql,v 1.10 2002/06/27 05:10:22 itp Exp $
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%  This file contains all procedures for manipulating annotations
%%  on proof nodes.  The soundness of proofs does not depend upon
%%  annotations, but they are useful for storing additional information
%%  about how the proof was constructed.  Eg. tactic calls, user
%%  commands or comments.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

#include "../lib/ergo.h"

?- new_global_fact(annotation, 2, []).


annotations header
  'Annotations on proof nodes'
extends
  concepts
doc
  'Annotations allow additional user-defined information to be',nl,
  'associated with each proof node.',
  para,
  'The soundness of proofs does not depend upon annotations,',nl,
  'but they are useful for storing additional information about how',nl,
  'the proof was constructed.  Eg. tactic calls, user commands or',nl,
  'comments.'.


general command
  new_annotation(Type, Doc)
doc
  'This command declares a new kind of proof node annotation.',
  para,
  'Type must be an atom or a compound term whose arguments describe',nl,
  'the types of the annotations arguments.  That is, Type is similar to',nl,
  xref('mode declarations',global,modeshelp),
  ', but without the @/?/+/- prefixes on the type names.',nl,
  para,
  'Doc must be the name of a documentation node that describes the',nl,
  'new kind of annotation.',
  para,
  'It is not good style to include proof node names in annotations,',nl,
  'or any other kind of data that will become meaningless when a proof',nl,
  'tree is abstracted into a tactic.',
  para,
  'Example:  new_annotation(usercmd(string), usercmd_annotation).'
extends
  annotations,
  general_cmds
mode
  new_annotation(@ground, @atom).

new_annotation(Type, Doc) :-
    check_atom(Doc),
    check_or_error(
        ground(Type),
        ['illegal (non-ground) type description: ',q(Type)]),
    functor(Type, F, A),
    functor(Skel, F, A),
    check_or_error(
        \+ query_global(annotation(Skel, _)),
        [q(F/A),' is already declared as an annotation name.']),
    assert_global(annotation(Type,Doc)).



proof command
  annotate(PNode, Ann)
doc
  xinc(global,annotate_doc)
extends
  proof_tree_cmds,
  annotations
mode
  annotate(@nodeid, @annotation),
  annotate(@nodeid, @list(annotation)).

rule_constraint command
  annotate(PNode, Ann)
doc
  xinc(global,annotate_doc)
extends
  rule_constraint_cmds
mode
  annotate(@nodeid, @annotation),
  annotate(@nodeid, @list(annotation)).

annotate_doc doc_fragment
  'Adds the annotation(s) Ann to the proof node PNode.',nl,
  'Ann can be a single annotation, or a list of annotations.',nl,
  'Each annotation in Ann must have been declared using ',
    xref('new_annotation/2',global,'general:new_annotation/2'),'.'.


annotate(PNode, Ann) :-
    list(Ann),
    !,
    map(annotate_1(PNode), [Ann]).
annotate(PNode, Ann) :-
    annotate_1(PNode, Ann).

annotate_1(PNode, Ann) :-
    PRE(annotate, is_legal_annotation(Ann)),
    \+ ip_lookup(update_annotations,no),
    handle_event(reannotate(PNode, Ann)),
    !.
annotate_1(PNode, Ann) :-
    ip_lookup('Annotation@List', PNode, List0),
    ( var(List0) ->
        List0 = [Ann]
    ;
        ip_set('Annotation@List', PNode, [Ann|List0])
    ).

is_legal_annotation(Ann) :-
    std_nonvar(Ann),
    functor(Ann, F, A),
    functor(Skel, F, A),
    query_global(annotation(Skel, _)).

library command
  annotation(PNode, Ann)
extends 
  proof_tree_cmds
doc
  'True when Ann unifies with one of the existing annotations',nl,
  'on proof node PNode.'
mode
  annotation(@nodeid, ?annotation).

annotation(PNode, Ann) :-
    ip_lookup('Annotation@List', PNode, List),
    list(List),
    annotation_search2(List, Ann).

annotation_search2([Ann|AS], Ann).
annotation_search2([_|AS], Ann) :-
    annotation_search2(AS, Ann).



library command
  annotations(PNode, AnnList)
extends 
  proof_tree_cmds
doc
  'AnnList is a list of all the annotations on proof node PNode.',nl,
  'This is similar to the ',
     xref('annotation/2',global,'library:annotation/2'),
     ' command',nl,
  'but returns the list of all annotations, rather than one by one.'
mode
  annotations(@nodeid, ?list(annotation)).

annotations(PNode, AnnList) :-
    ip_lookup('Annotation@List', PNode, List),
    ( list(List) ->
        AnnList = List
    ;
        AnnList = []
    ).


