%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% This file contains all procedures for manipulating proof trees.
%%
%% CVS $Id: proof.ql,v 1.46 2006/02/06 00:52:12 itp Exp $
%%
%%
%%  Proof Data Structures
%%  =====================
%%
%%  A proof tree is represented by a set of named tuples
%%  of the form:
%%
%%    proofnode(ParentId, Term, Thy, Rule, RuleType, Children, Timestamp)
%%
%%  The 'identity' of each proof node is an integer and each node
%%  is accessed via an implicit parameter called 'Proof@Node'.
%%  This is an array implicit parameter, indexed via the node identities.
%%  The ParentId of a root node is `root' or (if a postulate is being
%%  proved) `postulate(...)'.
%%
%%  The 'Next@Proof@Id' persistent counter is incremented each
%%  time a proof node is allocated.  
%%
%%  A node has been proved iff Thy is nonvar.  Once it has been
%%  proved, Thy:Rule gives the name of the rule that was used to
%%  prove it and Children is a closed list of the identities of any
%%  child proof nodes.  The Thy, Rule and Children components should
%%  all be variables in unproven (open) nodes.  Note that for rules
%%  whose names can take parameters, Rule may be the short name
%%  or the full name (ie. including parameters), depending upon
%%  which name was specified when the proof step was performed.
%%
%%  Other implicit parameter data structures store context terms
%%  and user-defined annotations for each node.
%%
%%  The 'Proof@Disabled@Rules' implicit parameter stores the short
%%  names of all rules that are not allowed to be used in the
%%  current proofs.
%%
%%  Well-formedness of proof terms
%%  ==============================
%%
%%  A proof term is well-formed iff it is constructed from known symbols
%%  of the current theory, with each symbol having the correct arity.
%%
%%  This proof engine does not check the well-formedness of terms
%%  during proofs, but does check that each term in the resulting
%%  inference rule is well-formed.  This means that tactics or
%%  user commands can introduce weird terms into a proof, but this
%%  will cause an error if that term ends up in the final rule.
%%  In the case where the weird term appears only within some subproof,
%%  not in the final rule, we could in principle replace it by any
%%  legal ground term to obtain a completely well-formed proof of
%%  the same rule.  Thus, the proof is still sound.  More formally,
%%  there is an interpretation function from the proof containing
%%  new symbols back into a proof of the unextended theory.
%%
%%  This result (or conjecture!) relies on the following restrictions
%%  on the side conditions of inference rules:
%%
%%    A) They must be monotonic with respect to instantiation.
%%       (eg. nonvar(T) would be unsound, but ground(T) is okay).
%%
%%    B) They must check that a term is well-formed if they are
%%       relying on the syntactic structure of that term.
%%       (eg. the constraint functor(T,_,5) is not sound without
%%            an additional check that the outermost functor of T
%%            is well-formed.  However, it is not necessary to check
%%            that the arguments of T are well-formed, since the truth
%%            of this constraint does not rely upon those arguments).
%%
%%  The set of available constraints is designed to ensure that
%%  these properties are satisfied.  However, writers of new constraint
%%  procedures need to be aware of these requirements.
%%  This approach avoids the expensive checks that Ergo 4 used to
%%  do every time a variable was instantiated.
%%
%%  Proof variable temperatures
%%  ===========================
%%  In ordinary (non-postulate) proofs, all the variables in the initial
%%  conclusion of the proof are frozen, to prevent accidental instantiation.
%%  However, tactics are free to freeze or thaw any of the variables
%%  in the proof at any time.  This means that tactics that match
%%  against proof terms should be careful not to instantiate thawed
%%  variables, unless that is intended.  Systematic use of std_nonvar/1
%%  etc. at each level is a useful technique for avoiding accidental
%%  instantiation.  
%%
%%  In postulate proofs, not only are all the variables frozen, but they
%%  also have a delay added to them which causes failure, with an error
%%  message, if they are instantiated.  Thus, though it is allowable to 
%%  thaw variables in a postulate proof, it is pointless, since they 
%%  cannot be instantiated without failure.
%%  
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


proof command
  proof_start(Formula, Constraints, Children)
doc
  'This starts a proof, with Formula as the term at the root',nl,
  'node of the proof and Constraints as the initial set of constraints.',nl,
  'Children is the list of unproved nodes in the new proof.'
extends
  theory_extension_cmds
mode
  proof_start(?thy(term), +rule_constraint, ?list(nodeid)).

proof_start(Concl, Constraints, Children) :-
    freeze_term(Concl),
    proof_start_internal(Concl, Constraints, Children, root).

proof_start_internal(Concl, Constraints, [Root], RootInfo) :-
    check_current_theory(Thy),
    std_nonvar(RootInfo),
    check_current_theory(_),
    check_term(Concl),
%% TO DO: In proof_start_internal, move these into a separate procedure,
%% to allow multiple proofs to coexist simultaneously.
%% Ray's last comment: See PMP task regarding flexible proof construction.
    context_init,
    global_state_set('Proof@Next@Id', 0),
    global_state_set('Proof@Timestamp', 0),
    ip_set('Proof@Disabled@Rules', 0, 0),  %% just create the i.p.
    %%
    %% Now a special version of proof_new_node(Root, ..., Concl),
    %% which puts RootInfo into the root node.
    get_proof_size(Size),
    ip_array_init('Proof@Node', Size),
    global_state_increment('Proof@Next@Id', Root),
    ip_set('Proof@Node', Root, proofnode(RootInfo, Concl, _, _, _, _, _)),
    context_proof_init(Root),
    ip_set(in_proof, true),
    ip_set(proofnames, [Thy:no_name]),
    with_context_delaying(Constraints).


proof command
  proof_start_postulate(Name, Children)
doc
  'This starts a proof of the postulate called Name.',
  para,
  'Name must be the name (may include parameters) of an',nl,
  'existing postulate in the current theory.',
  para,
  'Children will be a singleton list containing the name of the',nl,
  'root node of the new proof.  This node corresponds to the bottom',nl,
  'sequent of the postulate.  Tactics and/or a person must then build',nl,
  'a proof tree whose only unproved nodes are the same as the upper',nl,
  'sequents (if any) of the postulated inference rule.  Once this',nl,
  'has been achieved, the proof_complete_postulate command can be',nl,
  'called to mark the postulate as proved.'
extends
  theory_extension_cmds
mode
  proof_start_postulate(@rulename, ?list(nodeid)).

proof_start_postulate(Name0, [Root]) :-
    check_current_theory(Thy),
    check_std_nonvar(Name0),
    functor(Name0, Name, _),
    check_or_error(
        query_theory(infrule(Name, _, Concl, Root, Seqs,
                             Before, After), Thy),
        ['no such rule: ',q(Thy:Name)]),
    check_or_error(
        \+ query_theory(infrule_status(Name, _)),
        ['rule ',q(Name),' is not a postulate.']),
    %%
    extract_context_terms(Before, CTerms),
    freeze_term(CTerms, CTermVars),
    map(proof_never_instantiate, [CTermVars]),
    check_or_error(
        proof_start_internal(
            Concl, Before, [Root],
            postulate(Name,Seqs,Vars,IdMap,
                      CDist,CNFI,CRest,CContext)),
        ['constraints of postulate ',q(Name),' failed: ',code(Before)]),
    check_or_error(
	call_predicate(After),
        ['constraints of postulate ',q(Name),' failed: ',code(After)]),
    %%
    %% Freeze all variables, *and* add delays to make doubly sure that
    %% that cannot be instantiated.  Tactics can thaw variables, but
    %% they cannot remove the delays!  So, the delays guarantee that
    %% the postulate is not specialised via instantiation.
    %%
    freeze_term([Concl,Seqs,Before,After], Vars),
    debug(proof, ['postulate vars = ',q(Vars)]),
    map(proof_never_instantiate, [Vars]),
    %%
    %% Now that the Before and After constraints have been executed, we collect
    %% Qu-Prolog's idea of all the constraints associated with the
    %% postulate into CDist, CNFI, CRest and CContext.
    %% We do exactly the same after the proof is complete to check that
    %% there are no extra constraints.  (Adding constraints is another 
    %% way of over-specialising the postulate).
    %%
    retry_delays,
    collect_constraints(Vars, CDist, CNFI, C3),
    proof_check_special_constraints(C3, CRest),
    IdMap = [proofnode(Root)=_ | _],
    context_constraints(Root, IdMap, CContext),
    %% 
    %% Set in_proof info
    %%
    ip_set(in_proof, true),
    ip_set(proofnames, [Thy:Name]),
    %%
    %% Now disable dependent rules to prevent circular proofs.
    %%
    dependent_rules([Thy:Name], IllegalRules),
    debug(proof, ['illegal rules for ',q(Thy:Name),' are: ',q(IllegalRules)]),
    proof_disable_rules([Thy:Name|IllegalRules]).

extract_context_terms(true, []).
extract_context_terms((context_search(_,_,_,T),Constrains), [T|CTerms]) :-
    !,
    extract_context_terms(Constrains, CTerms).
extract_context_terms((_, Constrains), CTerms) :-
    extract_context_terms(Constrains, CTerms).

proof_never_instantiate(V) :-
    delay_until(bound(V), proof_instantiation_check(V)).

proof_instantiation_check(V) :-
    handle_error(
     ['naughty, naughty!  You thawed one of the variables in the',nl,
      'postulate, then instantiated it to: ',thy(V),nl,
      'This is illegal because it over-specialises the postulate.']).


%% private proof_check_special_constraints(Constraints0, Constraints).
%%
%%  Constraints is the subset of Constraints0 that should be stored
%%  as side-conditions of an inference rule.
%%
%%  Certain boring/irrelevant constraints are discarded.
%%
proof_check_special_constraints([], []).
%
%  Delayed unifications were not allowed once, but there seems to be no
%  reason why unifications cannot be delayed like other constraints.
%
% proof_check_special_constraints([A=B|_], _) :-
%     !,
%     handle_error(['unresolved unification: ',q(A=B)]).
proof_check_special_constraints([Delay|In], Out) :-
    Delay = delay_until(bound(V), proof_instantiation_check(V)),
    !,
    proof_check_special_constraints(In, Out).
proof_check_special_constraints([D|In], [D|Out]) :-
    proof_check_special_constraints(In, Out).


proof command
  proof_disable_rules(Rules)
extends
  proof_tree_cmds
doc
  'Prevents all the rules in Rules from being used in future proof steps',nl,
  'within the current proof.  Rules must be a list of rule names, each',nl,
  'of the form Thy:ShortName.'
mode
  proof_disable_rules(@list(rulename)).

proof_disable_rules([]).
proof_disable_rules([Thy:Name|Rules]) :-
    ancestor(Thy),
    atom(Name),
    !,
    %% The same rule name may appear in several theories, so 
    %% until QP allows implicit parameters to be indexed by ground
    %% terms, we map each shortname to a *list* of Thy:Name terms.
    ip_lookup('Proof@Disabled@Rules', Name, Val),
    ( var(Val) ->
        Val = [Thy:Name]
    ;
        list(Val),
        ip_set('Proof@Disabled@Rules', Name, [Thy:Name|Val])
    ),
    proof_disable_rules(Rules).
proof_disable_rules([Rule|_]) :-
    handle_error(['proof_disable_rules: illegal rule name ',q(Rule)]).


proof command
  is_disabled_rule(Thy,Rule)
extends
  proof_tree_cmds
doc
  'Succeeds if Thy:Rule is disabled from being used in the current proof.',nl,
  'For example, during the proof of a postulated rule, all rules that',nl,
  'depend upon that rule are disabled so that circular proofs between',nl,
  'rules cannot be created.  Unfortunately, there is no way of displaying',nl,
  'all the disabled rules as yet.'
mode
  is_disabled_rule(@theory,@shortrulename).

is_disabled_rule(Thy, Name) :-
    atom(Thy),
    atom(Name),
    ip_lookup('Proof@Disabled@Rules', Name, Result),
    nonvar(Result),
    member(Thy:Name,Result),
    !.

proof command
  proof_step(Thy, Rule, Id, Children)
extends
  proof_tree_cmds
doc
  'Proves node Id, by applying the rule Thy:Rule.',nl,
  'Children is the list of child node names that Thy:Rule produces.'
mode
  proof_step(?theory, ?rulename, @nodeid, ?list(nodeid)).

proof_step(Thy, Rule, Id, Children) :-
    ip_lookup('Proof@Node', Id, Node),
    Node = proofnode(_, Term, Thy, Rule, RuleType, Children, Timestamp),
    rule_apply(Thy, Rule, Term, Id, Children),
    ip_lookup('rule@unify@flag', Flag),
    ( Flag == sunify ->
        RuleType = srule
    ;
        RuleType = rule
    ),
    proof_non_circular(Thy, Rule),
    global_state_increment('Proof@Timestamp', Timestamp).

proof_non_circular(Thy, Rule) :-
    functor(Rule, F, _),
    ip_lookup('Proof@Disabled@Rules', F, Result),
    nonvar(Result),
    member(Thy:Rule,Result),
    !,
    fail.
proof_non_circular(_, _).


library command
  proof_new_node(Id, Parent, Term)
extends 
  proof_tree_cmds
doc
  'This routine constructs a new proof node.',
  para,
  'This procedure is for use by the core prover only, so is not allowed',nl,
  'within the user-supplied constraints of inference rules.',
  para,
  'It copies all the parent context lists to the constructed node,',nl,
  'so should *not* be used to construct root nodes of proof trees.'
mode
  proof_new_node(?nodeid, @nodeid, ?thy(term)).

proof_new_node(Id, Parent, Term) :-
    integer(Parent),
    global_state_increment('Proof@Next@Id', Id),
    ip_set('Proof@Node', Id, proofnode(Parent, Term, _, _, _, _, _)).


proof command
  proof_complete(Name,Root,OpenNodes,Constraints)
doc
  'Converts a proof into an inference rule, called Name, and writes',nl,
  'that new inference rule into the current theory.',
  para,
  'Name is the name of the new inference rule.  It must be an atom',nl,
  'or a compound term whose functor is an atom.  In the latter case,',nl,
  'the arguments are treated as being parameters of the resulting rule.',
  para,
  'Root is the name of the root node of a proof.  That is, a node',nl,
  'that was created by the proof_start/3 procedure.',
  para,
  'OpenNodes is a list of the names of all unproved nodes in the proof.',nl,
  'For example, if OpenNodes=[56,45] and Root=1, then the resulting',nl,
  'inference rule will be:',
  para,
  begin(verb),
  '    Seq56,',nl,
  '    Seq45',nl,
  '    ----------',nl,
  '    Seq1',nl,
  end(verb),
  nl,
  'where Seq45, Seq56 and Seq1 stand for the sequents at proof nodes',nl,
  '56, 45 and 1, respectively.',nl,
  'OpenNodes is typically used to order the antecedents of the rule and',nl,
  'perhaps to check that the proof does not contain other unproved nodes.',nl,
  'If it is a variable, then it is instantiated to contain all the',nl,
  'unproved nodes of the proof in some arbitrary order.',
  para,
  'Constraints is a comma-separated sequence of ''rule_constraint''',nl,
  'commands.  These commands will be added as constraints on the',nl,
  'resulting rule, in addition to other constraints that are already',nl,
  'associated with the proof.  If no extra constraints are needed, it',nl,
  'should be ''true''.'
extends
  save_close_treecmds
mode
  proof_complete(@std_nonvar, @nodeid, ?list(nodeid), +constraints).

proof_complete(Name, Root, OpenNodes, Constraints0) :-
    check_or_error(
        ( integer(Root),
          ip_lookup('Proof@Node', Root, Node),
          compound(Node),
          Node = proofnode(RootInfo, Concl, _, _, _, _, _) ),
        ['illegal proof node name: ',q(Root)]),
    check_or_error(
        std_nonvar(RootInfo),
        ['proof node ',q(Root),' is not the root node of a proof.']),
    check_current_theory(Thy),
    check_or_error(
        std_nonvar(Name),
        ['illegal name for rule: ',q(Name)]),
    functor(Name, F, _),
    check_legal_rule_name(F),
    check_or_error(
          \+ rule_query(Thy, F, _, _, _, _),
        ['there is already a rule called: ',q(Thy:F)]),
    proof_check_open_nodes(Root, OpenNodes),
    check_or_error(
        std_nonvar(Constraints0),
        ['illegal rule constraints: ',q(Constraints0)]),
    %%
    %% All checking complete, so now we build the inference rule.
    %%
    %% IdMap accumulates the mapping from proof node names and
    %% context term names to the corresponding identity variables
    %% of the inference rule.
    IdMap = [proofnode(Root)=RootId | _],
    map(proof_build_seq(Root,IdMap), [OpenNodes, Seqs]),
    debug(proof, ['built seqs: ',q(Seqs),nl,
                  'IdMap = ',q(IdMap)]),
    %%
    %% Collect all the constraints that are relevant to the new
    %% inference rule.
    %%
    collect_allvars([Concl,RootId,Seqs], Vars),
    proof_constraints_internal(Root, Vars, IdMap,
                                CDist, CNFI, CRest, CContext),
    append([CDist,CNFI,CRest,CContext], Constraints1L),
    list_to_comma(Constraints1L, Constraints1),
    prolog_code_diff((Constraints0,Constraints1), true, Constraints),
    %%
    %% Now rule_add checks the user-supplied constraints.
    %%
    rule_add(Name, Concl, RootId, Seqs, Constraints),
    proof_rules_used(Root, RulesUsed),
    assert_theory(infrule_status(F, RulesUsed)).


proof_build_seq(Root, IdMap, Node, Seq) :-
    proof_node_term(Node, Term),
    search_insert(proofnode(Node)=NodeId, IdMap),
    context_names(CNames),
    proof_build_cspecs(CNames, Node, Root, IdMap, CSpecs),
    Seq = seq(NodeId, CSpecs, Term).

proof_build_cspecs([], _, _, _, []).
proof_build_cspecs([CName|CNS], Node, Root, IdMap, CSpecs) :-
    context_diff(Node, Root, CName, CDiff),
    ( CDiff = (use(CName,Root)) ->
        %% like rule_add, we don't record these default context specs.
        CSpecs = CSpecs2
    ;
        context_diff_to_spec(CDiff, IdMap, CSpec),
        CSpecs = [CName=CSpec | CSpecs2]
    ),
    proof_build_cspecs(CNS, Node, Root, IdMap, CSpecs2).


proof command
  proof_complete_postulate(Root),
  proof_complete_postulate(Root, Unproved)
doc
  'This completes the proof of a postulate.',
  para,
  'Root must be the root node of a proof of a postulate.',nl,
  'If the proof has correctly discharged that postulate, without',nl,
  'instantiating variables or adding extra constraints, then the',nl,
  'postulate will be marked as proved.',
  para,
  'Unproved is a list of the names of all unproved nodes in the proof,',nl,
  'in an order that corresponds to the order of the upper sequents',nl,
  'of the postulated inference rule.  The one argument version sets',nl,
  'Unproved=[].'
extends
  save_close_treecmds
mode
  proof_complete_postulate(@proofnode),
  proof_complete_postulate(@proofnode, ?list(proofnode)).

proof_complete_postulate(Root) :-
    proof_complete_postulate(Root, []).

proof_complete_postulate(Root, OpenNodes) :-
    check_or_error(
        ( integer(Root),
          ip_lookup('Proof@Node', Root, Node),
          compound(Node),
          Node = proofnode(RootInfo, _, _, _, _, _, _) ),
        ['illegal proof node name: ',q(Root)]),
    check_or_error(
        RootInfo = postulate(Name,Seqs,Vars,IdMap,
                             CDist,CNFI,CRest,CContext),
        ['proof node ',q(Root),
         ' is not the root node of a proof of a postulate.']),
    debug(proof, ['original seqs = ',q(Seqs),nl,
                  'original Vars = ',q(Vars),nl,
                  'original IdMap = ',q(IdMap)]),
    proof_check_open_nodes(Root, OpenNodes),
    %%
    %% Now we build the inference rule, and check that it unifies
    %% with the original postulated rule.
    %%
    %% IdMap accumulates the mapping from proof node names and
    %% context term names to the corresponding identity variables
    %% of the inference rule.
    map(proof_build_seq(Root,IdMap), [OpenNodes, Seqs2]),
    debug(proof, ['built seqs: ',q(Seqs2),nl,
                  'IdMap = ',q(IdMap)]),
    check_or_error(
        Seqs2 = Seqs,
        ['unproved nodes of the proof: ',q(Seqs2),nl,
         'do not match postulate premises: ',q(Seqs)]),
    %%
    %% Collect all the constraints that are relevant to the proof
    %%
    proof_constraints_internal(Root, Vars, IdMap,
                               ProofDist, ProofNFI, ProofRest, ProofContext),
    proof_check_code_list_equal(ProofRest,    CRest),
    proof_check_code_list_equal(ProofDist,    CDist),
    proof_check_code_list_equal(ProofNFI,     CNFI),
    proof_check_code_list_equal(ProofContext, CContext),
    proof_rules_used(Root, RulesUsed),
    msg(postulate, 2, ['rulesused = ',q(RulesUsed)]),
    assert_theory(infrule_status(Name, RulesUsed)).

proof_check_code_list_equal(Proof, Postulate) :-
    debug(proof, ['comparing proof constraints: ',q(Proof),nl,
                  'with postulated constraints: ',q(Postulate)]),
    list_to_comma(Proof, ProofCode),
    list_to_comma(Postulate, PostulateCode),
    prolog_code_diff(ProofCode, PostulateCode, Diff),
    Diff = true.


%% proof_rules_used(Node, RuleList).
%%
%%  RuleList is all the rules used in the proof subtree given by Node.
%%
%%  mode proof_rules_used(@nodeid, ?list(rulename)).
%%
proof_rules_used(PNode, RulesUsed) :-
    proof_rules_used2(Used, PNode),
    open_to_closed(Used),
    RulesUsed = Used.

proof_rules_used2(Used, PNode) :-
    proof_node_rule_and_children(PNode, Thy, Rule, _, Children),
    functor(Rule, F, _),
    !,
    search_insert(Thy:F, Used),
    map(proof_rules_used2(Used), [Children]).
proof_rules_used2(_, _).
    
  

%%%%%%%%%%%%%%%%%%%% Query Procedures for Proof Nodes %%%%%%%%%%%%%%%%%%%%

proof command
  proof_node_term(Id, Pred),
  proof_node_terms(Ids, Preds)
doc
  'Pred is the main term (the conclusion) of the proof node Id.'
extends
  query_treecmds
mode
  proof_node_term(@nodeid, ?thy(term)),
  proof_node_terms(@list(nodeid), ?list(thy(term))).

proof_node_term(Id, Pred) :-
    ip_lookup('Proof@Node', Id, Node),
    compound(Node),
    Node = proofnode(_, Pred, _, _, _, _, _).

proof_node_terms(N, T) :- var(N),!,T=[].
proof_node_terms([], []).
proof_node_terms([Id|Ids], [Pred|Preds]) :-
    proof_node_term(Id, Pred),
    proof_node_terms(Ids, Preds).


proof command
  proof_root(Node)
doc
  'Succeeds when Node is the root node of a proof'
extends
  query_treecmds
mode
  proof_root(@nodeid).

proof_root(Node) :-
    ip_lookup('Proof@Node', Node, proofnode(Parent, _, _, _, _, _, _)),
    std_nonvar(Parent).

proof command
  proof_node_timestamp(Node,Timestamp)
doc
  'Succeeds when Timestamp is the timestamp for a closed node Node.'
extends
  query_treecmds
mode
  proof_node_timestamp(@nodeid, ?integer).

proof_node_timestamp(Node, Timestamp) :-
    ip_lookup('Proof@Node', Node, proofnode(_, _, _, _, _, _, Timestamp)),
    atomic(Timestamp).

proof command
  proof_root_postulate(Node, Thy, Name)
doc
  'Succeeds when Node is the root node of a proof of postulate Thy:Name.',nl,
  'Name is always the short name of the postulate (no parameters).',nl,
  'Currently, Thy is always the current theory, but in the future',nl,
  'it may be a subtheory of the current theory.'
extends
  query_treecmds
mode
  proof_root_postulate(@nodeid, ?theory, ?rulename).

proof_root_postulate(Node, Thy, Name) :-
    ip_lookup('Proof@Node', Node, proofnode(Info, _, _, _, _, _, _)),
    std_nonvar(Info),
    functor(Info, postulate, _),
    arg(1, Info, Name),
    global_state_lookup(current_dir,Thy).


proof command
  proof_node_parent(Id, Parent)
doc
  'Parent is the name of the parent node of the proof node Id.'
extends
  query_treecmds
mode
  proof_node_parent(@nodeid, ?nodeid).

proof_node_parent(Id, Parent) :-
    ip_lookup('Proof@Node', Id, Node),
    Node = proofnode(Parent, _, _, _, _, _, _),
    integer(Parent).


proof command
  proof_node_children(Id, Children)
doc
  'Children is the list of subproof node names associated with the',nl,
  'proof node Id.  This fails if Id refers to an open node.'
extends
  query_treecmds
mode
  proof_node_children(@nodeid, ?list(nodeid)).

proof_node_children(Id, Children) :-
    ip_lookup('Proof@Node', Id, Node),
    compound(Node),
    Node = proofnode(_, _, Thy, _, _, Children, _),
    atom(Thy).


proof command
  proof_node_rule(Id, Thy, Rule, RuleType)
doc
  'True if Thy:Rule was used to discharge proof node Id.',nl,
  'This fails if Id refers to an open node.',nl,
  'RuleType is the type of rule tactic used - i.e. rule or srule.'
extends
  query_treecmds
mode
  proof_node_rule(@nodeid, ?theory, ?rulename, ?atom).

proof_node_rule(Id, Thy, Rule, RuleType) :-
    ip_lookup('Proof@Node', Id, Node),
    Node = proofnode(_, _, Thy2, Rule, RuleType, _, _),
    atom(Thy2),
    Thy = Thy2.


proof command
  proof_node_rule_and_children(Id, Thy, Rule, RuleType, Children)
doc
  'True if Thy:Rule was used to discharge proof node Id and',nl,
  'Children is the list of subproof node names produced by that rule.',nl,
  'This fails if Id refers to an open node.',nl,
  'RuleType is the type of rule tactic used - i.e. rule or srule.'
extends
  query_treecmds
mode
  proof_node_rule_and_children(@nodeid, ?theory, ?rulename, ?atom, 
                               ?list(nodeid)).

proof_node_rule_and_children(Id, Thy, Rule, RuleType, Children) :-
    ip_lookup('Proof@Node', Id, Node),
    Node = proofnode(_, _, Thy2, Rule, RuleType, Children, _),
    atom(Thy2),
    Thy = Thy2.


proof command
  proof_node_context(Id, Contexts)
doc
  'Contexts is a list of CName=CTermIds for the context of Id.',
  para,
  'WARNING: some of the terms returned may have constraints',nl,
  'associated with them, such as x not_free_in A.  These constraints',nl,
  'may prevent a context term from being used to discharge a subproof',nl,
  'even though it looks as if the context term should match the',nl,
  'conclusion of the subproof.',nl,
  'The constraints are not returned or recorded by this procedure.',nl,
  'So, you should regard the results of this procedure as giving',nl,
  'only a rough indication (a superset) of what context terms are',nl,
  'available for use.  The only way to be certain that a context term',nl,
  'can be used to prove a given conclusion is to apply the appropriate',nl,
  'inference rule.'
extends
  query_treecmds
mode
  proof_node_context(@nodeid, @list(compound)).

proof_node_context(Id, Contexts) :-
    ip_lookup('Proof@Node', Id, Node),
    compound(Node),
    findall(CName=CDescs,
            ( is_context_name(CName),
              findall(CTermId,
                      context_search(CName, Id, CTermId, _),
                      CDescs)
            ),
            Contexts).

%%%%%%%%%%%%%%%%%%%% Query Procedures for Proof Trees %%%%%%%%%%%%%%%%%%%%

proof command
  proof_open_nodes(Id, OpenNodes)
doc
  'OpenNodes is the list of node names of all open nodes in the',nl,
  'subtree whose root is Id.'
extends
  query_treecmds
mode
  proof_open_nodes(@nodeid, ?list(nodeid)).

proof_open_nodes(Id, OpenNodes) :-
    proof_open_nodes2([Id], OpenNodes, []).

proof_open_nodes2([], L, L).
proof_open_nodes2([Id|Ids], OpenNodes0, OpenNodes) :-
    ip_lookup('Proof@Node', Id, Node),
    Node = proofnode(_, _, Thy, _, _, Children, _),
    (var(Thy) ->
        OpenNodes0 = [Id|OpenNodes1]
    ;
        proof_open_nodes2(Children, OpenNodes0, OpenNodes1)
    ),
    proof_open_nodes2(Ids, OpenNodes1, OpenNodes).


%% proof_check_open_nodes(Root, Open).
%%
%%  Succeeds with Open being exactly the open proof nodes of Root.
%%  This is similar to proof_open_nodes, but the intial value of
%%  Open may be a variable, or a closed list of all the open nodes
%%  in some desired order (whereas proof_open_nodes always returns
%%  them in some fixed order).
%%
%%  Generates error messages if Open does not contain exactly
%%  the open nodes of Root.
%%
%%  mode proof_check_open_nodes(@proofnode, ?list(proofnode)).
%%
proof_check_open_nodes(Root, Open) :-
    var(Open),
    !,
    proof_open_nodes(Root, Open).
proof_check_open_nodes(Root, Open) :-
    closed_list(Open),
    !,
    proof_open_nodes(Root, ActualOpen),
    diff_list(ActualOpen, Open, ExtraOpen),
    check_or_error(
        ExtraOpen = [],
        ['proof nodes ',q(ExtraOpen),' are not proved yet.']),
    diff_list(Open, ActualOpen, Proved),
    check_or_error(
        Proved = [],
        ['the unproved nodes of node ',q(1),' do not include ',q(Proved)]).
proof_check_open_nodes(_, Open) :-
    handle_error([q(Open),' is not a valid list of proof nodes.']).


proof command
  proof_constraints(Nodes,      Dist, NFI, User, Context),
  proof_constraints(Root, Vars, Dist, NFI, User, Context)
doc
  'These commands return lists of all the various delayed constraints',nl,
  'associated with the proof, such as (x not_free_in A).',
  para,
  'Root must be the root node of a proof and Vars can be any',nl,
  'subset of the object variables and metavariables that appear',nl,
  'anywhere within that proof.',
  para,
  'The Nodes form of this command is perhaps easier to use, but provides',nl,
  'less control over which variables you are interested in.',nl,
  'Nodes must be a list of proof nodes, and the Root node of the',nl,
  'proof must be the first one in the list.  Typically, Nodes will',nl,
  'contain all the open nodes of the proof, as well as the Root node.',
  para,
  'Dist is a list of distinctness (x not_free_in y) constraints.',nl,
  'These show which object variables are assumed to be distinct',nl,
  '(that is, not unifiable).',
  para,
  'NFI is a list of (x not_free_in Term) constraints.',
  para,
  'User is a list of other constraints on the proof terms, including',nl,
  'user-defined rule_constraint predicates.  For example: ',nl,
  'delay_until(ground(X),integer(X)) might appear in this list.',
  para,
  'Context is a list of context-related constraints, such as',nl,
  'context_search(CName, Root, IdVar, Term).  These reflect assumptions',nl,
  'about the context terms that will be available when the current rule',nl,
  'has been proved and is being applied in future proofs.'
extends
  query_treecmds
mode
  proof_constraints(@list(nodeid),
                    -list(rule_constraint),
                    -list(rule_constraint),
                    -list(rule_constraint),
                    -list(rule_constraint)),
  proof_constraints(@nodeid, @list(var),
                    -list(rule_constraint),
                    -list(rule_constraint),
                    -list(rule_constraint),
                    -list(rule_constraint)).

proof_constraints(Nodes, CDist, CNFI, CUser, CContext) :-
    closed_list(Nodes),
    Nodes = [Root|_],
    proof_node_terms(Nodes, Terms),
    collect_allvars(Terms, Vars),
    proof_constraints(Root, Vars, CDist, CNFI, CUser, CContext).

proof_constraints(Root, Vars, CDist, CNFI, CRest, CContext) :-
    check_or_error(
        ( integer(Root),
          ip_lookup('Proof@Node', Root, Node),
          compound(Node),
          Node = proofnode(RootInfo, _, _, _, _, _, _),
          std_nonvar(RootInfo) ),
        ['proof node ',q(Root),' is not the root node of a proof.']),
    IdMap = [proofnode(Root)=_ | _],
    proof_constraints_internal(Root, Vars, IdMap,
                               CDist, CNFI, CRest, CContext),
    %% Callers are presumably not interested in seeing the variable
    %% names that will be used as sequent identities in the inference
    %% rule that will be generated from this proof.  So, here we
    %% instantiate those variables to the actual nodes of the current
    %% proof, to make it clear which node the various constraints
    %% are associated with (especially context constraints).
    once(map(proof_make_idmap_concrete, [IdMap])).


proof_make_idmap_concrete(proofnode(Id)=Id).


proof_constraints_internal(Root, Vars, IdMap,
                            CDist, CNFI, CRest, CContext) :-
    retry_delays,
    collect_constraints(Vars, CDist, CNFI, C3),
    proof_check_special_constraints(C3, CRest),
    context_constraints(Root, IdMap, CContext).

proof command
  theorem_constraints(Vars, Dist, NFI, Others)
doc
  'Get the constraint information associated with the theorem being proved.',nl,
  'Vars is the list of variables in the statement of the theorem,',nl,
  'Dist is the list of distinctness constraints,',nl,
  'NFI is the list of not_free_in constraints',nl,
  'and Others is the list of other constraints.'
extends
  query_treecmds
mode
  theorem_constraints(-list(var),
                    -list(rule_constraint),
                    -list(rule_constraint),
                    -list(rule_constraint)).

theorem_constraints(Vars, Dist, NFI, Others) :-
    ip_lookup('Proof@Node',1, Root),
    Root = proofnode(RootInfo, _, _, _, _, _, _),
    RootInfo = _(_Name,_Seqs,Vars,_IdMap,
                 Dist,NFI,Others,_CContext).

proof command
  proof_root_sequents(Sequents)
doc
  'Get the Sequents of the root node (node 1).'
extends
  query_treecmds
mode
  proof_root_sequents(-list(sequent)).

proof_root_sequents(Sequents) :-
    ip_lookup('Proof@Node',1, Root),     
    Root = proofnode(RootInfo, _, _, _, _, _, _),     
    ( RootInfo = root ->    
        Sequents = []     
    ;   
        RootInfo = _(_Name,Sequents,_Vars,_IdMap,  
                     _CDist,_CNFI,_CRest,_CContext)
    ).
 
proof_size parameter
  [set_proof_size, get_proof_size]
doc
  'This parameter is used to set the sizes of various internal',nl,
  'data structures that are used when constructing proof trees.',nl,
  'Constant time access/update of these data structures can be maintained',nl,
  'if this parameter is set to approximately the expected number of',nl,
  'nodes in the proof tree. (The default is 128.)'. 
 
get_proof_size(Val) :-
    global_state_lookup('Proof@ProofSize', Val).
 
set_proof_size(Val) :-
    0 =< Val,
    global_state_set('Proof@ProofSize', Val).
 
 
 
?- set_proof_size(128).                                                          
