%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: rule.ql,v 1.31 2006/02/06 00:52:12 itp Exp $

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Procedures for adding inference rules to the current theory.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Implementation Notes and Rationale
% ==================================
%
% There are at least three representations of Ergo inference rules.
% (Note that theorems/axioms are simply inference rules that have
% only a single sequent, so the following comments apply to them to).
%
%  1. The pretty syntax defined by tactics/prettyrule.ql.
%
%     Eg.  rule fred(Arg) ===
%                  PNode1 ::: hyp+++[A] ---> C
%                  ------------------------------
%                  PNode0 :::              A => C
%            provided
%              blah(Arg,PNode1).
%
%     Rationale: Concrete syntax that gives maximal readability for users.
%     
%  2. The lower level interface (`abstract syntax') of rule_add/5 below.
%     This is the same format that is returned by rule_query.
%
%     Eg.  rule_add(fred(Arg), 'thy.=>'(A,C), PNode0, 
%                   [ seq(PNode1, [hyp=add(_,A,use(hyp,PNode0))], C) ],
%                   blah(Arg,PNode1)
%                  ).
%
%     Rationale: this is easier to process than representation 1.
%     Also, we want Ergo to be independent of the precise concrete
%     syntax used by a particular user interface (each user interface
%     can potentially define new concrete syntaxes for rules).
%
%  3. A private executable representation, where the upper sequents of
%     rules are represented as a series of constraint procedures which
%     build new proof nodes, modify the context at those nodes, etc.
%
%     Eg.  infrule(fred,             % short name (unique within theory)
%                  fred(Arg:Type),        % full name (= short name + args).
%                  'thy.=>'(A,C),    % conclusion of lower sequent
%                  PNode0,           % identity variable for the node the
%                                      rule will be applied to.
%                  [PNode1])         % list of children's identity vars.
%               :-
%                  % constraint/node-building code.
%                  % (lots of low level ip_lookup's and ip_set's)...
%
%     Rationale: The primary purpose of this representation is to
%     achieve maximum efficiency for rule_apply.  It may change at any
%     time, if we find more efficient representations.
%
%     Important: This internal representation should never be seen
%     or used by end-users.
%
%
%  The theory database stores each rule in representation 2 *and* 3.
%  This is partly because for efficiency of rule application it is important
%  that the bodies (ie. non-head part) of representation 3 clauses are 
%  compiled, and Qu-Prolog provides no way of querying compiled clauses.
%  Thus representation 2 is needed to implement rule_query.  However,
%  even if we could query representation 3, it might be quite hard
%  to transform it back into representation 2, which is the desired
%  user view of rules.  Note that when rules are saved to a file,
%  it is representation 2 that is obtained from the database, but that
%  is then transformed (by rule_compile) into representation 3 in the 
%  saved file (in addition to representation 2).


database command
  rule_make_fundamental(Name)
doc
  'Marks the postulated rule, Name, as being a fundamental rule.',nl,
  'Fundamental rules are rules that cannot be proven, because they',nl,
  'form the axiomatic basis of the theory.',
  para,
  'This command is only allowed when the current theory is open,',nl,
  'because it extends the axiomatic basis of the theory.'
extends
  theory_extension_cmds
mode
  rule_make_fundamental(@std_nonvar).

rule_make_fundamental(Name) :-
    compound(Name),
    functor(Name,F,_),
    rule_make_fundamental(F,Name).
rule_make_fundamental(Name) :-
    atom(Name),
    rule_make_fundamental(Name,_).

rule_make_fundamental(ShortName,FullName) :-
    check_open(Thy),
    check_or_error(
        rule_query(Thy, ShortName, FullName, _, _, _, _),
        ['there is no postulate called ',q(Thy:ShortName)]),
    check_or_error(
        \+ query_theory(infrule_status(ShortName, _)),
        ['rule ',q(Thy:ShortName),' is already fundamental/proven.']),
    assert_theory(infrule_status(ShortName,fundamental)).



database command
  rule_add(Name, Concl, ConclId, SeqList, Constraints)
doc
  'Adds an inference rule called Name to the current theory.',nl,
  'The rule is added as a postulate.  It can be proved, or marked',nl,
  'as being a `fundamental'' rule later.',
  para,
  'Name may be an atom or a compound term.  However, in both cases',nl,
  'the outermost functor of Name must be unique within the current theory.',
  para,
  'Concl is the conclusion (succedent) of the bottom sequent of the',nl,
  'inference rule and ConclId is a variable that will be instantiated',nl,
  'to the name of the proof node that the inference rule is applied to.',
  para,
  'SeqList must be a closed list of seq(Id,Contexts,Concl) triples.',nl,
  'These triples are the upper sequents of the inference rule.',nl,
  'Contexts is a list of ContextName=CSpec terms.  See ',
  xref('CSpec',global,cspec),'.',
  para,
  'Constraints is a comma-separated series of procedures which will',nl,
  'be evaluated when the rule is used to extend a proof.',nl,
  'All of them must succeed (possibly by adding delayed constraints)',nl,
  'for an application of the rule to succeed.',
  para,
  'If SeqList = [S1,...,Sn], the meaning of the rule is that any',nl,
  'instance of Concl that satisfies Constraints can be proved by proving',nl,
  'the corresponding instances of S1...Sn.'
extends
  theory_extension_cmds
mode
  rule_add(@nonvar, @thy(term), @var, @list(sequent), @rule_constraint).

rule_add(Name, Concl, ConclId, Seqs0, Constraints) :-
    debug(rule, [q(rule_add(Name, Concl, ConclId, Seqs0, Constraints))]),
    check_current_theory(Thy),
    check_or_error(
        std_nonvar(Name),
        ['illegal rule name: ',q(Name)]),
    functor(Name, NameF, _),
    check_legal_rule_name(NameF),
    check_or_error(
        \+ rule_query(Thy, NameF, _, _, _, _),
        ['there is already a rule called ',q(Thy:NameF)]),
    %%
    %% To enable tighter well-formedness checking, we keep track
    %% of all variables that will eventually be instantiated to proof
    %% node names and context term names and check that they are not
    %% confused and that they do not appear within theory terms.
    %% The info/N structure is used to pass this 'type-checking'
    %% information to/from auxillary procedures.  CIds and ThyVars
    %% are open lists that accumulate the variables that are used for
    %% each purpose.
    %%
    Info = info(Thy:NameF,ThyVars,CIds,ConclId,NodeIds),
    rule_check_theory_term(Concl, Info),
    rule_check_seqs(Seqs0, Info, Seqs, SeqIds),
    debug(rule, ['checked seqs.  Info=',q(Info)]),
    %%
    %% Now check that each proof node has a distinct name variable.
    NodeIds = [ConclId|SeqIds],
    check_distinct_metavars(NodeIds),
    %%
    %% Check Constraints and partition them according to when
    %% they should be executed (Before/After construction of child nodes).
    %%
    rule_check_constraints(Constraints,Info,Before,After),
    code_tail(Before, true),
    code_tail(After, true),
    debug(rule, ['checked constraints.  Info=',q(Info),nl,
                 '  Before=',q(Before),nl,
                 '  After=',q(After)]),
    %%
    %% Now check that variables are not used in conflicting ways.
    %%
    open_to_closed(CIds),
    open_to_closed(ThyVars),
    intersect_list(NodeIds, CIds,    NodeAndCIds),
    intersect_list(NodeIds, ThyVars, NodeAndThy),
    intersect_list(CIds,    ThyVars, CIdsAndThy),
    check_or_error(
	NodeAndCIds = [],
        [q(NodeAndCIds),' used in conflicting ways ',
         '(to name proof nodes and context terms).']),
    check_or_error(
	NodeAndThy = [],
        [q(NodeAndThy),' used in conflicting ways ',
         '(as proof node names and theory terms).']),
    check_or_error(
	CIdsAndThy = [],
        [q(CIdsAndThy),' used in conflicting ways ',
         '(as context term names and theory terms).']),
    %%
    %% All checking completed!
    %%
    %% Now compile the rule into the efficient rule_apply form.
    %%
    rule_compile(ConclId, (rule_unify(Concl, BConcl), Before), 
                 After, Seqs, SeqIds, Code),
    debug(rule, ['compiled code to: ',q(Code)]),
    %%
    assert_theory(infrule(NameF,Name,Concl,ConclId,Seqs,Before,After), Thy),
    mode_from_rule(Name, RName, _),
    assert_theory(infrule_apply(NameF,RName,BConcl,ConclId,SeqIds), Code, Thy).


%% private rule_check_seqs(Seqs0, Info, Seqs, SeqIds).
%%
%%  Applies as many checks as possible to Seqs0, which is a list
%%  of sequents as passed to rule_add, and returns an equivalent
%%  list of sequents (in a more canonical form) in Seqs.
%%  SeqIds is the list of proof node identity variables in Seqs0.
%%
rule_check_seqs([], _, [], []).
rule_check_seqs([seq(Id, CSpecs0, Concl)|Seqs0], Info,
                [seq(Id, CSpecs,  Concl)|Seqs], [Id|Ids]) :-
    std_var(Id),
    !,
    rule_check_theory_term(Concl, Info),
    sort(CSpecs0, CSpecs1),
    rule_check_cspecs(CSpecs1, Info, CSpecs),
    rule_check_seqs(Seqs0, Info, Seqs, Ids).
rule_check_seqs([Seq|_], Info, _, _) :-
    rule_error(Info,['contains illegal sequent: ',q(Seq)]).

rule_check_cspecs([], _, []).
rule_check_cspecs([CName=CSpec|CS], Info, Out) :-
    CSpec = use(CName, ParentId),
    arg(4, Info, ParentId),
    !,
    %% copying the parent's context is the default behaviour, so we elide
    %% all such specs, thus obtaining a more canonical rule representation.
    rule_check_cspecs(CS, Info, Out).
rule_check_cspecs([CName=CSpec|CS0], Info, Out) :-
    check_is_context_name(CName),
    !,
    rule_check_cspec(CSpec, Info),
    rule_check_cspecs(CS0, Info, CS),
    Out = [CName=CSpec|CS].
rule_check_cspecs([Bad|_], Info, _) :-
    rule_error(Info, ['bad context spec: ',q(Bad)]).

rule_check_cspec(empty, _).
rule_check_cspec(use(CName,Node), Info) :-
    check_is_context_name(CName),
    arg(4, Info, ParentId),
    ( Node == ParentId ->
        true
    ;
        rule_error(Info, ['wrong node name in context spec: ',
                          q(use(CName,Node))])
    ).
rule_check_cspec(add(Id,T,CSpec), Info) :-
    rule_check_cid(Id, Info),
    rule_check_theory_term(T, Info),
    rule_check_cspec(CSpec, Info).
rule_check_cspec(delete(Id,T,CSpec), Info) :-
    rule_check_cid(Id, Info),
    rule_check_theory_term(T, Info),
    rule_check_cspec(CSpec, Info).
rule_check_cspec(filter(Func,CSpec), Info) :-
    rule_check_filter_func(Func, Info),
    rule_check_cspec(CSpec, Info).
%%rule_check_cspec(transform(Func,CSpec), Info) :-
%%    rule_check_transform_func(Func, Info),
%%    rule_check_cspec(CSpec, Info).


rule_check_cid(CId, Info) :-
    std_var(CId),
    !,
    arg(3, Info, CIds),
    search_insert(CId, CIds).
rule_check_cid(CId, Info) :-
    rule_error(Info, ['bad context term name: ',q(CId)]).


rule_check_filter_func(nfi(Var), Info) :-
    Var ?= _x,
    !,
    rule_check_theory_term(Var, Info).
rule_check_filter_func(Func, Info) :-
%% TO DO: Add a table of these filter functions in rule_check_filter_func.
    rule_error(Info, ['unknown context filter function: ',q(Func)]).

%rule_check_transform_func(_Func, _Info).
    

%% private rule_check_theory_term(Term, Info).
%%
%%  Checks that Term is a well-formed term in the current theory.
%%  Info (from rule_add) allows some additional checks to be done.
%%
%%  mode rule_check_theory_term(@thy(term), @compound).
%%
rule_check_theory_term(Term, Info) :-
    check_term(Term),
    collect_metavars(Term, Vars),
    arg(2, Info, ThyVars),
    rule_insert_vars(Vars, ThyVars).

rule_insert_vars([], _).
rule_insert_vars([V|VS], Open) :-
    search_insert(V, Open),
    rule_insert_vars(VS, Open).


%% rule_check_constraints(Constraints,Info,Before,After).
%%
%%  Check Constraints and partition it according to when each
%%  constraint should be executed.  Constraints is required to
%%  be a comma-separated sequence of calls to known constraint
%%  procedures.  They are partitioned into Before and After.
%%  Those constraints in Before can/should be
%%  executed before the child proof nodes have been built, whereas
%%  constraints in After must be executed after they have been built.
%%
%%  mode rule_check_constraints(@rule_constraint,+compound,?code,?code).
%%
rule_check_constraints(Constraints,_,_,_) :-
    Constraints == true,
    !.
rule_check_constraints(Constraints,Info,Before,After) :-
    std_nonvar(Constraints),
    Constraints = (C1,C2),
    !,
    rule_check_constraints(C1,Info,Before,After),
    rule_check_constraints(C2,Info,Before,After).
rule_check_constraints(Cmd0,Info,Before,After) :-
    std_nonvar(Cmd0),
    %% Cmd0 may be a user-supplied command (without a theory prefix),
    %% or an implementation (extracted via QPs collect_constraints).
    %% So we try both lookups.
    (
        query_global_or_ancestors(rule_constraint(Cmd0,Mode,Impl,_)),
        Cmd = Cmd0
    ;
        query_global_or_ancestors(rule_constraint(Cmd,Mode,Cmd0,_)),
        Impl = Cmd0
    ),
    !,
    ( rule_check_mode(Mode, Cmd, Info, Cmd2) ->
        true
    ;
        rule_error(Info, ['bad arguments to: ',q(Cmd)])
    ),
    ( rule_early_constraint(Cmd), Cmd=Cmd2 ->
        debug(rule, ['emitting into Before: ',q(Impl)]),
        rule_code_emit(Impl, Before),
        debug(rule, ['gives Before = ',q(Before)])
    ; Cmd=Cmd2 ->
        debug(rule, ['emitting into After: ',q(Impl)]),
	rule_code_emit_unchanged(Impl, After),
        debug(rule, ['gives After = ',q(After)])
    ; %% Cmd2 args have been transformed, so we must recompute Impl
        query_global_or_ancestors(rule_constraint(Cmd2,Mode,Impl2,_)),
        debug(rule, ['emitting changed into After: ',q(Impl2)]),
	rule_code_emit_unchanged(Impl2, After),
        debug(rule, ['gives After = ',q(After)])
    ).
rule_check_constraints(C,Info,_,_) :-
    rule_error(Info, ['bad constraint procedure: ',code(C)]).


%% private rule_check_mode(Mode, Cmd, Info, CmdOut).
%%
%%  This checks that Cmd contains sensible arguments.
%%  CmdOut is usually the same as Cmd, but may have some arguments
%%  transformed.  For example, delay_until(Cond,Code) gets transformed
%%  to delay_until(Cond,Impl), because its second argument has type
%%  rule_constraint, so may contain constraint commands that need
%%  to be expanded into Qu-Prolog code.  For example, if check(S)
%%  is a theory-specific constraint defined in theory 'abc', then
%%
%%      delay_until(ground(S), check(S))
%%
%%  will get expanded into something like
%%
%%      delay_until(ground(S), 'abc:check'(S)).
%%
%%  There is nothing special about delay_until/2 except the way it is
%%  declared.  Theory designers may define other higher-order constraint
%%  procedures, simply by including an argument of type 'rule_constraint'.
%%
%%  mode rule_check_mode(@list(compound),+rule_constraint,
%%                             +compound,-rule_constraint).
%%
rule_check_mode([Mode], Cmd, Info, CmdOut) :-
    !,
    functor(Mode, F, A),
    functor(Cmd, F, A),
    functor(CmdOut, F, A),
    rule_check_constraint_args(A, Mode, Cmd, Info, CmdOut).
rule_check_mode(_Modes, Cmd, _Info, _) :-
    msg(rule_add, ['WARNING: constraint ',q(Cmd),' has multiple modes.',nl,
                   'Arguments have not been type checked.']).


rule_check_constraint_args(0, _, _, _, _) :-
    !.
rule_check_constraint_args(N, Mode, Cmd, Info, CmdOut) :-
    N > 0,
    arg(N, Mode, MType),
    arg(N, Cmd, Val),
    arg(N, CmdOut, ValOut),
    ( MType = -Type ->
        check_or_error(
            std_var(Val),
            ['arg ',q(N),' of constraint ',q(Cmd),' must be an output.'])
    ;
        arg(1, MType, Type)
    ),
    ( rule_check_arg_type(Type, Val, Info, ValOut) ->
        true
    ;
        rule_error(Info, ['type error at arg ',q(N),' of constraint ',q(Cmd)])
    ),
    N1 is N - 1,
    rule_check_constraint_args(N1, Mode, Cmd, Info, CmdOut).


rule_check_arg_type(rule_constraint, Val, Info, ValOut) :-
    !,
    rule_check_constraints(Val, Info, Before, After),
    %% now combine (conjoin) Before and After, to give ValOut.
    code_tail(Before, After),
    code_tail(After, true),
    ValOut = Before.
rule_check_arg_type(Type, Val, Info, Val) :-
    %% Other types leave Val unchanged, so just use rule_check_arg_type/3.
    rule_check_arg_type(Type, Val, Info).

rule_check_arg_type(theory, Val, _) :-
    (atom(Val) ->
        check_is_theory_name(Val)
    ;
        std_var(Val)
    ).
rule_check_arg_type(thy(_), Val, Info) :-
    !,
    rule_check_theory_term(Val, Info).
rule_check_arg_type(nodeid, Val, Info) :-  %% For backwards compatibility.
    !,
    rule_check_arg_type(proofnode, Val, Info).
rule_check_arg_type(proofnode, Val, Info) :-
    !,
    arg(5, Info, NodeIds),
    %% This is perhaps a little strong, requiring Val to be the name
    %% of an upper or lower sequent of this rule.  In theory, other
    %% proof node names could come in via rule arguments etc...
    member_eq(Val, NodeIds).
rule_check_arg_type(cname, Val, _) :-
    !,
    (atom(Val) ->
        check_is_context_name(Val)
    ;
        std_var(Val)
    ).
rule_check_arg_type(ctermid, Val, Info) :-
    !,
    rule_check_cid(Val, Info).
rule_check_arg_type(Type, Val, _) :-
    debug(rule, ['WARNING: no typecheck for type ',q(Type),'  Val=',q(Val)]).


%% rule_early_constraint(Command).
%%
%%  Command is a rule_constraint command that can safely be
%%  tested before the child nodes of the rule are built.
%%  (This gives better efficiency, by allowing rules to fail
%%  earlier, before the work of creating the child nodes).
%%
%%  This must not succeed if Command overwrites existing data
%%  structures (See the note in rule_compile).
%%
%%  mode rule_early_constraint(?compound).
%%
rule_early_constraint(not_free_in(_,_)).
rule_early_constraint(is_not_free_in(_,_)).
rule_early_constraint(defined(_,_)).
rule_early_constraint(context_search(_,_,_,_)).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%  `Compilation' of inference rules.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% rule_compile(ConclId, Before, After, Seqs, SeqIds, Code).
%%
%%  Code is Qu-Prolog code that calls all the constraint procedures
%%  in the Before and After lists and (inbetween those two)
%%  builds all the sequents in Seqs.
%%
%%  mode rule_compile(@var, ?code, @code, @list(sequent), ?list(var), -code).
%%
rule_compile(_, Before, After, _, _, _) :-
    (list(Before) ; list(After)),
    !,
    handle_error(['obsolete infrule representation.  Rebuild your theories.']).
rule_compile(ConclId, Before, After, Seqs, SeqIds, Code) :-
    %%
    %% Each of the predicates that build Seqs is inserted into Code.
    %% Initially Code contains just the Before predicates.
    %% Code is treated a bit like a set, in that predicates that are
    %% already in Code (using the custom equality, rule_predicate_equal/2)
    %% are not inserted again.  This relies on the fact that none of the
    %% predicates that we are inserting overwrite data structures that are
    %% used by the other predicates.  The predicates that rule_compile_seq
    %% inserts all satisfy this property and rule_early_constraint
    %% ensures that the predicates in Before satisfy it too.
    %% (but After contains arbitrary user constraints).
    %%
%    rule_code_inline(Before, BeforeIn),
    rule_code_inline(Before, BeforeIn),
    rule_code_emit(BeforeIn, Code),
    map(rule_compile_seq(ConclId,Code), [Seqs,SeqIds]),
    code_tail(Code, After).


code_tail(Code, Tail) :-
    var(Code),
    !,
    Tail = Code.
code_tail((_,Code), Tail) :-
    code_tail(Code, Tail).


rule_compile_seq(ConclId, OpenCode, seq(Id,CSpecs,Term), Id) :-
    %% An inline version of proof_new_node(Id, ConclId, Term).
    rule_code_emit((
                    global_state_increment('Proof@Next@Id', Id),
                    ip_set('Proof@Node', Id, proofnode(ConclId,Term,_,_,_,_,_))
                   ), OpenCode),
    rule_compile_cspecs(CSpecs, ConclId, Id, OpenCode).

rule_compile_cspecs([], _, _, _).
rule_compile_cspecs([CName=CSpec|CS], ConclId, Id, OpenCode) :-
    rule_compile_cspec(CSpec, OpenCode, Cref),
    rule_code_emit(ip_set(CName,Id,Cref), OpenCode),
    rule_compile_cspecs(CS, ConclId, Id, OpenCode).

rule_compile_cspec(empty, _, 0).
rule_compile_cspec(use(CName,Node), OpenCode, Cref) :-
    rule_code_emit((
                    context_lookup_cref(CName,Node,Cref)
                   ),
                   OpenCode).
rule_compile_cspec(add(I,T,CSpec), OpenCode, Cref) :-
    rule_compile_cspec(CSpec, OpenCode, Cref0),
    rule_code_emit((
                    global_state_increment('Context@Cref', Cref),
                    global_state_increment('Context@TermId', I),
                    ip_set('Context@Forest', Cref, forest_add(I,T,Cref0))
                   ),
                   OpenCode).
rule_compile_cspec(delete(I,T,CSpec), OpenCode, Cref) :-
    rule_compile_cspec(CSpec, OpenCode, Cref0),
    rule_code_emit((
                    ip_lookup('Context@Forest', Cref0, CNode0),
                    context_search2(CNode0, I, T),
                    global_state_increment('Context@Cref', Cref),
                    ip_set('Context@Forest', Cref, forest_delete(I,Cref0))
                   ),
                    OpenCode).
rule_compile_cspec(filter(Func,CSpec), OpenCode, Cref) :-
    rule_compile_cspec(CSpec, OpenCode, Cref0),
    rule_code_emit((
                    global_state_increment('Context@Cref', Cref),
                    ip_set('Context@Forest', Cref, forest_filter(Func,Cref0))
                   ),
                    OpenCode).
%%rule_compile_cspec(transform(Func,CSpec), OpenCode, Cref) :-
%%    rule_compile_cspec(CSpec, OpenCode, Cref0),
%%    rule_code_emit((
%%                    global_state_increment('Context@Cref', Cref),
%%                    ip_set('Context@Forest', Cref, forest_transform(Func,Cref0))
%%                   ),
%%                    OpenCode).


%% private rule_code_inline(Code, CodeIn)
%%
%% Inlines context_search(CName, PNode, CId, CTerm) to
%%    context_lookup_cref(CName, PNode, Cref),
%%    ip_lookup('Context@Forest', Cref, CNode),
%%    context_search2(CNode, CId, CTerm)
%%
rule_code_inline((C1,C2), (C1In, C2In)) :-
    !,
    rule_code_inline(C1, C1In),
    rule_code_inline(C2, C2In).
rule_code_inline(context_search(CName, PNode, CId, CTerm), Code) :-
    !,
    Code = 
      (
        context_lookup_cref(CName, PNode, Cref),
        ip_lookup('Context@Forest', Cref, CNode),
        context_search2(CNode, CId, CTerm)
      ).
rule_code_inline(Code, Code).

%% private rule_code_emit(Call, OpenCode).
%%
%%  Similar to QP's search_insert/2, but works on comma-separated
%%  code rather than lists, and uses a special form of
%%  equality (mostly ==, but with some arguments unified instead).
%%  If Call contains multiple predicates (comma-separated), they
%%  are each inserted in turn.
%%
%%  OpenCode is a comma-separated, right associated collection
%%  of predicate calls, whose tail is a variable.
%%
%%  mode rule_code_emit(+nonvar, ?nonvar).
%%
rule_code_emit(true, _) :-
    !.
rule_code_emit((C1,C2), OpenCode) :-
    !,
    rule_code_emit(C1, OpenCode),
    rule_code_emit(C2, OpenCode).
rule_code_emit(Call, OpenCode) :-
    var(OpenCode),
    !,
    OpenCode = (Call,_).
rule_code_emit(Call, (Call2,_)) :-
    rule_predicate_equal(Call,Call2),
    !.
rule_code_emit(Call, (_,OpenCode)) :-
    rule_code_emit(Call, OpenCode).


%% private.
%%
%%  This determines how much code sharing we get between
%%  the context construction code of different sequents.
%%         
rule_predicate_equal(ip_lookup(A,I,V),
                     ip_lookup(A2,I2,V2)) :-
    !,
    A == A2,
    I == I2,
    V = V2.
rule_predicate_equal(context_lookup_cref(A,I,V),
                     context_lookup_cref(A2,I2,V2)) :-
    !,
    A == A2,
    I == I2,
    V = V2.
rule_predicate_equal(context_search2(A,I,V),
                     context_search2(A2,I2,V2)) :-
    !,
    A == A2,
    I == I2,
    check_or_error(V = V2, ['Context search with the same ID but different terms.']).
rule_predicate_equal(Call, Call2) :-
    Call == Call2.


%% private rule_code_emit_unchanged(Call, Collected).
%%
%%  This is the same as rule_code_emit, but it preserves
%%  the order of predicates exactly (because it is inserting
%%  arbitrary user-defined predicates, which may have side-effects).
%%
%%  Collected is a comma-separated, right associated collection
%%  of predicate calls, whose tail is a variable.
%%
%%  mode rule_code_emit_unchanged(+nonvar, ?nonvar).
%%
rule_code_emit_unchanged(Call, Collected) :-
    code_tail(Collected, Tail),
    rule_code_emit_unchanged2(Call, Tail, _).

rule_code_emit_unchanged2(true, Tail0, Tail) :-
    !,
    Tail = Tail0.
rule_code_emit_unchanged2((A,B), Tail0, Tail) :-
    !,
    rule_code_emit_unchanged2(A, Tail0, Tail1),
    rule_code_emit_unchanged2(B, Tail1, Tail).
rule_code_emit_unchanged2(C, Tail0, Tail) :-
    Tail0 = (C,Tail).



%% rule_apply(Thy, Name, Concl, Id, Children).
%% rule_apply(Thy, Name, FullName, Concl, Id, Children).
%%  Applies the given rule to the proof node Id.
%%  Children is the resulting new proof nodes.
%%
%%  mode rule_apply(?theory,?rulename,?thy(term),@proofnode,
%%                   ?list(proofnode)).
%%  mode rule_apply(?theory,?srulename,?rulename,?thy(term),@proofnode,
%%                   ?list(proofnode)).
%%
/*
rule_apply(Thy, Name, Concl, Id, Children) :-
    compound(Name),
    !,
    functor(Name, F, _),
    ancestor(Thy),
    atom_concat2(Thy, '/L/infrule_apply', Pred),
    call_predicate(Pred(F, Name, Concl, Id, Children)).
rule_apply(Thy, Name, Concl, Id, Children) :-
    ancestor(Thy),
    atom_concat2(Thy, '/L/infrule_apply', Pred),
    call_predicate(Pred(Name, _, Concl, Id, Children)).
*/

rule_apply(Thy, Name, Concl, Id, Children) :-
    compound(Name),
    !,
    functor(Name, F, _),
    query_ancestors(infrule_apply(F, Name, Concl, Id, Children), Thy).
rule_apply(Thy, Name, Concl, Id, Children) :-
    query_ancestors(infrule_apply(Name, _, Concl, Id, Children), Thy).

rule_apply(Thy, Name, FullName, Concl, Id, Children) :-
    query_ancestors(infrule_apply(Name, FullName, Concl, Id, Children), Thy).

%% rule_query(Thy,            Name, Concl, ConclId, Seqs, Constraints).
%% rule_query(Thy, ShortName, Name, Concl, ConclId, Seqs, Constraints).
%%
%%  True iff there is an inference rule in Thy called Name.
%%  The other arguments have the same meaning as they do in rule_add.
%%
%%  mode rule_query(?theory, ?nonvar,
%%                  ?var, ?thy(term), ?list(sequent), ?rule_constraint).
%%  mode rule_query(?theory, ?atom, ?nonvar,
%%                  ?var, ?thy(term), ?list(sequent), ?rule_constraint).
%%
rule_query(Thy, Name, Concl, Id, Seqs, Constraints) :-
    compound(Name),
    !,
    mode_from_rule(TypedName, Name, _),
    functor(Name, F, _),
    rule_query(Thy, F, TypedName, Concl, Id, Seqs, Constraints).
rule_query(Thy, Name, Concl, Id, Seqs, Constraints) :-
    rule_query(Thy, Name, _, Concl, Id, Seqs, Constraints).

rule_query(Thy, F, Name, Concl, Id, Seqs, Constraints) :-
    query_ancestors(infrule(F, Name, Concl, Id, Seqs, Before, After), Thy),
    Constraints = (Before,After).

%% Extract the types and terms from the name of a theorem
%% E.G. mode_from_rule(foo(arg1:type1, arg2:type2), 
%%                     foo(arg1, arg2), foo(type1, type2)).

mode_from_rule(TypedRule, Rule, Mode) :-
    (atom(TypedRule);atom(Rule)),!,
    Rule = TypedRule,
    Mode = TypedRule.
mode_from_rule(TypedRule, Rule, Mode) :-
    compound(TypedRule),!,
    functor(TypedRule, Name, N),
    functor(Rule, Name, N),
    functor(Mode, Name, N),
    mode_from_rule_args(1, N, TypedRule, Rule, Mode).
mode_from_rule(TypedRule, Rule, Mode) :-
    check_or_error(compound(Rule), 
          ['Wrong types for call', q(mode_from_rule(TypedRule, Rule, Mode))]),
    functor(Rule, Name, N),
    functor(TypedRule, Name, N),
    functor(Mode, Name, N),
    mode_from_rule_args(1, N, TypedRule, Rule, Mode).

mode_from_rule_args(A, N, _TypedRule, _Rule, _Mode) :-
    A > N, !.
mode_from_rule_args(A, N, TypedRule, Rule, Mode) :-
    arg(A, TypedRule, ':'(R,M)),
    arg(A, Rule, R),
    arg(A, Mode, M),
    A1 is A+1,
    mode_from_rule_args(A1, N, TypedRule, Rule, Mode).
    

database command
  add_context_name(Name)
extends
  theory_extension_cmds
doc
  'Defines Name as a new kind of context.',nl,
  'The semantics of that context is determined purely by the',nl,
  'inference rules that manipulate it.'
mode
  add_context_name(@atom).

add_context_name(Name) :-
    check_atom(Name),
    check_or_error(
        \+ query_ancestors(context_name(Name)),
        [q(Name),' is already declared as a context name.']),
    assert_theory(context_name(Name)).


is_context_name(N) :-
    query_ancestors(context_name(N)).


check_is_context_name(N) :-
    query_ancestors(context_name(N)),
    !.
check_is_context_name(N) :-
    handle_error(['unknown context name: ',q(N)]).


%% private rule_error(Info, ErrorMessage).
%%
%%  Same as handle_error, but displays rule name too.
%%
%%  mode rule_error(@compound, @writelist).
%%
rule_error(Info, Err) :-
    arg(1, Info, Thy:Name),
    handle_error(['in rule ',q(Thy),':',q(Name),', '|Err]).




%%  This checks that there is a bijection between the user-specified
%%  form of the rule constraint and its implementation form.
%%  This is desirable, because rule_check_constraints above
%%  translates back from Impl to Call, in order to check the
%%  well-formedness of the constraint and check modes etc.
%%  If it cannot do this, rule_add will fail with an error, which
%%  means that you cannot prove any theorems that have this constraint
%%  as a delayed side-condition.  In many cases, this may be acceptable,
%%  but we give a warning here at the time the rule_constraint is
%%  defined.
%%
%%  At the moment we have quite a strict test.
%%
rule_constraint_check_impl(Call, Impl) :-
    Call =.. [CFunc | CArgs],
    Impl =.. [IFunc | IArgs],
    atom(CFunc),
    atom(IFunc),
    CArgs == IArgs,
    !.
rule_constraint_check_impl(Call, _) :-
    %% 
    writelist(['WARNING: rule_constraint ',code(Call),
               ' has a non-standard/complex implementation.',nl,
               'This means you must discharge this side-condition',
               ' before completing proofs.',nl]).


rule_unify(A,B) :-
    ip_lookup('rule@unify@flag', Flag),
    Flag == sunify,
    structural_unify(A,B),!.
rule_unify(A,A).

