%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: abbrev.ql,v 1.100 2004/07/28 03:26:50 itp Exp $

#include "../prover/lib/ergo.h"

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% This file provides abbreviation (or syntactic sugaring)
%% facilities that can be used by user interface tactics.
%% Two abbrevation facilities are provided: 
%%
%%   - a renaming mechanism for defining abbreviations for particular
%%     terms.  For example:  '<=>'(A,B)  ===  'ergo.equiv'(A,B).
%%
%%   - a set of 'visible theories' which allows constant, functor and
%%     quantifier names to be specified without an explicit theory
%%     prefix.  For example, if 'ab' is a visible theory, then 'ab.cd'
%%     will be read and written as just 'cd' (provided that this does
%%     not introduce ambiguity because '<X>.cd' is also a legal name
%%     where <X> is also in the set of visible theories).
%%
%% The abbreviations and set of visible theories can be changed at
%% any time, even if the current theory is closed.
%%
%% This file also adds a write-event handler to catch any theory
%% terms that are written, so that they can be abbreviated.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Implementation notes
% ====================
%
% 1) Visible Name Cache
% =====================
%  This file maintains a cache of asserted facts of the
%  form:  visible_cache(Name, Arity, InternalName),
%  where InternalName is the fully qualified name corresponding
%  to Name if Name is not ambiguous, and InternalName = 'ambiguous'
%  if Name is ambiguous (that is, is defined in more than one
%  currently visible theory.  Arity=when(_) if Name is the name
%  of a parameterized constant, otherwise it is a normal arity list.
%
%  Needless to say, this cache is updated (by retracting relevant
%  entries) each time the set of visible theories changes (this includes
%  all cd commands), or a new name is added to any of the visible theories.
%  The first few listener hooks defined below do this.
%
% 2) [De]Abbreviate Algorithm
% ===========================
%  Deabbreviate does a bottom-up pass over the given term, whereas
%  abbreviate does a top-down pass.  The latter is necessary so
%  that the RHSs of abbreviations are matched against the original
%  Ergo internal terms.  The algorithms for these two procedures are
%  fairly subtle in order to handle higher order terms sensibly.
%  They must also avoid instantiating non-frozen variables.
%
% 3) Inheritance of visible_theory/1 and abbrev/4 facts
% =====================================================
%  At the moment, the inherited_visible_theories and
%  inherited_abbrevs links between theories are non-transitive.
%  It would not be hard to change this to transitive,
%  by recursing up through the tree of links, either on each
%  query (slow!) or each time the link structure changes (however,
%  it is not very easy to notice changes to links between ancestor
%  theories at the moment).


%% setup_abbreviations.
%%
%%  This procedure is typically called when building an
%%  entirely new Ergo system from an empty global.erg.
%%  It sets up the abbreviation mechanisms, ready for use by
%%  other tactic packages.
%%  If it is called multiple times, the first call does the
%%  setup job and following calls have no additional effect.
%%
%%  mode setup_abbrevations.
%%
setup_abbreviations :-
    query_global(request(deabbreviate_with_type(_,_,_),_)),
    !.
setup_abbreviations :-
    theory_fact(inherited_abbrevs,1),
    !,
    setup_abbreviations_1.
setup_abbreviations :-
    %% declare all saved facts and events used by this package.
    new_theory_fact(inherited_abbrevs,1,[save(true),assert(a)]),
    new_theory_fact(inherited_visible_theories,1,[save(true),assert(a)]),
    new_theory_fact(abbrev,4,[save(true),assert(a)]),
    new_theory_fact(visible_theory,1,[save(true),assert(a)]),
    add_announcement(
        visible_theory(_,_,_),
        ['visible_theory(CurrThy,Thy,Action).',nl,
         'Action is add or delete.',nl]),
    %% We could perhaps combine these two deabbreviate requests.
    add_request(
        deabbreviate(_,_,_,_),
        ['deabbreviate(In,Thy,Arity,Out).',nl,
         'Do custom deabbreviation of In->Out.',nl]),
    add_request(
        abbreviate(_,_,_,_,_),
        ['abbreviate(In,Thy,Arity,Depth,Out).',nl,
         'Do custom abbreviation of In->Out.',nl]),
    setup_abbreviations_1.

%% deabbreviate_with_type is a new request type (Oct 1996).
%% Allow old global.erg files that lack this fact to work properly by checking
%% for it separately.
%% TO DO: MOVE THIS INTO setup_abbreviations FOR THE NEXT MAJOR RELEASE.
setup_abbreviations_1 :-
    add_request(
        deabbreviate_with_type(_,_,_),
        ['deabbreviate_with_type(Type,In,Out).',nl,
         'Deabbreviate In of type Type giving Out.']).

?- setup_abbreviations.
   %% Must call setup here to ensure that visible_theory announcement
   %% is defined before we try to add a listener to it.
   %% A more clumsy alternative would be to hook our listener directly in
   %% to every point in this file where visible_theory announcements
   %% are made.

?- dynamic(visible_cache/3).

?- add_listener(
    %% This catches all cd commands.  Thy is the new theory.
    theory(_Thy, selected),
    ( abolish(visible_cache/3),
      dynamic(visible_cache/3))).

?- add_listener(
    %% This catches all asserts of new symbols into the Ergo database.
    assert(Thy,symbol(Sym,Arity,_)),
    delete_cache_entry(Sym,Arity,Thy)).

?- add_listener(
    %% This catches adds/deletes to the set of visible theories.
    visible_theory(_CurrThy,Thy,_Action),
    delete_cache_entries(Thy)).


delete_cache_entry(param_const(IName,_), Arity, Thy) :-
    !,
    check_or_error(
        Arity = when(_),  %% should always be true.
        ['param_const ',q(IName),' has strange arity: ',q(Arity)]),
    internal_name(IName,Thy,Name),
    retractall(visible_cache(Name,when(_),_)).
delete_cache_entry(IName, Arity, Thy) :-
    internal_name(IName,Thy,Name),
    retractall(visible_cache(Name,Arity,_)).


delete_cache_entries(Thy) :-
    (
        query_theory(symbol(Sym,Arity,_), Thy),
        delete_cache_entry(Sym, Arity, Thy),
        fail
    ;
        true
    ),
    !.


general command
  enable_abbreviation,
  disable_abbreviation
extends
  sugar_cmds
doc
  'Enable/disable automatic abbreviation of theory terms on output.'
mode
  enable_abbreviation,
  disable_abbreviation.

enable_abbreviation :-
    add_handler(
        write(Stream, Depth, thy(Thy,Term)),
        abbreviate_and_write(Stream, Term, Thy, Depth)),
    add_handler(
        write(Stream, Depth, code(Thy,Code)),
        abbreviate_and_write(Stream, Code, Thy, Depth)).

disable_abbreviation :-
    delete_handlers(
        write(_,_,_),
        abbreviate_and_write(_,_,_,_)).

?- enable_abbreviation.

abbreviate_and_write(Stream, Term0, Thy, Depth) :-
    abbreviate(Term0, Thy, [], Depth, Term),
    write_term(Stream, Term,
               [op_table(Thy), obvar_prefix_table(Thy), 
                quoted(true), remember_name(true)]).



database command
  abbreviation('==='(E,I))
header
  'abbreviation E === I.'
extends
  sugar_cmds
mode
  abbreviation(@compound)
implementation
  (I = provided(I2,Cond) ->
      add_abbrev(E,I2,Cond)
  ;
      add_abbrev(E,I,true)
  )
doc
  'Defines E as an abbreviation for I.',
  para,
  'I may be of the form: Term provided Cond, where Cond is',nl,
  '''input'' for input-only abbreviations, or some Prolog code',nl,
  'to perform a conversion between E and I.  However, there are',nl,
  'various constraints on such code which not yet described here.'.

database command
  delete_abbrev(Abbrev),
  delete_abbrev(Abbrev,Term)
extends
  sugar_cmds
mode
  delete_abbrev(?term),
  delete_abbrev(?term, ?thy(term))
doc
  'Deletes the given abbreviation.',nl,
  'It checks that Abbrev and Term are not both variables, since that',nl,
  'would delete an arbitrary abbreviation.'.

database command
  add_abbrevs(Thy)
extends
  sugar_cmds
mode
  add_abbrevs(@theory)
doc
  'Copies all abbreviations in theory Thy into the current theory.'.

database command
  delete_abbrevs(Thy)
extends
  sugar_cmds
mode
  delete_abbrevs(@theory)
doc
  'Deletes all abbreviations that were copied from theory Thy.',nl,
  'If Thy is a variable, ALL abbreviations will be deleted.'.

database command
  add_visible_theory(Thy)
extends
  sugar_cmds
mode
  add_visible_theory(@theory)
doc
  'Adds Thy to the set of visible theories.',nl,nl,
  'Eg. visible_theory(t) allows ''t.c'' to be written as ''c''.'.

database command
  delete_visible_theory(Thy)
extends
  sugar_cmds
mode
  delete_visible_theory(@theory)
doc
  'Deletes Thy from the set of visible theories.'.

%% add_abbrev(ETerm, ITerm).
%% add_abbrev(ETerm, ITerm, Body).
%%
%%  Add an abbreviation to the current theory, which defines ETerm
%%  to be syntactic sugar for ITerm.  There are two kinds of
%%  abbreviations that can be created by these procedures:
%%
%%  Definitional abbreviations:  These correspond to definitions
%%    of new constants or functions at the object theory level,
%%    except that they never appear within theorems etc. because they
%%    are unfolded upon input (see deabbreviate/4).  They are folded
%%    back in upon output (see abbreviate/4) iff they correspond to
%%    a simple renaming.  That is, if ETerm and ITerm are identical
%%    except for the name of the outermost functor/quantifier.
%%    For definitional abbreviations, ETerm and ITerm must satisfy
%%    the same constraints as in a definition, ETerm === ITerm, and
%%    Body must be 'true' or 'input', where 'input' means that the
%%    abbreviation will be unfolded upon input, but never folded back
%%    in upon output.
%%
%%  Parameterized constant abbreviations:  These define the concrete
%%    syntax that is used to represent a given set of parameterized
%%    constants.  Such abbreviations are always unfolded upon input and
%%    folded upon output.  ETerm may be any term (including a variable),
%%    ITerm must be of the form param_const(Name,Val), and Body must be
%%    Prolog code that translates ETerm to ITerm and vice versa.
%%
%%  mode add_abbrev(@term, @term).
%%  mode add_abbrev(@term, @term, @nonvar).
%%
add_abbrev(E, I) :-
    add_abbrev(E, I, true).

add_abbrev(Head, Term, Body) :-
    (
        Body == true
    ;
        Body == input
    ),
    !,
    check_current_theory(Thy),
    deabbreviate(Term, Thy, [], Term2),
    check_definition(Head, Term2, _Name, _Arity, NFIs),
    check_or_error(
        NFIs == true,
        ['abbreviation is too complex.  It generates side conditions:',nl,
         code(Thy,NFIs)]),
    abbrev_kind(Head, Term2, Body, Kind),
    add_abbrev2(Head, Term2, Kind, Thy, Thy).
add_abbrev(Head, Term, Code) :-
    check_current_theory(Thy),
    deabbreviate(Term, Thy, [], Term2),
    check_or_error(
        is_param_const(Term2),
        ['illegal param_const abbreviation.',nl,
         'Since code is not ''true'', the right hand side must be',nl,
         'a parameterized constant.  But it is: ',thy(Term2)]),
    add_abbrev2(Head, Term2, when(Code), Thy, Thy).


%% private abbrev_kind(LHS, RHS, Body, Kind).
%%
%%  This takes a definitional abbreviation and determines whether
%%  it should be an input-only abbreviation (Kind=input), or an
%%  input-output abbreviation (Kind=input_output).
%%
%%  Kind is set to input_output if every possible subterm path
%%  into LHS leads to the identical subterm in RHS, and neither
%%  side is a quantifier.  (The latter restriction could perhaps
%%  be dropped if abbreviate/4 was made more sophisticated).
%%
%%  mode abbrev_kind(@term, @thy(term), @code, ?atom).
%%
abbrev_kind(_, _, input, Kind) :-
    !,
    Kind = input.
abbrev_kind(LHS, RHS, _, Kind) :-
    \+ quant(RHS),
    \+ quant(LHS),
    \+ (Path = [_|_],
        subterm_path(LHS, Path, T1),
        \+ (subterm_path(RHS, Path, T2),
            T1 == T2)),
    !,
    Kind = input_output.
abbrev_kind(_, _, _, input).


%% private abbrev_input_test(Kind).
%% private abbrev_output_test(Kind).
%%
%%  These check that Kind is an allowed input/output abbreviation,
%%  respectively.  This includes calling any code that computes
%%  the input from the output (or vice versa).
%%
abbrev_input_test(input).
abbrev_input_test(input_output).
abbrev_input_test(true).  %% for backward compatibility.
abbrev_input_test(when(Code)) :- call_predicate(Code).

abbrev_output_test(input_output).
abbrev_output_test(when(Code)) :- call_predicate(Code).


%% private add_abbrev2(ETerm, ITerm, Kind, FromThy, ToThy).
%%
%%  This procedure is used to implement add_abbrev/3 and add_abbrevs/1
%%  and should not be called directly by users.
%%  It assumes that ETerm, ITerm and Kind define a valid abbreviation
%%  and adds that abbreviation to ToThy, marking it as having come from
%%  FromThy.  If the abbreviation would clash with existing abbreviations
%%  to ToThy, error messages are generated and the procedure fails.
%%  FromThy should be an ancestor of ToThy.
%%
%%  mode add_abbrev2(@term, @term, @nonvar, @atom, @atom).
%%
add_abbrev2(E, I, Kind, FromThy, ToThy) :-
    nonvar(E),
    query_abbrev(E2,I2,Kind2,FromThy2,ToThy),
    nonvar(E2),
    E = E2,
    !,
    ( I=I2, Kind=Kind2, FromThy=FromThy2 ->
        true
    ;
        handle_error([thy(E),' is already defined as an abbreviation.'])
    ).
add_abbrev2(E, I, Kind, FromThy, ToThy) :-
    Kind = when(_),
    nonvar(I),
    query_abbrev(E,I,Kind,FromThy2,ToThy),
    !,
    ( FromThy=FromThy2 ->
        true
    ;
        handle_error([thy(E),' === ',q(I),
                      ' is already defined as an abbreviation.'])
    ).
add_abbrev2(E, I, Kind, FromThy, ToThy) :-
    tactic_assert_theory(abbrev(E,I,Kind,FromThy),ToThy).


%% delete_abbrev(ETerm).
%% delete_abbrev(ETerm, ITerm).
%%
%%  Delete the abbreviation that matches ETerm (and ITerm, if given).
%%  This can be used to delete abbreviations that were added directly
%%  to the current theory, or abbreviations that were copied from some
%%  other theory (but see also delete_abbrevs/1 for the latter case).
%%
%%  mode delete_abbrev(+nonvar).
%%  mode delete_abbrev(+nonvar, ?nonvar).
%%  mode delete_abbrev(?nonvar, +nonvar).
%%
delete_abbrev(E) :-
    delete_abbrev(E,_).

delete_abbrev(E,I) :-
    check_current_theory(Thy),
    check_or_error(
        (
            nonvar(E)
        ;
            nonvar(I)
        ),
        [q(delete_abbrev(E,I)),' is not sufficiently specific.']),
    deabbreviate(I,I2),
    tactic_retract_theory(abbrev(E,I2,_,_), Thy).


%% add_abbrevs(Thy).
%% add_all_abbrevs(Thy).
%%
%%  Copy abbreviations from theory Thy into the current theory,
%%  discarding any that clash with existing abbreviations in the
%%  current theory.  A warning message will be produced for each
%%  abbreviation that is discarded.
%%
%%  Note that add_abbrevs/1 copies only those abbreviations that
%%  were defined in Thy, whereas add_all_abbrevs/1 also copies
%%  any abbreviations that have been copied into Thy from other
%%  theories.
%%
%%  mode add_abbrevs(@atom).
%%  mode add_all_abbrevs(@atom).

add_abbrevs(FromThy) :-
    check_current_theory(Thy),
    check_strict_ancestor(FromThy),
    %% The inherited abbreviations must be added at the front of the current
    %% list (linear visibility), but in the original order.  So we'll use
    %% asserta after reversing the original list.
    findall(abbrev(E,I,Body,FromThy),
            query_theory(abbrev(E,I,Body,FromThy), FromThy),
            Abbrevs),
    add_abbrevs1(Abbrevs, Thy).

add_all_abbrevs(FromThy) :-
    check_current_theory(Thy),
    check_strict_ancestor(FromThy),
    findall(abbrev(E,I,Body,OrigThy),
            query_theory(abbrev(E,I,Body,OrigThy), FromThy),
            Abbrevs),
    add_abbrevs1(Abbrevs, Thy).

%% private add_abbrevs1(Abbrevs, Thy).
%% 
%%  mode add_abbrevs1(@list(compound), @theory).
%% 
add_abbrevs1([], _).
add_abbrevs1([abbrev(E,I,Body,FromThy)|Abbrevs], Thy) :-
    add_abbrevs1(Abbrevs, Thy),
    add_abbrev2(E,I,Body,FromThy,Thy).

%% inherit_all_abbrevs(Thy).
%%
%%  Inherits all the abbreviations of Thy into the current theory.
%%  It does this by sharing the visible_theory facts of Thy,
%%  rather than copying them into the current theory.
%%  Thy must be an ancestor of the current theory.
%%
%%  mode inherit_all_abbrevs(@theory).
%%
inherit_all_abbrevs(FromThy) :-
    check_current_theory(CD),
    check_strict_ancestor(FromThy),
    ( query_theory(inherited_abbrevs(FromThy), CD) ->
        msg(abbreviations, ['theory ',w(CD),' already inherits the',
                            ' abbreviations of theory ',w(FromThy)])
    ;
        tactic_assert_theory(inherited_abbrevs(FromThy), CD)
    ).


%% delete_abbrevs(Thy).
%%
%%  Delete all abbreviations that were copied from theory Thy.
%%  If Thy is a variable, *all* abbreviations in the current theory
%%  will be deleted.  If Thy is the name of the current theory, all
%%  abbreviations added directly to the current theory will be deleted,
%%  but abbreviations copied from ancestor theories will be left unchanged.
%%
%%  mode delete_abbrevs(@atom).
%%
delete_abbrevs(FromThy) :-
    check_current_theory(Thy),
    tactic_retractall_theory(abbrev(_,_,_,FromThy), Thy).



%% add_visible_theory(Thy).
%%
%%  Adds Thy to the set of visible theories.
%%  Thy be an ancestor of the current theory.
%%  If Thy is already visible, this command succeeds silently.
%%
%%  mode add_visible_theory(@atom).
%%
add_visible_theory(Thy) :-
    check_current_theory(CD),
    check_ancestor(Thy),  %% includes subtheories
    (query_visible_theory(Thy,CD) ->
        true
    ;
        tactic_assert_theory(visible_theory(Thy), CD),
        announce_event(visible_theory(CD, Thy, add))
    ).


%% add_all_visible_theories(Thy).
%%
%%  Makes all theories that are visible in Thy visible in the
%%  current theory.
%%  Thy must be an ancestor of the current theory.
%%
%%  mode add_all_visible_theories(@theory).
%%
add_all_visible_theories(FromThy) :-
    check_current_theory(_CD),
    check_strict_ancestor(FromThy),
    \+ (query_visible_theory(Thy,FromThy),
        add_visible_theory(Thy),
        fail).


database command
  inherit_all_visible_theories(Thy)
extends
  sugar_cmds
doc
  'Makes all theories that are visible in Thy visible in the',nl,
  'current theory.  Does this by sharing the visible_theory/1',nl,
  'facts of Thy, rather than copying them into the current theory.',nl,
  'Thy must be an ancestor of the current theory.'
mode
  inherit_all_visible_theories(@theory).

inherit_all_visible_theories(FromThy) :-
    check_current_theory(CD),
    check_strict_ancestor(FromThy),
    ( query_theory(inherited_visible_theories(FromThy), CD) ->
        msg(visible_theories, ['theory ',q(CD),' already inherits the',
                        ' visible_theory/1 facts of theory ',q(FromThy)])
    ;
        tactic_assert_theory(inherited_visible_theories(FromThy), CD)
    ),
    %% Now we must announce all the new visible theories, to
    %% allow the visible-name cache to be maintained etc.
    (
        query_theory(visible_theory(NewT), FromThy),
        announce_event(visible_theory(CD, NewT, add)),
        fail
    ;
        true
    ).


%% delete_visible_theory(Thy).
%%
%%  Deletes Thy from the set of visible theories.
%%
%%  mode delete_visible_theory(@atom).
%%
delete_visible_theory(Visible) :-
    check_current_theory(Thy),
    is_theory_name(Visible),
    tactic_retract_theory(visible_theory(Visible), Thy),
    announce_event(visible_theory(Thy, Visible, delete)).



%% deabbreviate(ETerm, ITerm).
%% deabbreviate(ETerm, Thy, ITerm).
%% deabbreviate(ETerm, Thy, Arity, ITerm).
%%
%%  ITerm is the internal form of ETerm, with all abbreviations expanded
%%  and theory names prefixed to all identifiers that do not already
%%  contain theory prefixes.  If an identifier has more than one possible
%%  interpretation, an error message is produced and deabbreviate fails.
%%
%%  If the Thy argument is not given, the abbreviations of the current
%%  theory are used.  If the Arity argument is not given, the arity of
%%  ETerm is assumed to be [].  All these procedures assume that the
%%  visible name cache is consistent with Thy.
%%
%%  mode deabbreviate(@term, -term).
%%  mode deabbreviate(@term, @theory, -term).
%%  mode deabbreviate(@term, @theory, @arity, -term).
%%
deabbreviate(In, Out) :-
    global_state_lookup(current_dir,Thy),
    deabbreviate(In, Thy, [], Out).

deabbreviate(In, Thy, Out) :-
    deabbreviate(In, Thy, [], Out).

deabbreviate(T0, Thy, Arity, T) :-
    handle_event(deabbreviate(T0, Thy, Arity, T)),
    !,
    %% yuk.  see deabbreviate_rulename.
    ip_lookup(deabbrev_error_flag, ErrorFlag),
    ( ErrorFlag = yes ->
        check_term(T)
    ;
        true
    ).
deabbreviate(T0, Thy, Arity, T) :-
    abbrev_thylist(Thy, ThyList),
    deabbreviate2(T0, ThyList, Arity, T).

deabbreviate2(T0, Thy, Arity, T) :-
    sub(T0),
    !,
    substitute(T0, Subs0, A0),
    deabbrev_subslists(Subs0, Thy, Subs),
    deabbreviate2(A0, Thy, Arity, A),
    substitute(T, Subs, A).
deabbreviate2(T0, _, _, T) :-
    var(T0),
    !,
    T = T0.
deabbreviate2(T0, _, _, T) :-
    obvar(T0),
    !,
    T = T0.
deabbreviate2('...', _, _, _) :-
    %% '...' represents an elided subterm, so we replace each
    %% such elided subterm by a fresh variable.
    %% Note that abbreviate/N is careful to only elide subterms 
    %% that can sensibly be replaced by variables.
    !.
deabbreviate2(T0, Thy, Arity, T) :-
    atomic(T0),
    !,
    (
        is_internal_name(T0),
        T = T0
    ;
        Arity == [],
        %% try to deabbreviate T0.
        query_abbrev_in_theories(Thy,T0,T,Kind,_),
        abbrev_input_test(Kind)
    ;
        list(Arity),
        ip_lookup(deabbrev_error_flag, ErrorFlag),
        ErrorFlag == no,
        \+ (atom(T0), in_visible_cache(T0, Arity, _)),
        T = T0        
    ;
        list(Arity),
        %% try to prefix T0.
        prefix_name(T0, Arity, T)
    ;
        \+ list(Arity),
        handle_error(['unknown name ''',w(T0),''' of arity ',q(Arity),'.'])
    ),
    !.
%% PJR - fix for terms like X(Y)
deabbreviate2(T0, Thy, _, T) :-
    compound(T0),
    functor(T0, F0, N),
    var(F0),
    !,
    functor(T, F0, N),
    deabbrev_args(N, T0, Thy, T).
deabbreviate2(T0, _, _, T) :-
    is_param_const(T0),
    !,
    param_const(T0, Name0, Value),
    ( is_internal_name(Name0) ->
        Name = Name0
    ;
        prefix_name(Name0, when(_), Name)
    ),
    param_const(T, Name, Value).
deabbreviate2(@Spec, _Thy, Arity, T) :-
    Arity == [],
    !,
    deabbrev_term_spec(Spec, T).
deabbreviate2(T0, Thy, Arity, T) :-
    compound(T0),
    %% and T0 is not a param_const or @term_spec.
    !,
    functor(T0, F0, N),
    functor(T1, F1, N),
    deabbrev_args(N, T0, Thy, T1),
    (
        is_internal_name(F0),
        F1 = F0,
        T = T1
    ;
        Arity == [],
        %% look for abbreviation matching F0(NewArgs).
        F1 = F0,
        query_abbrev_in_theories(Thy,T1,T,Kind,_),
        abbrev_input_test(Kind)
    ;
        list(Arity),
        deabbreviate2(F0, Thy, [N|Arity], F1),
        T = T1
    ),
    !.
deabbreviate2(T0, Thy, Arity, T) :-
    quant(T0),
    !,
    %% This follows a similar structure to the compound case above.
    quantify(T0, Q0, V0, B0),
    deabbreviate2(B0, Thy, Arity, B1),
    deabbrev_bvars(V0, Thy, V1),
    quantify(T1, Q1, V1, B1),
    (
        is_internal_name(Q0),
        Q1 = Q0,
        T = T1
    ;
        Arity == [],
        %% look for abbreviation matching quant(Q0,V1,B1).
        Q1 = Q0,
        query_abbrev_in_theories(Thy,T1,T,Kind,_),
        abbrev_input_test(Kind)
    ;
        list(Arity),
        bound_vars_to_arity(V1,QArity),
        deabbreviate2(Q0, Thy, [quant(QArity)|Arity], Q1),
        T = T1
    ),
    !.
deabbreviate2(T0, _, Arity, _) :-
    handle_error(['unknown term ',q(T0),' of arity ',q(Arity)]).


%% private deabbrev_args(NumArgs, In, Thy, Out).
%%
%%  Transforms the arguments of In to give Out.
%%  Assumes that all arguments have arity [] (ie. are terms).
%%
%%  mode deabbrev_args(@integer, @compound, @theory, +compound).
%%
deabbrev_args(0, _, _, _) :-
    !.
deabbrev_args(N, T0, Thy, T) :-
    N > 0,
    arg(N, T0, A0),
    arg(N, T, A),
    deabbreviate2(A0, Thy, [], A),
    N1 is N - 1,
    deabbrev_args(N1, T0, Thy, T).


%% private deabbrev_subslists(In, Thy, Out).
%% private deabbrev_subs(In, Thy, Out).
%%
%%  Transform the range of all substitutions within In to give Out.
%%  Assumes that all such terms have arity [].
%%
%%  mode deabbrev_subslists(@list(list(compound)),@atom,?list(list(compound))).
%%  mode deabbrev_subslists(@list(compound),@atom,?list(compound)).
%%
deabbrev_subslists([], Thy, []).
deabbrev_subslists([S0|S0s], Thy, [S|Ss]) :-
    deabbrev_subs(S0, Thy, S),
    deabbrev_subslists(S0s, Thy, Ss).

deabbrev_subs([], Thy, []).
deabbrev_subs([T0/V|S0s], Thy, [T/V|Ss]) :-
    deabbreviate2(T0, Thy, [], T),
    deabbrev_subs(S0s, Thy, Ss).


%% private deabbrev_bvars(In, Thy, Out).
%%
%%  Transform the types of a list of bound variables.
%%  It assumes that all types have arity [].
%%
%%  mode deabbrev_bvars(+list(term), @theory, ?list(term)).
%%
deabbrev_bvars(V, _, O) :- var(V), !, O = V.
deabbrev_bvars([], _, []).
deabbrev_bvars([B0|B0s], Thy, [B|Bs]) :-
    ( obvar(B0) ->
	B = B0
    ;
	B0 = (V:T0),
	B = (V:T),
	deabbreviate2(T0, Thy, [], T)
    ),
    deabbrev_bvars(B0s, Thy, Bs).




%% deabbreviate_relation(R0,R).
%%
%%  Passes the binary function R0 through deabbreviate/2, to give R.
%%
%%  mode deabbreviate_relation(@atom, ?atom).
%%
deabbreviate_relation(rev(R0),R) :-
    !,
    deabbreviate_relation(R0,R1),
    R = rev(R1).
deabbreviate_relation(R0,R) :-
    check_atom(R0),
    functor(T0,R0,2),
    deabbreviate(T0,T),
    T =.. [R,Arg1,Arg2],
    var(Arg1),
    var(Arg2).


%% deabbreviate_list(F0,F).
%%
%%  Maps deabbreviate/2 over all the terms in F0, giving F.
%%
%%  mode deabbreviate_list(@list(term), ?list(term)).
%%
deabbreviate_list([],[]).
deabbreviate_list([T0|TS0], [T|TS]) :-
    deabbreviate(T0,T),
    deabbreviate_list(TS0,TS).


%% deabbreviate_relation_map(F0,F).
%%
%%  Maps deabbreviate_relation/2 over all the relation pairs 
%%  (R1--->R2) in F0, giving F.
%%
%%  mode deabbreviate_relation_map(@list(compound), ?list(compound)).
%%
deabbreviate_relation_map([],[]).
deabbreviate_relation_map(['--->'(R0,S0)|RS0],['--->'(R,S)|RS]) :-
    deabbreviate_relation(R0,R),
    deabbreviate_relation(S0,S),
    deabbreviate_relation_map(RS0,RS).


%% deabbreviate_term_map(F0,F).
%%
%%  Maps deabbreviate/2 over all the term pairs (T1--->T2) in F0, giving F.
%%
%%  mode deabbreviate_term_map(@list(compound), ?list(compound)).
%%
deabbreviate_term_map(id,id).
deabbreviate_term_map([],[]).
deabbreviate_term_map([id|RS0],[id|RS]) :-
    !,
    deabbreviate_term_map(RS0,RS).
deabbreviate_term_map(['--->'(R0,S0)|RS0],['--->'(R,S)|RS]) :-
    deabbreviate(R0,R),
    deabbreviate(S0,S),
    deabbreviate_term_map(RS0,RS).


%% deabbreviate_interpretation(InstIn, AbThy, InstOut).
%%
%%  Deabbreviate the LHS and RHS of each maplet in InstIn to give InstOut.
%%  AbThy is the name of the abstract theory and is added as a prefix
%%  to any LHS functor/constant/quantifier that does not already have
%%  an explicit theory prefix.
%%
%%  mode deabbreviate_interpretation(@list(term), +theory, ?list(term)).
%%
deabbreviate_interpretation([], _, []).
deabbreviate_interpretation(['--->'(LIn,RIn)|In], AbThy,
                            ['--->'(LOut,ROut)|Out]) :-
    LIn = param_const(Name,Val),
    !,
    ( internal_name(Name, _, _) ->
        LOut = param_const(Name,Val)
    ;
        internal_name(IntName, AbThy, Name),
        LOut = param_const(IntName,Val)
    ),
    deabbreviate(RIn, ROut),
    deabbreviate_interpretation(In, AbThy, Out).
deabbreviate_interpretation(['--->'(LIn,RIn)|In], AbThy,
                            ['--->'(LOut,ROut)|Out]) :-
    msg(deabbrev_interpretation, 3, ['deabbreviating ',q(--->(LIn,RIn))]),
    defn_lhs(LIn, Name0, Arity, A1, A2, A3),
    ( internal_name(Name0, Thy, _) ->
        Name = Name0,
        LOut = LIn
    ;
        internal_name(Name, AbThy, Name0),
        defn_lhs(LOut, Name, Arity, A1, A2, A3),
        Thy = AbThy
    ),
    deabbreviate(RIn, ROut),
    deabbreviate_interpretation(In, AbThy, Out).


%% deabbreviate_instantiation(InstIn, AbThy, InstOut).
%%
%%  Deabbreviate the LHS of each maplet in InstIn to give InstOut.
%%  AbThy is the name of the abstract theory and is added as a prefix
%%  to any LHS functor/constant/quantifier that does not already have
%%  an explicit theory prefix.
%%
%%  mode deabbreviate_instantiation(@list(term), +theory, ?list(term)).
%%
deabbreviate_instantiation([], _, []).
deabbreviate_instantiation(['--->'(LIn,RIn)|In], AbThy,
                            ['--->'(LOut,ROut)|Out]) :-
    LIn = param_const(Name,Val),
    !,
    ( internal_name(Name, _, _) ->
        LOut = param_const(Name,Val)
    ;
        internal_name(IntName, AbThy, Name),
        LOut = param_const(IntName,Val)
    ),
    deabbreviate(RIn, ROut),
    deabbreviate_instantiation(In, AbThy, Out).
deabbreviate_instantiation(['--->'(LIn,RIn)|In], AbThy,
                            ['--->'(LOut,ROut)|Out]) :-
    defn_lhs(LIn, Name0, Arity, A1, A2, A3),
    ( internal_name(Name0, Thy, _) ->
        LOut = LIn
    ;
        internal_name(Name, AbThy, Name0),
        defn_lhs(LOut, Name, Arity, A1, A2, A3),
        Thy = AbThy
    ),
    deabbreviate(RIn, ROut),
    deabbreviate_instantiation(In, AbThy, Out).




%% deabbreviate_cmd(Modes, Cmd0, Cmd).
%%
%%  Constructs Cmd from Cmd0 by deabbreviating all arguments
%%  that need deabbreviating (that is, contain theory terms).
%%  It uses Modes to determine which arguments contain theory terms.
%%  If Modes is a list of possible mode specifications, it chooses
%%  the first mode specification in that list that seems to match
%%  the actual arguments of Cmd0 (see first_matching_mode/3).
%%
%%  mode deabbreviate_cmd(@mode, @cmd, ?cmd).
%%
deabbreviate_cmd(Modes, Cmd0, Cmd) :-
    list(Modes),
    !,
    functor(Cmd0,F,A),
    functor(Cmd,F,A),
    first_matching_mode(Modes, Cmd0, M1),
    deabbrev_cmd_args(A,M1,Cmd0,Cmd).
deabbreviate_cmd(Mode, Cmd0, Cmd) :-
    %% \+ list(Mode)
    functor(Mode,F,A),
    functor(Cmd0,F,A),
    functor(Cmd,F,A),
    deabbrev_cmd_args(A,Mode,Cmd0,Cmd).
    
deabbrev_cmd_args(0, _, _, _) :-
    !.
deabbrev_cmd_args(N, Mode, Cmd0, Cmd) :-
    N > 0,
    arg(N, Mode, M),
    arg(N, Cmd0, C0),
    arg(N, Cmd,  C),
    mode_io_type(M, _, MT),
    deabbreviate_with_type(MT, C0, C),
    %% could check actual argument types here too.
    N1 is N - 1,
    deabbrev_cmd_args(N1, Mode, Cmd0, Cmd).


%% deabbreviate_cmd_relaxed(Modes, Cmd0, Cmd).
%%
%%  Same as deabbreviate_cmd except that type checking with respect 
%%  to mode declarations is relaxed so that all parameters that as 
%%  variables satisfies_mode (see first_matching_mode/4).
%%
%%  mode deabbreviate_cmd_relaxed(@mode, @cmd, ?cmd).
%%
deabbreviate_cmd_relaxed(Modes, Cmd0, Cmd) :-
    list(Modes),
    !,
    functor(Cmd0,F,A),
    functor(Cmd,F,A),
    first_matching_mode(Modes, Cmd0, M1, relax),
    deabbrev_cmd_args(A,M1,Cmd0,Cmd).
deabbreviate_cmd_relaxed(Mode, Cmd0, Cmd) :-
    %% \+ list(Mode)
    functor(Mode,F,A),
    functor(Cmd0,F,A),
    functor(Cmd,F,A),
    deabbrev_cmd_args(A,Mode,Cmd0,Cmd).
    
%% deabbreviate_with_type(Type, Term0, Term).
%%
%%  Type describes what kind of value Term0 represents.
%%  Type typically comes from a mode declaration of some procedure.
%%
%%  mode deabbreviate_with_type(@ground, @term, ?term).
%%
deabbreviate_with_type(Type, C0, C) :-
    handle_event(deabbreviate_with_type(Type, C0, C)),
    !.
deabbreviate_with_type(rulename, C0, C) :-
    !,
    deabbreviate_rulename(C0, C).
deabbreviate_with_type(hypname, C0, C) :-
    !,
    deabbreviate_rulename(C0, C).
deabbreviate_with_type(thy(_), C0, C) :-
    !,
    deabbreviate(C0, C).
deabbreviate_with_type(relation, C0, C) :-
    !,
    deabbreviate_relation(C0, C).
deabbreviate_with_type(query, C0, C) :-
    !,
    deabbreviate_query(C0, C).
deabbreviate_with_type(list(T), C0, C) :-
    !,
    deabbrev_list_of(C0, T, C).
deabbreviate_with_type(term_spec, C0, C) :-
    !,
    deabbrev_term_spec(C0, C).
deabbreviate_with_type(_, C, C).


deabbrev_list_of(L0, _, L) :-
    var(L0),
    !,
    L = L0.
deabbrev_list_of([], _, []).
deabbrev_list_of([A0|L0], T, [A|L]) :-
    deabbreviate_with_type(T, A0, A),
    deabbrev_list_of(L0, T, L).

deabbrev_term_spec(Term0, Term) :-
    ( var(Term0) ->
        Term = Term0
    ; Term0 == focus ->
        is_focus(Term)
    ; Term0 == goal ->
        is_goal(Term),
        Term \== unknown
    ; Term0 = Hyp:Name, Hyp == hyp ->
        deabbreviate_with_type(hypname, Hyp:Name, HypSpec),
        is_hyp(HypSpec, Term)
    ; Term0 = Hyp(Name), Hyp == hyp ->
        deabbreviate_with_type(hypname, Hyp:Name, HypSpec),
        is_hyp(HypSpec, Term)
    ; Term0 = Focus(Path), Focus == focus ->
        is_focus(F),
        subterm_path(F, Path, Term)
    ; Term0 = Goal(Path), Goal == goal ->
        is_goal(G),
        G \== unknown,
        subterm_path(G, Path, Term)
    ; Term0 = Hyp(Name, Path), Hyp == hyp ->
        deabbreviate_with_type(hypname, Hyp:Name, HypSpec),
        is_hyp(HypSpec, H),
        subterm_path(H, Path, Term)
    ;
        deabbreviate_with_type(thy(term), Term0, Term)
    ).


%% deabbreviate_rulename(RuleName0, RuleName).
%%
%%  Deabbreviates any theory terms that may appear in RuleName0,
%%  (eg. if RuleName0 = hyp:Term or an Ergo 5 rulename with parameters).
%%
%%  mode deabbreviate_rulename(@rulename, ?rulename).
%%
deabbreviate_rulename(C0, C) :-
    compound(C0),
    C0 =.. [':', Hyp, Patt0],
    Hyp == hyp,
    \+ integer(Patt0),
    !,
    deabbreviate_with_type(thy(term), Patt0, Patt),
    C = hyp:Patt.
deabbreviate_rulename(C0, C) :-
    compound(C0),
    C0 =.. [':', Thy, Rule],
    atom(Thy), %% a supplied theory
    !,
    deabbreviate_rulename(Rule, DRule),
    C = (Thy:DRule).
deabbreviate_rulename(C0, C) :-
    compound(C0),
    functor(C0, Fun, N),
    Fun \== (:),
    !,
    functor(C, Fun, N),
    rule_query(_Thy, Fun, FullName,_,_,_,_),
    mode_from_rule(FullName, C0, Mode),
    deabbrev_rule_args(N, Mode, C0, C).
deabbreviate_rulename(C0, C0).

deabbrev_rule_args(0, _, _, _) :-
    !.
deabbrev_rule_args(N, Mode, Cmd0, Cmd) :-
    N > 0,
    arg(N, Mode, M),
    arg(N, Cmd0, C0),
    arg(N, Cmd,  C),
    deabbreviate_with_type(M, C0, C),
    %% could check actual argument types here too.
    N1 is N - 1,
    deabbrev_rule_args(N1, Mode, Cmd0, Cmd).

%% deabbreviate_query(Query0, Query).
%%
%%  Deabbreviates any theory terms that appear within Query0,
%%  which must be a legal use/fwd/rule query.
%%
%%  mode deabbreviate_query(?query, -query).
%%

%% TO DO: Move deabbreviate_query to handler in usefwd.
deabbreviate_query([], Q) :-
    !,
    Q = [].
deabbreviate_query(C, Q) :-
    is_chars(C),
    !,
    Q = C.
deabbreviate_query([C0|CS0], Q) :-
    !,
    deabbrev_query_constraint(C0, C),
    Q = [C|CS],
    deabbreviate_query(CS0, CS).
deabbreviate_query(and(A0,B0), Q) :-
    !,
    deabbreviate_query(A0, A),
    deabbreviate_query(B0, B),
    Q = and(A,B).
deabbreviate_query(C0, C) :-
    deabbrev_query_constraint(C0, C).


deabbrev_query_constraint(C0, C) :-
    atom(C0),
    !,
    C = C0.
deabbrev_query_constraint(C0, C) :-
    C0 = F(Arg0),
    constraint_arg1_type(F, Type),
    !,
    deabbreviate_with_type(Type, Arg0, Arg),
    C = F(Arg).
deabbrev_query_constraint(C0, C) :-
    C0 = F(Arg0, Arg1),
    constraint_arg2_type(F, Type),
    !,
    deabbreviate_with_type(Type, Arg1, Arg),
    C = F(Arg0, Arg).
deabbrev_query_constraint(Thys:Name0, C) :-
    !,
    deabbreviate_rulename(Thys:Name0, C).
deabbrev_query_constraint(c(Rules0), C) :-
    list(Rules0),
    !,
    deabbreviate_with_type(list(rulename), Rules0, Rules),
    C = c(Rules).
deabbrev_query_constraint(c(Rule0), C) :-
    !,
    deabbreviate_rulename(Rule0, Rule),
    C = c(Rule).
deabbrev_query_constraint(conjs(Terms0), C) :-
    !,
    deabbreviate_with_type(list(thy(term)), Terms0, Terms),
    C = conjs(Terms).
deabbrev_query_constraint(C, C).


constraint_arg1_type(i,         rulename).
constraint_arg1_type(o,        rulename).
constraint_arg1_type(concl,        term_spec).
constraint_arg1_type(prem,        term_spec).
constraint_arg1_type(subterm,      term_spec).
%% For backwards compatibility.
constraint_arg1_type(input_name,   rulename).
constraint_arg1_type(input,        rulename).
constraint_arg1_type(output,       rulename).

constraint_arg2_type(context,        term_spec).


%% private prefix_name(In, Arity, Out).
%%
%%  Constructs Out from In by adding a theory prefix.
%%  In must not already have a theory prefix.
%%  Raises an error and fails if In is unknown or ambiguous.
%%
%%  Arity may be when(_) or a normal arity list.
%%
%%  mode prefix_name(@atom, @arity, ?atom).
%%
prefix_name(In, Arity, Out) :-
    atom(In),
    in_visible_cache(In, Arity, Result),
    !,
    (Result = ambiguous(L) ->
        arity_description(Arity, Desc),
        handle_error(['the ',w(Desc),' ''',w(In),''' of arity ',
                      q(Arity),' is ambiguous.',nl,
                      'Possible meanings include: ',q(L)])
    ;
        Out = Result
    ).
prefix_name(In, Arity, _) :-
    arity_description(Arity, Desc),
    handle_error(['unknown ',Desc,' ''',w(In),''' of arity ',q(Arity),'.']).



%% private in_visible_cache(Short, Arity, Meaning).
%%
%%  Short must be an atom without a theory prefix.
%%  Meaning will be 'ambiguous' if Short has multiple meanings, given
%%  the current set of visible theories, or the fully prefixed
%%  name corresponding to Short if there is exactly one visible
%%  theory that defines Short with the given Arity.
%%  Fails silently if Short and Arity do not correspond to any
%%  known symbols.  Arity may be when(_) or a normal arity list.
%%
%%  If it is successful, it also updates the visible_cache/3 entries
%%  to include this entry.
%%
%%  mode in_visible_cache(@atom, @arity, ?atom).
%%
in_visible_cache(Short, Arity, Meaning) :-
    visible_cache(Short, Arity, Result),
    !,
    Meaning = Result.
in_visible_cache(Short, Arity, Meaning) :-
    global_state_lookup(current_dir,CD),
    findall(IName,
            abbrev_internal_name(Arity,CD,Short,IName),
            [First|Others]),
    ( Others = [] ->
        Meaning = First
    ;
        Meaning = ambiguous([First|Others])
    ),
    asserta(visible_cache(Short, Arity, Meaning)).


%% private abbrev_internal_name(Arity,Thy,ShortName,FullName).
%%
%%  FullName is 'VisThy.ShortName', where VisThy is a theory that
%%  is visible from theory Thy.  Arity may be a when/1 term, which
%%  means that FullName is the name of a parameterized constant,
%%  otherwise Arity must be a normal arity list, which means that
%%  FullName is the name of a normal constant/function/quantifier.
%%
%%  mode abbrev_internal_name(@arity, @theory, @atom, ?atom).
%%
abbrev_internal_name(when(_),CD,Short,IName) :-
    query_visible_theory(Thy, CD),
    internal_name(IName, Thy, Short),
    query_theory(symbol(param_const(IName,_),_,_), Thy).
abbrev_internal_name(Arity,CD,Short,IName) :-
    list(Arity),
    query_visible_theory(Thy, CD),
    internal_name(IName, Thy, Short),
    query_theory(symbol(IName,Arity,_), Thy).



%% abbreviate(In, Out).
%% abbreviate(In, Thy, Out).
%% abbreviate(In, Thy, Arity, Out).
%% abbreviate(In, Thy, Arity, Depth, Out).
%%
%%  Out is the external form of In, with all subterms that
%%  match abbreviations replaced by those abbreviations and
%%  explicit theory prefixes within names removed where it
%%  does not result in ambiguity.  It also simplifies substitutions
%%  as much as possible (without generating new bound variable names)
%%  and elides subterms below Depth levels (Depth=0 means no elision).
%%
%%  If Thy is not specified, the abbreviations of the current theory
%%  are used.  If Arity is not specified, the arity of In is assumed
%%  to be [].  If a Depth argument is not given, Depth=0 is used.
%%  All these procedures assume that the visible name cache
%%  is consistent with Thy.  (For the moment, that means that Thy
%%  *must* be the current theory).
%%
%%  mode abbreviate(@term, -term).
%%  mode abbreviate(@term, @theory, -term).
%%  mode abbreviate(@term, @theory, @arity, -term).
%%  mode abbreviate(@term, @theory, @arity, @integer, -term).
%%
abbreviate(In, Out) :-
    abbreviate(In, _Thy, [], 0, Out).

abbreviate(In, Thy, Out) :-
    abbreviate(In, Thy, [], 0, Out).

abbreviate(In, Thy, Arity, Out) :-
    abbreviate(In, Thy, Arity, 0, Out).

abbreviate(In, Thy, Arity, Depth, Out) :-
    handle_event(abbreviate(In, Thy, Arity, Depth, Out)),
    !.
abbreviate(In, Thy, Arity, Depth, Out) :-
    global_state_lookup(current_dir,Thy),
    list(Arity),
    integer(Depth),
    !,
    (Depth > 0 ->
        Depth1 = Depth
    ;
        Depth1 = 999999
    ),
    collect_thawvars(In, Unfrozen),
    freeze_list(Unfrozen),
    abbrev_thylist(Thy, ThyList),
%% PJR - simplify the term up front WRT substitutions
    (abbrev_subs_raw ->
        InS = In
    ;
        simplify_term(In, InS)
    ),
    abbreviate2(InS, ThyList, Arity, Depth1, Out),
    thaw_list(Unfrozen).
abbreviate(_, _, Arity, _, _) :-
    \+ list(Arity),
    !,
    handle_error(['invalid arity argument to abbreviate/5: ',q(Arity)]).
abbreviate(_, _, _, Depth, _) :-
    \+ integer(Depth),
    !,
    handle_error(['invalid depth argument to abbreviate/5: ',q(Depth)]).
abbreviate(_, Thy, _, _, _) :-
    !,
    handle_error(['abbreviate/5 must use current theory, not: ',q(Thy)]).


simplify_substitutions parameter
  [set_simplify_substitutions, get_simplify_substitutions]
extends
  parameters
doc
  'The simplify_substitutions parameter determines how substitutions',nl,
  'are displayed when theory terms are printed.',nl,
  'If it is set to false, then all substitutions are displayed exactly',nl,
  'where they appear in the term that is being output.',nl,
  'If it is set to true, then substitutions are pushed into the term',nl,
  'and evaluated as much as possible.  This usually results in more',nl,
  'readable output.',nl,
  'The false setting is mostly used for analysing difficult problems',nl,
  'involving substitutions.'.

?- dynamic(abbrev_subs_raw/0).

set_simplify_substitutions(false) :-
    ( abbrev_subs_raw ->
        true
    ;
        asserta(abbrev_subs_raw)
    ).
set_simplify_substitutions(true) :-
    retractall(abbrev_subs_raw).

get_simplify_substitutions(F) :-
    ( abbrev_subs_raw ->
        F = false
    ;
        F = true
    ).


abbreviate2(T0, Thy, Arity, Depth, T) :-
    sub(T0),
    abbrev_subs_raw,
    !,
    substitute(T0, Subs0, A0),
    Depth1 is Depth - 1,
    Depth2 is Depth - 2,
    abbrev_subslists(Subs0, Thy, Depth2, Subs),
    abbreviate2(A0, Thy, Arity, Depth1, A),
    substitute(T, Subs, A).
abbreviate2(T0, _, _, _, T) :-
    var(T0),
    \+ sub(T0),
    !,
    T = T0.
abbreviate2(T0, _, _, _, T) :-
    obvar(T0),
    \+ sub(T0),
    !,
    T = T0.
abbreviate2(T0, Thy, Arity, Depth, T) :-
    is_internal_name(T0),
    !,
    (
        Arity == [],
        %% try to abbreviate T0.
        query_abbrev_in_theories(Thy,T,T0,Kind,_),
        abbrev_output_test(Kind)
    ;
        list(Arity),
        %% try to deprefix T0.
        deprefix_name(T0, Arity, T)
    ;
        Depth > 0,
        T = T0
    ;
        T = '...'
    ),
    !.
abbreviate2(T0, _, _, _, T) :-
    atomic(T0),
    !,
    T = T0.
abbreviate2(T0, Thy, Arity, Depth, T) :-
    is_param_const(T0),
    !,
    (
        Arity == [],
        %% try to abbreviate T0.
        query_abbrev_in_theories(Thy,T,T0,Kind,_),
        abbrev_output_test(Kind)
    ;
        %% try to deprefix its name.
        Depth > 0,
        param_const(T0, Name0, Value),
        deprefix_name(Name0, when(_), Name),
        param_const(T, Name, Value)
    ;
        Depth > 0,
        T = T0
    ;
        T = '...'
    ),
    !.
abbreviate2(_, _, _, Depth, T) :-
    PRE(abbreviate2, integer(Depth)),
    Depth =< 0,
    !,
    T = '...'.
abbreviate2(T0, Thy, Arity, Depth, T) :-
    compound(T0),
    %% and T0 is not a param_const.
    !,
    Depth1 is Depth - 1,
    ( Arity == [],
      %% look for abbreviation matching T0.
      query_abbrev_in_theories(Thy,T1,T0,Kind,_),
      abbrev_output_test(Kind) ->
        functor(T1, F, N),
        functor(T, F, N),
        abbrev_args(N, T1, Thy, Depth1, T)
    ; list(Arity) ->
        functor(T0, F0, N),
        abbreviate2(F0, Thy, [N|Arity], Depth1, F1),
        functor(T, F1, N),
        abbrev_args(N, T0, Thy, Depth1, T)
    ;
        functor(T0, F0, N),
        functor(T, F0, N),
        abbrev_args(N, T0, Thy, Depth1, T)
    ).
abbreviate2(T0, Thy, Arity, Depth, T) :-
    quant(T0),
    %% quantifier abbreviations are not allowed on output (yet!), so
    %% we just abbreviate the components of the quantifier, taking great
    %% care to evaluate (and push in) substitutions where possible.
    !,
    substitute(T0, Subs0, T1),  %% Subs0 may be empty.
    quantify(T1, Q0, V0, B0),
    % PJR - simplification already done
    %abbrev_simplify_subs(Subs0, Q0, V0, Subs1),
    ( Subs0 = [] ->
        Q1 = Q0,
        V1 = V0,
        B1 = B0,
        Subs = Subs0,
        QDepth = Depth
    ; bound_vars_nfi_subs(V0, Subs0, V1) ->
        %% We can push Subs1 inside the quantifier, thus transforming
        %% Subs0(!!Q0 V0 B0) into (!!(Subs0 Q0) V1 (Subs0 B0)),
        %% (V1 is V0 with Subs applied to each type).
        %% This simplification is possible because all the bound
        %% variables in V1 are not_free_in Subs1.
        substitute(QS, Subs0, Q0),
        substitute(BS, Subs0, B0),
        ( abbrev_subs_raw ->
            Q1 = QS,
            B1 = BS
        ;
	    simplify_term(QS, Q1),
	    simplify_term(BS, B1)
        ),
        Subs = [],
        QDepth = Depth
    ;
        Q1 = Q0,
        V1 = V0,
        B1 = B0,
        SubsDepth is Depth - 2,
        abbrev_subslists(Subs0, Thy, SubsDepth, Subs),
        QDepth is Depth - 1
    ),
    %% After doing this, we abbreviate Q1, B1 and the types in V1.
    Depth1 is QDepth - 1,
    Depth2 is QDepth - 2,
    abbreviate2(B1, Thy, Arity, Depth1, B2),
    abbrev_bvars(V1, Thy, Depth2, V2),
    ( list(Arity) ->
        bound_vars_to_arity(V2,QArity),
        abbreviate2(Q1, Thy, [quant(QArity)|Arity], Depth1, Q2)
    ;
        Q2 = Q1
    ),
    quantify(T2, Q2, V2, B2),
    substitute(T, Subs, T2).
abbreviate2(T0, Thy, Arity, Depth, T) :-
    sub(T0),
    %% \+ quant(T0).
    !,
    substitute(T0, Subs0, A0),
    Depth1 is Depth - 1,
    Depth2 is Depth - 2,
    abbrev_subslists(Subs0, Thy, Depth2, Subs),
    abbreviate2(A0, Thy, Arity, Depth1, A),
    substitute(T, Subs, A).
abbreviate2(T0, _, Arity, _, _) :-
    handle_error(program,['abbreviating ',q(T0),' of arity ',q(Arity)]).



%% private abbrev_simplify_subs(Subs0, Quant, BoundVars, Subs).
%%
%%  Subs is an equivalent, but simplified, substitution to Subs0,
%%  under the assumption that the substitutions are to be applied
%%  to !!Quant BoundVars _.
%%
%%  WARNING: The implementation of this relies on the fact that
%%  simplify_term/3 never pushes substitutions inside quantified terms.
%%
abbrev_simplify_subs([], _, _, Subs) :-
    !,
    Subs = [].
abbrev_simplify_subs(Subs0, Q0, V0, Subs) :-
    quantify(Skel, Q0, V0, _),
    substitute(SubSkel, Subs0, Skel),
    simplify_term(SubSkel, SubSkel2, true),
    !,
    substitution(SubSkel2, Subs).
abbrev_simplify_subs(Subs0, _, _, Subs0).


%% private bound_vars_nfi_subs(V0, Subs, V1).
%%
%%  Succeeds iff (forall x in dom V0 . x is_not_free_in Subs).
%%  V1 is the same as V0, but with Subs applied to each type.
%%
bound_vars_nfi_subs(BV, _, _) :-
    var(BV),
    !,
    fail.
bound_vars_nfi_subs([], _, Out) :-
    !,
    Out = [].
bound_vars_nfi_subs([V|B0s], Subs, Out) :-
    obvar(V),
    !,
    V is_not_free_in Subs,
    bound_vars_nfi_subs(B0s, Subs, Bs),
    Out = [V|Bs].
bound_vars_nfi_subs([V:T0|B0s], Subs, Out) :-
    V is_not_free_in Subs,
    substitute(T, Subs, T0),
    bound_vars_nfi_subs(B0s, Subs, Bs),
    Out = [V:T|Bs].


%% private abbrev_args(NumArgs, In, Thy, Depth, Out).
%%
%%  Transforms the arguments of In to give Out.
%%  Assumes that all arguments have arity [] (ie. are terms).
%%
%%  mode abbrev_args(@integer, @compound, @theory, @integer, +compound).
%%
abbrev_args(0, _, _, _, _) :-
    !.
abbrev_args(N, T0, Thy, Depth, T) :-
    N > 0,
    arg(N, T0, A0),
    arg(N, T, A),
    abbreviate2(A0, Thy, [], Depth, A),
    N1 is N - 1,
    abbrev_args(N1, T0, Thy, Depth, T).


%% private abbrev_subslists(In, Thy, Depth, Out).
%% private abbrev_subs(In, Thy, Depth, Out).
%%
%%  Transform the range of all substitutions within In to give Out.
%%  Assumes that all such terms have unknown arity.
%%
%%  mode abbrev_args(@integer, @compound, @integer, +compound).
%%
abbrev_subslists([], Thy, _, []).
abbrev_subslists([S0|S0s], Thy, Depth, [S|Ss]) :-
    abbrev_subs(S0, Thy, Depth, S),
    abbrev_subslists(S0s, Thy, Depth, Ss).

abbrev_subs([], Thy, _, []).
abbrev_subs([T0/V|S0s], Thy, Depth, [T/V|Ss]) :-
    abbreviate2(T0, Thy, [], Depth, T),
    abbrev_subs(S0s, Thy, Depth, Ss).


%% private abbrev_bvars(In, Thy, Depth, Out).
%%
%%  Transform the types of a list of bound variables.
%%  It assumes that all types have arity [].
%%
%%  mode abbrev_bvars(?list(term), @theory, @integer, -list(term)).
%%
abbrev_bvars(BV, _, _, Out) :-
    var(BV),
    !,
    Out = BV.
abbrev_bvars([], _, _, Out) :-
    !,
    Out = [].
abbrev_bvars([B0|B0s], Thy, Depth, Out) :-
    obvar(B0),
    !,
    abbrev_bvars(B0s, Thy, Depth, Bs),
    Out = [B0|Bs].
abbrev_bvars([V:T0|B0s], Thy, Depth, Out) :-
    abbreviate2(T0, Thy, [], Depth, T),
    abbrev_bvars(B0s, Thy, Depth, Bs),
    Out = [V:T|Bs].



%% private deprefix_name(In, Arity, Out).
%%
%%  Succeeds iff In is an internal name and Out is a corresponding
%%  non-ambiguous name for In with the theory prefix removed.
%%  Fails silently if the theory prefix of In can/should not be removed.
%%
%%  Arity may be when(_) or a normal arity list.
%%
%%  mode deprefix_name(@atom, @arity, ?atom).
%%
deprefix_name(In, Arity, Out) :-
    internal_name(In, _, Out),
    in_visible_cache(Out, Arity, In).


abbreviate_code(In, Out) :-
    atom(In),
    !,
    Out = In.
abbreviate_code(In, Out) :-
    compound(In),
    functor(In, F, _),
    atom(F),
    !,
    %% now we can unify against outermost functor of In.
    abbreviate_code2(In, Out).
abbreviate_code(In, In).


abbreviate_code2((I1,I2),(O1,O2)) :-
    !,
    abbreviate_code(I1,O1),
    abbreviate_code(I2,O2).
abbreviate_code2((I1;I2), (O1;O2)) :-
    !,
    abbreviate_code(I1,O1),
    abbreviate_code(I2,O2).
abbreviate_code2((I1->I2), (O1->O2)) :-
    !,
    abbreviate_code(I1,O1),
    abbreviate_code(I2,O2).
abbreviate_code2('\+'(I1), '\+'(O1)) :-
    !,
    abbreviate_code(I1,O1).
abbreviate_code2(not_free_in(Var,Expr0), not_free_in(Var,Expr)) :-
    !,
    abbreviate(Expr0, Expr).
abbreviate_code2(In, In).

general command
  deabbrev(T)
extends
  sugar_cmds
mode
  deabbrev(@thy(term))
doc
  'Display a term with all abbreviations unfolded.',nl,
  'This is a useful way of seeing the unambiguous internal',nl,
  'form of any term that is legal in the current theory.'.

proof command
  deabbrev_focus
extends
  sugar_cmds
mode
  deabbrev_focus
doc
  'Display the current focus with all abbreviations unfolded.',nl,
  'This is a useful way of seeing the unambiguous internal',nl,
  'form of the current focus.'.

deabbrev_focus :-
    is_focus(F),
    deabbrev(F).

deabbrev(T) :-
    check_current_theory(_),
    deabbreviate(T,T2),
    writelist(['Deabbreviated term is: ',q(T2),'.', nl]).


%% private query_abbrev(E,I,Body,OrigThy,FromThy).
%%
%%  Similar to query_theory(abbrev(E,I,Body,OrigThy), FromThy),
%%  but it also searches for inherited abbreviations.
%%
%%  mode query_abbrev(?term,?term,?term,?theory,@theory).
%%
query_abbrev(E,I,Body,OrigThy,FromThy) :-
    query_theory(abbrev(E,I,Body,OrigThy), FromThy).
query_abbrev(E,I,Body,OrigThy,FromThy) :-
    query_theory(inherited_abbrevs(InheritFrom), FromThy),
    query_theory(abbrev(E,I,Body,OrigThy), InheritFrom).


%% private query_abbrev_in_theories(Thys,E,I,Body,OrigThy).
%%
%%  Equivalent to member(T,Thys),query_theory(abbrev(E,I,Body,OrigThy), T).
%%
%%  mode query_abbrev_in_theories(@list(theory),?term,?term,?term,?theory).
%%
query_abbrev_in_theories([Thy|_],E,I,Body,OrigThy) :-
    query_theory(abbrev(E,I,Body,OrigThy), Thy).
query_abbrev_in_theories([_|Thys],E,I,Body,OrigThy) :-
    query_abbrev_in_theories(Thys,E,I,Body,OrigThy).


%% private abbrev_thylist(CurrThy, AbbrevThys).
%%
%%  AbbrevThys is the list of all theories that (may) contain
%%  abbreviation facts for CurrThy.
%%
%%  mode abbrev_thylist(@theory, ?list(theory)).
%%
abbrev_thylist(none, []) :- !.
abbrev_thylist(Thy, ThyList) :-
    query_theory(abbrev(_,_,_,_), Thy),
    !,
    ThyList = [Thy|Parents],
    findall(Parent, query_theory(inherited_abbrevs(Parent),Thy), Parents).
abbrev_thylist(Thy, ThyList) :-
    %% Thy does not contain any abbreviations!
    findall(Parent, query_theory(inherited_abbrevs(Parent),Thy), ThyList).


%% private query_visible_theory(Thy,FromThy).
%%
%%  Similar to query_theory(visible_theory(Thy), FromThy),
%%  but it also searches for inherited visible_theories.
%%
%%  mode query_visible_theory(?theory,@theory).
%%
query_visible_theory(Thy,FromThy) :-
    query_theory(visible_theory(Thy), FromThy).
query_visible_theory(Thy,FromThy) :-
    query_theory(inherited_visible_theories(InheritFrom), FromThy),
    query_theory(visible_theory(Thy), InheritFrom).


%% Abbreviate the representation of sequents as found in the root
%% node info structure (the rule premises).
%%

abbrev_seqs([], []).
abbrev_seqs([Seq|Seqs], [ASeq|ASeqs]) :-
    abbrev_seq(Seq, ASeq),
    abbrev_seqs(Seqs, ASeqs).

abbrev_seq(seq(_, [], R), R).
abbrev_seq(seq(_, Lists, R), '--->'(Hyps, R)) :-
    abbrev_seq_hyps_lists(Lists, Hyps).

abbrev_seq_hyps_lists([Name = H], Hyps) :-
    abbrev_seq_hyps(H, Name, Hyps).
abbrev_seq_hyps_lists([Name = H|Lists], '&&&'(Hyps1,Hyps)) :-
    abbrev_seq_hyps(H, Name, Hyps1),
    abbrev_seq_hyps_lists(Lists, Hyps).

abbrev_seq_hyps(use(Name,_), Name, Name).
abbrev_seq_hyps(H, Name, Hyp) :-
    H = add(_,_,_),
    !,
    abbrev_seq_add(H, Name, [], Hyp).
abbrev_seq_hyps(H, Name, Hyp) :-
    H = delete(_,_,_),
    !,
    abbrev_seq_del(H, Name, [], Hyp).
abbrev_seq_hyps(filter(C,H), Name, '???'(Hyp, C)) :-
    abbrev_seq_hyps(H, Name, Hyp).

abbrev_seq_add(add(_, H, Rest), Name, SoFar, Hyp) :-
    !,
    abbrev_seq_add(Rest, Name, [H|SoFar], Hyp).
abbrev_seq_add(R, Name, SoFar, '+++'(Hyp, SoFar)) :-
    abbrev_seq_hyps(R, Name, Hyp).

abbrev_seq_del(delete(_, H, Rest), Name, SoFar, Hyp) :-
    !,
    abbrev_seq_del(Rest, Name, [H|SoFar], Hyp).
abbrev_seq_del(R, Name, SoFar, '---'(Hyp, SoFar)) :-
    abbrev_seq_hyps(R, Name, Hyp).

/*
abbrev_seq(seq(_, [], R), R).
abbrev_seq(seq(_, [hyp = H], R), '--->'(Hyps, R)) :-
    abbrev_seq_hyps(H, Hyps).

abbrev_seq_hyps(use(hyp,_), hyp).
abbrev_seq_hyps(H, Hyp) :-
    H = add(_,_,_),
    !,
    abbrev_seq_add(H, [], Hyp).
abbrev_seq_hyps(H, Hyp) :-
    H = delete(_,_,_),
    !,
    abbrev_seq_del(H, [], Hyp).
abbrev_seq_hyps(filter(C,H), '???'(Hyp, C)) :-
    abbrev_seq_hyps(H, Hyp).

abbrev_seq_add(add(_, H, Rest), SoFar, Hyp) :-
    !,
    abbrev_seq_add(Rest, [H|SoFar], Hyp).
abbrev_seq_add(R, SoFar, '+++'(Hyp, SoFar)) :-
    abbrev_seq_hyps(R, Hyp).

abbrev_seq_del(delete(_, H, Rest), SoFar, Hyp) :-
    !,
    abbrev_seq_del(Rest, [H|SoFar], Hyp).
abbrev_seq_del(R, SoFar, '---'(Hyp, SoFar)) :-
    abbrev_seq_hyps(R, Hyp).
*/

