%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: gumtree.ql,v 1.80 2006/02/06 00:52:12 itp Exp $

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% This file defines the 'gumtree' user interface for proofs.
%%
%% It is a command-line based interface for the Ergo 5
%% tree-structured prover.  The information displayed after each
%% command is configurable via the `gumtree_display' parameter.
%%
%% This user interface maintains a `history' list of all 
%% commands that have succeeded in contributing to the proof, in
%% the order that they were entered.  This history list corresponds
%% to the Prolog choicepoints that are available to the user.
%% For example, each undo command undoes the effect of the most
%% recent history entry.
%% There is also a commit(N) command for removing
%% alternatives from a sequence of history entries (similar to a cut
%% in Prolog) and an undo(N) command for undoing back to a given
%% point in the history.  Each entry on the history list also
%% becomes an annotation in the proof tree, so that a chronological
%% record of the user's exact input commands is stored with each proof.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

#include "../prover/lib/ergo.h"

%% setup_gumtree.
%%
%%  Initialises the database for this interface.
%%
%%  mode setup_gumtree.
%%
setup_gumtree :-
    %% Declare all events and saved facts.
    (command_table(gumtree, _) ->
        true
    ;
        %% declare session-specific things
        new_annotation(usercmd(integer,nodeid,list(nodeid),list(nodeid),chars),
                       usercmd_annotation),
        new_annotation(name(integer,atom),
                       name_annotation),
        new_command_table(gumtree, ['Commands specific to the gumtree UI.'])
    ),
    ( global_fact(gumtree_display,1) ->
        true
    ;
        %% declare persistent things
        new_global_fact(gumtree_display,1,[save(true)])
    ),
    %% discard usercmd and name annotations when rerunning
    add_handler(reannotate(_,usercmd(_,_,_,_,_)), true),
    add_handler(reannotate(_,name(_,_)), true),
    %% (de)abbreviation hooks
    require_global_tactic_file(abbrev),
    add_handler(deabbreviate_with_type(tactic,T0,T),
                gumtree_deabbrev_tactic_args(T0, T)),
    ( query_global(request(pre_prompt_cmd(Root, Nodes),_))
      ->
        true
      ;
        add_request(pre_prompt_cmd(Root, Nodes), ['Commands to run before gumtree prompt.'])
    ),
    require_global_tactic_file(show).

?- setup_gumtree.

pre_prompt_cmd(Root, Nodes) :-
    handle_event(pre_prompt_cmd(Root, Nodes)),!.
pre_prompt_cmd(_,_).

usercmd_annotation header
  'usercmd(CmdNum, Node, InNodes, OutNodes, Cmd)'
extends
  annotations
doc
  'When this annotation is attached to a proof node, it means that',nl,
  'the user successfully executed Cmd at that proof node.',
  para,
  'CmdNum is a positive integer that is incremented upon each successful',nl,
  'user command, so it provides a record of the chronological order',nl,
  'in which the commands were executed.',
  para,
  'Node, InNodes and OutNodes give historical information about which',nl,
  'proof nodes were affected by Cmd.  They contain proof node names',nl,
  'from the proof tree in which Cmd was originally executed.  After',nl,
  'a proof has been rerun, these node names will typically be unrelated',nl,
  'to the node names of the current proof tree.',nl,
  'InNodes is the list of nodes that was passed to Cmd and OutNodes is',nl,
  'the list of nodes that was returned by Cmd.  Node is the name of the',nl,
  'proof node to which this annotation was first attached, so it is',nl,
  'usually a member of InNodes.  It is included to provide a link from',nl,
  'the current proof tree to the node names of the original proof tree',nl,
  'of Cmd.',
  para,
  'Cmd is the exact character string that was entered by the user,',nl,
  'but with :::/2 commands stripped off, because they merely determine',nl,
  'which proof nodes the annotation is attached to.  Similarly,',nl,
  'commands that just move around the tree, or display information,',nl,
  'are not recorded as usercmd annotations.'.

%% TO DO: remove some of the following limitations of usercmd_annotation.
%%  (The idea is to enable an accurate proof script to be
%%  regenerated just from the usercmd annotations).
%% 1. if undo commands are used to force a tactic to return alternative
%%    solutions, this is not reflected very well in the usercmd annotations.
%%    (though it does record the correct input and output node lists).
%% 2. commands that add constraints or instantiate variables are
%%    not yet recorded as usercmd annotations?



gumtree_display parameter
  [set_gumtree_display, get_gumtree_display]
extends
  parameters
doc
  'This parameter specifies what should be displayed each time the',nl,
  'Gumtree proof interface prompts for a command.',
  para,
  'The parameter must be one of the following terms:',nl,
  begin(list),
  item(0),' silent     -- produces no output,',nl,
  item(1),' curr(Proc) -- calls Proc(CurrentUnprovedNodes),',nl,
  item(2),' open(Proc) -- calls Proc(AllUnprovedNodes),',nl,
  end(list),
  'Currently there is a restriction that Proc is either an atom, or the',nl,
  'term `application(P)''; later this will be relaxed to allow an',nl,
  'arbitrary higher-order goal.',nl,
  'The parameter value is saved in the global.erg file.'.

set_gumtree_display(Term) :-
    ground(Term),
    check_or_error(
        legal_gumtree_display_spec(Term),
        ['illegal gumtree display value: ',q(Term)]),
    retractall_global(gumtree_display(_)),
    assert_global(gumtree_display(Term)).

legal_gumtree_display_spec(silent).
legal_gumtree_display_spec(curr(P)) :- current_predicate(P/1).
%% It would be nice if current_predicate/1 did this itself.
legal_gumtree_display_spec(curr(application(P))) :- current_predicate(P/1).
legal_gumtree_display_spec(open(P)) :- current_predicate(P/1).
legal_gumtree_display_spec(open(application(P))) :- current_predicate(P/1).

get_gumtree_display(Proc) :-
    query_global(gumtree_display(Proc)).
    

database command
  prove(Pred)
doc
  'Starts a proof, using the Ergo 5 tree-structured prover and',nl,
  'the `gumtree'' user interface.  See ',
  xref('Proof Commands',global,gumtree_cmds),'.'
extends 
  theory_extension_cmds
mode
  prove(?thy(term)).

prove(Pred) :-
    check_current_theory(Thy),
    proof_start(Pred, true, Open),
    Open = [Root],
    writelist_to_chars(['prove(',thy(Pred),').',nl], PCmd),
    annotate(Root, usercmd(0,Root,[],Open,PCmd)),
    catch(gumtree0(Root, Thy, [ucmd(0,[],Open,start,PCmd)]),
          abort_proof(_),
          writelist(stdout, ['Exited Proof', nl])).

database command
  prove_rule(Name),
  prove_rule(Name, Concl)
doc
  'Starts a proof of a postulated rule, using the Ergo 5 tree-structured',nl,
  'prover and the `gumtree'' user interface.  See ',
  xref('Proof Commands',global,gumtree_cmds),'.',nl,
  'The two argument form unifies Concl with the conclusion of the',nl,
  'postulate.  This is typically used to choose meaningful names for',nl,
  'the proof variables.  It should not be used to instantiate the',nl,
  'postulate, since that would cause the proof to fail later.'
extends 
  theory_extension_cmds
mode
  prove_rule(+rulename),
  prove_rule(+rulename, ?thy(term)).

prove_rule(Name) :-
    prove_rule(Name, _).

prove_rule(Name, Concl) :-
    check_current_theory(Thy),
    proof_start_postulate(Name, Open),
    Open = [Root],
    proof_node_term(Root, Concl1),
    structural_unify(Concl, Concl1),
    writelist_to_chars([q(prove_rule(Name)),'.',nl], PCmd),
    annotate(Root, usercmd(0,Root,[],Open,PCmd)),
    catch(gumtree0(Root, Thy, [ucmd(0,[],Open,start,PCmd)]),
          abort_proof(_),
          writelist(stdout, ['Exited Proof', nl])).


%% gumtree0(RootNode, Theory, History).
%% gumtree1(CurrentNodes, History, Result).
%%
%%  gumtree0/3 is the initial entry point and should only be called
%%  at the beginning of the proof.
%%
%%  gumtree1/3 is the usual recursive entry point.
%%
%%  gumtree2/3 is only for internal use by gumtree/3.
%%
%%  WARNING: the implementation of these procedures and gumtree_result
%%           is quite subtle because they must maintain a precise
%%           relationship between the Prolog choicepoint stack and
%%           the explicit History stack, while implementing the
%%           equivalent of a Prolog `ancestor cut'.
%%
gumtree0(Root, Thy, History) :-
    ip_set(gumtree_root_node, Root),
    announce_event(proof(Thy, the_only_proof, start([Root]))),
    gumtree_premises_display,
    gumtree1([Root], History, _).
        
gumtree1(Cur, History, Result) :-
    (  get_parameter(elves_type, ET),
       ET \= none, 
       query_global(gumtree_tactic_declaration(elves, _, _)) ->
         append(Cur, CurrT, CurrD),
         gumtree_2_predicate(elves, CurrD, CurrT, Output, OutT, NewTac),
         catch(
                ( ip_set(in_tactic, true),
                  call_predicate(NewTac),
                  ip_set(in_tactic, fail),
                  OutT = []
                ),
                abort_tactic(_),
                Output = Cur
         )
    ;
         Output = Cur
    ),
    gumtree1_aux(Output, History, Result).

gumtree1_aux(Cur, History, Result) :-
    ip_lookup(gumtree_root_node, Root),
    proof_open_nodes(Root, Open),
    diff_list(Open, Cur, NonCur),
    ( Open = [] ->
        Prompt = ['complete> ']
    ;
      NonCur = [] ->
        Prompt = [proof_nodes(Cur),'::: ']
    ;
        Prompt = ['(', proof_nodes(NonCur), ') ', proof_nodes(Cur),'::: ']
    ),
    global_state_lookup(current_dir,CurrThy),
    History = [ucmd(ThisCmdNum,_,_,_,_)|_],
    repeat,
    pre_prompt_cmd(Root, Cur),
    gumtree_display(Cur),
    announce_event(proof(CurrThy, the_only_proof, gumtree_display(Open,Cur))),
    gumtree_get_command_top(Prompt, CurrThy, Cmd0, Text0),
    gumtree_deabbrev_cmd(Table, Cmd0, Cur, Impl),
    debug(gumtree, ['doing cmd: ',q(Impl)]),
    std_nonvar(Impl),
    ( Impl = undo ->
        !, Result = undo
    ;
      Impl = undo(Level) ->
        !, Result = undo(Level)
    ;
      Impl = retry ->
        !, fail
    ;
        gumtree2(Impl, Table, CurrThy, Text0, Cur, History, Result, ThisCmdNum)
    ).

gumtree_get_command_top(Prompt, Cur, Cmd, Text) :-
    gumtree_get_command(Prompt, Cur, Cmd0, Text0),
    ( Cmd0 = silent_cmd(SCmd) ->
        call_silent_command(SCmd),
        gumtree_get_command_top([], Cur, Cmd, Text)
      ;
        Cmd = Cmd0,
        Text = Text0
    ).


call_silent_command(Cmd) :-
    ( gumtree_deabbrev_cmd(general, Cmd, _, Imp), call_predicate(Imp) ->
        fail
    ;
        increment_error_count,
        writelist(['failed: ', chars(Cmd),nl]),
        fail
    ).
call_silent_command(_).
        

gumtree2(Impl, Table, CurrThy, Text0, Cur, History, Result, ThisCmdNum) :-
    gumtree_do_cmd(Impl, Table, CurrThy, Text0, Cur, History, Result2),
    debug(gumtree, [q(ThisCmdNum),': do_cmd result is ',q(Result2)]),
    gumtree_result(Result2, ThisCmdNum, Cur, History, Result),
    !,
    ( Result == undo ->
        writelist(['undone: ', chars(Text0),nl]),   
        fail
    ;
        true
    ).

%% This is the same as get_command/4, but removes any ::: commands
%% from the Text form of the command.
%%.
gumtree_get_command(Prompt, CurrThy, Cmd0, Text) :-
    get_command(Prompt, CurrThy, Cmd0, Text0),
    ( std_compound(Cmd0), Cmd0 = ':::'(_,_) ->
        ( gumtree_node_selection(Text0, Text1) ->
            Text = Text1
        ;
            handle_error(['cannot parse term before ::: in: ',chars(Text0)])
        )
    ;
        Text = Text0
    ).

%% We use DCG notation to define a recognizer that skips whitespace
%% comments, integers, variables and lists containing integers and
%% variables, then checks for ":::".  The io_... recognizers
%% are defined in the core Ergo library.
%%
gumtree_node_selection  --> node_spec_chars, ":::", io_whitespace.

node_spec_chars --> node_spec_char, !, node_spec_chars.
node_spec_chars --> "".

node_spec_char  --> io_comment.
node_spec_char  --> io_whitechar.
node_spec_char  --> "]".
node_spec_char  --> ",".
node_spec_char  --> "*".
node_spec_char  --> "_".
node_spec_char  --> "[".
node_spec_char  --> [C], {node_spec_char(C)}.

node_spec_char(C) :- 0'0 =< C, C =< 0'9.
node_spec_char(C) :- 0'A =< C, C =< 0'Z.
node_spec_char(C) :- 0'a =< C, C =< 0'z.


gumtree_result(commit(CmdNum,Label,Cur),ThisCmdNum,Cur0,History,Result) :-
    PRE('gumtree_result - commit', (integer(ThisCmdNum), integer(CmdNum))),
    ThisCmdNum < CmdNum,
    !,
    debug(gumtree, ['starting new gumtree at level ',q(ThisCmdNum)]),
    NextNum is ThisCmdNum + 1,
    %% Cur0 ---> Cur is the best we can generate as an approximation
    %% to the input and output nodes of the new `compound' command.
    ( gumtree1(Cur, [ucmd(NextNum,Cur0,Cur,commit,Label)|History], Result) ->
        true
    ;
        increment_error_count,
        writelist(['failed: ', chars(Label),nl]),
        fail
    ).
gumtree_result(undo(CmdNum), ThisCmdNum, _, _, Result) :-
    PRE('gumtree_result - undo(_)', (integer(ThisCmdNum), integer(CmdNum))),
    ThisCmdNum < CmdNum,
    !,
    debug(gumtree, ['failing due to ',q(undo(CmdNum))]),
    Result = undo.
gumtree_result(undo, _, _, _, Result) :-
    !,
    debug(gumtree, ['failing due to undo']),
    Result = undo.
gumtree_result(Result, _, _, _, Result) :-
    debug(gumtree, ['succeeded with result = ',q(Result)]).


%% gumtree_deabbrev_cmd(Table, Cmd0, CurrNodes, Impl).
%%
%%  This deabbreviates the arguments of a gumtree command
%%  and returns the code that should be executed for that command.
%%  Note that a gumtree command can be a tactic, or a tactic
%%  preceded by an explicit selection of some proof nodes,
%%  or a general command (like help, displaying information etc.).
%%
%%  mode gumtree_deabbrev_cmd(?atom, +term, +list(nodeid), -code).

gumtree_deabbrev_cmd(Table, :::(InNodes0,Tac0), _, Impl) :-
    !,
    gumtree_deabbrev_node_list(InNodes0, InNodes),
    Table = gumtree,
    gumtree_deabbrev_tactic_args(Tac0, Tac),
    Impl = call_gumtree_tactic(InNodes,Tac).
gumtree_deabbrev_cmd(Table, SelNodes0, _, Impl) :-
    gumtree_deabbrev_node_list(SelNodes0, SelNodes),
    !,
    Table = gumtree,
    Impl = select_gumtree_nodes(SelNodes).
gumtree_deabbrev_cmd(gumtree, Cmd0, Cur, Impl) :-
    query_global(gumtree_tactic_declaration(Cmd0, Mode, _)),
    debug(gumtree, ['deabbreviating: ', q(Cmd0),nl,'with mode: ',q(Mode)]),
    deabbreviate_cmd_relaxed(Mode, Cmd0, Cmd),
    !,
    %%%gumtree_2_predicate(Cmd, GCmd),
    Impl = call_gumtree_tactic(Cur,Cmd).
gumtree_deabbrev_cmd(Table, Cmd0, Cur, Impl) :-
    %% No nodes explicitly selected, so we default to Cur.
    check_is_legal_cmd(Cmd0),
    functor(Cmd0, F, A),
    functor(Cmd, F, A),
    debug(gumtree, ['Trying to find cmd: ',q(F/A)]),
    member(Table, [gumtree, general, proof]),
    find_command(Cmd, Table, Thy, Mode, Impl0, _),
    (  Mode = [arity(Mode0,_)] ->
         Mode1 = [Mode0]
    ;
       Mode1 = Mode
    ),
    debug(gumtree, ['Found ',q(Table),' cmd: ',q(Thy),':',q(Cmd)]),
    debug(gumtree, ['deabbreviating: ', q(Cmd0),nl,'with mode: ',q(Mode)]),
    deabbreviate_cmd_relaxed(Mode1, Cmd0, Cmd),
    debug(gumtree, ['deabbreviated to: ', q(Cmd),nl]),
    !,
    debug(gumtree, ['impl0: ', q(Impl0), nl]), 
    ( Table=gumtree, Impl0=call_gumtree_tactic(Tac) ->
        Impl = call_gumtree_tactic_basic(Cur,Tac)
    ;
        %% Not a gumtree tactic, so we ignore Cur.
        Impl = Impl0
    ),
    debug(gumtree, ['impl: ', q(Impl), nl]). 
gumtree_deabbrev_cmd(Table, Cmd0, _, _) :-
    %% To give better error messages here, we see if Cmd0 exists.
    std_nonvar(Cmd0),
    functor(Cmd0, F, A),
    functor(Cmd, F, A),
    ( member(Table, [gumtree, general, proof]),
      find_command(Cmd, Table, Thy, Mode, _, _) ->
        handle_error(['bad arguments in command: ',q(Cmd0),nl,
                      '   Expecting: ',q(Thy),':',q(Mode)])
    ;
        handle_error(['unknown command: ', code(Cmd0)])
    ).
                     
        

%% gumtree_deabbrev_tactic_args(Tac0, Tac).
%%
%%  This deabbreviates the arguments of a call to a gumtree tactic.
%%  Since the tactic call may appear in the middle of a sequence
%%  of other tactics, some variables may not be instantiated now,
%%  but will/might be later when the tactic is actually executed.
%%  For that reason, we allow variables to be tactics, and are
%%  relaxed about checking instantiatedness of arguments.
%%
%%  mode gumtree_deabbrev_tactic_args(?tactic, -tactic).

gumtree_deabbrev_tactic_args(Tac0, Tac) :-
    var(Tac0),
    !,
    Tac = Tac0.
gumtree_deabbrev_tactic_args(Tac0, Tac) :-
    check_is_legal_cmd(Tac0),
    %% now find the first match of F/A in the gumtree command table.
    query_global(gumtree_tactic_declaration(Tac0, Mode, _)),
    !,
    debug(gumtree, ['Found tactic: ',q(_Thy),':',q(Tac0),nl,
                    ' with mode: ',q(Mode)]),
    ( deabbreviate_cmd_relaxed(Mode, Tac0, Tac) ->
        true
    ;
        handle_error(['bad arguments in command: ',q(Tac0)])
    ).
gumtree_deabbrev_tactic_args(Tac0, Tac) :-
    check_is_legal_cmd(Tac0),
    %% now find the first match of F/A in the gumtree command table.
    functor(Tac0, F, A),
    functor(Tac, F, A),
    functor(TacTmp, F, A),
    find_command(TacTmp, gumtree, Thy, Mode, _, _),
    !,
    ( Mode = [arity(Mode0,_)] ->
        Mode1 = [Mode0]
    ;
        Mode1 = Mode
    ),
    debug(gumtree, ['Found tactic: ',q(Thy),':',q(TacTmp),nl,
                    ' with mode: ',q(Mode)]),
    ( deabbreviate_cmd_relaxed(Mode1, Tac0, Tac1) ->
        true
    ;
        handle_error(['bad arguments in command: ',q(Tac0)])
    ),
    gumtree_deabbrev_fix_call(F, Tac1, Tac).   
gumtree_deabbrev_tactic_args(Tac, deabbrev_and_call_gumtree_tactic(Tac)).

gumtree_deabbrev_fix_call('|', Tac1, Tac) :-
    !,
    gumtree_deabbrev_fix_call_args(Tac1, Tac).
gumtree_deabbrev_fix_call('if_tac', Tac1, Tac) :-
    !,
    gumtree_deabbrev_fix_call_args(Tac1, Tac).
gumtree_deabbrev_fix_call(';', Tac1, Tac) :-
    !,
    gumtree_deabbrev_fix_call_args(Tac1, Tac).
gumtree_deabbrev_fix_call('.', Tac1, Tac) :-
    !,
    gumtree_deabbrev_fix_call_args(Tac1, Tac).
gumtree_deabbrev_fix_call('!', Tac1, Tac) :-
    !,
    gumtree_deabbrev_fix_call_args(Tac1, Tac).
gumtree_deabbrev_fix_call('fails', Tac1, Tac) :-
    !,
    gumtree_deabbrev_fix_call_args(Tac1, Tac).
gumtree_deabbrev_fix_call('succs', Tac1, Tac) :-
    !,
    gumtree_deabbrev_fix_call_args(Tac1, Tac).
gumtree_deabbrev_fix_call(_, Tac, Tac).

gumtree_deabbrev_fix_call_args(Tac1, Tac) :-
    functor(Tac1, F, N),
    functor(Tac, F, N),
    gumtree_deabbrev_fix_call_args1(N, Tac1, Tac).

gumtree_deabbrev_fix_call_args1(0, _Tac1, _Tac) :- !.
gumtree_deabbrev_fix_call_args1(N, Tac1, Tac) :- 
    arg(N, Tac1, A),
    ( var(A) ->
        arg(N, Tac, deabbrev_and_call_gumtree_tactic(A))
    ; compound(A) ->
        functor(A, F, _),
        gumtree_deabbrev_fix_call(F, A, A1),
        arg(N, Tac, A1)
    ;
        arg(N, Tac, A)
    ),
    N1 is N-1,
    gumtree_deabbrev_fix_call_args1(N1, Tac1, Tac).

%% gumtree_deabbrev_node_list(Nodes0, Nodes).
%%
%%  This deabbreviates an explicit selection of a set of open nodes.
%%
%%  mode gumtree_deabbrev_node_list(+term, -list(nodeid)).

gumtree_deabbrev_node_list(InNodes0, InNodes) :-
    ( integer(InNodes0) ->
        InNodes = [InNodes0]
    ; InNodes0 == (*) ->
        ip_lookup(gumtree_root_node, Root),
        proof_open_nodes(Root, InNodes)
    ; list(InNodes0), map(integer, [InNodes0]) ->
        InNodes = InNodes0
    ;
        fail
    ).
    
%% private gumtree_display(CurrentNodes).
%%
%%  Optionally calls some display procedure.
%%
%%  mode gumtree_display(@list(nodeid)).
%%
gumtree_display(Curr) :-
    query_global(gumtree_display(Spec)),
    ( Spec = silent ->
        true
    ; Spec = curr(P) ->
        P(Curr)
    ; Spec = open(P) ->
        ip_lookup(gumtree_root_node, Root),
        proof_open_nodes(Root, Open),
        P(Open)
    ;
        fail
    ),
    !.
gumtree_display(_).

%%
%% gumtree_get_abbrev_premises(Premises)
%%
%% Return the premises of the rule in abbreviated form.
%%
gumtree_get_abbrev_premises(Premises) :-
    proof_root_sequents(Seqs),
    abbrev_seqs(Seqs, Premises).

%% private gumtree_premises_display.
%%
%% Display premises at proof start (if not silent)
%%
gumtree_premises_display :-
    query_global(gumtree_display(Spec)),
    Spec = silent,
    !.
gumtree_premises_display :-
    gumtree_get_abbrev_premises(Premises),
    (   Premises = []
        ->
            true
        ;
	    writelist([nl, 'Premises:',nl]),
	    gumtree_premises_display(Premises)
    ).

gumtree_premises_display([]) :-
    writelist([nl]).
gumtree_premises_display([S|Seqs]) :-
    writelist([thy(S),nl]),
    gumtree_premises_display(Seqs).


%% gumtree_do_cmd(Cmd, Table, Thy, PrettyCmd, Cur, History, Result).
%%
%%  Executes Cmd.
%%
%%  If Cmd succeeds, then it (usually) recurses, with History extended
%%  by adding PrettyCmd onto the front.  Before the recursive call
%%  to gumtree/3, a choicepoint is left so that `undo' commands within
%%  the recursive call produce failure messages.  After the first
%%  successful exit from the recursive call (typically due to a
%%  commit(Level) or undo(Level) command from that level or deeper)
%%  all alternative solutions are cut away.
%%  
%%  However, non-proof commands whose effects are `persistent' (that is,
%%  are not undone by failure -- eg. I/O and database updates) are
%%  treated differently because we do not want to add them to the history.
%%  They are failed over, so that control returns to the current recursive
%%  level and history is not extended.
%%
%%  If Cmd fails, a message is printed using PrettyCmd, which should
%%  be a prettier (string) equivalent of Cmd.
%%
%%  mode gumtree_do_cmd(+code, @atom, @theory, +chars,
%%                      @list(nodeid), @list(nodeid), @list(code), -nonvar).
%%
gumtree_do_cmd(undo, _, _, _, _, _, Result) :-
    !,
    Result = undo.
gumtree_do_cmd(undo(Level), _, _, _, _, _, Result) :-
    !,
    check_or_error(
        (integer(Level), 0 =< Level),
	['The level of undo must be a non-negative integer, not ', q(Level)]
    ),
    Result = undo(Level).
gumtree_do_cmd(commit(Level), _, _, _, Cur, History, Result) :-
    !,
    check_or_error(
        ( 0 < Level,
          member(ucmd(Level,_,_,_,Cmd), History)),
        ['there is no command number ',q(Level)]),
    %% create a one line summary of the committed commands.
    ( append(Cmd2,[10|_],Cmd) ->
        true
    ;
        Cmd2 = Cmd
    ),
    append(Cmd2," ...\n",Cmd3),
    Result = commit(Level,Cmd3,Cur).
gumtree_do_cmd(commit(Level,Label), _, _, _, Cur, History, Result) :-
    !,
    check_or_error(
        ( list(Label),
          map(integer, [Label]),
          append(Label," ...\n",Label2) ),
        ['illegal second argument of commit: ',q(Label)]),
    check_or_error(
        ( 0 < Level,
          member(ucmd(Level,_,_,_,_), History)),
        ['there is no command number ',q(Level)]),
    Result = commit(Level,Label2,Cur).
gumtree_do_cmd(quit_proof, _, Thy, _, _, _, _) :-
    !,
    throw(abort_proof(Thy:thy_only_proof)).
gumtree_do_cmd(quit_tactic, _, Thy, _, _, _, _) :-
    !,
    throw(abort_tactic(Thy:thy_only_proof)).
gumtree_do_cmd(suspend_proof, _, Thy, _, _, _, _) :-
    suspend_proof(user_interface),
    %% We cd back to OrigThy in case the user has changed theories.
    cd(Thy).
gumtree_do_cmd(show_history, _, _, _, _, History, _) :-
    !,
    reverse(History, History2),
    member(ucmd(Num,In,Out,_,Cmd), History2),
    writelist([right(3,[q(Num)]),' ',
               proof_nodes(In),' ---> ',proof_nodes(Out),nl,
               '    ',chars(Cmd)]),
    fail.
gumtree_do_cmd(Cmd, general, _, PCmd, _, _, _) :-
    !,
    % General commands are executed in a repeat-fail style.
    debug(gumtree, ['calling general command: ', q(Cmd)]),
    ( call_predicate(Cmd) ->
        fail
    ;
        increment_error_count,
        writelist(['failed: ', chars(PCmd),nl]),
        fail
    ).
gumtree_do_cmd(call_passive_cmd(Cmd), _, _, PCmd, _, _, _) :-
    !,
    % Passive commands are executed in a repeat-fail style.
    debug(gumtree, ['calling passive cmd: ', q(Cmd)]),
    ( call_predicate(Cmd) ->
        fail
    ;
        increment_error_count,
        writelist(['failed: ', chars(PCmd),nl]),
        fail
    ).

gumtree_do_cmd(select_gumtree_nodes(Nodes),
               _, _Thy, PCmd, In, History, Result) :-
    %% leave a choice point for the failure message.
    debug(gumtree, ['recording cmd in tree and history']),
    History = [ucmd(CmdNum0,_,_,_,_)|_],
    CmdNum is CmdNum0 + 1,
    check_or_error(diff_list(Nodes, In, _Rest),
                   ['can only select open nodes.']),
    diff_list(In, Nodes, _Def),
    gumtree1(Nodes, 
            [ucmd(CmdNum,In,Nodes,node_sel,PCmd)|History],
            Result),
    !.
gumtree_do_cmd(call_gumtree_tactic(Input,Tac),
               _, _Thy, PCmd, In, History, Result) :-
    %% leave a choice point for the failure message.
    debug(gumtree, ['calling tactic: ', q(Tac)]),
    check_or_error(diff_list(Input, In, []),
                   ['can only select open nodes.']),
    diff_list(In, Input, []),
    append(Input, InputT, InputD),
    ( query_global(gumtree_tactic_declaration(Tac, _, _)) ->
      gumtree_2_predicate(Tac, InputD, InputT, Output, OutT, NewTac)
    ; 
      Tac =.. TacList,
      append(TacList, [InputD, InputT, Output, OutT], NewTacList),
      NewTac =.. NewTacList
    ),
    catch(
           ( ip_set(in_tactic, true),
             call_predicate(NewTac),
             ip_set(in_tactic, fail)
           ),
           abort_tactic(_),
           fail
    ),
    OutT = [],
    debug(gumtree, ['recording cmd in tree and history']),
    History = [ucmd(CmdNum0,_,_,_,_)|_],
    CmdNum is CmdNum0 + 1,
    annotate_input_nodes(Input, CmdNum, Input, Output, PCmd),
    gumtree1(Output, 
            [ucmd(CmdNum,Input,Output,tactic,PCmd)|History],
            Result),
    !.
gumtree_do_cmd(call_gumtree_tactic_basic(Input,Tac),
               _, _Thy, PCmd, In, History, Result) :-
    %% leave a choice point for the failure message.
    debug(gumtree, ['calling tactic: ', q(Tac)]),
    check_or_error(diff_list(Input, In, []),
                   ['can only select open nodes.']),
    diff_list(In, Input, []),
    append(Input, InputT, InputD),
    catch(
           ( ip_set(in_tactic, true),
             call_gumtree_tactic(Tac, InputD, InputT, Output, OutT),
             ip_set(in_tactic, fail)
           ),
           abort_tactic(_),
           fail
    ),
    OutT = [],
    debug(gumtree, ['recording cmd in tree and history']),
    History = [ucmd(CmdNum0,_,_,_,_)|_],
    CmdNum is CmdNum0 + 1,
    annotate_input_nodes(Input, CmdNum, Input, Output, PCmd),
    gumtree1(Output, 
            [ucmd(CmdNum,Input,Output,tactic,PCmd)|History],
            Result),
    !.    
gumtree_do_cmd('--->'(L),
                   _, _, PCmd, Cur0, History, Result) :-
    %% leave a choice point for the failure message.
    ( var(L) ->
        List = [L]
    ;
        List = L
    ),
    append(List, _, Cur0),
    debug(gumtree, ['recording ''--->'' in history']),
    History = [ucmd(CmdNum0,_,_,_,_)|_],
    CmdNum is CmdNum0 + 1,
    gumtree1(Cur0, [ucmd(CmdNum,List,List,cmd,PCmd)|History], Result),
    !.
gumtree_do_cmd(Cmd, _, _Thy, PCmd, Cur0, History, Result) :-
    %% TO DO: MB 7/9/1998.
    %% This alternative may eventually disappear, once all useful
    %% functionality is available through the tactic language?
    %%
    %% Leave a choice point for the failure message.
    %% Don't consider this alternative if Cmd has already matched above.
    functor(Cmd, CmdF, _),
    \+ member(CmdF, [call_gumtree_tactic,'--->',select_gumtree_nodes,
                     call_gumtree_tactic_basic]),
    %%writelist(['is this obsolete?  Cmd = ',q(Cmd),nl]),
    debug(gumtree, ['calling command: ', q(Cmd)]),
    call_predicate(Cmd),
    %% Since we don't know much about what Cmd has done, we must
    %% recompute the entire set of open nodes.  Thus, we put
    %% Cur0 ---> Open into Cmd's history entry.
    %ip_lookup(gumtree_root_node, Root),
    %proof_open_nodes(Root, Open),
    History = [ucmd(CmdNum0,_,_,_,_)|_],
    CmdNum is CmdNum0 + 1,
    gumtree1(Cur0, [ucmd(CmdNum,Cur0,Cur0,cmd,PCmd)|History], Result),
    !.
gumtree_do_cmd(_, _, _Thy, PCmd, _, _, _) :-
    increment_error_count,
    writelist(['failed: ', chars(PCmd),nl]),
    fail.


annotate_input_nodes([], _, _, _, _).
annotate_input_nodes([N|NS], CmdNum, In, Out, Cmd) :-
    annotate(N, usercmd(CmdNum, N, In, Out, Cmd)),
    annotate_input_nodes(NS, CmdNum, In, Out, Cmd).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Documentation for this User Interface.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

gumtree_cmds header
  'Proof commands (The ''GumTree'' User Interface for Ergo 5 proofs)'
extends
  master_help_node
doc
  '`Gumtree'' is a user interface for creating Ergo 5 proofs.',nl,
  'The number in the prompt is the number of the first open (non-proved)',nl,
  'node in the current proof tree.  However, you may choose to work on',nl,
  'any open node.',
  para,
  'The following commands and groups of commands are available:',nl.


gumtree_user_tactics header
  'User Defined Gumtree Tactics'
extends 
  gumtree_cmds
doc
  'The current user defined gumtree tactics are:',nl.

proof_tree_cmds header
  'Proof Tree Commands'
extends
  gumtree_cmds
doc
  'These commands allow you to extend and manage the current proof tree.',nl,
  'Most of the commands take a ''nodeid'' argument (an integer)',nl,
  'to determine which node of the proof tree they will be applied to.',
  para,
  'The proof tree commands are:',nl.

save_close_treecmds header
  'Save/Close/Read Proof Commands'
extends
  gumtree_cmds
doc
  'These commands are for:',nl,
  begin(list),
  item('- '),'saving the current proof staring at a given node to a file',
  item('- '),'converting the current proof to an inference rule (or theorem)',
  item('- '),'quiting the current proof without saving anything',
  item('- '),'reading a proof file',
  end(list),nl.
  
tactic_combine_cmds header
  'Tactic Combinator Commands'
extends
  gumtree_cmds
doc
  'These commands are for combining several tactics in parallel or',nl,
  'in sequence.',nl.
  
prooftree_info_cmds header
  'Proof Information Commands'
extends
  gumtree_cmds
doc
  'The following commands viewing information about the current proof.',nl.
  
query_treecmds header
  'Query Procedures'
extends
  gumtree_cmds
doc
  'Query Procedures for Proof Nodes and Proof Tree.',nl.
  
var_treecmds header
  'Variable Instantiation Commands'
extends
  gumtree_cmds
doc
  'These commands control the instantiation of proof variables',nl,
  '(both meta-variables and object-variables).  The proof variables',nl,
  'are split into two classes: `fixed variables'' and `new variables''.',
  para,
  'Fixed variables are those that appear in the original theorem.',nl,
  'They can only be instantiated manually, not automatically.',
  para,
  'New variables have typically been introduced during the proof.',nl,
  'They may be instantiated automatically during proof steps.',nl.

gumtree command
  undo
extends
  proof_tree_cmds
doc
  'Undoes the last proof-related command.'
mode
  undo.

gumtree command
  retry
extends
  proof_tree_cmds
doc
  'Finds the next solution to the last proof-related command',nl,
  'if it has multiple solutions.'
mode
  retry.

gumtree command
  undo(CmdNum)
extends
  proof_tree_cmds
doc
  'Undoes all commands back to (and including) the one numbered CmdNum.'
mode
  undo(@integer).


gumtree command
  '--->'(Names)
header
  '---> Names'
extends
  proof_tree_cmds
doc
  'This names the first N open nodes, where N is the length of Names.',nl,
  'Names must be a list of variables or a single variable (N=1).',
  para,
  'This command is equivalent to:  * ::: [names(Names), skip].'
mode
  '--->'(-nodeid),
  '--->'(+list(nodeid)).

gumtree command
  commit(CmdNum),
  commit(CmdNum,Label)
extends
  proof_tree_cmds
doc
  'Removes all undo points and alternative solutions between the current',nl,
  'command and the command numbered CmdNum.  This means that those',nl,
  'commands cannot be undone individually, only as a block.',
  para,
  'The history display changes to reflect this by packaging up all the',nl,
  'commands after CmdNum into a single history entry marked as `commit:''.',nl,
  'If Label is given, it will be displayed as the `action'' part of the',nl,
  'new history entry.',
  para,
  'The primary reason for using this command is to reduce the size',nl,
  'of the history list or to commit to the current solution of some',nl,
  'non-deterministic command.'
mode
  commit(@integer),
  commit(@integer, @string).


gumtree command
  history
extends
  proof_tree_cmds
doc
  'Display the history of proof-related commands, with the most recent',nl,
  'command last.  An `undo'' command will undo the effect of the last',nl,
  'command in the command history.'
implementation
  show_history
mode
  history.

gumtree command
  quit_proof
extends
  save_close_treecmds
doc
  'Quits out of the current proof without saving anything.'
mode
  quit_proof.

gumtree command
  end_of_file
doc
  %% We don't want end_of_file to be too drastic (eg. abandon), since
  %% it is fairly easy to generate it by mistake.  However, if we get 
  %% it on standard input, we must eventually exit the proof (and the
  %% prover), otherwise we will just process an infinite stream of
  %% end_of_file commands.  So, we make it equivalent to `quit_proof'.
  %% (but don't advertise this fact).
  'Same as quit_proof.'
implementation
  quit_proof
mode
  end_of_file.

gumtree command
  suspend
doc
  'Suspend the current proof and start up a new theory-level user',nl,
  'interface.  After that interface is exited, the current theory will',nl,
  'be reselected and the suspended proof will be resumed.'
mode
  suspend
implementation
  suspend_proof.

?- assert(elves_type(none)).

elves_type parameter
  [set_elves_type, get_elves_type]
doc
  'This determines if the elves tactic is run before each user',nl,
  'interaction in the Gumtree interpreter.',
  para,
  'If elves_type === none then no elves are run.',br,
  'Otherwise the elves tactic is run, provided it is defined',
  para,
  'See jprop/elf.gum for the default definition of the elves tactic.' 
  
extends
  parameters.

set_elves_type(X) :-
    ( elves_type(X) ->
        true
    ;
       retract(elves_type(_)),
       assert(elves_type(X))
    ).

get_elves_type(Type) :-
    elves_type(Type).


