%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: operator.ql,v 1.44 2002/06/27 05:10:26 itp Exp $

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% This file provides procedures for adding and deleting global
%% and theory-specific operators.  It supports prefix, infix, postfix
%% and 'quant' operators and defines three kinds of associativity
%% for infix operators: leftassoc, rightassoc and nonassoc.
%%
%% The theory-specific operators of a given theory can be changed
%% at any time, even when that theory is closed.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

#include "../prover/lib/ergo.h"

?- dynamic(operators_declared/1).


%% setup_operators.
%%
%%  This procedure sets up the operator mechanisms, ready 
%%  for use by other tactic packages.
%%  If it is called multiple times, the first call does the
%%  setup job and following calls have no additional effect.
%%
%%  mode setup_operators.
%%
setup_operators :-
    theory_fact(inherited_operators,1),
    !.
setup_operators :-
    %% Declare all saved fact types.
    new_theory_fact(inherited_operators,1,[save(true)]),
    new_global_fact(global_operator,3,[save(true)]),
    new_theory_fact(operator,4,[save(true)]).


?- setup_operators.


%% enable_obvar_prefix(Thy).
%%
%% Enable all the declared object variable prefixes in Thy.
%%
%% mode enable_obvar_prefix(@atom).
%%
enable_obvar_prefix(Thy) :-
    query_theory(object_vars(Prefix), Thy),
    obvar_prefix(Thy, Prefix),
    fail.
enable_obvar_prefix(_).

%% obvar_prefix_table_inherit(FromThy).
%%
%%  Include all object variables declared in FromThy into the current theory.
%%
%%  mode obvar_prefix_table_inherit(@theory).
%% 
obvar_prefix_table_inherit(FromThy) :-
    query_theory(object_vars(Prefix), FromThy),
    declare_object_vars(Prefix),
    fail.
obvar_prefix_table_inherit(_) :-
    global_state_lookup(current_dir,Thy),
    enable_obvar_prefix(Thy).

%% enable_operators(Thy).
%%
%%  If Thy=global, this procedure declares a Qu-Prolog operator
%%  table called 'global' containing all the global operators.
%%  Otherwise, Thy must be the name of a loaded theory and this
%%  procedure declares an operator table called Thy that inherits 
%%  from the 'global' operator table and contains all the operators 
%%  of Thy.
%%
%%  This is typically called whenever a new theory is entered (via cd/1).
%%  For efficiency, it records which operator tables have been set up
%%  so that the overhead of building each operator table only occurs once.
%%
%%  It does *not* change the Qu-Prolog current operator table,
%%  because all read commands should specify an explicit operator table
%%  (eg. get_command/3).  However, it does call op_table/1 at the moment,
%%  because QP doesn't set up new operator tables properly (with the default
%%  operator ',') otherwise.
%%
%%  mode enable_operators(@atom).
%%
enable_operators(Thy) :-
    operators_declared(Thy),
    !.
enable_operators(global) :-
    !,
    op_table(Curr),
    op_table(global),  %% must call this to initialize the operator table.
    op_table(Curr),
    msg(operators, ['setting up global operators.']),
    (
        query_global(global_operator(Name, Kind, Prec)),
        op(global, Prec, Kind, Name),
        fail
    ;
        true
    ),
    !,
    assert(operators_declared(global)).
enable_operators(Thy) :-
    op_table(Curr),
    op_table(Thy),  %% must call this to initialize the operator table.
    op_table(Curr),
    op_table_inherit(Thy,global),
    (
        query_theory(inherited_operators(FromThy), Thy),
        enable_operators(FromThy),
        op_table_inherit(Thy, FromThy),
        fail
    ;
        msg(operators, ['setting up operators for theory ',q(Thy),'.']),
        query_theory(operator(Name,Kind,Prec,_), Thy),
        op(Thy,Prec,Kind,Name),
        fail
    ;
        true
    ),
    !,
    assert(operators_declared(Thy)).


%% add_global_operator(Prec,Kind,Name).
%%
%%  Adds an operator to the 'global' part of the Ergo database.
%%  Fails (and generates an error message) if the new operator clashes
%%  with existing global operators.
%%
%%  Global operators are implicitly inherited by all theories,
%%  so should be used sparingly.  There is currently no checking to
%%  ensure that Name is not already defined as a theory specific
%%  operator (such checking would not be foolproof anyway, because 
%%  not all theories are loaded or accessible at any given time).
%%
%%  mode add_global_operator(@integer, @atom, @atom).
%%
add_global_operator(Name, Kind, Prec) :-
    %% for backwards compatibility only.
    atom(Name),
    !,
    writelist(['Warning: obsolete ',q(add_global_operator(Name,Kind,Prec))]),
    add_global_operator(Prec, Kind, Name).
add_global_operator(Prec, Kind, Name) :-
    check_atom(Name),
    check_or_error(
        (integer(Prec), 0 < Prec, Prec =< 1200),
        ['illegal precedence ',q(Prec),'.  (Must be 1..1200)']),
    check_operator_kind(Kind, Kind2),
    check_no_global_operator_clash(Name,Kind2),
    tactic_assert_global(global_operator(Name, Kind2, Prec)),
    op(global, Prec, Kind2, Name).

check_no_global_operator_clash(Name, Kind) :-
    query_global(global_operator(Name,Kind2,_)),
    kind_incompatible(Kind, Kind2),
    !,
    handle_error([q(Name),' is already a global operator. (',q(Kind2),').']).
check_no_global_operator_clash(_, _).


%% delete_global_operator(Name).
%% delete_global_operator(Name,Kind).
%%
%%  Deletes the operator called Name from the global part of the database.
%%
%%  mode delete_global_operator(@atom).
%%  mode delete_global_operator(@atom,@atom).
%%
delete_global_operator(Name) :-
    delete_global_operator(Name,_).
delete_global_operator(Name,Kind) :-
    check_atom(Name),
    tactic_retract_global(global_operator(Name,Kind,_)),
    op(global, 0, Kind, Name).



database command
  add_operator(Prec,Kind,Name)
extends
  sugar_cmds
mode
  add_operator(@integer, @atom, @atom)
doc
  'Add an operator, called Name, to the current theory.',nl,
  'Note that this command only affects the concrete syntax of the theory',nl,
  'and it is not necessary for Name to be a function or quantifier of the',nl,
  'current theory before it can be declared as an operator.',
  para,
  'Kind can be: prefix, postfix, nonassoc, leftassoc, rightassoc or quant.',
  para,
  'Prec defines the precedence of the operator.  It can be any number',nl,
  'between 1 and 1200, where smaller numbers mean tighter binding.',nl,
  'Most user-defined operators should have a precedence less than 990,',nl,
  'because precedences above that are reserved for the prover itself.',
  para,
  'This command fails (and generates an error message) if the new operator',nl,
  'clashes with existing global operators or with existing operators of',nl,
  'the current theory.'.

add_operator(Name, Kind, Prec) :-  %% For backwards compatibility only.
    atom(Name),
    integer(Prec),
    !,
    writelist(['Warning: obsolete ',q(add_operator(Name,Kind,Prec))]),
    add_operator(Prec,Kind,Name).
add_operator(Prec, Kind, Name) :-
    check_current_theory(Thy),
    check_atom(Name),
    check_or_error(
        (integer(Prec), 0 < Prec, Prec =< 1200),
        ['illegal precedence ',q(Prec),'.  (Must be 1..1200)']),
    check_operator_kind(Kind, Kind2),
    add_operator2(Name, Kind2, Prec, Thy, Thy).


%% private add_operator2(Name, Kind, Prec, FromThy, ToThy).
%%
%%  This procedure is used to implement add_operator/3 and add_operators/1
%%  and should not be called directly by users.
%%  It assumes that Name, Kind and Prec define a valid operator 
%%  and adds that operator to ToThy, marking it as having come from
%%  FromThy.  If the operator would clash with existing operators in
%%  ToThy, error messages are generated and the procedure fails.
%%  FromThy should be an ancestor of ToThy.
%%  Kind should be one of: fy, yf, xfx, xfy, yfx or quant.
%%
%%  mode add_operator2(@atom, @atom, @integer, @atom, @atom).
%%
add_operator2(Name, Kind, Prec, FromThy, ToThy) :-
    query_global(global_operator(Name,Kind1,Prec1)),
    kind_incompatible(Kind1, Kind),
    %% We allow global operators to be overridden, but give
    %% a warning the first time that they are added.
    FromThy = ToThy,
    operator_kind(ErgoKind,Kind),
    operator_kind(ErgoKind1,Kind1),
    msg(operators, ['WARNING: overriding ',
                    q(global_operator(Prec1,ErgoKind1,Name)),nl,
                    '         with ',q(operator(Prec,ErgoKind,Name))]),
    fail.
add_operator2(Name, Kind, Prec, FromThy, ToThy) :-
    query_operator(Name,Kind1,Prec1,FromThy1, ToThy),
    kind_incompatible(Kind1, Kind),
    !,
    ( Prec=Prec1, FromThy=FromThy1 ->
        true
    ;
        handle_error([q(Name),' is already defined as an operator.'])
    ).
add_operator2(Name, Kind, Prec, FromThy, ToThy) :-
    tactic_assert_theory(operator(Name, Kind, Prec, FromThy), ToThy),
    op(ToThy, Prec, Kind, Name).


%% private kind_incompatible(Kind1, Kind2).
%%
%%  Succeed if operators of kind Kind1 and Kind2 can not coexist safetly.
%%  Kind1 and Kind2 should be one of: fy, yf, xfx, xfy, yfx or quant.
%%
%%  mode kind_incompatible(@atom, @atom).
%%
kind_incompatible(A, B) :-
    member(A, [fy, yf, quant]),
    member(B, [fy, yf, quant]).
kind_incompatible(A, B) :-
    member(A, [xfx, xfy, yfx]),
    member(B, [xfx, xfy, yfx]).


%% operator_kind(ErgoKind, QPKind).
%% check_operator_kind(ErgoKind, QPKind).
%%
%%  ErgoKind is the Ergo name for the Qu-Prolog associativity
%%  specification QPKind as used by op/3.
%%  
%%  mode operator_kind(?atom, ?atom).
%%  mode check_operator_kind(+atom, ?atom).
%%
operator_kind(prefix,     fy).
operator_kind(postfix,    yf).
operator_kind(nonassoc,   xfx).
operator_kind(leftassoc,  yfx).
operator_kind(rightassoc, xfy).
operator_kind(quant,      quant).

check_operator_kind(ErgoKind, QPKind) :-
    operator_kind(ErgoKind, QPKind),
    !.
check_operator_kind(ErgoKind, _) :-
    handle_error(['unknown kind of operator: ',q(ErgoKind)]).


database command
  add_operators(Thy),
  add_all_operators(Thy)
extends
  sugar_cmds
mode
  add_operators(@theory),
  add_all_operators(@theory)
doc
  'Copy operators from theory Thy into the current theory.',nl,
  'Operators that clash with existing operators of the current theory will',nl,
  'generate an error message and will not be added to the current theory.',
  para,
  'Note that add_operators/1 copies only those operators that',nl,
  'were originally defined in Thy, whereas add_all_operators/1 also copies',nl,
  'any operators that have been copied into Thy from other theories.',nl,
  'That is, add_all_operators/1 is transitive, whereas add_operators/1 is',nl,
  'non-transitive.'.

add_operators(FromThy) :-
    check_current_theory(Thy),
    check_strict_ancestor(FromThy),
    \+ (query_operator(N,A,P,FromThy, FromThy),
        add_operator2(N,A,P,FromThy,Thy),
        fail).

add_all_operators(FromThy) :-
    check_current_theory(Thy),
    check_strict_ancestor(FromThy),
    \+ (query_operator(N,A,P,OrigThy, FromThy),
        add_operator2(N,A,P,OrigThy,Thy),
        fail).


database command
  inherit_all_operators(Thy)
extends
  sugar_cmds
mode
  inherit_all_operators(@theory)
doc
  'Inherits all operators and obvar prefixes from theory Thy.',nl,
  'Thy must be an ancestor of the current theory.',
  para,
  'This inheritance of operators is not transitive.  That is, if Thy',nl,
  'inherits operators from some other theory, this command does not',nl,
  'inherit them into the current theory.  (If you wish to check that',nl,
  'Thy does not inherit operators, simply check that it does not contain',nl,
  'any inherited_operators/1 facts.)',
  para,
  'The difference between this command and the ',
    xref('add_all_operators/1',global,'database:add_all_operators/1'),nl,
  'command, is that this one shares the operators of Thy, which means',nl,
  'that future changes to the operators of Thy will automatically be',nl,
  'reflected in the current theory, whereas add_all_operators/1 makes',nl,
  'a local copy of all the operators from Thy, thus allowing the operators',nl,
  'of the current theory to diverge from those of Thy in the future.',nl,
  'Also, operator inheritance links between theories cannot be deleted,',nl,
  'whereas operators that are copied can be deleted using the',nl,
    xref('delete_operator/2',global,'database:delete_operator/2'),
  ' or ',
    xref('delete_operators/1',global,'database:delete_operators/1'),
  ' commands.'.

inherit_all_operators(FromThy) :-
    check_current_theory(Thy),
    check_strict_ancestor(FromThy), %% because ancestors are definitely loaded.
    ( query_theory(inherited_operators(FromThy), Thy) ->
        msg(operators, ['theory ',w(Thy),' already inherits the operators',
                        ' of theory ',w(FromThy)])
    ;
        tactic_assert_theory(inherited_operators(FromThy), Thy),
        enable_operators(FromThy),
        enable_obvar_prefix(FromThy),
        op_table_inherit(Thy, FromThy),
        obvar_prefix_table_inherit(FromThy)
    ).

database command
  delete_operator(Name),
  delete_operator(Name, Kind)
extends
  sugar_cmds
mode
  delete_operator(@atom),
  delete_operator(@atom, ?atom)
doc
  'Delete the theory-specific operator called Name.',nl,
  'This can be used to delete operators that were added directly',nl,
  'to the current theory, or operators that were copied from some',nl,
  'other theory (see also ',
    xref('delete_operators/1',global,'database:delete_operators/1'),
  ').'.

delete_operator(Name) :-
    delete_operator(Name, _).

delete_operator(Name, Kind) :-
    check_current_theory(Thy),
    check_atom(Name),
    tactic_retract_theory(operator(Name,Kind,_,_), Thy),
    op(Thy, 0, Kind, Name).



database command
  delete_operators(Thy)
extends
  sugar_cmds
mode
  delete_operators(?theory)
doc
  'Delete all operators that were copied from theory Thy.',nl,
  'If Thy is a variable, ALL operators of the current theory',nl,
  'If Thy is the name of the current theory, all operators added',nl,
  'directly to the current theory will be deleted, but operators',nl,
  'copied from ancestor theories will be left unchanged.'.

delete_operators(FromThy) :-
    check_current_theory(Thy),
    ( var(FromThy) ->
        true
    ;
        check_is_theory_name(FromThy)
    ),
    tactic_retractall_theory(operator(_,_,_,FromThy), Thy).


%% private query_operator(Name,Assoc,Prec,FromThy,Thy).
%%
%%  Similar to query_theory(operator(Name,Assoc,Prec,FromThy), Thy),
%%  but it also searches for inherited operators.
%%
%%  mode query_operator(?atom,?atom,?integer,?theory,@theory).
%%
query_operator(Name,Assoc,Prec,FromThy, Thy) :-
    query_theory(operator(Name,Assoc,Prec,FromThy), Thy).
query_operator(Name,Assoc,Prec,FromThy, Thy) :-
    query_theory(inherited_operators(InheritFrom), Thy),
    query_theory(operator(Name,Assoc,Prec,FromThy), InheritFrom).


?- add_listener(prover(startup),      enable_operators(global)).
?- add_listener(theory(Thy,selected), enable_operators(Thy)).
?- add_listener(theory(Thy,selected), enable_obvar_prefix(Thy)).
?- add_listener(need_operators(Thy),  (enable_operators(Thy),
                                       enable_obvar_prefix(Thy))).
